
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <glib.h>
#include <sys/stat.h>
#include "utils.h"
#include "callbacks.h"
#include "al.h"
#include "folder.h"
#include "imap.h"
#include "defs.h"
#include "pop_recv.h"
#include "account.h"
#include "procheader.h"
#include "smtp.h"
#include "smtp_errorcodes.h"
#include "send_message.h"
#include "al_support.h"
#include "engine_errno.h"
#include "htmltotextconv.h"
#include "mmc_hlr.h"
#include "smime_security.h"
#include "procmime.h"

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "intl.h"

#if CL_USE
#include "cl.h"
#endif
/* Globals */
extern gboolean gl_mmc_status_flag;
/* defined in utils.c */

extern gint engine_errno;
/* defined in utils.c */

 /* globals for smime defined in utils.c */
extern guint gl_cert_id;
extern gchar *gl_cert_pass;

/* List of populating folder*/
static GList *gl_lst = NULL;

/* static globals */
AlCallback *gl_al_cb_ptr = NULL;
PrefsAccount *gl_cur_account = NULL;
gboolean gl_al_folder_cancel = FALSE;

/* Static functions */
static PrefsAccount *al_get_account_info(MsgInfo * msginfo);
static void al_remove_queued_list(const gchar * account_name, ListType type);
static gint al_folder_rename_msgs(const gchar * folderdir,
				  const gchar * old_account, const gchar * new_account);

/**
  This function frees FolderAttachInfo structures. 

  @param FolderAttachInfo - structure to be freed
  @param data - Pointer data which will be NULL
  @return void
*/
static void al_free_list(FolderAttachInfo * folder_attach, gpointer data);

/** Sends messages to UI in chunks of NUM_MAILS
 
    @param type Type of message to be sent
    @param folder_list list of msg information to be sent to UI
*/
static void al_folder_send_max_msg(gint type, GSList * folder_list);


/**
  This function will be called by upper layer to register the call backs.
  All the structures are allocated by the caller but freed by the engine 
  during cleanup.
  @param smtp_callback    pointer to the smtp callback structure
  @param pop3_callback    pointer to the pop3 callback structure
  @param imap4_callback    pointer to the imap4 callback structure
  @return  TRUE if the registration is success,FALSE otherwise
 */
gboolean
al_register_to_engine(SmtpCallback * smtp_callback,
		      Pop3Callback * pop3_callback,
		      Imap4Callback * imap4_callback,
		      AcctMgmtCallback * acct_callback,
		      MmcHlrCallback * mmc_hlr_callback, AlCallback * al_callback)
{
	gboolean rv = TRUE;

	g_print("\n al register to engine called\n");
	if ((!smtp_callback) && (!pop3_callback) && (!imap4_callback)
	    && (!acct_callback) && (!mmc_hlr_callback) && (!al_callback)) {
		engine_errno = ENGINE_ERROR_REG_FAILED;
		rv = FALSE;
	} else {
		/* Register the callbacks to sending module */
		smtp_init(smtp_callback);
		/* Register the callbacks to receiving modules(POP3 & IMAP) */
		imap4_init(imap4_callback);
		pop3_init(pop3_callback);
#if USE_SSL
		/* Initialize the SSL layer */
		if (ssl_init() != SSL_SUCCESS) {
			engine_errno = ENGINE_ERROR_SSL_INIT_FAILED;
			rv = FALSE;
		}
#endif
		/* initialize al callbacks for ending AL errors */
		if (al_callback == NULL) {
			engine_errno = ENGINE_ERROR_AL_INIT_FAILED;
			rv = FALSE;
		} else {
			gl_al_cb_ptr = al_callback;
			engine_error_code_cb = al_callback->al_error_code_cb;
		}

		/* Initialize the account list */
		if (account_init(acct_callback) != AC_OK) {
			/* Even Account Management error occured engine can continue */
			debug_print("Account Management Error \n");
			engine_errno = ENGINE_ERROR_ACCOUNT_INIT;
			rv = FALSE;
		}
		/* Initialize the folder module & check standard folders */
		if (init_folder() == FOLDER_INITIALIZATION_ERROR) {
			debug_print("Folder initialization Error \n");
			if (engine_errno == 0) {
				engine_errno = ENGINE_ERROR_FOLDER_INIT;
			}
			return FALSE;
		}
		/* Initialize the folder lists */
		if (folder_init_lists() == FALSE) {
			engine_errno = ENGINE_ERROR_FOLDER_INIT_LIST;
			rv = FALSE;
		}
		/* Initialize the mmc callbacks & gnome VFS */
		if (mmc_hlr_init(mmc_hlr_callback) == FALSE) {
			engine_errno = ENGINE_ERROR_MMC_HLR_INIT_FAILED;
			rv = FALSE;
		}
#if CL_USE
		/* Initialize the connectivity layer */
		if (cl_init(al_iap_cancel) != CL_OK) {
			debug_print("cl_init failed\n");
			engine_errno = ENGINE_ERROR_CL_INIT_FAILED;
			rv = FALSE;
		}
		debug_print("cl_init succeded\n");
#endif


		if (smime_init() == FALSE) {
			debug_print("Unable to open CM store\n");
			engine_errno = ENGINE_ERROR_SMIME_INIT_FAILED;
			rv = FALSE;
		}
		debug_print("smime_init succeded\n");

	}

	return rv;
}

/**
  This function is called by upper layer to send a mail to SMTP server.
  This function gets the account information based on the email settings
  in msg info structure and send mail to the list specified.
  @param message    body of the message to be send
  @param msginfo    contains msg header, attachment list,email settings,
  this is allocated and freed by the caller
  @return nothing
 */

void al_send_mail(MsgInfo * msginfo, const gchar * message)
{
	PrefsAccount *prefs_account = NULL;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return);
	set_engine_errno(0);


	prefs_account = al_get_account_info(msginfo);
	/* bug fix 10158 */
	gchar *folder_path = NULL;
	folder_path = folder_get_message_file_path(msginfo->header->msgid);
	if (folder_path && is_file_exist(folder_path) == TRUE) {
		if (unlink(folder_path) != 0) {
			g_warning("unlink: failed to delete %s: %s", folder_path, strerror(errno));
		}
		SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
	}
	g_free(folder_path);
	/* bug fix 10158 */

	if (prefs_account) {
		if (send_mail(message, msginfo, prefs_account) == SM_OK) {
			SMTP_SET_MSG_CB(SM_OK, NULL, NULL);
		}
	} else {
		debug_print("[%s]:[%d] : PrefsAccount is NULL  \n", __FUNCTION__, __LINE__);
		SMTP_SET_MSG_CB(SMTP_ACCOUNT_NULL, NULL, NULL);
	}
}


/**
  This function is called to sort selected mails based on account names 
  and callthe respective  send function based on folder type.
  @param msgid_list    list of message ids( format of message id is
  accountname_foldername_subfoldername_messagenumber(a number unique 
  to a directory & account name) )
  @return SM_OK(0) on success otherwise returns status of smtp operation
 */

gint al_send_selected_mails(gint foldertype, GSList * msgid_list)
{
	gint rv = 0;
	set_engine_errno(0);

	msgid_list = folder_sort_msglist(msgid_list, FALSE);
	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);
	switch (foldertype) {
	case OPR_SEND_OUTBOX_MAILS:
		rv = send_outbox_mails(msgid_list);
		break;
	case OPR_SEND_DRAFT_MAILS:
		rv = send_draft_mails(msgid_list);
		break;
	case OPR_RESEND_MAILS:
		rv = send_resend_mails(msgid_list);
		break;
	default:
		debug_print("Invalid folder type selected");
	}
	return rv;
}

/**
  This function will be called by upper layer to cencel the mails in 
  outbox folder.
  @param msgid_list    List of message ids to be cancelled.The format 
  of message id is accountname_foldername_subfoldername_messagenumber. 
  @return nothing
 */

void al_send_cancel_mails(const GSList * msgid_list)
{

#if CL_USE
	cl_cancel_iap_conn();
#endif

	send_cancel_mails(msgid_list);
	engine_errno = ENGINE_STATUS_SEND_RECV_CANCEL;
}

/**
  This function is called to  resend specifed mails in the sent folder
  @param msgid_list    list of message ids( format of message id is
  accountname_foldername_subfoldername_messagenumber(a number unique 
  to a directory & account name) )
  @return none
 */
void al_resend_mails(const GSList * msgid_list)
{
	set_engine_errno(0);
	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);
	if (msgid_list) {
		send_resend_mails(msgid_list);
	}
}

/**
  This funtion scalled to sends mails stored in outbox folder
  @param account    account name which mail(s) to be sent 
  @param msgid    message id.Incase of message id NULL send all the mails 
  with specified account.The format of message id is 
  accountname_foldername_subfoldername_messagenumber.
  @param keepconnalive    is set to TRUE if same connection to be 
  used or set to FALSE
  @return none
 */
void al_send_outbox_mail(const gchar * account, const gchar * msgid, gboolean keepconnalive)
{
	gint total = 0;
	gint tosend = 0;

	set_engine_errno(0);
	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);
	if (account != NULL) {
		if (account_check_for_valid_account(account) == FALSE) {
			return;
		}
	}
	if (msgid != NULL) {
		total = 1;
		tosend = 1;
	}
	send_outbox_mail(account, msgid, keepconnalive, tosend, total, TRUE);
}

/**
  This function is to send the list of mails saved in the draft folder
  @param msgid_list message id list of mails to be sent from draft folder.
  The format of message id is 
  accountname_foldername_subfoldername_messagenumber.
  @return none
 */
void al_send_draft_mails(const GSList * msgid_list)
{
	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);

	if (msgid_list) {
		send_draft_mails(msgid_list);
	}
}


/**
  This function is used to save the mail in the draft folder
  @param msginfo    message information,allocated and freed by the caller
  @param message   body of the message
  @param flag    is used send the mail from draft folder (or) save the mail 
  @param old_id  gives ID with which message was saved earlier.
  @return  none 
 */
void
al_send_save_draft_mail(MsgInfo * msginfo, const gchar * message, gboolean flag, gchar * old_id)
{
	PrefsAccount *prefs_account = NULL;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(FALSE);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return);
	IS_PARAM_NULL(msginfo->header, SMTP_MESSAGE_INFO_NULL, return);
	set_engine_errno(0);

	prefs_account = al_get_account_info(msginfo);
	if (prefs_account) {
		if (send_save_draft_mail(message, msginfo, prefs_account, flag, old_id) == SM_OK) {
			debug_print("Message saved successfully\n");
		}
	} else {
		prefs_account = account_get_dummy_account();
		if (send_save_draft_mail(message, msginfo, prefs_account, 0, old_id) == SM_OK) {
			debug_print("Message saved successfully\n");
		}
	}
}

/**
  This function is get all account information. This is called by upper
  layers.  It creates a list of account_name field from all the account 
  lists.
  @param account_name_list   OUT parameter,contains account list. If this
  parameter already contains some account list it will be freed first and
  updated with existing account information.
  @return total number of accounts contains in the returned list.
 */
gint al_get_account_list(GSList ** acct_list)
{
	if (acct_list == NULL) {
		return FALSE;
	}
	return account_get_list(acct_list);
}

/**
  This function is tp get specified type information of a particular 
  account. The retrived information is updated in supplied prefs 
  account structure.
  @param account_name    name of the account
  @param type    type of settings required.
  @param prefs_account    OUT  structure of containing the account
  settings requested,this has to be freed by the caller
  @return TRUE on success or FALSE on failure
 */
gboolean
al_get_account_settings(const gchar * account_name, SettingType type, PrefsAccount ** prefs_account)
{
	if (prefs_account == NULL) {
		return FALSE;
	}
	return (account_get_settings((gchar *) account_name, type, prefs_account));
}

/**
  This function is set specified type information of a particular account.
  The retrived information is updated in supplied prefs account structure.
  This function also updates the account resource file with the new
  settings. This function is called by upper layers. 
  @param type    type of settings required.
  @param prefs_account    structure containing the account settings.
  @return TRUE on success or FALSE on failure
 */
gboolean al_set_account_settings(SettingType type, PrefsAccount * prefs_account)
{
	PrefsAccount *account = NULL;
	gint recvtype = 0;
	set_engine_errno(0);

	if (prefs_account == NULL) {
		return FALSE;
	}
	if (prefs_account->user_settings.account_name == NULL) {
		return FALSE;
	}
	account = account_get_account(prefs_account->user_settings.account_name);
	if (account != NULL) {
		if (account->user_settings.session != NULL
		    && account->receive_settings.protocol == PROTO_IMAP4) {
			imap_session_disconnect(account->user_settings.session, TRUE);
			account->user_settings.session = NULL;
		}
		recvtype = account->receive_settings.recv_type;
	}

	if (account_set_settings(prefs_account->user_settings.account_name, type, prefs_account)) {
		if ((prefs_account->user_settings.session != NULL)
		    && (prefs_account->receive_settings.protocol == PROTO_IMAP4)) {

			imap_session_disconnect(prefs_account->user_settings.session, TRUE);
			prefs_account->user_settings.session = NULL;
		}
		if (recvtype != 0) {
			if (prefs_account->receive_settings.recv_type != recvtype) {
				/* In case receive type is changed all folder time 
				 * stamps are touched. This is required because cache 
				 * should be re-generated. If cache file is 
				 * re-generated new recv type will be reflected 
				 * which is required by UI to show appropriate icon 
				 * in list view
				 */
				folder_change_time(NULL);

			}
		}
		return TRUE;
	}
	return FALSE;
}


/**
  This function is to delete the account from account list and updates
  the account resource file. This function also deletes the mails in 
  Inbox and Sent folders, releted deleted account. This function also
  moves the mails in Outbox, Draft folder mails to default account
  This function is called by above layers. 
  @param account_name    name of the account
  @return TRUE on success or FALSE on failure
 */
gboolean al_delete_account(const gchar * account_name)
{
	PrefsAccount *prefs_account = NULL;
	gchar *default_account = NULL;
	set_engine_errno(0);

	if (account_name == NULL) {
		return FALSE;
	}
	if (account_delete((gchar *) account_name) == FALSE) {
		return FALSE;
	}
	/* Delete the all msgs in Inbox Folder */
	al_folder_delete_msgs(INBOX_DIR, account_name, FALSE);

	/* Move the Draft and OutBox mails to Default acccount. 
	 * Incase of Default account null move them with account name 'default'
	 */
	prefs_account = account_get_default_account();

	if (prefs_account != NULL) {

		if (prefs_account->user_settings.session != NULL) {
			session_destroy(prefs_account->user_settings.session);
			prefs_account->user_settings.session = NULL;
		}

		if ((prefs_account->user_settings.account_name != NULL)) {
			default_account = prefs_account->user_settings.account_name;
		} else {
			default_account = "default";
		}
	} else {
		default_account = "default";
	}
	al_folder_rename_msgs(DRAFT_DIR, account_name, default_account);
	al_folder_rename_msgs(OUTBOX_DIR, account_name, default_account);
	return TRUE;
}

/**
  This function is to rename the existing account with new account 
  @param old_account    name of the account to be renamed
  @param new_account    name of the new account
  @return TRUE on success or FALSE on failure
 */
gboolean al_rename_account(const gchar * old_account, const gchar * new_account)
{
	Folder *dir = NULL;
	GList *folder_list = NULL;
	set_engine_errno(0);

	if ((old_account == NULL) || (new_account == NULL)) {
		return FALSE;
	}
	if (strcmp(old_account, new_account) == 0) {
		debug_print(" Rename failed : old name new name are same \n");
		return FALSE;
	}
	if (account_rename(old_account, new_account) == FALSE) {
		debug_print("Acccount rename failed \n");
		return FALSE;
	}

	/* Touch archival folder on renaming
	 * for cache generation
	 */
	dir = folder_find_obj_by_name(ARCHIVE);
	if (dir != NULL) {
		folder_list = g_list_append(folder_list, dir);
		folder_change_time(folder_list);
		folder_list = g_list_remove(folder_list, dir);
		g_list_free(folder_list);
	}

	al_folder_rename_msgs(INBOX_DIR, old_account, new_account);
	al_folder_rename_msgs(OUTBOX_DIR, old_account, new_account);
	al_folder_rename_msgs(DRAFT_DIR, old_account, new_account);
	al_folder_rename_msgs(SENT_DIR, old_account, new_account);

	return TRUE;
}


/**
  This function is to delete messages in the existing account files
  @param folder_name    name of the folder files to renamed   
  @param account_name    name of the account in which files to be renamed
  @param required        whether call back needs to be sent to UI or not
  @return FOLDER_SUCCESS on success or FOLDER_FAIL on failure
 */

gint al_folder_delete_msgs(const gchar * foldername, const gchar * account_name, gboolean required)
{
	GSList *msgheader_list = NULL;
	GSList *curheader = NULL;
	GSList *msglist = NULL;
	GSList *curmsg = NULL;
	MsgHeader *msgheader = NULL;
	gchar *msgid_account_name = NULL;
	GSList *msgid_list = NULL;
	set_engine_errno(0);
	set_scheduling_op(FALSE);

	if ((foldername) == NULL || (account_name == NULL)) {
		debug_print(" Invalid arguments \n");
		return FOLDER_DELETE_ERROR;
	}
	if (check_folder_perms(foldername, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	/* Remove the files related to deleted account from Inbox */
	msgheader_list = folder_populate_folder((gchar *) foldername);
	/* If there are no messages in the Inbox,donot consider it as error */
	if (msgheader_list == NULL) {
		return FOLDER_SUCCESS;
	}
	for (curheader = msgheader_list; curheader != NULL; curheader = curheader->next) {
		msgheader = (MsgHeader *) curheader->data;
		if ((msgheader == NULL) || (msgheader->msgid == NULL)) {
			continue;
		}
		msgid_account_name = get_account_name(msgheader->msgid);
		if (msgid_account_name == NULL) {
			continue;
		}
		if (strcmp(account_name, msgid_account_name) == 0) {
			msgid_list = g_slist_append(msgid_list, g_strdup(msgheader->msgid));
		}
		g_free(msgid_account_name);
	}
	if (msgid_list != NULL) {
		al_folder_remove_msgs((gchar *) foldername, msgid_list, FALSE, required);
		slist_free_strings(msgid_list);
	}
	/* Free the header list returned by folder populate folder function */
	while (msgheader_list != NULL) {
		procmsg_msgheader_free((MsgHeader *) msgheader_list->data);
		msgheader_list = g_slist_remove(msgheader_list, msgheader_list->data);
	}
	/* Delete the other files like .mhdr, .0, .1 files from the folder */
	msglist = folder_get_msgid_files((gchar *) foldername, NULL);
	/* other files may not exist */
	if (msglist == NULL) {
		return FOLDER_SUCCESS;
	}
	for (curmsg = msglist; curmsg != NULL; curmsg = curmsg->next) {
		if (curmsg->data != NULL) {
			msgid_account_name = get_account_name((gchar *) curmsg->data);
			if (msgid_account_name == NULL) {
				continue;
			}
			if (strcmp(account_name, msgid_account_name) == 0) {
				folder_remove_msg(foldername, (gchar *) curmsg->data);
			}
			g_free(msgid_account_name);
		}
	}
	/* Free the msg list */
	slist_free_strings(msglist);
	g_slist_free(msglist);

	return FOLDER_SUCCESS;
}

/**
  This function is to rename the existing account files with new account 
  in the specified folder
  @param folderdir    name of the folder files to renamed   
  @param old_account    name of the account in which files to be renamed
  @param new_account    name of the new account which files are replaced
  @return FOLDER_SUCCESS on success or FOLDER_FAIL or error number on failure
*/

static gint
al_folder_rename_msgs(const gchar * folderdir, const gchar * old_account, const gchar * new_account)
{
	GSList *msginfo_list = NULL, *msglist = NULL;
	GSList *cur = NULL, *curmsg = NULL;
	MsgHeader *msgheader = NULL;
	gchar *msgid_account_name = NULL, *tmp = NULL;
	gchar *oldmsgid_path = NULL, *newmsgid_path = NULL, *newmsgid = NULL;
	gboolean msgunread = FALSE;
	set_engine_errno(0);

	if ((folderdir == NULL) || (old_account == NULL)
	    || (new_account == NULL)) {
		debug_print(" Inavlid input args ");
		return FOLDER_FAIL;
	}

	if (check_folder_perms(folderdir, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	/*The files related to old account in the folderdir  */
	msginfo_list = folder_populate_folder((gchar *) folderdir);
	/* there may not be any messages in the Inbox */
	if (msginfo_list == NULL) {
		return FOLDER_SUCCESS;
	}
	for (cur = msginfo_list; cur != NULL; cur = cur->next) {
		msgheader = (MsgHeader *) cur->data;
		if ((msgheader == NULL) || (msgheader->msgid == NULL)) {
			continue;
		}
		msgid_account_name = get_account_name(msgheader->msgid);
		if (msgid_account_name == NULL) {
			continue;
		}
		if (strcmp(msgid_account_name, old_account) == 0) {
			tmp = strchr(msgheader->msgid, MSGID_SEPARATOR);
			if (tmp != NULL) {
				newmsgid = g_strdup_printf("%s%s", new_account, tmp);
				newmsgid_path = folder_get_message_file_path(newmsgid);
				oldmsgid_path = folder_get_message_file_path(msgheader->msgid);
				msgunread = is_message_unread(msgheader->msgid);
				if (newmsgid_path && oldmsgid_path &&
				    (move_file(oldmsgid_path, newmsgid_path, FALSE) != 0)) {
					if (msgunread == TRUE) {
						folder_message_del_list(msgheader->msgid, UNREAD);
						folder_message_add_list(newmsgid, UNREAD);
					}
					folder_delete_list_msg(msgheader->msgid, TRUE);
				} else {
					debug_print("Rename of file = %s failed \n", oldmsgid_path);
				}
				g_free(oldmsgid_path);
				g_free(newmsgid_path);
				g_free(newmsgid);
			}
		}
		g_free(msgid_account_name);
	}
	/* Free the header list returned by folder populate folder function */
	while (msginfo_list) {
		procmsg_msgheader_free((MsgHeader *) msginfo_list->data);
		msginfo_list = g_slist_remove(msginfo_list, msginfo_list->data);
	}

	/* Move the other files like .mhdr, .0, .1 files from the folder */
	msglist = folder_get_msgid_files((gchar *) folderdir, NULL);
	for (curmsg = msglist; curmsg != NULL; curmsg = curmsg->next) {
		if (curmsg->data == NULL) {
			continue;
		}
		msgid_account_name = get_account_name((gchar *) curmsg->data);
		if (msgid_account_name == NULL) {
			continue;
		}
		if (strcmp(old_account, msgid_account_name) == 0) {
			tmp = strchr((gchar *) curmsg->data, MSGID_SEPARATOR);
			if (tmp != NULL) {
				newmsgid = g_strdup_printf("%s%s", new_account, tmp);
				newmsgid_path = folder_get_message_file_path(newmsgid);
				oldmsgid_path = folder_get_message_file_path(curmsg->data);
				if (!newmsgid_path || !oldmsgid_path ||
				    (move_file(oldmsgid_path, newmsgid_path, FALSE) < 0)) {
					debug_print("Rename of file = %s failed \n", oldmsgid_path);
				}
				g_free(newmsgid);
				g_free(newmsgid_path);
				g_free(oldmsgid_path);
			}
		}
		g_free(msgid_account_name);
	}
	/* Free the msg list */
	slist_free_strings(msglist);
	g_slist_free(msglist);
	account_rename_uidl_file(old_account, new_account);
	return FOLDER_SUCCESS;
}

/**
  This function will be called by upper layer to receive the mails.
  @param account_name    account name  in which receive operation 
  is performed.
  @return none
*/
void al_recv_mail(const gchar * account_name, gboolean scheduled_op)
{
	PrefsAccount *account = NULL;
	gint first_uid = 0, last_uid = 0;
	Folder *inbox_folder = NULL;
	set_engine_errno(0);

	CHECK_RETURN_IF_FAIL(account_name != NULL);
	/*set whether recv mail is thr scheduling operation */
	set_scheduling_op(scheduled_op);

	if (is_inbox_perm_exist() == FALSE) {
		/* send error though callback */
		gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_NO_FOLDER_PERM);
		set_scheduling_op(FALSE);
		return;
	}
	/* delete the messages,which are pending */
	al_remove_queued_list((gchar *) account_name, DELETE);
	/* Mark \Seen  flag for the  messages,which are pending */
	al_remove_queued_list((gchar *) account_name, SEEN);

	account = (PrefsAccount *) account_get_account_cb((gchar *)
							  account_name);

	if (account == NULL) {
		/* send error though callback */
		gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_ACCOUNT_NOT_FOUND);
		set_scheduling_op(FALSE);
		return;
	}
	account->user_settings.scheduled_op = scheduled_op;
	if (account->receive_settings.protocol == PROTO_IMAP4) {
		if (account->receive_settings.recent_mails > 0) {
			first_uid = last_uid = IMAP_FIRST_UID;
		} else if (account->receive_settings.recent_mails == 0) {
			first_uid = last_uid = IMAP_ALL_UID;
		} else if (account->receive_settings.recent_mails < 0) {
			first_uid = last_uid = IMAP_NO_UID;
		}
		imap_get_headers(account, INBOX_DIR, first_uid, last_uid);

		if (al_sup_sync_imap(account, first_uid, last_uid) != FOLDER_SUCCESS) {
			/* send error though callback */
			gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_OPERATION);
			set_scheduling_op(FALSE);
			return;
		}

	} else if (account->receive_settings.protocol == PROTO_POP3) {

		inbox_folder = folder_find_obj_by_name(INBOX_DIR);

		if (inbox_folder != NULL) {
			if (folder_scan(inbox_folder) == FOLDER_SCAN_ERROR) {
				/* send error though callback */
				gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_INVALID_FOLDER);
				set_scheduling_op(FALSE);
				return;
			}
		} else {
			/* send error though callback */
			gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_NO_MEM);
			set_scheduling_op(FALSE);
			return;
		}

		pop3_recv_mail(account, NULL, FOLDER_OP_NONE, FALSE, NULL);
	}
}

/**
 This function is used to set the seen flag on the server , for a imap mail
 after its read.
 @param msgid is the gchar pointer to the msgid for which the seen flag need to be updated.

*/
void al_imap_set_seen(const gchar * msgid)
{
	PrefsAccount *account = NULL;
	gchar *account_name = NULL;
	account_name = (gchar *) get_account_name(msgid);

	if (account_name != NULL) {
		account = (PrefsAccount *) account_get_account_cb(account_name);
		g_free(account_name);
	}
	if (account != NULL) {
		if (account->receive_settings.protocol == PROTO_IMAP4) {
			al_sup_imap_set_seen(msgid, account);
		}
	}
	return;
}

/**
  This function will be called by upper layer to receive 
  and/or decode the message body.
  @param msgid    id of the message for which the header/message has 
  already been downloaded.The format of message id is
  accountname_foldername_subfoldername_messagenumber.
  @param complete_download    indicates the message has to be completey i
  downloaded,irrespective of the retrieval type. 
  @param signer_email_id - email ID of the sender
  @param recip_email_id - email ID of the recip
  @return mimeinformation of the message,which has to be freed by the caller 
*/
MimeInfo *al_recv_msg_body(const gchar * msgid, gboolean complete_download,
			   gchar * signer_email_id, gchar * recip_email_id,
			   SMimeType smime_type, gboolean read_flag,
			   gboolean edit_flag, gboolean get_html_flag,
			   guint cert_id, gchar * password)
{
	MimeInfo *mimeinfo = NULL;
	gchar *filename = NULL;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	engine_errno = ENGINE_ERROR_DEFAULT;
	gboolean archive_flag = FALSE;
	gchar *header_path = NULL;
	gboolean is_unread = FALSE;
	gchar *fold_name = NULL;

	gl_cert_id = cert_id;
	gl_cert_pass = password;
	set_engine_errno(0);
	set_scheduling_op(FALSE);

	engine_errno = ENGINE_ERROR_DEFAULT;

	if (msgid == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
	if (check_msgid(msgid) == FALSE) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
    /** check for archive folder **/
	if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
		return NULL;
	}

	account_name = (gchar *) get_account_name(msgid);

	if (account_name != NULL) {
		account = (PrefsAccount *) account_get_account_cb(account_name);
		g_free(account_name);
	}
	/* even if the account does not exist,decode the  file if it exists */
	filename = folder_get_message_file_path(msgid);
	if (filename == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		debug_print(" engine_errno : %d \n", engine_errno);
		return NULL;
	}

	/* set true only if the folder is inbox and message is unread */
	fold_name = get_fold_name(msgid);
	if (fold_name
	    && g_strncasecmp(fold_name, INBOX_DIR, MAX_STR_LEN(fold_name, INBOX_DIR)) == 0) {
		if (is_message_unread(msgid) == TRUE) {
			is_unread = TRUE;
		}
	}
	g_free(fold_name);


	if (is_file_entry_exist(filename) == TRUE) {
		if (smime_type == SMIME_NONE) {
			mimeinfo = procmime_get_mimeinfo(msgid, filename, edit_flag, get_html_flag);
			g_free(filename);
		} else {
			mimeinfo = al_sup_get_smime_message_info(filename,
								 signer_email_id,
								 recip_email_id, msgid);
			g_free(filename);
		}

		if (del_arch_file(msgid, archive_flag) == AL_SUP_FAIL) {
			procmime_mimeinfo_free_all(mimeinfo);
			return NULL;
		}
		if (mimeinfo == NULL) {
			return NULL;
		}
		if ((mimeinfo->mime_type == MIME_MULTIPART)
		    || (edit_flag == TRUE)) {
			mimeinfo->inline_images =
			    al_support_get_inline_imageinfo_list(mimeinfo, msgid);
		}
		return mimeinfo;
	} else if (account == NULL) {
		g_free(filename);
		return NULL;
	}
	if (account->receive_settings.protocol == PROTO_IMAP4) {
		mimeinfo =
		    al_sup_recv_body_imap(account, msgid, complete_download,
					  smime_type, signer_email_id,
					  recip_email_id, edit_flag, get_html_flag);
		if (is_file_entry_exist(filename) == TRUE) {
			header_path = get_file_name(msgid, HEAD);
			unlink(header_path);
			g_free(header_path);
		}
		g_free(filename);
		if (mimeinfo == NULL) {
			return NULL;
		}
		procmime_set_is_downloaded(mimeinfo, msgid);

		return mimeinfo;

	} else if (account->receive_settings.protocol == PROTO_POP3) {

		pop3_recv_mail(account, msgid, FOLDER_OP_NONE, FALSE, NULL);

		/* the mime information is returned by the callback */
		/* change unread to read status only if read flag is true */
		if ((read_flag == TRUE)
		    && (engine_errno == ENGINE_ERROR_DEFAULT)) {
			if (folder_message_del_list(msgid, UNREAD) != FOLDER_SUCCESS) {
				engine_errno = ENGINE_ERROR_FOLDER_MESSAGE_DEL_LIST_FAILED;
				return NULL;
			}
		}
		g_free(filename);
		return NULL;
	} else {
		engine_errno = ENGINE_ERROR_INVALID_PROTOCOL;
		g_free(filename);
		return NULL;
	}

}

/**
  This function gets the multipart body of a message,
  which can be displayed in a browser
  @param mimeinfo    pointer to the multipart mime information,
  which has to be separated
  This has to be allocated and freed by the caller
  @param msgid    message id of the message.The format of message id is
  accountname_foldername_subfoldername_messagenumber.
  @return name of the file in which the contents are stored,
  this has to be freed by the caller
 */
gchar *al_recv_multipart_body(MimeInfo * mimeinfo, const gchar * msgid, const gchar * default_font)
{
	gchar *mhtml_file = NULL;
	gchar *tmp_file = NULL;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	FILE *outfp = NULL;
	gint ret_val = AL_SUP_FAIL;
	gchar *filename = NULL;
	gboolean archive_flag = FALSE;

	/*smime changes */
	gchar *final_file = NULL;

	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);
	if (msgid == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
    /** check for archive folder **/
	if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
		return NULL;
	}
	/* get the mimefile, even if account does not exist */
	filename = folder_get_message_file_path(msgid);
	if (filename == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		return NULL;
	}
	if (is_file_exist(filename) == TRUE) {

		/* check if message is smime and decrypt the message */
		final_file = return_file_if_smime(msgid);
		if (final_file == NULL) {
			engine_errno = ENGINE_ERROR_MSG_FILE_DOES_NOT_EXIST;
			g_free(filename);
			return NULL;
		}


		/* get the mimefile, even if account does not exist */
		mhtml_file = al_sup_extract_mhtml(msgid, mimeinfo, final_file, default_font);
		g_free(filename);
		remove_smime_file(final_file);
		g_free(final_file);
		if (del_arch_file(msgid, archive_flag) == AL_SUP_FAIL) {
			g_free(mhtml_file);
			return NULL;
		}
		return mhtml_file;

	} else {
		g_free(filename);
		account_name = (gchar *) get_account_name((gchar *) msgid);

		if (account_name != NULL) {
			account = (PrefsAccount *) account_get_account_cb(account_name);
			g_free(account_name);
		}
		if (account != NULL) {

			if (account->receive_settings.protocol == PROTO_IMAP4) {
				/* Handle this case differently */

				mhtml_file = g_strdup_printf("%s%c%s%c%s%c%s",
							     get_mail_dir(),
							     G_DIR_SEPARATOR, INBOX_DIR,
							     G_DIR_SEPARATOR, msgid,
							     SECTION_SEPARATOR, MHTML_EXTENSION);

				if (is_file_exist(mhtml_file) == TRUE) {
					tmp_file = procmime_get_tmp_mhtml_file(msgid, NULL);
					if (!tmp_file || (procmime_check_for_style_tag(mhtml_file,
										       tmp_file,
										       default_font)
							  == AL_SUP_FAIL)) {
						g_free(tmp_file);
						g_free(mhtml_file);
						return NULL;
					}

					g_free(mhtml_file);
					return tmp_file;
				}
				/* open the output file */
				outfp = fopen(mhtml_file, "wb");

				if (outfp == NULL) {
					if (errno == ENOSPC) {
						send_engine_error_to_ui(errno);
					} else {
						engine_errno = ENGINE_ERROR_OUT_FILE;
					}
					g_free(mhtml_file);
					return NULL;
				}
				ret_val = imap_get_mhtml_body(mimeinfo, outfp, msgid, account);

				fclose(outfp);
				if (ret_val == IMAP_SUCCESS) {
					tmp_file = procmime_get_tmp_mhtml_file(msgid, NULL);
					if (!tmp_file || (procmime_check_for_style_tag(mhtml_file,
										       tmp_file,
										       default_font)
							  == AL_SUP_FAIL)) {

						g_free(tmp_file);
						g_free(mhtml_file);
						return NULL;
					}
					g_free(mhtml_file);
					return tmp_file;
				} else {
					g_free(mhtml_file);
					return NULL;
				}
			}

		}
		/* this is case of error */
		engine_errno = ENGINE_ERROR_MSG_FILE_DOES_NOT_EXIST;
		return NULL;
	}
}

/**
  This function will be called by upper layer to decode the parts.
  @param msgid    message id of the message.The format of message id is
  accountname_foldername_subfoldername_messagenumber.
  @param mimeinfo   the mimeinformation of the part or attachment,
  this has to be allocated and freed by the caller
  @return Message information,this has to be freed by caller
 */
MsgInfo *al_recv_msg_attachment(const gchar * msgid, MimeInfo * mimeinfo)
{

	MsgInfo *msginfo = NULL;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	gchar *file_name = NULL;
	gboolean archive_flag = FALSE;

	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);

	if ((msgid == NULL) || (mimeinfo == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
    /** check for archive folder **/
	if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
		return NULL;
	}

	account_name = (gchar *) get_account_name(msgid);

	if (account_name == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_MSGID;
		return NULL;
	}

	account = (PrefsAccount *) account_get_account_cb(account_name);
	g_free(account_name);

	if (account != NULL) {
		/* the attachment is downloaded and decoded selectively */
		file_name = folder_get_message_file_path(msgid);
		if (!file_name || is_file_exist(file_name) == FALSE) {
			msginfo = al_sup_get_imap_selective_attachment(mimeinfo, msgid, account);
			g_free(file_name);
			return msginfo;
		}
		g_free(file_name);
	}

	/* The attachment is already downloaded */
	if (mimeinfo->mime_type == MIME_MESSAGE_RFC822) {

		msginfo = al_sup_rfc822_attach(mimeinfo, msgid);

	} else {

		/* This is normal attachment */
		msginfo = al_sup_get_attachment_info(mimeinfo, msgid);
	}

	/* delete temporary archive file */
	if ((del_arch_file(msgid, archive_flag) == AL_SUP_FAIL)
	    && (msginfo != NULL)) {
		procmsg_msginfo_free(msginfo);
		return NULL;
	}
	return msginfo;
}

/**
  This function will cancel receiving of the mails.
  @param account_name    name of the account to receive mails 
  @return none
 */
void al_cancel_recv_mail(const gchar * account_name)
{
	PrefsAccount *account = NULL;
#if CL_USE
	cl_cancel_iap_conn();
#endif
	CHECK_RETURN_IF_FAIL(account_name != NULL);

	account = (PrefsAccount *) account_get_account_cb((gchar *)
							  account_name);

	if (account == NULL) {
		return;
	}
	if (account->receive_settings.protocol == PROTO_IMAP4) {
		imap_recv_cancel();
	} else if (account->receive_settings.protocol == PROTO_POP3) {
		pop3_recv_cancel();

	}
	engine_errno = ENGINE_STATUS_SEND_RECV_CANCEL;
}

/**
  This function creates specified folder
  @param name    folder name which is to be created
  @param path    folder path of the folder name to create
  @param type    folder type 
  @return success or error code
 */

gint al_folder_create(const gchar * name, const gchar * path, FolderType type)
{
	set_engine_errno(0);
	if (is_dir_perm_exist(path, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	return folder_create(name, path, type, FALSE);
}

/**
  This function deletes specified folder 
  @param name    folder name to de deleted
  @return success or error code
 */
gint al_folder_delete(const gchar * name)
{
	if (check_folder_perms(name, FALSE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	return folder_delete(name);
}

/**
  This function renames specified folder 
  @param oldfolder    folder name to be renamed
  @param newfolder    new folder name 
  @return success or error code
 */
gint al_folder_rename(const gchar * oldfolder, const gchar * newfolder)
{
	set_engine_errno(0);
	if (check_folder_perms(oldfolder, FALSE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	return folder_rename(oldfolder, newfolder);
}

/**
  This function moves specified message in the folder
  @param srcfolder    source folder name where msg exists
  @param destfolder    folder name where message to be moved
  @param msgid    message id of message to be moved. The format of message
  id is accountname_foldername_subfoldername_messagenumber.
  @return success or error code
 */
gint al_folder_move_msg(const gchar * srcfolder, const gchar * destfolder, const gchar * msgid)
{
	GSList *msgid_list = NULL;
	gint ret_val = FOLDER_FAIL;
	gchar *account = NULL;
	PrefsAccount *prefs_account = NULL;
	gchar *file_path = NULL;

	gboolean is_unread = FALSE;
	set_engine_errno(0);
	set_scheduling_op(FALSE);

	if ((srcfolder == NULL) || (destfolder == NULL)
	    || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return ENGINE_ERROR_INVALID_PARAM;
	}

	if (check_folder_perms(srcfolder, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	if (check_folder_perms(destfolder, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}

	if ((g_strncasecmp(srcfolder, INBOX_DIR, MAX_STR_LEN(srcfolder, INBOX_DIR))
	     == 0)) {
		//|| (g_strncasecmp(destfolder, INBOX_DIR, MAX_STR_LEN(srcfolder, INBOX_DIR)) == 0)) {
		if (is_message_unread(msgid) == TRUE) {
			is_unread = TRUE;
		}
	}
	account = get_account_name(msgid);
	prefs_account = account_get_account(account);
	if (prefs_account != NULL) {

		/* Download the messages from the server */
		file_path = folder_get_message_file_path(msgid);
		if (!file_path || (al_sup_folder_recv_msg(prefs_account, msgid, srcfolder,
							  TRUE, file_path) == AL_SUP_FAIL)) {
			g_free(account);
			g_free(file_path);
			return engine_errno;
		}
		g_free(file_path);
	}
	/* Add message to the list */
	msgid_list = g_slist_append(msgid_list, g_strdup(msgid));
	if (msgid_list != NULL) {

		ret_val = folder_move_msgs(srcfolder, destfolder, msgid_list);
		if (msgid_list != NULL) {
			slist_free_strings(msgid_list);
		}
		if (ret_val == FOLDER_SUCCESS) {

			/* delete the messages from the list */
			ret_val = folder_update_unread(msgid, destfolder, TRUE);

			folder_delete_list_msg(msgid, TRUE);
		}

	} else {
		ret_val = FOLDER_MOVE_MSG_ERROR;
	}
	g_free(account);
	return ret_val;
}

/**
  This function copies specified message in the folder
  @param srcfolder    source folder name where msg exists
  @param destfolder    folder name where message to be copied
  @param msgid    message id in which message is to be copied.
  The format of message id is
  accountname_foldername_subfoldername_messagenumber. 
  @return success or error code
 */
gint al_folder_copy_msg(const gchar * srcfolder, const gchar * destfolder, const gchar * msgid)
{
	GSList *msgid_list = NULL;
	gint ret_val = FOLDER_FAIL;
	gchar *account = NULL;
	gchar *file_path = NULL;
	PrefsAccount *prefs_account = NULL;

	gboolean is_unread = FALSE;

	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);

	if ((srcfolder == NULL) || (destfolder == NULL)
	    || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return ENGINE_ERROR_INVALID_PARAM;
	}
	if (check_folder_perms(srcfolder, FALSE) == FALSE) {
		return FOLDER_NO_PERMS;
	}
	if (check_folder_perms(destfolder, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}

	if (is_message_unread(msgid) == TRUE) {
		is_unread = TRUE;
	}
	account = get_account_name(msgid);
	prefs_account = account_get_account(account);
	if (prefs_account != NULL) {
		/* Download the messages,which are not received */
		file_path = folder_get_message_file_path(msgid);
		if (!file_path || (al_sup_folder_recv_msg(prefs_account, msgid,
							  srcfolder, FALSE,
							  file_path) == AL_SUP_FAIL)) {
			g_free(account);
			g_free(file_path);
			return engine_errno;
		}
		g_free(file_path);
	}
	/* Add message to the list */
	msgid_list = g_slist_append(msgid_list, g_strdup(msgid));

	if (msgid_list != NULL) {
		ret_val = folder_copy_msgs(srcfolder, destfolder, msgid_list);
		slist_free_strings(msgid_list);
		g_slist_free(msgid_list);
		if (ret_val == FOLDER_SUCCESS) {
			ret_val = folder_update_unread(msgid, destfolder, FALSE);

		}
	} else {
		ret_val = FOLDER_COPY_MSG_ERROR;
	}

	g_free(account);
	return ret_val;
}

/**
  This function moves specified messages in the folder
  @param srcfolder    source folder name where messages exist
  @param destfolder    folder name where messages are to be moved
  @param msglist    GSList of message ids in which message are to be moved.
  This list has to be allocated and freed by the caller.The format of 
  message id is accountname_foldername_subfoldername_messagenumber. 
  @return success or error code
 */
gint al_folder_move_msgs(const gchar * srcfolder, const gchar * destfolder, GSList * msglist)
{
	GSList *folder_add_list = NULL;
	GSList *folder_update_list = NULL;
	GSList *folder_change_list = NULL;
	GSList *sorted_list = NULL;
	gint succ_val = AL_SUP_SUCCESS;
	gint total = 0;
	gint cur_count = 0;
	gint err_count = 0;

	gl_al_folder_cancel = FALSE;
	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);
	if ((srcfolder == NULL) || (destfolder == NULL)
	    || (msglist == NULL)) {
		FOLDER_SET_MSG_CB(FOLDER_MSG_ERROR);
		return AL_SUP_FAIL;
	}

	if (check_folder_perms(srcfolder, FALSE) == FALSE) {
		FOLDER_SET_MSG_CB(FOLDER_NO_PERMS);
		return AL_SUP_FAIL;
	}
	if (check_folder_perms(destfolder, TRUE) == FALSE) {
		FOLDER_SET_MSG_CB(FOLDER_NO_PERMS);
		return AL_SUP_FAIL;
	}
	total = g_slist_length(msglist);


	FOLDER_PROGRESS_CB(cur_count, total);

	sorted_list = folder_sort_msglist(msglist, FALSE);

	succ_val = al_sup_accountwise_op(sorted_list, &folder_add_list,
					 &folder_update_list, TRUE, total,
					 srcfolder, destfolder, &folder_change_list);

	FOLDER_SET_MSG_CB(FOLDER_COMPLETE);

	err_count = total - g_slist_length(folder_add_list);


	if ((succ_val == AL_SUP_NOT_AVAILABLE) && (total == 1)) {

		if (gl_al_cb_ptr != NULL) {
			if (gl_al_cb_ptr->al_folder_msg_cb != NULL)
				gl_al_cb_ptr->al_folder_msg_cb(FOLDER_MSG_DELETED_ERROR,
							       "%s", msglist->data);
		}
	} else if (succ_val == AL_SUP_FAIL) {

		if (err_count < 0) {

			FOLDER_SET_ERROR_CB(FOLDER_MOVE_MSG_ERROR, total);

		} else {

			FOLDER_SET_ERROR_CB(FOLDER_MOVE_MSG_ERROR, err_count);
		}
	}
	/* send the list of messages under any condition */
	al_folder_send_max_msg(LIST_FOLDER_ADD, folder_add_list);
	procmsg_msgheader_list_free(folder_add_list);

	/* send the list of messages to delete */
	al_folder_send_max_msg(LIST_FOLDER_DEL, folder_update_list);
	procmsg_msgheader_list_free(folder_update_list);

	/*send list of messages which have been downloaded but not moved
	 * as cancel was pressed before moving
	 */
	if (folder_change_list != NULL) {
		al_folder_send_max_msg(LIST_FOLDER_UPDATE, folder_change_list);
		procmsg_msgheader_list_free(folder_change_list);
	}

	return AL_SUP_SUCCESS;
}

/**
  This function copies specified messages in the folder
  @param srcfolder    source folder name where messages exist
  @param destfolder    folder name where messages are to be copied
  @param msglist    List of message ids in which messages are be copied.
  This list has to be allocated and freed by the caller.The format of 
  message id is accountname_foldername_subfoldername_messagenumber. 
  @return success or error code
 */

gint al_folder_copy_msgs(const gchar * srcfolder, const gchar * destfolder, GSList * msglist)
{
	GSList *folder_add_list = NULL;
	GSList *folder_update_list = NULL;
	GSList *folder_del_list = NULL;
	GSList *sorted_list = NULL;
	gint succ_val = AL_SUP_SUCCESS;
	gint total = 0;
	gint cur_count = 0;
	gl_al_folder_cancel = FALSE;
	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);
	if ((srcfolder == NULL) || (destfolder == NULL)
	    || (msglist == NULL)) {
		FOLDER_SET_MSG_CB(FOLDER_MSG_ERROR);
		return AL_SUP_FAIL;
	}
	if (check_folder_perms(srcfolder, FALSE) == FALSE) {
		FOLDER_SET_MSG_CB(FOLDER_NO_PERMS);
		return AL_SUP_FAIL;
	}
	if (check_folder_perms(destfolder, TRUE) == FALSE) {
		FOLDER_SET_MSG_CB(FOLDER_NO_PERMS);
		return AL_SUP_FAIL;
	}
	total = g_slist_length(msglist);

	FOLDER_PROGRESS_CB(cur_count, total);
	sorted_list = folder_sort_msglist(msglist, FALSE);
	succ_val = al_sup_accountwise_op(sorted_list, &folder_add_list,
					 &folder_update_list, FALSE, total,
					 srcfolder, destfolder, &folder_del_list);

	FOLDER_SET_MSG_CB(FOLDER_COMPLETE);

	/* send the list of messages under any condition */
	al_folder_send_max_msg(LIST_FOLDER_ADD, folder_add_list);
	procmsg_msgheader_list_free(folder_add_list);

	/* send the list of messages to update */
	if (strcmp(srcfolder, INBOX_DIR) == 0) {
		al_folder_send_max_msg(LIST_FOLDER_UPDATE, folder_update_list);
		procmsg_msgheader_list_free(folder_update_list);
		al_folder_send_max_msg(LIST_FOLDER_DEL, folder_del_list);
		procmsg_msgheader_list_free(folder_del_list);
	}

	if (succ_val != AL_SUP_SUCCESS) {

		FOLDER_SET_MSG_CB(FOLDER_COPY_MSG_ERROR);

	}

	return AL_SUP_SUCCESS;
}

/**
  This function removes specified message in the folder
  @param folder    name of the folder in which message to be removed exists
  @param msgid    message id which is to be removed.The format of message 
  id is accountname_foldername_subfoldername_messagenumber.
  @param flag    TRUE for local & remote deletion FALSE for only local deletion
  @return success(FOLDER_SUCCESS) or error code(FOLDER_REMOVE_MSGS_ERROR)
 */

gint al_folder_remove_msg(const gchar * folder, const gchar * msgid, gboolean flag)
{
	GSList *msgid_list = NULL;
	gint ret = 0;

	gboolean is_unread = FALSE;

	engine_errno = ENGINE_ERROR_DEFAULT;

	if ((folder == NULL) || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return FOLDER_REMOVE_MSGS_ERROR;
	}
	if (check_msgid(msgid) == FALSE) {
		engine_errno = ENGINE_ERROR_INVALID_MSGID;
		return FOLDER_REMOVE_MSGS_ERROR;
	}
	if (check_folder_perms(folder, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}

	msgid_list = g_slist_append(msgid_list, (gpointer) msgid);
	if (msgid_list == NULL) {
		return FOLDER_REMOVE_MSGS_ERROR;
	}

	if (g_strncasecmp(folder, INBOX_DIR, MAX_STR_LEN(folder, INBOX_DIR)) == 0) {
		if (is_message_unread(msgid) == TRUE) {
			is_unread = TRUE;
		}
	}
	ret = al_folder_remove_msgs(folder, msgid_list, flag, TRUE);

	return ret;
}

/**
  This function removes specified messages in the folder
  @param folder    name of the folder in which contains the message to be 
  removed.
  @param msglist    List of messages which is to be removed.
  This list has to be allocated and freed by the caller.The format of message id
  is accountname_foldername_subfoldername_messagenumber.
  @param flag    TRUE for local & remote deletion FALSE for only local deletion
  @param required TRUE if call back needs to be sent to UI FALSE otherwise
  @return success(FOLDER_SUCCESS) or error code(FOLDER_REMOVE_MSGS_ERROR)
 */

gint al_folder_remove_msgs(const gchar * folder, GSList * msglist, gboolean flag, gboolean required)
{
	GSList *folder_delete_list = NULL;
	GSList *folder_queued_list = NULL;
	GSList *sorted_list = NULL;
	GSList *folder_update_list = NULL;
	gint succ_val = FOLDER_SUCCESS;
	gint cur_count = 0;
	gint total = 0;

	engine_errno = ENGINE_ERROR_DEFAULT;
	gl_al_folder_cancel = FALSE;

	if ((folder == NULL) || (msglist == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		FOLDER_SET_MSG_CB(FOLDER_REMOVE_MSGS_ERROR);
		return FOLDER_REMOVE_MSGS_ERROR;
	}
	if (check_folder_perms(folder, TRUE) == FALSE) {
		FOLDER_SET_MSG_CB(FOLDER_NO_PERMS);
		return FOLDER_NO_PERMS;
	}
	/* adjust the list in sorting order */
	sorted_list = folder_sort_msglist(msglist, FALSE);

	total = g_slist_length(msglist);

	FOLDER_DEL_PROGRESS_CB(cur_count, total);

	succ_val = al_sup_rem_accountwise(sorted_list, &folder_delete_list,
					  &folder_update_list, total, folder, flag);

	FOLDER_SET_MSG_CB(FOLDER_COMPLETE);

	/* send the list of messages under any condition */
	if (required == TRUE) {
		al_folder_send_max_msg(LIST_FOLDER_DEL, folder_delete_list);
	}
	procmsg_msgheader_list_free(folder_delete_list);

	if ((strcmp(folder, INBOX_DIR) == 0)
	    && (engine_errno == ENGINE_ERROR_CONNECT_SERVER)) {
		folder_queued_list = folder_get_queued_complete_list();
		/* send the list of remote messages failed to connect */
		al_folder_send_max_msg(LIST_FOLDER_DEL, folder_queued_list);
		procmsg_msgheader_list_free(folder_queued_list);
		FOLDER_SET_MSG_CB(FOLDER_REMOVE_SERVER_MSGS_ERROR);
		return FOLDER_REMOVE_SERVER_MSGS_ERROR;

	}
	/* send the list of messages to update */
	if (strcmp(folder, INBOX_DIR) == 0) {
		al_folder_send_max_msg(LIST_FOLDER_UPDATE, folder_update_list);
		procmsg_msgheader_list_free(folder_update_list);
	}

	if (succ_val != FOLDER_SUCCESS) {

		FOLDER_SET_MSG_CB(FOLDER_REMOVE_MSGS_ERROR);
		return FOLDER_REMOVE_MSGS_ERROR;
	} else {
		FOLDER_SET_MSG_CB(FOLDER_SUCCESS);

	}
	return FOLDER_SUCCESS;
}

/**
  This function removes all the messages in the specified folder
  @param folder    folder name in which messages are to be removed
  @param flag    set to true to remove both local & remote, false to 
  remove only local
  @return success or error code
 */

gint al_folder_remove_all_msg(const gchar * folder, gboolean flag)
{
	gint ret = FOLDER_FAIL;

	GSList *msgid_list = NULL;
	GSList *msginfo_list = NULL;
	GSList *cur = NULL;
	MsgHeader *msgheader = NULL;

	gl_al_folder_cancel = FALSE;
	engine_errno = ENGINE_ERROR_DEFAULT;
	if (folder == NULL) {
		return ENGINE_ERROR_INVALID_PARAM;
	}
	if (check_folder_perms(folder, TRUE) == FALSE) {
		return FOLDER_NO_PERMS;
	}

	msginfo_list = folder_populate_folder(folder);
	if (msginfo_list == NULL) {
		return FOLDER_REMOVE_MSGS_ERROR;
	}

	for (cur = msginfo_list; cur != NULL; cur = cur->next) {
		msgheader = (MsgHeader *) cur->data;
		if ((msgheader) && (msgheader->msgid != NULL)) {
			msgid_list = g_slist_append(msgid_list, g_strdup(msgheader->msgid));
		}
	}

	ret = al_folder_remove_msgs(folder, msgid_list, flag, TRUE);
	if (msgid_list != NULL) {
		slist_free_strings(msgid_list);
	}
	while (msginfo_list) {
		procmsg_msgheader_free((MsgHeader *) msginfo_list->data);
		msginfo_list = g_slist_remove(msginfo_list, msginfo_list->data);
	}

	return ret;
}

/**
  This function returns the list of folders
  @param folderlist    OUT parameter, list of folders are returned as out 
  parameter. This list has to be freed by the calling function.
  @return folder error code or (FOLDER_SUCCESS)success
 */

gint al_folder_get_list(GSList ** folderlist)
{
	GList *cur = NULL;
	GList *folder_list = NULL;
	Folder *folder = NULL;

	if (folderlist == NULL) {
		return FOLDER_READ_ERROR;
	}

	folder_list = folder_get_list();

	if (folder_list == NULL) {
		return FOLDER_READ_ERROR;
	}
	for (cur = folder_list; cur != NULL; cur = cur->next) {
		*folderlist = g_slist_append(*folderlist, cur->data);
		folder = (Folder *) cur->data;
	}
	CHECK_RETURN_VAL_IF_FAIL(*folderlist != NULL, FOLDER_READ_ERROR);
	return FOLDER_SUCCESS;
}

/**
  This function populates the messages in the particular folder
  @param name    folder name where message header list to returned 
  @param index    index of the message id list
  @return list of message headers.This memory has to be freed by 
  the calling function.
 */
GSList *al_populate_folder(const gchar * name, gint index)
{
	GSList *cur = NULL;
	gchar *account_name = NULL;
	gchar *message_file_name = NULL;
	PrefsAccount *account = NULL;
	MsgHeader *header = NULL;
	gboolean archive_flag = FALSE;
	GSList *list = NULL;

	set_engine_errno(0);
	if (index == 0) {
		gl_lst = NULL;
		gl_lst = folder_get_file_list(name);
		if (gl_lst == NULL) {
			engine_errno = ENGINE_ERROR_FILE_ERROR;
			return NULL;
		}
	}

	list = folder_populate_from_list(gl_lst, index, name);
    /** All messages populated hence it is NULL*/
	if (list == NULL) {

		g_list_foreach(gl_lst, (GFunc) gnome_vfs_file_info_unref, NULL);
		g_list_free(gl_lst);
		gl_lst = NULL;
		return NULL;
	}

	for (cur = list; cur != NULL; cur = cur->next) {
		header = (MsgHeader *) cur->data;
		account_name = (gchar *) get_account_name(header->msgid);
		if (check_archive(header->msgid, &archive_flag) == AL_SUP_FAIL) {
			while (list != NULL) {
				procmsg_msgheader_free((MsgHeader *) list->data);
				list = g_slist_remove(list, list->data);
			}
			g_list_foreach(gl_lst, (GFunc) gnome_vfs_file_info_unref, NULL);
			g_list_free(gl_lst);
			gl_lst = NULL;
			return NULL;
		}
		if (account_name == NULL) {
			while (list != NULL) {
				procmsg_msgheader_free((MsgHeader *) list->data);
				list = g_slist_remove(list, list->data);
			}
			g_list_foreach(gl_lst, (GFunc) gnome_vfs_file_info_unref, NULL);
			g_list_free(gl_lst);
			gl_lst = NULL;
			return NULL;
		}
		account = (PrefsAccount *) account_get_account_cb(account_name);
		g_free(account_name);
		if (account == NULL) {
			header->msgtag.protocol = PROTO_INVALID;
			header->msgtag.fullydownloaded = TRUE;
			header->msgtag.bodydownloaded = TRUE;
			header->msgtag.remote_copy = FALSE;
			message_file_name = folder_get_message_file_path(header->msgid);
			if (message_file_name
			    && (procmime_check_attachments(message_file_name, FALSE)
				== TRUE)) {
				header->flags.perm_flags |= MSG_ATTACH;
			}
			header->msgtag.recvtype = BODY_ATTCH;
			g_free(message_file_name);
			continue;
		}

		header->msgtag.protocol = account->receive_settings.protocol;
		header->msgtag.recvtype = account->receive_settings.recv_type;

		/* set the header fields */
		procmime_folder_set_header(header, account);
		if (del_arch_file(header->msgid, archive_flag) == AL_SUP_FAIL) {
			while (list != NULL) {
				procmsg_msgheader_free((MsgHeader *) list->data);
				list = g_slist_remove(list, list->data);
			}
			g_list_foreach(gl_lst, (GFunc) gnome_vfs_file_info_unref, NULL);
			g_list_free(gl_lst);
			gl_lst = NULL;
			return NULL;
		}
	}

	return list;
}

/** 
  This function removes the attachment file 
  @param folder_attach_info_list    List of attachment information. 
  This list has to be allocated and freed by the calling function.
  @return on FOLDER_SUCCESS(success) or some error code on fail
 */
GSList *al_folder_remove_attachment(GSList * folder_attach_info_list)
{
	GSList *cur = NULL;
	GSList *ret_success_list = NULL;
	gint ret = 0;
	FolderAttachInfo *folder_attach_info = NULL;
	gchar *filepath = NULL;
	gchar *mime_file_name = NULL;

	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);

	if (folder_attach_info_list == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}

	/* Use only the first msgid, list belongs to single message */
	folder_attach_info = (FolderAttachInfo *)
	    folder_attach_info_list->data;

	if (folder_attach_info == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
	filepath = folder_get_message_file_path(folder_attach_info->msgid);
	if (filepath)
		mime_file_name =
		    g_strdup_printf("%s%c%s", filepath, SECTION_SEPARATOR, PURGE_SUFFIX);

	/* the file is fully downloaded,split the message to parts */
	if (filepath && is_file_entry_exist(filepath) == TRUE) {
		if (al_sup_split_to_parts(filepath, folder_attach_info->msgid) == AL_SUP_FAIL) {
			g_free(filepath);
			g_free(mime_file_name);
			return NULL;
		}
		/* remove the full file,as we now have parts */
		unlink(filepath);
	}
	g_free(filepath);

	for (cur = folder_attach_info_list; cur != NULL; cur = cur->next) {

		ret =
		    folder_remove_attachment(((FolderAttachInfo *) cur->data)->
					     msgid,
					     ((FolderAttachInfo *) cur->data)->section_number);

		if (ret == FOLDER_SUCCESS) {
			/* write the purged section number */
			procmime_write_purge_section(mime_file_name,
						     ((FolderAttachInfo *) cur->
						      data)->section_number);
			/* Add this to the return success list */
			ret_success_list = g_slist_append(ret_success_list,
							  g_strdup(((FolderAttachInfo *) cur->
								    data)->section_number));
		}
	}
	g_free(mime_file_name);
	return ret_success_list;
}

/**
  This function will be called by upperlayer to do cleanup 
  of temporary files  & global lists before the process exits. 
  @param none 
  @return none 
 */
void al_engine_cleanup(void)
{
	const gchar *tmp_dir = NULL;
	const gchar *attach_dir = NULL;

	tmp_dir = get_mime_tmp_dir();
	if (tmp_dir != NULL) {
		folder_remove_all_msg(tmp_dir);
		/* free the directory names used all over */
	}
	tmp_dir = get_tmp_dir();

	if (tmp_dir != NULL) {
		folder_remove_all_msg(tmp_dir);
		/* free the directory names used all over */
	}
	g_free(gl_al_cb_ptr);
	attach_dir = get_tem_attach_dir();
	gl_al_cb_ptr = NULL;
}

/**
  This function will be called by upperlayer to get engine global errno 
  @param none 
  @return engine error  number 
 */
gint al_get_engine_errno(void)
{
	gint err_no = AL_SUP_FAIL;
	err_no = get_engine_errno();
	return err_no;
}

/**
  This function will be called by to get account information  
  @param msginfo    pointer to message information,whose variables will be set
  to default account settings,if account of the message does not exist.
  @return pointer to PrefsAccount information,
  which should not be freed by the caller.
 */

static PrefsAccount *al_get_account_info(MsgInfo * msginfo)
{
	PrefsAccount *prefs_account = NULL;

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return NULL);

	if ((msginfo->settings) && (msginfo->settings->account)) {
		prefs_account = account_get_account(msginfo->settings->account);
		if (prefs_account == NULL) {
			debug_print("[%s]:[%d] : PrefsAccount Not found for %s \n",
				    __FUNCTION__, __LINE__, msginfo->settings->account);
		}
	}
	if (prefs_account == NULL) {
		prefs_account = account_get_default_account();
		if (prefs_account != NULL) {
			if (msginfo->settings == NULL) {
				msginfo->settings = g_new0(EmailSpecificSettings, 1);
				if (msginfo->settings == NULL) {
					SMTP_SET_MSG_CB(SMTP_INTERNAL_ERROR, NULL, NULL);
					return NULL;
				}
				msginfo->settings->account =
				    g_strdup(prefs_account->user_settings.account_name);
				msginfo->settings->options =
				    prefs_account->advanced_settings.email_sending;
				msginfo->settings->type =
				    prefs_account->advanced_settings.email_type;
				msginfo->settings->copy_to_own_address =
				    prefs_account->advanced_settings.copy_to_own_address;
			} else {
				if (msginfo->settings->account != NULL) {
					g_free(msginfo->settings->account);
					msginfo->settings->account =
					    g_strdup(prefs_account->user_settings.account_name);
				} else {
					msginfo->settings->account =
					    g_strdup(prefs_account->user_settings.account_name);
					msginfo->settings->options =
					    prefs_account->advanced_settings.email_sending;
					msginfo->settings->type =
					    prefs_account->advanced_settings.email_type;
					msginfo->settings->copy_to_own_address =
					    prefs_account->advanced_settings.copy_to_own_address;
				}
			}
		}
	}
	return prefs_account;
}

/**
  This function converts the HTML contents specified in the file name to 
  the normal text.
  @param filename    html file
  @param charset    character set
  @param convertedtext    file name contains the converted text
  @return boolean flag TRUE/ FALSE to indicate whether conversion was 
  successful.
 */

gboolean
al_html_to_text_converter(const gchar * filename, const gchar * charset, gchar ** convertedtext)
{
	if ((filename == NULL) || (convertedtext == NULL)) {
		return FALSE;
	}
	return (html_to_text_converter(filename, charset, convertedtext));
}

/**
  This function cleans up all the global memory
  @param none
  @return none
 */

void al_engine_cleanup_memory(void)
{

	smtp_cleanup();
	pop3_cleanup();
	imap_cleanup();
	account_cleanup();
	al_engine_cleanup();
	folder_cleanup();
	mmc_cleanup();
}

/**
  This function is called by Email Manager to get the no of mails present 
  in the outbox 
  @return none
 */

gint al_get_outbox_mail(void)
{
	GSList *flist = NULL;
	GSList *cur = NULL;
	gint count = 0;
	set_engine_errno(0);

	flist = folder_populate_folder(OUTBOX_DIR);
	if (flist == NULL) {
		return count;
	}
	for (cur = flist; cur != NULL; cur = cur->next) {
		if (cur->data == NULL) {
			continue;
		}
		count++;
	}
	while (flist) {
		procmsg_msgheader_free((MsgHeader *) flist->data);
		flist = g_slist_remove(flist, flist->data);
	}
	return count;
}

/**
  This function is called by Email Manager to update the new mail 
  information  
  in INBOX and outbox mail count 
  @return none
 */
void al_update_mail_list(void)
{
	GSList *acc_list = NULL;
	GSList *cur = NULL;

	set_engine_errno(0);
	al_get_account_list(&acc_list);
	if (acc_list == NULL) {
		debug_print("[%s]:[%d] No Account Exist ..\n", __FILE__, __LINE__);
		return;
	}
	for (cur = acc_list; cur != NULL; cur = cur->next) {
		if (cur->data == NULL) {
			continue;
		}
		al_send_recv_mail(cur->data, FALSE);
		if ((engine_errno == ENGINE_ERROR_IAP_CONNECT)
		    || (engine_errno == ENGINE_STATUS_SEND_RECV_CANCEL)) {
			break;
		}
#if CL_USE
		if (cl_current_iap_status(FALSE, 0) == FALSE) {
			break;
		}
#endif
	}
	engine_errno = 0;
	slist_free_strings(acc_list);
	g_slist_free(acc_list);
}

/**
  This function remove if any mails queued for deletion
  @param account_name    name of the account, for which mails have to be 
  deleted 
  @return none
 */
static void al_remove_queued_list(const gchar * account_name, ListType type)
{
	GSList *msgid_list = NULL;
	GSList *cur = NULL;
	PrefsAccount *account = NULL;
	GSList *tmp = NULL;
	GSList *delete_list = NULL;

	engine_errno = ENGINE_ERROR_DEFAULT;
	msgid_list = folder_get_queued_msgid_list(account_name, type);
	account = account_get_account(account_name);
	if (account == NULL) {
		log_message("account not found\n");
		return;
	}

	if (account->user_settings.passwd == NULL) {
		return;
	}

	if (account->receive_settings.protocol == PROTO_IMAP4) {

		for (cur = msgid_list; cur != NULL; cur = cur->next) {

			folder_delete_list_msg((gchar *) cur->data, FALSE);
		}
		UNSET_CB_FLAG;
		imap_expunge_msg(account, msgid_list, NULL, type, NULL, 0, NULL);
		SET_CB_FLAG;

	} else if (account->receive_settings.protocol == PROTO_POP3) {
		if (type == DELETE) {
			engine_errno = ENGINE_ERROR_DEFAULT;

			delete_list = al_sup_retr_msglist(msgid_list);

			if (delete_list != NULL) {

				UNSET_CB_FLAG;
				pop3_recv_mail(account, NULL, FOLDER_OP_DEL, TRUE, delete_list);
				SET_CB_FLAG;
			}

			cur = msgid_list;
			while (cur != NULL) {
				if (engine_errno != ENGINE_ERROR_CONNECT_SERVER) {
					tmp = cur->next;
					folder_delete_list_msg((gchar *) cur->data, FALSE);
					g_free(cur->data);
					msgid_list = g_slist_delete_link(msgid_list, cur);
					cur = tmp;
				} else {
					cur = cur->next;
				}
			}
		}
	}
	procmsg_free_retr_data_list(delete_list);
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);

	return;

}

/**
  This function gets message header information of message id 
  @param messageid_list    list of message ids,for which header
  information required.This is allocated  and freed by the caller. 
  @return list of pointer to the message header information,
  this list has to be freed by the caller. 
 */
GSList *al_get_msg_header(GSList * msgid_list)
{
	GSList *cur = NULL;
	GSList *msghdr_list = NULL;
	MsgHeader *msghdr = NULL;
	gchar *filepath = NULL;
	gchar *folder_name = NULL;
	gchar *part_file = NULL;
	gchar *msgid = NULL;
	MsgInfo *msginfo = NULL;
	gboolean archive_flag = FALSE;
	gchar *msgnum = NULL;

	set_engine_errno(0);
	set_scheduling_op(FALSE);
	if (msgid_list == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}
	for (cur = msgid_list; cur != NULL; cur = cur->next) {
		if (check_msgid(cur->data) == FALSE) {
			engine_errno = ENGINE_ERROR_INVALID_MSGID;
			continue;
		}
		folder_name = get_fold_name(cur->data);

		archive_flag = FALSE;

		if (folder_name && (g_ascii_strncasecmp(folder_name, ARCHIVE, strlen(ARCHIVE)) == 0)
		    && (CHECK_MMC_FLAG == TRUE)) {
			archive_flag = TRUE;
		}
		g_free(folder_name);
		filepath = folder_get_message_file_path(cur->data);
		if (filepath == NULL) {
			engine_errno = ENGINE_ERROR_INVALID_MSGID;
			continue;
		}
		if ((is_file_entry_exist(filepath) == FALSE)
		    && (archive_flag == FALSE)) {

			part_file = g_strdup_printf("%s%c%s", (gchar *) cur->data,
						    SECTION_SEPARATOR, HEADER_SECTION);

			filepath = folder_get_message_file_path(part_file);

			if (filepath && is_file_entry_exist(filepath) == TRUE) {
				msgid = part_file;
			} else {
				g_free(part_file);
				g_free(filepath);
				engine_errno = ENGINE_ERROR_INVALID_MSGID;
				continue;
			}
		} else {
			msgid = cur->data;
		}
		if (archive_flag == TRUE) {
			msgnum = g_strdup_printf("%d", get_msgid_num(msgid));
			g_free(filepath);
			filepath = folder_get_mounted_path(msgnum);
			msginfo = folder_get_info(msgnum, filepath, archive_flag);
			g_free(msgnum);

		} else {
			msginfo = folder_get_info(msgid, filepath, archive_flag);
		}
		g_free(filepath);
		g_free(part_file);
		if (msginfo == NULL) {
			engine_errno = ENGINE_ERROR_INVALID_MSGID;
			continue;

		}
		msghdr = msginfo->header;
		if (msghdr != NULL) {
			g_free(msghdr->msgid);
			msghdr->msgid = g_strdup(cur->data);
		}
		msginfo->header = NULL;
		procmsg_msginfo_free(msginfo);

		msghdr_list = g_slist_append(msghdr_list, (gpointer *) msghdr);
	}
	return msghdr_list;

}


/**
  This function to backup or restores the autosave mail from temp folder.  
  @param msgid    message id which is to be copied or restored.The format 
  of message id is accountname_foldername_subfoldername_messagenumber. 
  @param operation    backup/restore the autosave mail
  @return TRUE on success of operation or FALSE on failure 
 */

gint al_autosave_mail(const gchar * msgid, const gint operation)
{
	gchar *filename = NULL;
	gchar *tmp = NULL;
	gint rv = TRUE;

	set_engine_errno(0);
	if (check_msgid(msgid) == FALSE) {
		engine_errno = ENGINE_ERROR_INVALID_MSGID;
		return FALSE;
	}

	filename = folder_get_message_file_path(msgid);
	if (filename == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		debug_print(" al_autosave_mail engine_errno : %d \n", engine_errno);
		return FALSE;
	}

	tmp = g_strdup_printf("%s%c%s", get_tmp_dir(), G_DIR_SEPARATOR, msgid);

	if (operation == AUTOSAVE_BACKUP) {
		if (is_file_entry_exist(filename) != TRUE) {
			g_free(filename);
			g_free(tmp);
			engine_errno = ENGINE_ERROR_MSG_FILE_DOES_NOT_EXIST;
			return FALSE;
		}
		if (copy_file(filename, tmp, FALSE) != 0) {
			rv = FALSE;
		}
	} else if (operation == AUTOSAVE_RESTORE) {
		if (move_file(tmp, filename, TRUE) != 0) {
			rv = FALSE;
		}
	}
	g_free(filename);
	g_free(tmp);
	return rv;
}

/**
  This function is called when message is opened for editing purpose.
  So that when send/receive operation performed mails opened for 
  editing will not be sent.
  @param msgid    message id which is to be copied or restored.The format
  of message id is accountname_foldername_subfoldername_messagenumber. 
  @param option   to add or delete to the open editor mail list
  @return TRUE if able to open mail 
 */

gboolean al_editor_open_mail(const gchar * msgid, const gint option)
{
	if (check_msgid(msgid) == FALSE) {
		engine_errno = ENGINE_ERROR_INVALID_MSGID;
		return FALSE;
	}
	return send_editor_open_mail(msgid, option);
}


/**
  This function is used to get the size of folder 
  @param folder_name    folder name for which size is required 
  @return size    of the folder,AL_SUP_FAIL(-1) on error 
 */
gint al_get_folder_size(const gchar * folder_name)
{

	Folder *folder = NULL;
	gint size = 0;

	if (folder_name == NULL) {
		return AL_SUP_FAIL;
	}
	if (strcmp(folder_name, ARCHIVE) == 0) {


		size = folder_get_arch_size(NULL, TRUE);

	} else {
		folder = folder_find_obj_by_name(folder_name);
		if (folder == NULL) {
			return AL_SUP_FAIL;
		}
		if (folder->path == NULL) {
			return AL_SUP_FAIL;
		}
		size = folder_get_arch_size(folder->path, FALSE);
	}

	return size;
}

/**
  This function returns the account status.This is used to check whether 
  account is editable or not.
  @param account_name    name of the account to check the status   
  @return account status if account is found else AL_SUP_FAIL(-1) 
*/
AccntStatusFlags al_check_accnt_status(const gchar * account_name)
{
	PrefsAccount *account = NULL;

	if (account_name == NULL) {
		return AL_SUP_FAIL;
	}
	account = (PrefsAccount *) account_get_account_cb((gchar *)
							  account_name);
	if (account == NULL) {
		return AL_SUP_FAIL;
	}
	return account->advanced_settings.account_status;
}

gint al_prefs_account_get_list(GSList ** prefs_account_list)
{
	GSList *tmp_list = NULL;

	if (prefs_account_list == NULL) {
		return AL_SUP_FAIL;
	}
	tmp_list = account_get_prefs_account_list();
	if (tmp_list == NULL) {
		return AL_SUP_FAIL;
	}
	*prefs_account_list = tmp_list;
	return AL_SUP_SUCCESS;
}

/*MsgInfo **/
void
al_recv_multipart_attachment(const gchar * msgid, MimeInfo * mimeinfo,
			     gint attach_flag, const gchar * default_font)
{

	MsgInfo *msginfo = NULL;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	gchar *file_name = NULL;
	gboolean archive_flag = FALSE;

	engine_errno = ENGINE_ERROR_DEFAULT;
	set_scheduling_op(FALSE);

	if ((msgid == NULL) || (mimeinfo == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		gl_al_cb_ptr->al_error_code_cb(engine_errno);
		return;
	}
	if ((mimeinfo->mime_type != MIME_MESSAGE_RFC822)
	    || (mimeinfo->sub == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		gl_al_cb_ptr->al_error_code_cb(engine_errno);
		return;
	}
    /** check for archive folder **/
	if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
		gl_al_cb_ptr->al_error_code_cb(engine_errno);
		return;
	}

	account_name = (gchar *) get_account_name(msgid);

	if (account_name == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_MSGID;
		gl_al_cb_ptr->al_error_code_cb(engine_errno);
		return;
	}

	account = (PrefsAccount *) account_get_account_cb(account_name);
	g_free(account_name);

	if (account != NULL) {
		/* the attachment is downloaded and decoded selectively */
		file_name = folder_get_message_file_path(msgid);
		if (!file_name || is_file_exist(file_name) == FALSE) {
			msginfo =
			    al_sup_get_imap_mhtml_attachment(mimeinfo, msgid,
							     account, default_font);
			g_free(file_name);
			if ((msginfo != NULL) && (msginfo->header != NULL)) {
				msginfo->header->attach_tag = attach_flag;
			}
			gl_al_cb_ptr->al_recv_attachment_cb(msginfo, attach_flag);
			procmsg_msginfo_free(msginfo);
			return;
		}
		g_free(file_name);
	}

	/* The attachment is already downloaded */
	msginfo = al_sup_get_mhtml_attach(mimeinfo, msgid, default_font);

	/* delete temporary archive file */
	if ((del_arch_file(msgid, archive_flag) == AL_SUP_FAIL)
	    && (msginfo != NULL)) {
		procmsg_msginfo_free(msginfo);
		gl_al_cb_ptr->al_error_code_cb(engine_errno);
		return;
	}
	if ((msginfo != NULL) && (msginfo->header != NULL)) {
		msginfo->header->attach_tag = attach_flag;
	}
	gl_al_cb_ptr->al_recv_attachment_cb(msginfo, attach_flag);
	procmsg_msginfo_free(msginfo);
	return;
}


GSList *al_folder_remove_all_attachment(GSList * msgid_list)
{
	GSList *cur = NULL;
	GSList *succ_list = NULL;
	GSList *ret_list = NULL;
	GSList *attach_list = NULL;
	MimeInfo *hdr_mimeinfo = NULL;
	gchar *file_path = NULL;
	gchar *mime_file_name = NULL;
	gchar *msgid = NULL;
	gchar *folder_name = NULL;
	gint ret_count = 0;
	gint attach_count = 0;
	gboolean archive_flag = FALSE;
	FILE *infp = NULL;
	MimeInfo *mimeinfo = NULL;
	set_engine_errno(0);
	set_scheduling_op(FALSE);

	for (cur = msgid_list; cur != NULL; cur = cur->next) {

		msgid = cur->data;

		folder_name = get_fold_name(msgid);
		file_path = folder_get_message_file_path(msgid);
		if (file_path)
			mime_file_name = g_strdup_printf("%s%c%s", file_path,
							 SECTION_SEPARATOR, MIME_SUFFIX);

		if (folder_name && (strcmp(folder_name, INBOX_DIR) == 0)
		    && is_file_entry_exist(mime_file_name)) {
			g_free(file_path);
		} else {
			g_free(mime_file_name);
			mime_file_name = NULL;
			if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
				ret_list = g_slist_append(ret_list, g_strdup(msgid));
				continue;
			}
		}
		g_free(folder_name);
		/* anyway,the attachment is for non imap */
		if (mime_file_name == NULL) {

			if (file_path && ((infp = fopen(file_path, "rb")) == NULL)) {
				ret_list = g_slist_append(ret_list, g_strdup(msgid));
				g_free(file_path);
				continue;
			}
			mimeinfo = procmime_scan_message(infp, -1, 0, 0, NULL);
			fclose(infp);
			g_free(file_path);

			if (mimeinfo == NULL) {
				ret_list = g_slist_append(ret_list, g_strdup(msgid));
				continue;
			}
			attach_list = al_sup_get_all_attach_list(mimeinfo,
								 TRUE, &attach_count, msgid, FALSE);

			/* if error occurs,add to the return error list */
			if (al_remove_msg_attachments(msgid, attach_list) == AL_SUP_FAIL) {
				ret_list = g_slist_append(ret_list, g_strdup(msgid));
			}
			procmime_mimeinfo_free_all(mimeinfo);
			g_slist_free(attach_list);
			continue;
		}

		hdr_mimeinfo = procmime_read_mimeinfo(mime_file_name);
		procmime_set_is_downloaded(hdr_mimeinfo, msgid);
		g_free(mime_file_name);

		if (hdr_mimeinfo == NULL) {
			continue;
		}
		if (procmime_mime_attachment(hdr_mimeinfo) == FALSE) {
			ret_list = g_slist_append(ret_list, g_strdup(msgid));
			procmime_mimeinfo_free_all(hdr_mimeinfo);
			continue;
		}
		attach_count = 0;

		if (hdr_mimeinfo != NULL) {
			mimeinfo = hdr_mimeinfo;
		}
		attach_list = al_sup_get_all_attach_list(mimeinfo,
							 TRUE, &attach_count, msgid, TRUE);

		if (attach_list == NULL) {
			ret_list = g_slist_append(ret_list, g_strdup(msgid));
		}

		succ_list = al_folder_remove_attachment(attach_list);
		ret_count = 0;

		ret_count = g_slist_length(succ_list);


		if (ret_count != attach_count) {
			ret_list = g_slist_append(ret_list, g_strdup(msgid));
		}
		procmime_mimeinfo_free_all(hdr_mimeinfo);
		slist_free_strings(succ_list);
		g_slist_free(succ_list);
		g_slist_foreach(attach_list, (GFunc) al_free_list, NULL);
		g_slist_free(attach_list);
	}
	return ret_list;
}

static void al_free_list(FolderAttachInfo * folder_attach, gpointer data)
{
	if (folder_attach == NULL)
		return;

	g_free(folder_attach->msgid);
	g_free(folder_attach->section_number);
	return;
}

void al_folder_cancel_mail(void)
{
	gl_al_folder_cancel = TRUE;
	if (gl_cur_account != NULL) {
		if (gl_cur_account->receive_settings.protocol == PROTO_POP3) {
			pop3_recv_cancel();

		} else if (gl_cur_account->receive_settings.protocol == PROTO_IMAP4) {
			imap_recv_cancel();

		}
	}
}
gint al_set_temp_passwd(const gchar * account_name, const gchar * passwd, gboolean cancel)
{
	PrefsAccount *prefs_account = NULL;

	if (account_name == NULL) {
		return AL_SUP_FAIL;
	}
	prefs_account = (PrefsAccount *) account_get_account_cb((gchar *)
								account_name);

	if (prefs_account == NULL) {
		return AL_SUP_FAIL;
	}
	if (cancel == TRUE) {
		prefs_account->user_settings.cancel = TRUE;
		return AL_SUP_SUCCESS;
	}
	if (passwd == NULL) {
		prefs_account->user_settings.tmp_passwd = g_strdup(NO_PASSWD);
	} else {
		prefs_account->user_settings.tmp_passwd = g_strdup(passwd);
	}
	return AL_SUP_SUCCESS;
}

#if CL_USE
void al_iap_cancel(void)
{
	GSList *tmp_list = NULL;
	PrefsAccount *prefs_account = NULL;

	imap_recv_cancel();
	pop3_recv_cancel();
	send_cancel_mails(NULL);

	tmp_list = account_get_prefs_account_list();

	if (tmp_list == NULL) {
		return;

	}

	for (; tmp_list != NULL; tmp_list = tmp_list->next) {

		/* disconnect all imap sessions connected to IAP */
		prefs_account = (PrefsAccount *) tmp_list->data;

		if ((prefs_account->user_settings.session != NULL) &&
		    (prefs_account->receive_settings.protocol == PROTO_IMAP4)) {

			if (prefs_account == imap_get_cur_account()) {
				continue;
			}

			imap_session_disconnect(prefs_account->user_settings.session, TRUE);
			prefs_account->user_settings.session = NULL;
		}
	}
	engine_errno = ENGINE_STATUS_SEND_RECV_CANCEL;
}
#endif


void al_cleanup_sessions(void)
{

	imap_recv_cancel();
	pop3_recv_cancel();
	send_cancel_mails(NULL);
	gl_al_folder_cancel = TRUE;
	engine_errno = ENGINE_STATUS_SEND_RECV_CANCEL;
}

gint al_remove_msg_attachments(const gchar * msgid, GSList * rem_fpos_list)
{
	gchar *file_path = NULL;
	gchar *folder_name = NULL;
	gchar *tmp_file = NULL;
	gboolean archive_flag = FALSE;
	gint pos = -1;
	gint level = 0;
	gint ret_val = 0;
	glong fpos = 0;
	FILE *infp = NULL;
	FILE *outfp = NULL;
	MimeInfo *mimeinfo = NULL;
	GSList *cur = NULL;

	set_engine_errno(0);
	set_scheduling_op(FALSE);
	if (check_archive(msgid, &archive_flag) == AL_SUP_FAIL) {
		return AL_SUP_FAIL;
	}
	file_path = folder_get_message_file_path(msgid);

	if (!file_path || is_file_entry_exist(file_path) == FALSE) {
		g_free(file_path);
		return AL_SUP_FAIL;
	}
	if ((infp = fopen(file_path, "rb")) == NULL) {
		g_free(file_path);
		return AL_SUP_FAIL;
	}
	/* open the output file */
	tmp_file = g_strdup_printf("%s%c%4x%s",
				   get_tmp_dir(), G_DIR_SEPARATOR, (gint) msgid, msgid);

	outfp = fopen(tmp_file, "w");
	if (outfp == NULL) {
		send_engine_error_to_ui(errno);
		g_free(tmp_file);
		fclose(infp);
		g_free(file_path);
		return AL_SUP_FAIL;
	}

	mimeinfo = procmime_scan_message(infp, pos, level, fpos, NULL);

	if (mimeinfo == NULL) {
		g_free(tmp_file);
		fclose(infp);
		fclose(outfp);
		g_free(file_path);
		return AL_SUP_FAIL;
	}

	cur = rem_fpos_list;
	if (cur->data == NULL) {
		ret_val = AL_SUP_FAIL;
	}
	fpos = GPOINTER_TO_INT(cur->data);
	ret_val = al_sup_rem_part(infp, outfp, fpos, rem_fpos_list, mimeinfo, msgid);
	fclose(infp);
	fclose(outfp);

	procmime_mimeinfo_free_all(mimeinfo);
	if (ret_val == AL_SUP_FAIL) {
		g_free(file_path);
		return ret_val;
	}
	/* move the temporary file back to original file */
	if (archive_flag == TRUE) {
		folder_name = get_fold_name(msgid);
		g_free(file_path);
		if (folder_name) {
			/* get the archival file path */
			file_path = folder_get_file_path(folder_name, msgid, FALSE, NULL);
			g_free(folder_name);
		} else {
			g_free(tmp_file);
			return AL_SUP_FAIL;
		}

	}
	if (folder_mmc_move_file(tmp_file, file_path, TRUE) != FOLDER_SUCCESS) {

		g_free(tmp_file);
		g_free(file_path);
		return AL_SUP_FAIL;
	}
	g_free(file_path);
	g_free(tmp_file);
	if (del_arch_file(msgid, archive_flag) == AL_SUP_FAIL) {
		return AL_SUP_FAIL;
	}

	return ret_val;
}

gchar *al_folder_get_archive_path(void)
{
	gchar *archive_path = NULL;

	archive_path = folder_get_archive_path();

	return archive_path;
}

void al_send_recv_mail(const gchar * account_name, gboolean sched_op)
{
	engine_errno = ENGINE_ERROR_DEFAULT;

	al_send_outbox_mail(account_name, NULL, FALSE);
	if ((engine_errno == ENGINE_ERROR_IAP_CONNECT)
	    || (engine_errno == ENGINE_STATUS_SEND_RECV_CANCEL)) {
		return;
	}
	al_recv_mail(account_name, sched_op);
}

void al_send_get_msg_id(MsgInfo * msginfo, gboolean send_flag)
{
	PrefsAccount *prefs_account = NULL;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return);

	prefs_account = al_get_account_info(msginfo);
	if (prefs_account) {
		if (send_get_msg_id(msginfo, prefs_account, send_flag) == FALSE) {
			debug_print("Problem is finding message ID\n");
		}
	}

	else {
		prefs_account = account_get_dummy_account();
		if (send_get_msg_id(msginfo, prefs_account, send_flag) == FALSE) {
			debug_print("Problem is finding message ID\n");
		}
	}
}

gint al_retr_msgs(GSList * msg_list)
{
	gchar *msgid = NULL;
	GSList *cur = NULL;
	GSList *sorted_list = NULL;
	GSList *cur_account_list = NULL;
	gint succ_val = AL_SUP_SUCCESS;
	gint ret_val = AL_SUP_SUCCESS;
	gchar *account = NULL;
	gchar *prev_account = NULL;

	/* adjust the list in sorting order */
	sorted_list = folder_sort_msglist(msg_list, TRUE);
	set_scheduling_op(FALSE);

	for (cur = sorted_list; cur != NULL; cur = cur->next) {

		msgid = ((RetrMsgData *) cur->data)->msgid;

		if (check_msgid(msgid) == FALSE) {
			succ_val = AL_SUP_FAIL;
			continue;
		}
		account = get_account_name(msgid);

		if (prev_account == NULL) {

			prev_account = g_strdup(account);
		}
		if (strcmp(account, prev_account) != 0) {

			ret_val = al_sup_retr_list(cur_account_list, prev_account);
			if (ret_val == AL_SUP_FAIL) {
				succ_val = AL_SUP_FAIL;
			}

			g_free(prev_account);

			prev_account = account;

			g_slist_free(cur_account_list);

			cur_account_list = NULL;

			cur_account_list = g_slist_append(cur_account_list, cur->data);

			continue;

		}
		cur_account_list = g_slist_append(cur_account_list, cur->data);

		g_free(account);
	}
	ret_val = al_sup_retr_list(cur_account_list, prev_account);
	g_free(prev_account);

	if (ret_val == AL_SUP_FAIL) {
		succ_val = AL_SUP_FAIL;
	}
	g_slist_free(cur_account_list);
	return succ_val;
}

static void al_folder_send_max_msg(gint type, GSList * folder_list)
{
#if 0
	GSList *list = NULL;
	gint counter = 0, len = 0, length = 0;
#endif

	if (folder_list == NULL) {
		debug_print("Nothing to send to UI\n");
		return;
	}
#if 0
	len = g_slist_length(folder_list);
	for (counter = 0; counter < len; counter++) {
		list = g_slist_append(list, g_slist_nth_data(folder_list, counter));
		if ((g_slist_length(list) == NUM_MAILS)
		    || (counter == len - 1)) {
			FOLDER_MSG_LIST_CB(type, list);
			length = g_slist_length(list);

			/* Remove contents of the list but do not free it
			 * contents will be freed when folder_list is freed.
			 * List pointer is freed
			 */
			g_slist_free(list);
			list = NULL;
		}
	}
#endif
	FOLDER_MSG_LIST_CB(type, folder_list);
}
GSList *al_recv_multiple_attachments(const gchar * msgid, GSList * mimeinfo_list,
				     GSList * attach_tag_list)
{
	GSList *cur_node = NULL;
	gchar *file_path = NULL;
	GSList *msg_info_list = NULL;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	MsgInfo *msg_info = NULL;
	gint num = 0;
	gint index = 0;
	gchar *folder_name = NULL;
	gboolean archive_flag = FALSE, unretrieved = FALSE;
	MimeInfo *mime_info = NULL, *part_mimeinfo = NULL;

	folder_name = get_fold_name(msgid);
	set_scheduling_op(FALSE);

	if (folder_name && (strcmp(folder_name, ARCHIVE) == 0)) {
		archive_flag = TRUE;
	}
	g_free(folder_name);


	file_path = folder_get_message_file_path(msgid);

	/*check if something has to retrieved in the list of attachments */
	for (cur_node = mimeinfo_list; cur_node != NULL; cur_node = cur_node->next) {
		if (!((MimeInfo *) cur_node->data)->is_downloaded) {
			unretrieved = TRUE;
			break;
		}
	}

	if (!file_path || (!is_file_entry_exist(file_path) && !archive_flag && unretrieved)) {
		account_name = (gchar *) get_account_name(msgid);

		if (account_name == NULL) {
			engine_errno = ENGINE_ERROR_INVALID_MSGID;
			g_free(file_path);
			return NULL;
		}

		account = (PrefsAccount *) account_get_account_cb(account_name);
		g_free(account_name);
		if (account == NULL) {
			g_free(file_path);
			return NULL;
		}
		if (account->receive_settings.protocol == PROTO_POP3) {
			g_free(file_path);
			return NULL;
		}
		/* if the multiple attachments are to be downloaded for IMAP,use single connection */
		if (imap_fetch_all_attachments(msgid, account, mimeinfo_list) == IMAP_FAIL) {
			g_free(file_path);
			return NULL;

		}

	}
	/*update the mimeinfo if the mail is completely downloaded after the attachments are retrieved*/
        if(is_file_entry_exist(file_path) &&  unretrieved)
                mime_info = procmime_get_mimeinfo(msgid, file_path, FALSE, TRUE);
	g_free(file_path);

	for (cur_node = mimeinfo_list; cur_node != NULL; cur_node = cur_node->next, index++) {
		/*get the updated mimeinfo*/
		if(mime_info)
                        part_mimeinfo = al_sup_get_sec_mimeinfo(mime_info,
				       	((MimeInfo *)cur_node->data)->section_number);
                else 
                       part_mimeinfo = (MimeInfo *)cur_node->data;
                

		msg_info = al_recv_msg_attachment(msgid, part_mimeinfo);
		if (msg_info == NULL) {
			procmime_mimeinfo_free_all(mime_info);
			procmsg_msginfo_list_free(msg_info_list);
			g_slist_free(msg_info_list);
			return NULL;
		}
		num = GPOINTER_TO_INT(g_slist_nth_data(attach_tag_list, index));
		if (msg_info->header == NULL) {
			msg_info->header = g_new0(MsgHeader, 1);
		}
		msg_info->header->attach_tag = num;
		msg_info->header->msgid = g_strdup(msgid);
		msg_info_list = g_slist_append(msg_info_list, msg_info);
	}

	procmime_mimeinfo_free_all(mime_info);
	return msg_info_list;

}
