
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <glib.h>
#include <glib.h>
#include <sys/stat.h>
#include <string.h>
#include <strings.h>

#include "al.h"
#include "folder.h"
#include "imap.h"
#include "defs.h"
#include "procheader.h"
#include "procmime.h"
#include "al_support.h"
#include "utils.h"
#include "prefs_account.h"
#include "engine_errno.h"
#include "pop_recv.h"
#include "account.h"

#include "smime_security.h"

extern GSList *gl_unavailable_list;
/* Globals */
extern gint engine_errno;
/* defined in utils.c */

extern gboolean gl_callback_enable_flag;
/* defined in utils.c */

extern gboolean gl_mmc_status_flag;
/* defined in utils.c */

extern gint gl_total_count;
extern gint gl_cur_count;
/* defined in pop_recv.c */

/* static globals */
extern AlCallback *gl_al_cb_ptr;
extern PrefsAccount *gl_cur_account;
extern gboolean gl_al_folder_cancel;

/* Static functions */

static MsgInfo *al_sup_get_msginfo(FILE * fp,
				   const gchar * sub_filename,
				   MimeInfo * mimeinfo, const gchar * msgid);
static MsgInfo *al_sup_get_rfc822_msginfo(const gchar * enc_filename,
					  const gchar * msgid, const gchar * dec_filename);

static MsgInfo *al_sup_imap_get_attach(const gchar * enc_filename,
				       MimeInfo * mimeinfo, gboolean nested, const gchar * msgid);

static MsgInfo *al_sup_imap_rfc822_attach(const gchar * enc_filename,
					  const gchar * dec_filename, const gchar * msgid);

static gboolean al_sup_is_last_attachment(const gchar * msgid, const gchar * sec_num);

/**
  this function parses the mime header part, returns Content-ID value
  @param mime_info    mimeinformation 
  @param msgid message id
  @return Content-ID value (to be freed by the calling function)
*/
static gchar *al_sup_get_content_id(MimeInfo * mime_info, const gchar * msgid);

static gint al_sup_download_complete(PrefsAccount * prefs_account, const gchar * msgid);

/**
  This function is a callback function,whcih will be called by the imap 
  module to send progress information during folder operations.
  @param cur    current message number 
  @param total    total message number 
  @return none 
*/
static void al_sup_send_progress(gint cur, gint total);

/**
  This function gets the Mime information in form of 
  MimeInfo structure

  @param msgid - Message ID in which mimeinfo is filled
  @param filename - Email content file
  @return MimeInfo - Mime information
*/
static MimeInfo *al_sup_get_smime_mimeinfo(const gchar * msgid, const gchar * filename);


/**
  This function prepares the MimeInfo structure and fills all the required
  S/MIME fields in the mimeinfo structure

  @param msgid - ID of the email
  @param outfile - The decrypted/verified contents
  @param smime_type - Whether mail is signed/encrypted or both
  @param trust - Whether certificate is trusted or not
  @param saved - Whether certificate needs to be saved or not
  @param cert_path - certificate path in case cert is to be saved
  @param private_cert_state - Certificate state
  @param recip_cert_state - Certificate state
  @param signature - Whether signature is verified or not
  @param smime - type of smime mail
*/
static MimeInfo *al_sup_fill_mimeinfo(const gchar * msgid,
				      const gchar * outfile, gboolean trust,
				      gboolean saved, gchar * cert_path,
				      gint private_cert_state,
				      gint recip_cert_state,
				      gboolean signature, gint smime,
				      gint sender_cert_id, gint recip_cert_id);

/**
  This function performs operation on the list of messages from the local 
  directory & from server,if the message belongs to POP3 account  
  @param account    account name of the message 
  @param folder_append_list    OUT parameter,pointer to list of messages, 
                               on which operation is successfully performed 
  @param folder_update_list_ptr    OUT paramater,pointer to list of source 
                                   messages, on which operation is successfull 
  @param move_flag    indicates whether the operation is move 
  @param total    total message number 
  @param srcfolder    source folder for the operation  
  @param prefs_account account information of the messages in the list
  @param destfolder    destination folder for the operation  
  @param cur_ptr OUT parameter,pointer to the current message number  
  @param folder_del_list_ptr    OUT paramater,pointer to list of source 
                              messages, which have been deleted from server 
  @return AL_SUP_SUCCESS on success or AL_SUP_FAIL on failure 
*/
static gint al_sup_folder_pop3_op(GSList * account_list,
				  GSList ** folder_append_list_ptr,
				  GSList ** folder_update_list_ptr,
				  gboolean move_flag, gint total,
				  const gchar * srcfolder,
				  PrefsAccount * prefs_account,
				  const gchar * destfolder,
				  gint * cur_ptr, GSList ** folder_del_list_ptr);

/**
  This function creates a mhtml file for the message body part of 
  email attachment 
  @param msgid    message id of the message file
  @param enc_filename    filename of the message part 
  @return  the mhtml file in which the body of the email attachment 
  is stored or NULL on failure.
*/
static gchar *al_sup_get_imap_mimepart(const gchar * msgid,
				       const gchar * enc_filename, const gchar * default_font);

/**
  This function performs the remote delete opration for the account
  and frees the meory allocated to the list and account 
  @param prev_account    name of the account,to perform operation on.This will be freed  by this function. 
  @param folder_delete_list_ptr    OUT parameter,pointer to list of messages, 
  on which operation is successfully performed 
  @param cur_account_list    list of messages,sorted according to account name.
  This will be freed by this function. 
  @param total   total number of messages 
  @return FOLDER_SUCCESS or FOLDER_FAIL 
*/
static gint al_sup_del_remote_list_free(gchar * prev_account,
					GSList ** folder_delete_list_ptr,
					GSList * cur_account_list, gint total);


/**
  This function gives list of messages that are not available locally

  @param full_list - Contains list of msgids
  @return GSList - List of messages that are locally available
*/

/**
  Function updates the header info to be sent to UI 

  @param full_list - List of msgids
  @param hdr_list - List to update
  @return GSList* - List pointer containing msgheaders
*/
static
GSList *al_sup_update_available_list(GSList * full_list, GSList * hdr_list);

/**
  Function to return list of mails to be downloaded from the server

  @param full_list - List of all mails to be checked
  @param GSList - List of mails to be downloaded from server
*/
static
GSList *al_sup_unavailable_list(GSList * full_list);

/**
  Function updates the header info to be sent to UI

  @param msgid - msgid
  @param list - List to update
  @return list returned with header info
*/
static
GSList *al_sup_add_update_list(gchar * msgid, GSList * list);

/**
  Function to move and copy POP3 mails or normal mails

  @param move_flag - TRUE for move operation
  @param srcfolder - source folder path
  @param destfolder - dest folder path
  @param prefs_account - Account information
  @param account_list - List of mails to be moved/copied
  @param remote - Whether mail has remote copy or not
  @param folder_append_list_ptr - List of headers to be added in UI
  @param cur_ptr - Number of mails for which operation is completed
  @param total - total number of mails
  @param folder_update_list_ptr - List of updated headers
  @param folder_del_list_ptr - List of headers to be deleted
*/
static
gint al_sup_copy_move_pop_msg(gboolean move_flag, const gchar * srcfolder,
			      const gchar * destfolder, PrefsAccount * prefs_account,
			      GSList * account_list, gboolean remote,
			      GSList ** folder_append_list_ptr, gint * cur_ptr,
			      gint total, GSList ** folder_update_list_ptr,
			      GSList ** folder_del_list_ptr);


/**
  Check if list contains remote mails or fully downloaded mails

  @param full_list - List of message IDs
  @return TRUE if mails are remote else false
*/
static
gboolean al_sup_remote_mails(GSList * full_list);

/**
  This function performs a search through the mimeinformation 
  and return pointer to the mimeinformation node which is having the
  same file position as passed. 
  @param mimenfo mime information
  @param pos    filepoistion value to match 
  @return mime information 
*/
static
MimeInfo *al_sup_get_matching_mimeinfo(MimeInfo * mimeinfo, glong pos);

/**
  Function to return path where email attachment is saved

  @param fp - File pointer of the file containing email attachment
  @param mimeinfo - Mimeinformation of the email attachment
  @return gchar* - path of file containing the email attachment
*/
static
gchar *al_sup_get_email_attach_path(FILE * fp, MimeInfo * mimeinfo);

/**
  This function is used to unlink the file  

  @param  fname file name to be removed
  @return nothing 
*/
static 
void al_sup_unlink_file( gchar *fname);

/**
 Helper function to get the mimeinfo corresponding to the section num
 */
MimeInfo *
al_sup_get_sec_mimeinfo(MimeInfo *mimeinfo, const gchar * sec_num)
{
	MimeInfo *partinfo = NULL;
	partinfo = mimeinfo;

	while(partinfo && partinfo->section_number){
		if(strcmp(sec_num,partinfo->section_number) == 0){
			return partinfo;
		}
		partinfo = procmime_mimeinfo_next(partinfo);
	}
	return partinfo;
}

/**
This function downloads the entire mail & gets the attachment for the give 
section_number
*/
static MsgInfo*
get_attachment( PrefsAccount *account, const gchar *msgid, const gchar *file_path,
		const gchar *section_number)
{
	MimeInfo *mime_info = NULL;
	MsgInfo *msginfo = NULL;

	/*download completely & then get the attachment part*/
	al_sup_download_complete(account, msgid);
	if(!is_file_exist(file_path)){
                        return NULL;
	}
	mime_info = procmime_get_mimeinfo(msgid, file_path, FALSE, TRUE);
	mime_info = al_sup_get_sec_mimeinfo(mime_info,section_number);
	msginfo = al_sup_get_mhtml_attach(mime_info, msgid, NULL);
	procmime_mimeinfo_free_all(mime_info);
	return msginfo;
}
/**
  This function downloads the attachment selectively for IMAP and decodes it. 
  @param mimeinfo    the mimeinformation of the part or attachment            
  @param msgid    message id of the message
  @param account    the account information pointer in the global account list,
  for which the attachment has to be downloaded.Global account list created and 
  deleted by account management functions. 
  @return Message information pointer,this has to be freed by the caller
 */
MsgInfo *al_sup_get_imap_selective_attachment(MimeInfo * mimeinfo,
					      const gchar * msgid, PrefsAccount * account)
{
	gchar *enc_filename = NULL;
	gchar *dec_filename = NULL;
	MsgInfo *msginfo = NULL;
	gchar *file_path = NULL;
	gboolean nested = FALSE;
	gboolean complete_flag = FALSE;

	if ((msgid == NULL) || (mimeinfo == NULL) || (account == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}


	if (mimeinfo->parent != NULL) {
		if ((mimeinfo->parent->section_number != NULL) &&
		    (mimeinfo->section_number != NULL)) {


			if (strcmp(mimeinfo->parent->section_number, mimeinfo->section_number) == 0) {
				/* this is the case of nested */

				nested = TRUE;
			}
		}
	}
	file_path = folder_get_message_file_path(msgid);
	if (file_path == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		return NULL;
	}
	enc_filename = g_strdup_printf("%s%c%s",
				       file_path, SECTION_SEPARATOR, mimeinfo->section_number);

	if (is_file_entry_exist(enc_filename) == FALSE) {

		complete_flag = al_sup_is_last_attachment(msgid, mimeinfo->section_number);
		if(complete_flag == TRUE){
			/*download completely & then get the attachment part*/
			msginfo = get_attachment(account, msgid, file_path,mimeinfo->section_number);
			g_free(file_path);
		        g_free(enc_filename);
			return msginfo;
		}else {
			imap_fetch_msg_attachment(account, (gchar *) msgid,
					  mimeinfo->section_number, FALSE, NULL, mimeinfo, TRUE);
		}
	}

	if (mimeinfo->mime_type == MIME_MESSAGE_RFC822) {

		GET_DECODE_FILENAME(msgid, mimeinfo, dec_filename);

		msginfo = al_sup_imap_rfc822_attach(enc_filename, dec_filename, msgid);
		if (msginfo == NULL) {
			g_free(dec_filename);
		}
	} else {
		msginfo = al_sup_imap_get_attach(enc_filename, mimeinfo, nested, msgid);
		complete_flag = al_sup_is_last_attachment(msgid, mimeinfo->section_number);
	}

	g_free(enc_filename);
	if (complete_flag == TRUE) {
		al_sup_download_complete(account, msgid);
	}
	return msginfo;
}

/**
  This function downloads the attachment selectively for IMAP and decodes it. 
  @param mimeinfo the mimeinformation of the part or attachment.           
  @param msgid message id of the message.
  @return pointer to the attachment information,caller has to free the 
   memory allocated by this function 
 */
MsgInfo *al_sup_get_attachment_info(MimeInfo * mimeinfo, const gchar * msgid)
{
	gchar *enc_filename = NULL;
	gchar *decoded_file = NULL;
	AttachInfo *attach_info = NULL;
	MsgInfo *msginfo = NULL;

	gchar *final_file = NULL;
	engine_errno = ENGINE_ERROR_DEFAULT;

	if ((mimeinfo == NULL) || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}

	enc_filename = folder_get_message_file_path(msgid);
	if (is_file_entry_exist(enc_filename) == FALSE) {
		g_free(enc_filename);
		engine_errno = ENGINE_ERROR_FILE_NOT_FOUND;
		return NULL;
	}

	/* check if message is smime and decrypt the message */
	final_file = return_file_if_smime(msgid);
	if (final_file == NULL) {
		engine_errno = ENGINE_ERROR_FILE_NOT_FOUND;
		g_free(enc_filename);
		return NULL;
	}

	/* Decode the attachment file */
	decoded_file = procmime_get_part(final_file, mimeinfo, FALSE);
	g_free(enc_filename);

	msginfo = g_new0(MsgInfo, 1);

	if (msginfo == NULL) {
		engine_errno = ENGINE_ERROR_NO_MEM;
		return NULL;
	}

	attach_info = g_new0(AttachInfo, 1);

	if (attach_info == NULL) {
		g_free(decoded_file);
		engine_errno = ENGINE_ERROR_NO_MEM;
		procmsg_msginfo_free(msginfo);
		return NULL;
	}
	attach_info->type = DIRECT;
	attach_info->filepath = decoded_file;
	attach_info->msgid = g_strdup(msgid);
	attach_info->attachmentname = g_strdup(mimeinfo->name);

	msginfo->attach_list = g_slist_append(msginfo->attach_list, (gpointer)
					      attach_info);
	remove_smime_file(final_file);
	g_free(final_file);

	return msginfo;
}

/**
  This function will be called by AL layer to decode the message
  if the receive type is BODY for imap.
  @param pointer to the account structure in the global account list,
  for which the attachment has to be downloaded.
  Global account list created and deleted by account management functions.  
  @param msgid message id of the message
  @return mimeinfo the mimeinformation of the body, the memory allocated
  by this function for mime information has to be freed by the caller
 */
MimeInfo *al_sup_get_imap_body(PrefsAccount * account, const gchar * msgid,
			       gboolean edit_flag, gboolean get_html_flag)
{
	gint retval = AL_SUP_SUCCESS;
	gchar *dec_filename = NULL;
	gchar *mhtml_file = NULL;
	MimeInfo *partinfo = NULL;
	gchar *enc_filename = NULL;
	MimeInfo *mimeinfo = NULL;
	FILE *infp = NULL;
	gchar *mime_file_name = NULL;
	gchar *file_path = NULL;
	gboolean part_body = TRUE;
	gboolean mhtml_body = FALSE;
	Session *session = NULL;
	gboolean session_cancelled = FALSE;

	file_path = folder_get_message_file_path(msgid);
	if (file_path == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		return NULL;
	}
	mime_file_name = g_strdup_printf("%s%c%s", file_path, SECTION_SEPARATOR, MIME_SUFFIX);

	if (is_file_entry_exist(mime_file_name) == FALSE) {
		mimeinfo = imap_fetch_msg_body_structure(account, (gchar *) msgid, &session);
		if(imap_is_session_cancelled()){
                        if(is_file_entry_exist(mime_file_name))
                               unlink(mime_file_name);
                       session_cancelled = TRUE;
               }
	} else {
		mimeinfo = procmime_read_mimeinfo(mime_file_name);
	}
	g_free(mime_file_name);
	if (mimeinfo == NULL) {
		g_free(file_path);
		/* disconnect the session */
		if(session)
	               imap_session_complete(session);
		return NULL;
	}

	partinfo = procmime_get_text_ptr(mimeinfo, FALSE);

	if (partinfo == NULL) {
		g_free(file_path);
		/* disconnect the session */
		if(session)
	               imap_session_complete(session);
		return mimeinfo;
	}


	enc_filename = g_strdup_printf("%s%c%s",
				       file_path, SECTION_SEPARATOR, partinfo->section_number);

	if (is_file_entry_exist(enc_filename) == FALSE) {
		mhtml_file = g_strdup_printf("%s%c%s%c%s%c%s",
					     get_mail_dir(), G_DIR_SEPARATOR,
					     INBOX_DIR, G_DIR_SEPARATOR,
					     msgid, SECTION_SEPARATOR, MHTML_EXTENSION);
		if (is_file_entry_exist(mhtml_file) == TRUE) {

			mhtml_body = TRUE;
			part_body = FALSE;

		} else if (procmime_is_smime(mimeinfo) == TRUE) {
			imap_cmd_fetch_msg_body((gchar *) msgid, account, FALSE,
						TRUE, TRUE, session, TRUE);
			procmime_mimeinfo_free_all(mimeinfo);
			/* read the one from file */
			mimeinfo = procmime_get_mimeinfo(msgid, file_path, edit_flag, get_html_flag);
			part_body = FALSE;

		} else if (procmime_mime_attachment(mimeinfo) == TRUE) {
			imap_fetch_msg_attachment(account, (gchar *) msgid,
						  partinfo->section_number,
						  TRUE, session, mimeinfo, TRUE);

			/*remove the file when session is cancled*/
			if(imap_is_session_cancelled())
				if(is_file_entry_exist(mhtml_file))
					unlink(mhtml_file);
					               
			if (is_file_entry_exist(mhtml_file) == TRUE) {
				mhtml_body = TRUE;
			} else {
				part_body = TRUE;
			}

		} else {
			imap_cmd_fetch_msg_body((gchar *) msgid, account, FALSE,
						TRUE, FALSE, session, TRUE);
			procmime_mimeinfo_free_all(mimeinfo);
			/* read the one from file */
			mimeinfo = procmime_get_mimeinfo(msgid, file_path, edit_flag, get_html_flag);
			part_body = FALSE;

		}
	}
	g_free(file_path);

	if (mhtml_body == TRUE) {

		mimeinfo->filename = decode_mhtml_file(mhtml_file,msgid,
							      edit_flag, get_html_flag);

	} else if (part_body == TRUE) {

		infp = fopen(enc_filename, "r+");
		if (infp == NULL) {
			g_free(enc_filename);
			g_free(dec_filename);
		        g_free(mhtml_file);
			return NULL;
		}
		GET_DECODE_FILENAME(msgid, partinfo, dec_filename);
		retval = procmime_get_content(partinfo, infp, dec_filename, part_body, edit_flag);
		fclose(infp);
		if (retval != AL_SUP_FAIL) {
			g_free(mimeinfo->filename);
			mimeinfo->filename = dec_filename;
		} else {
			g_free(dec_filename);
			procmime_check_and_set_filename(mimeinfo,NULL);
		}
	}
	g_free(enc_filename);
	g_free(mhtml_file);

	if (folder_message_del_list((gchar *) msgid, UNREAD) != FOLDER_SUCCESS) {
		engine_errno = ENGINE_ERROR_FOLDER_MESSAGE_DEL_LIST_FAILED;
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	return mimeinfo;
}

/**
  This function will be called by AL layer to decode the attachment if the
  attachment type is RFC822.
  @param mimeinfo    the mimeinformation of the body 
  @param msgid    message id of the message
  @return pointer to the message information of the attachment,
  which has to be freed by caller 
 */
MsgInfo *al_sup_rfc822_attach(MimeInfo * mimeinfo, const gchar * msgid)
{
	MsgInfo *msginfo = NULL;
	gchar *sub_filename = NULL;
	FILE *fp = NULL;
	MimeInfo *partinfo = NULL;
	MimeInfo *sub_mimeinfo = NULL;
	gchar *final_file = NULL;

	engine_errno = ENGINE_ERROR_DEFAULT;
	if ((mimeinfo == NULL) || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}

	/* This is a case of message type of attachment */
	sub_mimeinfo = mimeinfo->sub;
	if (sub_mimeinfo == NULL) {
		engine_errno = ENGINE_ERROR_NULL_SUB_MIME;
		return NULL;
	}
	partinfo = procmime_get_text_ptr(sub_mimeinfo, FALSE);
	GET_SUB_DECODE_FILENAME(msgid, sub_filename);


	final_file = return_file_if_smime(msgid);
	if (final_file == NULL) {
		engine_errno = ENGINE_ERROR_MSG_FILE_DOES_NOT_EXIST;
		g_free(sub_filename);
		return NULL;
	}

	/* decode the message file */
	fp = fopen(final_file, "r+");

	if (fp == NULL) {
		g_free(sub_filename);
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}
	if (partinfo != NULL) {
		if (procmime_get_content(partinfo, fp, sub_filename, FALSE, FALSE) < 0) {
			fclose(fp);
			g_free(sub_filename);
			g_free(final_file);
			engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
			return NULL;
		}
		rewind(fp);
	} else {

		g_free(sub_filename);
		sub_filename = NULL;
	}
	msginfo = al_sup_get_msginfo(fp, sub_filename, mimeinfo, msgid);

	fclose(fp);
	if (msginfo == NULL) {
		g_free(sub_filename);
	}
	remove_smime_file(final_file);
	g_free(final_file);

	return msginfo;
}

/**
  This function will be called by AL layer to get the mhtml part of file 
  @param mimeinfo    pointer to the mime information.This has to be 
  allocated and freed by the caller. 
  @param fp    pointer to the message file 
  @param outfp    pointer to the output file 
  @param root_flag    indicates whether the beginning of the part to check 
  @return whether the contents are written successfully(AL_SUP_SUCCESS)
  or not(AL_SUP_FAIL) 
 */
gint
al_sup_get_mhtml_contents(MimeInfo * mimeinfo, FILE * fp, FILE * outfp,
			  gboolean root_flag, gboolean copy_hdrs, const gchar * default_font)
{
	gchar buf[BUFFSIZE];
	gchar *boundary = NULL;
	gint boundary_len = 0;
	gboolean boundary_copy = TRUE;

	if ((mimeinfo == NULL) || (outfp == NULL)) {
		return AL_SUP_FAIL;
	}

	memset(buf, '\0', sizeof(buf));


	/* copy other headers and data */
	if ((mimeinfo->fpos != 0) && (copy_hdrs == TRUE)) {
		procheader_copy_headers(fp, outfp);
	}
	if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) {
		log_message((const gchar *) ("Mime:Error in contents of the file"));
		return AL_SUP_FAIL;
	}
	if ((mimeinfo->parent != NULL)
	    && (mimeinfo->parent->boundary != NULL)) {
		boundary = mimeinfo->parent->boundary;
		boundary_len = strlen(boundary);
		if (root_flag == TRUE) {
			boundary_copy = FALSE;
		}
	} else if (mimeinfo->boundary != NULL) {
		boundary = mimeinfo->boundary;
		boundary_len = strlen(boundary);
	}
	while ((fgets(buf, sizeof(buf), fp) != NULL) &&
	       (!boundary || !IS_BOUNDARY(buf, boundary, boundary_len))) {

		fputs(buf, outfp);
	}
	/* write the boundary */
	if ((boundary != NULL) && (boundary_copy == TRUE)) {
		fputs(buf, outfp);
	}

	return AL_SUP_SUCCESS;
}

/**
  This function will be called by AL layer to get the mhtml part of file 
  @param mimeinfo    pointer to the mime information 
  @param infp    pointer to the message file 
  @param outfp    pointer to the output file 
  @param root_flag    flag to indicate whether the beginning of 
  the part to check. 
  @return whether the contents are written successfully(AL_SUP_SUCCESS)
  or not(AL_SUP_FAIL) 
 */
gint
al_sup_get_mhtml_file(MimeInfo * mimeinfo, FILE * infp, FILE * outfp,
		      gboolean root_flag, const gchar * default_font)
{
	gchar buf[BUFFSIZE];
	gboolean recursive_flag = FALSE;

	if ((mimeinfo == NULL) || (outfp == NULL) || (infp == NULL)) {
		return AL_SUP_FAIL;
	}
	memset(buf, '\0', sizeof(buf));

	while (mimeinfo != NULL) {
		if (mimeinfo->mime_type == MIME_MULTIPART) {
			/* write the body part to the file */
			if (al_sup_get_mhtml_contents
			    (mimeinfo, infp, outfp, root_flag, TRUE, default_font)
			    == AL_SUP_FAIL) {

				return AL_SUP_FAIL;
			}
			recursive_flag = TRUE;
			if (mimeinfo->parent != NULL) {
				if (mimeinfo->parent->boundary != NULL) {
					recursive_flag = FALSE;

				}
			}
			if (recursive_flag == TRUE) {
				if (al_sup_get_mhtml_file
				    (mimeinfo->children, infp, outfp,
				     FALSE, default_font) == AL_SUP_FAIL) {
					return AL_SUP_FAIL;
				}
			}

		} else {
			/* write the body part to the file */
			if (al_sup_get_mhtml_contents
			    (mimeinfo, infp, outfp, root_flag, TRUE, default_font) == AL_SUP_FAIL) {
				return AL_SUP_FAIL;
			}
		}
		if (root_flag == FALSE) {
			mimeinfo = mimeinfo->next;
		} else {
			break;
		}
	}
	return AL_SUP_SUCCESS;
}

/**
  This function decodes the message body,for IMAP  
  @param account    pointer to the account structure in global account list,
   for which the attachment has to be downloaded.Global account list created and
    deleted by account management functions. 
  @param msgid     pointer to the message id 
  @param complete_download     indicates whether the message 
  has to be completely downloaded 
  @param signer_email_id - Senders email ID
  @param recip_email_id - recipient email ID
  @return NULL on failure or mimeinfo read from file.
  The memory allocated for return value has to be freed by calling function 
 */
MimeInfo *al_sup_recv_body_imap(PrefsAccount * account, const gchar * msgid,
				gboolean complete_download, SMimeType smime_type,
				gchar * signer_email_id, gchar * recip_email_id,
				gboolean edit_flag, gboolean get_html_flag)
{
	MimeInfo *mimeinfo = NULL;
	gchar *filename = NULL;

	if (smime_type != SMIME_NONE) {
		complete_download = TRUE;
		al_sup_download_complete(account, msgid);
		filename = folder_get_message_file_path(msgid);

		/*check if mail is downloaded, if not return with out further process*/
		if(!is_file_exist(filename))
			return NULL;
		if (NULL == (mimeinfo = al_sup_get_smime_message_info(filename,
								      signer_email_id,
								      recip_email_id, msgid))) {
			debug_print("Mimeinfo NULL\n");
			g_free(filename);
			return NULL;
		}
		g_free(filename);
		return mimeinfo;
	}


	else if (complete_download == TRUE) {
		al_sup_download_complete(account, msgid);
		filename = folder_get_message_file_path(msgid);
		if(!filename) {
			engine_errno = ENGINE_ERROR_INVALID_FOLDER;
			return NULL;
		}

		mimeinfo = procmime_get_mimeinfo(msgid, filename, edit_flag, get_html_flag);
		g_free(filename);
		return mimeinfo;
	}
	if (account->receive_settings.recv_type == HEADERS) {
		mimeinfo = al_sup_get_imap_body(account, msgid, edit_flag, get_html_flag);

		return mimeinfo;
	}
	if (account->receive_settings.recv_type == BODY_ATTCH) {
		filename = g_strdup_printf("%s%c%s%c%s%c%s",
					   get_mail_dir(), G_DIR_SEPARATOR,
					   INBOX_DIR, G_DIR_SEPARATOR, msgid,
					   SECTION_SEPARATOR, MSG_BODY_SEC);

		if (is_file_entry_exist(filename) == TRUE) {
			g_free(filename);
			mimeinfo = al_sup_get_imap_body(account, msgid, edit_flag, get_html_flag);
			return mimeinfo;
		}
		g_free(filename);
		filename = folder_get_message_file_path(msgid);

		if (is_file_entry_exist(filename) == FALSE) {
			imap_cmd_fetch_msg_body((gchar *) msgid, account, FALSE,
						TRUE, FALSE, NULL, TRUE);
		}


	} else if (account->receive_settings.recv_type == BODY) {
		filename = folder_get_message_file_path(msgid);

		if (is_file_entry_exist(filename) == FALSE) {
			mimeinfo = al_sup_get_imap_body(account, msgid, edit_flag, get_html_flag);

			if (folder_message_del_list((gchar *) msgid, UNREAD) != FOLDER_SUCCESS) {
				engine_errno = ENGINE_ERROR_FOLDER_MESSAGE_DEL_LIST_FAILED;
				procmime_mimeinfo_free_all(mimeinfo);
				return NULL;
			}
			return mimeinfo;
		}
	}
	mimeinfo = procmime_get_mimeinfo(msgid, filename, edit_flag, get_html_flag);
	g_free(filename);
	return mimeinfo;
}

/**
  This function decodes the message body,for IMAP  
  @param msgid    pointer to the message id 
  @param mimeinfo    pointer to the mime information 
  @return mhtml file.NULL on failure .
  The filename returned must be freed by the calling function. 
 */
gchar *al_sup_extract_mhtml(const gchar * msgid, MimeInfo * mimeinfo,
			    const gchar * filename, const gchar * default_font)
{
	struct stat s;
	FILE *outfp = NULL;
	FILE *fp = NULL;
	gint ret_val = 0;
	gchar *mhtml_file = NULL;


	if (stat(filename, &s) < 0) {
		engine_errno = ENGINE_ERROR_FILE_STAT_FAILED;
		return NULL;
	}
	if ((s.st_mode & S_IRUSR) == 0) {
		gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_NO_FILE_PERM);
		return NULL;
	}

	fp = fopen(filename, "r");

	if (fp == NULL) {
		return NULL;
	}

	/* open the output file */
	mhtml_file = g_strdup_printf("%s%c%s%c%s",
				     get_tmp_dir(), G_DIR_SEPARATOR, msgid,
				     SECTION_SEPARATOR, MHTML_EXTENSION);

	outfp = fopen(mhtml_file, "wb");
	if (outfp == NULL) {
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else
			engine_errno = ENGINE_ERROR_OUT_FILE;
		g_free(mhtml_file);
		fclose(fp);
		return NULL;
	}
	ret_val = al_sup_get_mhtml_file(mimeinfo, fp, outfp, TRUE, default_font);

	fclose(outfp);
	fclose(fp);

	if (ret_val == AL_SUP_FAIL) {
		g_free(mhtml_file);
		engine_errno = ENGINE_ERROR_FILE_CONTENTS;
		unlink(mhtml_file);
		return NULL;
	}
	return mhtml_file;
}

/**
  This function downloads the message to be moved 
  @param account    pointer to the account information in global account list,
  for which the attachment has to be downloaded.Global account list created and
  deleted by account management functions. 
  @param msgid    pointer to the message id 
  @param srcfolder    source folder name,in which the message is present 
  @param move_msg_flag    indicates whether operation is moving or copying 
  @return AL_SUP_SUCCESS(0) on success AL_SUP_FAIL(-1) on failure 
 */
gint
al_sup_folder_recv_msg(PrefsAccount * prefs_account, const gchar * msgid,
		       const gchar * srcfolder, gboolean move_msg_flag, const gchar * file_path)
{
	FolderOpernType operation = FOLDER_OP_COPY;


	if ((prefs_account == NULL) || (msgid == NULL)
	    || (srcfolder == NULL)) {
		return AL_SUP_FAIL;
	}

	if (file_path == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		return AL_SUP_FAIL;
	}
	if (prefs_account->receive_settings.protocol == PROTO_POP3) {
		if (move_msg_flag == TRUE) {
			operation = FOLDER_OP_MOVE;
		}
		UNSET_CB_FLAG;
		pop3_recv_mail(prefs_account, msgid, operation, FALSE, NULL);
		SET_CB_FLAG;

		if (engine_errno == ENGINE_ERROR_POP_MSG_DELETE) {
			return AL_SUP_NOT_AVAILABLE;
		}
	}
	return AL_SUP_SUCCESS;

}

/**
  This function removes the message on the server 
  @param account    pointer to the account information in global account list,
  for which the attachment has to be downloaded.Global account list created and
  deleted by account management functions. 
  @param rem_msgid    pointer to the message id 
  @param local_copy_flag    indicates whether message local or remote 
  @param total    total number of messages 
  @return AL_SUP_SUCCESS(0) on success AL_SUP_FAIL(-1) on failure 
 */
gint
al_sup_folder_rem_msg(PrefsAccount * prefs_account, GSList * cur_list,
		      GSList ** folder_succ_list, gint total)
{
	gchar *file_name = NULL;
	GSList *tmp = NULL;
	gchar *rem_msgid;
	gint cur_count = 0;
	GSList *del_list = NULL;
	gchar *header_file = NULL;

	if ((prefs_account == NULL) || (cur_list == NULL)) {
		return FOLDER_REMOVE_MSGS_ERROR;
	}
	engine_errno = ENGINE_ERROR_DEFAULT;
	if (prefs_account->receive_settings.protocol == PROTO_IMAP4) {

		UNSET_CB_FLAG;
		imap_expunge_msg(prefs_account, cur_list, folder_succ_list, DELETE,
				 &cur_count, total, al_sup_send_progress);
		SET_CB_FLAG;

	} else if (prefs_account->receive_settings.protocol == PROTO_POP3) {
		gl_cur_account = prefs_account;

		del_list = al_sup_retr_msglist(cur_list);

		/* will have to discard globals for this and pass it 
		 * through pop3_recv_mail
		 */
		gl_cur_count = cur_count;
		gl_total_count = total;

		/*delete all mails at one go */
		UNSET_CB_FLAG;
		pop3_recv_mail(prefs_account, NULL, FOLDER_OP_DEL, FALSE, del_list);
		SET_CB_FLAG;

		for (tmp = cur_list; tmp != NULL; tmp = tmp->next) {

			rem_msgid = tmp->data;
			file_name = folder_get_message_file_path(rem_msgid);
			if (file_name != NULL) {
				header_file = g_strdup_printf("%s%c%s", file_name, DOT, MSG_HDR);
				if (folder_succ_list != NULL) {
					/*if message not deleted: header or body
					 *  dont add it to delete list
					 */
					if ((is_file_entry_exist(file_name) == TRUE) ||
					    (is_file_entry_exist(header_file) == TRUE)) {
						continue;
					}
					folder_delete_list_msg(rem_msgid, TRUE);
					folder_message_del_list(rem_msgid, UNREAD);

					*folder_succ_list =
					    folder_append_header_list(*folder_succ_list,
								      tmp->data, NULL,
								      prefs_account, TRUE,
								      NULL, NULL);
				}
			}
			g_free(file_name);
			file_name = NULL;
			g_free(header_file);
			header_file = NULL;
		}
		gl_cur_account = NULL;
		gl_cur_count = 0;
		gl_total_count = 0;
		procmsg_free_retr_data_list(del_list);

	}
	return FOLDER_SUCCESS;
}

/**
  This function gets thei decoded rfc822 imap message attachment   
  @param enc_filename    name of the message file 
  @param dec_filename    name of the file to store decoded contents 
  @param msgid    message id of the message 
  @return pointer to the message information containing the attachment info.
  The memory allocated for this message Information has to be freed i
  by calling function. 
 */
static MsgInfo *al_sup_imap_rfc822_attach(const gchar * enc_filename,
					  const gchar * dec_filename, const gchar * msgid)
{
	MimeInfo *partinfo = NULL;
	MimeInfo *sub_mimeinfo = NULL;
	MsgInfo *msginfo = NULL;
	FILE *fp;
	gint pos = -1;
	gint level = 0;
	glong fpos = 0;
	GSList *file_pos_list = NULL;
	gchar *final_file = NULL;


	/* decode the message file */
	fp = fopen(enc_filename, "r+");

	if (fp == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}

	sub_mimeinfo = (MimeInfo *) procmime_scan_message(fp, pos, level, fpos, file_pos_list);

	if (sub_mimeinfo == NULL) {
		fclose(fp);
		engine_errno = ENGINE_ERROR_MIME_INVALID;
		return NULL;
	}
	partinfo = procmime_get_text_ptr(sub_mimeinfo, FALSE);
	if (partinfo != NULL) {
		rewind(fp);
		if (procmime_get_content(partinfo, fp, dec_filename, FALSE, FALSE) < 0) {
			fclose(fp);
			engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
			return NULL;
		}
		final_file = (gchar *) dec_filename;
	} else {

		final_file = NULL;
	}
	fclose(fp);
	msginfo = al_sup_get_rfc822_msginfo(enc_filename, msgid, final_file);

	procmime_mimeinfo_free_all(sub_mimeinfo);

	return msginfo;
}

/**
  This function gets the decoded imap message attachment   
  @param filename    name of the file containing attachment 
  @param mimeinfo    pointer to the mimeinformation of the  
  @param nested    this flag indicates whether it is nested attachment  
  @param msgid    message identifier of the message 
  @return pointer to the message information containing the attachment
  information. This has to be freed by the calling function. 
 */
static MsgInfo *al_sup_imap_get_attach(const gchar * enc_filename, MimeInfo * mimeinfo,
				       gboolean nested, const gchar * msgid)
{
	MsgInfo *msginfo = NULL;
	gchar *decoded_file = NULL;
	AttachInfo *attach_info = NULL;

	msginfo = g_new0(MsgInfo, 1);

	if (msginfo == NULL) {
		engine_errno = ENGINE_ERROR_NO_MEM;
		return NULL;
	}
	/* check whether the attachment is a nested one */
	if (nested == TRUE) {
		decoded_file = procmime_get_part(enc_filename, mimeinfo, FALSE);
	} else {
		decoded_file = procmime_get_part(enc_filename, mimeinfo, TRUE);
	}

	attach_info = g_new0(AttachInfo, 1);

	if (attach_info == NULL) {
		procmsg_msginfo_free(msginfo);
		g_free(decoded_file);
		engine_errno = ENGINE_ERROR_NO_MEM;
		return NULL;
	}

	attach_info->type = DIRECT;
	attach_info->filepath = decoded_file;
	attach_info->msgid = g_strdup(msgid);
	attach_info->attachmentname = g_strdup(mimeinfo->name);

	msginfo->attach_list = NULL;
	msginfo->attach_list = g_slist_append(msginfo->attach_list, (gpointer) attach_info);

	return msginfo;
}

/**
  This function gets the decoded message information for email attachment   
  @param fp    pointer to the message file,to be opened and closed by the
  calling function. 
  @param sub_filename    name of the file containing attachment 
  @param mimeinfo    pointer to the mimeinformation of the message.
  This is allocated and freed by the calling function. 
  @param msgid    message id of the file  
  @return pointer to the message information containing the attachment 
  information. This has to be freed by the calling function. 
*/
static MsgInfo *al_sup_get_msginfo(FILE * fp, const gchar * sub_filename,
				   MimeInfo * mimeinfo, const gchar * msgid)
{
	MsgInfo *msginfo = NULL;
	AttachInfo *attach_info = NULL;
	MsgFlags flags;
	gchar buf[BUFFSIZE];
	gchar *email_path = NULL;
	struct stat s;

	email_path = al_sup_get_email_attach_path(fp, mimeinfo);

	/* get the message header information */
	if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}

	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (buf[0] == '\r' || buf[0] == '\n') {
			break;
		}
	}
	msginfo = procheader_parse_stream(fp, flags, TRUE);
	if (msginfo == NULL) {
		engine_errno = ENGINE_ERROR_PARSE_FAIL;
		return NULL;
	}

	msginfo->attach_list = NULL;
	attach_info = g_new0(AttachInfo, 1);

	if (attach_info == NULL) {
		procmsg_msginfo_free(msginfo);
		engine_errno = ENGINE_ERROR_NO_MEM;
		g_free(email_path);
		return NULL;
	}

	g_free(msginfo->header->msgid);
	msginfo->header->msgid = g_strdup(msgid);
	attach_info->type = EMAIL;
	attach_info->filepath = (gchar *) sub_filename;
	attach_info->email_path = (gchar *) email_path;
	attach_info->msgid = g_strdup(msgid);
	attach_info->attachmentname = NULL;
	msginfo->attach_list = g_slist_append(msginfo->attach_list, (gpointer) attach_info);
	/*fill the size and creation time*/
        if(stat(email_path,&s) == 0){
                msginfo->header->size = s.st_size;
                msginfo->header->mtime = s.st_mtime;
        }
	return msginfo;
}

/**
  This function gets the message information of rfc822 type for imap   
  @param enc_filename    name of the file containing attachment 
  @param msgid    message identifier  
  @param dec_filename    name of the file for storing decoded attachment 
  @return pointer to the message information containing the attachment 
  information. This has to be freed by the calling function. 
 */
static MsgInfo *al_sup_get_rfc822_msginfo(const gchar * enc_filename, const gchar * msgid,
					  const gchar * dec_filename)
{
	gint msg_num = 0;
	struct stat s;
	MsgFlags flags;
	MsgInfo *msginfo = NULL;
	AttachInfo *attach_info = NULL;
	gchar *email_path = NULL;

	msg_num = get_msgid_num(msgid);

	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;

	msginfo = (MsgInfo *) procheader_parse_file(enc_filename, flags, TRUE);

	if (msginfo == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
		return NULL;
	}
	if (stat(enc_filename, &s) < 0) {
		engine_errno = ENGINE_ERROR_FILE_STAT_FAILED;
		return NULL;
	}

	msginfo->header->msgnum = msg_num;
	msginfo->header->size = s.st_size;
	msginfo->header->mtime = s.st_mtime;

	/* Set message id & account information */
	msginfo->header->msgid = g_strdup(msgid);
	msginfo->header->msgtag.remote_copy = FALSE;
	attach_info = g_new0(AttachInfo, 1);

	if (attach_info == NULL) {
		procmsg_msginfo_free(msginfo);
		engine_errno = ENGINE_ERROR_NO_MEM;
		return NULL;
	}

	email_path = get_tmp_file();
	copy_file(enc_filename, email_path, TRUE);

	attach_info->type = EMAIL;
	attach_info->filepath = (gchar *) dec_filename;
	attach_info->msgid = g_strdup(msgid);
	attach_info->attachmentname = NULL;
	attach_info->email_path = (gchar *) email_path;

	msginfo->attach_list = NULL;
	msginfo->attach_list = g_slist_append(msginfo->attach_list, (gpointer) attach_info);

	return msginfo;
}

/**
  This function checks if its the last attachment to be downloaded 
  @param msgid message id
  @param sec_num section number of the attachment 
  @return returns TRUE if it is the last attachment to be downloaded 
 */
static gboolean al_sup_is_last_attachment(const gchar * msgid, const gchar * sec_num)
{
	MimeInfo *partinfo = NULL;
	gchar *file_path = NULL;
	gchar *enc_filename = NULL;
	gchar *mime_file_name = NULL;
	MimeInfo *mimeinfo = NULL;

	if ((msgid == NULL) || (sec_num == NULL)) {
		return FALSE;
	}

	file_path = folder_get_message_file_path(msgid);
	if (file_path == NULL) {
		return FALSE;
	}
	mime_file_name = g_strdup_printf("%s%c%s%c%s%c%s",
					 get_mail_dir(),
					 G_DIR_SEPARATOR, INBOX_DIR,
					 G_DIR_SEPARATOR, msgid, SECTION_SEPARATOR, MIME_SUFFIX);

	mimeinfo = procmime_read_mimeinfo(mime_file_name);

	g_free(mime_file_name);

	if (mimeinfo == NULL) {
		g_free(file_path);
		return FALSE;
	}

	partinfo = mimeinfo;
	while (partinfo != NULL) {
		if (partinfo->section_number != NULL) {

			enc_filename = g_strdup_printf("%s%c%s",
						       file_path, SECTION_SEPARATOR,
						       partinfo->section_number);

			if (partinfo->content_disposition != NULL) {

				if (strncasecmp
				    (partinfo->content_disposition, ATTACHMENT,
				     strlen(ATTACHMENT)) == 0) {

					if ((is_file_entry_exist(enc_filename) == FALSE) &&
					    (strcmp(partinfo->section_number, sec_num) != 0)) {
						g_free(file_path);
						g_free(enc_filename);
						procmime_mimeinfo_free_all(mimeinfo);
						return FALSE;
					}
				}

			}

			g_free(enc_filename);
		}
		if(partinfo->sub)
			partinfo =  partinfo->next;
		else
			partinfo = procmime_mimeinfo_next(partinfo);
	}
	g_free(file_path);
	procmime_mimeinfo_free_all(mimeinfo);
	return TRUE;
}

/**
  This function downloads the complete message 
  @param prefs_account    pointer to the account information structure 
  in global account list.This list is maintained by account management
  module.  
  @param msgid    message id
  @return AL_SUP_SUCCESS(success) or AL_SUP_FAIL(failure) 
 */
static gint al_sup_download_complete(PrefsAccount * prefs_account, const gchar * msgid)
{

	gint ret_val = AL_SUP_SUCCESS;
	GSList *msgid_list = NULL;
	gchar *srcfolder = INBOX_DIR;
	gchar *file_name = NULL;

	msgid_list = folder_get_msgid_all_files(srcfolder, msgid, NULL, FALSE);

	imap_cmd_fetch_msg_body(msgid, prefs_account, TRUE, FALSE, FALSE, NULL, TRUE);
	imap_send_repopulate();
	file_name = folder_get_message_file_path(msgid);

	/* if file is downloaded successfully,delete parts of the mail */

	if ((is_file_entry_exist(file_name) == TRUE)
	    && (msgid_list != NULL)) {
		folder_remove_msgs(srcfolder, msgid_list);
	}
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);
	g_free(file_name);
	return ret_val;
}


/**
  This function handles the synchronisation of imap account
  @param account    pointer to account information in global
  account list, for which the attachment has to be downloaded.
  Global account list created and deleted by account management functions.  
  @param first    first msgid                
  @param last    last msgid                
  @return FOLDER_SUCCESS if messages can be successfully deleted,
  else error code 
 */
gint al_sup_sync_imap(PrefsAccount * account, gint first, gint last)
{
	gint ret_val = 0;

	if (engine_errno == ENGINE_ERROR_IMAP_SYNC) {
		ret_val = al_folder_delete_msgs(INBOX_DIR,
						account->user_settings.account_name, TRUE);
		if (ret_val != FOLDER_SUCCESS) {
			return ret_val;
		}
		imap_get_headers(account, INBOX_DIR, first, last);
	}
	return FOLDER_SUCCESS;
}


static
MimeInfo *al_sup_get_smime_mimeinfo(const gchar * msgid, const gchar * filename)
{
	FILE *fp = NULL;
	MimeInfo *mimeinfo = NULL;
	MsgInfo *msginfo = NULL;
	MsgFlags flags;
	struct stat s;
	gint msg_num = 0;
	gint pos = -1;
	gint level = 0;
	glong fpos = 0;
	GSList *file_pos_list = NULL;


	if (stat(filename, &s) < 0) {
		engine_errno = ENGINE_ERROR_FILE_STAT_FAILED;
		return NULL;
	}
	if ((s.st_mode & S_IRUSR) == 0) {
		gl_al_cb_ptr->al_error_code_cb(ENGINE_ERROR_NO_FILE_PERM);
		return NULL;
	}

	fp = fopen(filename, "r");

	if (fp == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}
	mimeinfo = (MimeInfo *) procmime_scan_message(fp, pos, level, fpos, file_pos_list);
	fclose(fp);

	if (mimeinfo == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}

	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;

	msginfo = (MsgInfo *) procheader_parse_file(filename, flags, TRUE);
	if (msginfo == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	if (S_ISREG(s.st_mode) == 0) {
		procmsg_msginfo_free(msginfo);
		procmime_mimeinfo_free_all(mimeinfo);
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}

	msg_num = get_msgid_num(msgid);
	msginfo->header->msgnum = msg_num;
	msginfo->header->size = s.st_size;
	msginfo->header->mtime = s.st_mtime;

	/* Set message id & account information */
	g_free(msginfo->header->msgid);
	msginfo->header->msgid = g_strdup(msgid);
	mimeinfo->filename = procmime_get_smime_text_content(msginfo->header, filename);
	procmsg_msginfo_free(msginfo);

	if (folder_message_del_list((gchar *) msgid, UNREAD) != FOLDER_SUCCESS) {
		engine_errno = ENGINE_ERROR_FOLDER_MESSAGE_DEL_LIST_FAILED;
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}
	return mimeinfo;
}

MsgInfo *al_sup_get_imap_mhtml_attachment(MimeInfo * mimeinfo,
					  const gchar * msgid,
					  PrefsAccount * account, const gchar * default_font)
{
	gchar *enc_filename = NULL;
	MsgInfo *msginfo = NULL;
	gchar *file_path = NULL;
	gchar *temp_filename = NULL;
	gboolean complete_flag = FALSE;

	if ((msgid == NULL) || (mimeinfo == NULL) || (account == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}


	file_path = folder_get_message_file_path(msgid);
	if (file_path == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FOLDER;
		return NULL;
	}
	enc_filename = g_strdup_printf("%s%c%s",
				       file_path, SECTION_SEPARATOR, mimeinfo->section_number);


	if (is_file_entry_exist(enc_filename) == FALSE) {

		complete_flag = al_sup_is_last_attachment(msgid, mimeinfo->section_number);
		if(complete_flag == TRUE){
			/*download completely & then get the attachment part*/
			msginfo = get_attachment(account, msgid, file_path,mimeinfo->section_number);
			g_free(file_path);
		        g_free(enc_filename);
			return msginfo;
		}else {
			imap_fetch_msg_attachment(account, (gchar *) msgid,
					  mimeinfo->section_number, FALSE, NULL, mimeinfo, TRUE);
		}
	}
	g_free(file_path);

	/* generate a temp filename */
	if (IS_MHTML_MESSAGE(mimeinfo->sub)) {
		temp_filename = procmime_get_tmp_mhtml_file(msgid, enc_filename);
		if (!temp_filename || procmime_check_for_style_tag(enc_filename,
						 temp_filename, default_font) == AL_SUP_FAIL) {
			g_free(temp_filename);
			g_free(enc_filename);
			return NULL;
		}
	} else {
		temp_filename = al_sup_get_imap_mimepart(msgid, enc_filename, default_font);
	}

	if (temp_filename == NULL) {
		return NULL;
	}
	/* No decoding is required in this case */
	msginfo = al_sup_get_rfc822_msginfo(enc_filename, msgid, temp_filename);
	al_sup_split_to_parts(enc_filename, msgid);

	if (msginfo == NULL) {
		g_free(enc_filename);
		g_free(temp_filename);
	}

	if (complete_flag == TRUE) {
		al_sup_download_complete(account, msgid);
	}
	return msginfo;
}


MsgInfo *al_sup_get_mhtml_attach(MimeInfo * mimeinfo, const gchar * msgid,
				 const gchar * default_font)
{
	MsgInfo *msginfo = NULL;
	gchar *out_filename = NULL;
	FILE *fp = NULL;
	FILE *outfp = NULL;
	gint ret = AL_SUP_FAIL;
	MimeInfo *partinfo = NULL;
	gchar *final_file = NULL;

	engine_errno = ENGINE_ERROR_DEFAULT;
	if ((mimeinfo == NULL) || (msgid == NULL)) {
		engine_errno = ENGINE_ERROR_INVALID_PARAM;
		return NULL;
	}

	/* check if S/MIME message and get file path */
	final_file = return_file_if_smime(msgid);
	if (final_file == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}

	/* decode the message file */
	fp = fopen(final_file, "r+");

	if (fp == NULL) {
		g_free(final_file);
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}
	out_filename = procmime_get_tmp_mhtml_file(msgid, NULL);
	if(!out_filename) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		g_free(final_file);
		fclose(fp);
		return NULL;
	}

	outfp = fopen(out_filename, "w");
	if (outfp == NULL) {
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else
			engine_errno = ENGINE_ERROR_OUT_FILE;
		g_free(out_filename);
		g_free(final_file);
		fclose(fp);
		return NULL;
	}
	partinfo = mimeinfo->sub;
	while (partinfo != NULL) {
		if (IS_MHTML_MESSAGE(partinfo) == TRUE) {
			break;
		}
		partinfo = partinfo->children;
	}
	ret = al_sup_get_mhtml_contents(partinfo, fp, outfp, TRUE, FALSE, default_font);
	fclose(outfp);

	if (ret == AL_SUP_FAIL) {
		fclose(fp);
		g_free(out_filename);
		engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
		return NULL;

	}
	msginfo = al_sup_get_msginfo(fp, out_filename, mimeinfo, msgid);

	if (msginfo == NULL) {
		fclose(fp);
		g_free(final_file);
		g_free(out_filename);
		engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
		return NULL;
	}
	fclose(fp);
	remove_smime_file(final_file);
	g_free(final_file);

	return msginfo;
}

GSList *al_support_get_inline_imageinfo_list(MimeInfo * mimeinfo, const gchar * msgid)
{
	GSList *inline_info_list = NULL;
	InlineImageInfo *inline_image_info = NULL;
	AttachInfo *attach_info = NULL;
	MsgInfo *msginfo = NULL;

	if (mimeinfo->children && g_ascii_strncasecmp
	    (mimeinfo->children->content_type, MULTIPART_TYPE, strlen(MULTIPART_TYPE)) != 0) {
		return NULL;
	}

	mimeinfo = procmime_mimeinfo_next(mimeinfo);
	while (mimeinfo != NULL) {
		if (mimeinfo->mime_type == MIME_IMAGE) {
			msginfo = al_sup_get_attachment_info(mimeinfo, msgid);
			if (msginfo != NULL) {
				if (msginfo->attach_list != NULL) {
					inline_image_info = g_new0(InlineImageInfo, 1);
					attach_info = (AttachInfo *) msginfo->attach_list->data;
					inline_image_info->image_path =
					    g_strdup(attach_info->filepath);
					inline_image_info->image_name =
					    mimeinfo->name != NULL ? g_strdup(mimeinfo->name) :
					    mimeinfo->filename != NULL ?
					    g_strdup(mimeinfo->filename) : NULL;
					inline_image_info->content_id =
					    al_sup_get_content_id(mimeinfo, msgid);
					inline_info_list =
					    g_slist_append(inline_info_list, inline_image_info);
				}
			}
		}
		mimeinfo = procmime_mimeinfo_next(mimeinfo);
	}
	return inline_info_list;
}

static gchar *al_sup_get_content_id(MimeInfo * mime_info, const gchar * msgid)
{
	gchar *message_file = NULL;
	FILE *fp = NULL;
	gchar buf[BUFFSIZE];
	gchar *contentid = NULL;
	gchar *tmp = NULL;
	gchar *tmp1 = NULL;

	memset(buf, '\0', sizeof(buf));
	message_file = folder_get_message_file_path(msgid);
	if (is_file_exist(message_file) == TRUE) {
		fp = fopen(message_file, "rb");
		if (fp == NULL) {
			debug_print("%s : %s :could not open file : %s \n", __FILE__,
				    __FUNCTION__, message_file);
			g_free(message_file);
			return NULL;
		}
		g_free(message_file);
		if (fseek(fp, mime_info->fpos, SEEK_SET) == 0) {
			while (fgets(buf, sizeof(buf), fp) != NULL) {
				if (buf[0] == '\r' || buf[0] == '\n') {
					break;
				}
				if (!g_strncasecmp(buf, CONTENTID, CONTENTIDLEN)) {
					debug_print("content id buff : %s \n", buf);
					tmp = strchr(buf, '<');
					tmp1 = strchr(buf, '>');
					contentid = tmp != NULL || tmp1 != NULL
					    ? g_strndup(tmp + 1, (tmp1 - tmp) - 1) : NULL;
					debug_print("contentid = %s \n", contentid);
				}
			}
			fclose(fp);
			return contentid;
		} else {
			debug_print("%s : %s :fseek failed file : %s \n", __FILE__,
				    __FUNCTION__, message_file);
			fclose(fp);
			return NULL;

		}
	} else {
		return NULL;
	}
}
static void al_sup_unlink_file( gchar *fname)
{
	if(is_file_exist(fname)) {
		unlink(fname);
	}
}
gint al_sup_split_to_parts(const gchar * filepath, const gchar * msgid)
{
	FILE *infp = NULL;
	FILE *outfp = NULL;
	MimeInfo *partinfo = NULL;
	MimeInfo *mimeinfo = NULL;
	gint section_number = 0;
	gint ret = AL_SUP_SUCCESS;
	gchar *part_filename = NULL;
	GSList *fileslist = NULL;

	infp = fopen(filepath, "r");
	if (infp == NULL) {
		engine_errno = ENGINE_ERROR_NO_FILE_PERM;
		return AL_SUP_FAIL;
	}
	mimeinfo = procmime_scan_message(infp, -1, 0, 0, NULL);

	rewind(infp);

	section_number = -1;
	partinfo = mimeinfo;

	if (mimeinfo == NULL) {
		engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
		return AL_SUP_FAIL;
	}

	while (partinfo != NULL) {
              if(section_number == -1){
		part_filename =
		    g_strdup_printf("%s%c%s", filepath, SECTION_SEPARATOR, MHTML_EXTENSION);
	      } else {
		part_filename =
		    g_strdup_printf("%s%c%d", filepath, SECTION_SEPARATOR, section_number);
	      }

		outfp = fopen(part_filename, "w+");

		if (outfp == NULL) {
			if (errno == ENOSPC) {
				send_engine_error_to_ui(errno);
			} else
				engine_errno = ENGINE_ERROR_NO_FILE_PERM;
			fclose(infp);
			g_free(part_filename);
			procmime_mimeinfo_free_all(mimeinfo);
			g_slist_foreach(fileslist, (GFunc) al_sup_unlink_file, NULL);
			g_slist_foreach(fileslist, (GFunc) g_free, NULL);
			g_slist_free(fileslist);
			return AL_SUP_FAIL;
		}
		if(section_number == -1){
			/*create the .mime file*/
			ret = al_sup_get_mhtml_file(mimeinfo, infp, outfp, TRUE, NULL);
			rewind(infp);
		}else {
			ret = procmime_copy_part(infp, outfp, partinfo, (section_number == 0)? TRUE : FALSE);
			if (partinfo == mimeinfo ) {
				partinfo = partinfo->children;
			} else {
				partinfo = partinfo->next;
			}
		}
		section_number++;
		if (ret == AL_SUP_FAIL) {
			engine_errno = ENGINE_ERROR_INVALID_FILE_CONTENT;
			fclose(infp);
			fclose(outfp);
			unlink(part_filename);
			g_free(part_filename);
			continue;
		}
		fclose(outfp);
		fileslist = g_slist_append(fileslist, part_filename);

	}
	procmime_mimeinfo_free_all(mimeinfo);
	g_slist_foreach(fileslist, (GFunc) g_free, NULL);
	g_slist_free(fileslist);
	ret = AL_SUP_SUCCESS;
	fclose(infp);
	return ret;
}

gint
al_sup_folder_list_op(GSList * account_list, const gchar * account,
		      const gchar * srcfolder, const gchar * destfolder,
		      gboolean move_flag, GSList ** append_list_ptr,
		      gint * cur_ptr, gint total, GSList ** update_list_ptr, GSList ** del_list_ptr)
{
	GSList *folder_append_list = NULL;
	gint imap_ret = IMAP_ERROR;
	gint succ_val = AL_SUP_SUCCESS;
	gboolean imap_flag = FALSE;
	PrefsAccount *prefs_account = NULL;

	folder_append_list = *append_list_ptr;

	imap_flag = FALSE;
	prefs_account = account_get_account(account);

	if (prefs_account != NULL) {
		if (prefs_account->receive_settings.protocol == PROTO_IMAP4) {
			imap_flag = TRUE;
		}
	}
	if (imap_flag == TRUE) {
		gl_cur_account = prefs_account;

		UNSET_CB_FLAG;
		if (move_flag == TRUE) {
			del_list_ptr = NULL;
		}
		folder_append_list = imap_move_copy_msgs(srcfolder, destfolder,
							 prefs_account,
							 account_list, move_flag,
							 folder_append_list,
							 &imap_ret, cur_ptr, total,
							 al_sup_send_progress,
							 update_list_ptr, del_list_ptr);

		SET_CB_FLAG;
		if (imap_ret != IMAP_SUCCESS) {
			if (imap_ret == IMAP_MESSAGE_NOT_AVAILABLE_IN_SERVER) {
				succ_val = AL_SUP_NOT_AVAILABLE;
			} else {
				succ_val = AL_SUP_FAIL;
			}
		}
		gl_cur_account = NULL;
		*append_list_ptr = folder_append_list;
		return succ_val;

	}
	/* if the account is NULL or pop3,perform the operation  */
	succ_val = al_sup_folder_pop3_op(account_list, append_list_ptr,
					 update_list_ptr, move_flag,
					 total, srcfolder, prefs_account,
					 destfolder, cur_ptr, del_list_ptr);

	return succ_val;
}
static void al_sup_send_progress(gint cur, gint total)
{
	FOLDER_PROGRESS_CB(cur, total);
}

static
MimeInfo *al_sup_fill_mimeinfo(const gchar * msgid, const gchar * outfile,
			       gboolean trust, gboolean saved,
			       gchar * cert_path, gint private_cert_state,
			       gint recip_cert_state, gboolean signature, gint smime,
			       gint sender_cert_id, gint recip_cert_id)
{
	MimeInfo *mimeinfo = NULL;

	CHECK_RETURN_VAL_IF_FAIL(msgid != NULL, NULL);
	CHECK_RETURN_VAL_IF_FAIL(outfile != NULL, NULL);
	CHECK_RETURN_VAL_IF_FAIL(cert_path != NULL, NULL);

	mimeinfo = al_sup_get_smime_mimeinfo(msgid, outfile);

	if (mimeinfo == NULL) {
		debug_print("Mimeinfo NULL\n");
		return NULL;
	}

	mimeinfo->trust = trust;
	mimeinfo->saving = saved;
	mimeinfo->cert_path = g_strdup(cert_path);
	mimeinfo->cert_state_sender = private_cert_state;
	mimeinfo->cert_state_recip = recip_cert_state;
	mimeinfo->signature_verify = signature;
	mimeinfo->smime_type = smime;
	mimeinfo->sender_cert_id = sender_cert_id;
	mimeinfo->recip_cert_id = recip_cert_id;

	return mimeinfo;
}

MimeInfo *al_sup_get_smime_message_info(const gchar * filename,
					const gchar * signer_email_id,
					const gchar * recip_email_id, const gchar * msgid)
{

	MimeInfo *mimeinfo = NULL;
	gchar outfile[BSIZE];
	gchar detached[BSIZE];
	gboolean saved = FALSE;
	gint private_cert_state = 0;
	gint recip_cert_state = 0;
	gboolean signature = TRUE;
	gboolean trust = FALSE;
	gchar cert_path[BSIZE];
	gint smime = 0;
	gint sender_cert_id = 0;
	gint recip_cert_id = 0;


	CHECK_RETURN_VAL_IF_FAIL(filename != NULL, NULL);
	CHECK_RETURN_VAL_IF_FAIL(msgid != NULL, NULL);

	if (0 != (engine_errno = folder_prepare_process_smime(outfile,
							      detached, cert_path, filename, &saved,
							      signer_email_id, recip_email_id,
							      &recip_cert_state,
							      &private_cert_state, &signature,
							      &trust, &smime, &sender_cert_id,
							      &recip_cert_id))) {
		debug_print(" engine_errno : %d \n", engine_errno);
		remove_smime_file(outfile);
		remove_smime_file(detached);
		remove_smime_file(cert_path);
		return NULL;
	}

	if (NULL == (mimeinfo = al_sup_fill_mimeinfo(msgid,
						     outfile, trust, saved, cert_path,
						     private_cert_state, recip_cert_state,
						     signature, smime, sender_cert_id,
						     recip_cert_id))) {
		debug_print("Mimeinfo NULL\n");
		remove_smime_file(outfile);
		remove_smime_file(detached);
		remove_smime_file(cert_path);
		return NULL;
	}

	remove_smime_file(outfile);
	remove_smime_file(detached);

	return mimeinfo;
}

gint
al_sup_accountwise_op(GSList * sorted_list, GSList ** folder_add_list_ptr,
		      GSList ** folder_update_list_ptr, gboolean move_flag,
		      gint total, const gchar * srcfolder,
		      const gchar * destfolder, GSList ** folder_del_list_ptr)
{
	GSList *cur = NULL;
	GSList *cur_account_list = NULL;
	gchar *account = NULL;
	gchar *prev_account = NULL;
	gint ret_val = AL_SUP_SUCCESS;
	gint succ_val = AL_SUP_SUCCESS;
	gint cur_count = 0;

	for (cur = sorted_list; cur != NULL; cur = cur->next) {

		if (gl_al_folder_cancel == TRUE) {
			break;
		}

		if (check_msgid(cur->data) == FALSE) {
			succ_val = AL_SUP_FAIL;
			continue;
		}
		/* remote deletion */
		account = get_account_name(cur->data);
		if (prev_account == NULL) {
			prev_account = g_strdup(account);
		}
		if (strcmp(account, prev_account) != 0) {


			ret_val = al_sup_folder_list_op(cur_account_list, prev_account,
							srcfolder, destfolder,
							move_flag, folder_add_list_ptr,
							&cur_count, total,
							folder_update_list_ptr,
							folder_del_list_ptr);
			g_free(prev_account);

			prev_account = g_strdup(account);

			if (ret_val != AL_SUP_SUCCESS) {

				succ_val = ret_val;
			}

			slist_free_strings(cur_account_list);

			g_slist_free(cur_account_list);

			cur_account_list = NULL;


		}
		cur_account_list = g_slist_append(cur_account_list, g_strdup(cur->data));

		g_free(account);
	}

	if (gl_al_folder_cancel == FALSE) {
		ret_val = al_sup_folder_list_op(cur_account_list, prev_account,
						srcfolder, destfolder, move_flag,
						folder_add_list_ptr, &cur_count,
						total, folder_update_list_ptr, folder_del_list_ptr);
	}

	g_free(prev_account);

	slist_free_strings(cur_account_list);

	g_slist_free(cur_account_list);

	gl_al_folder_cancel = FALSE;

	if (ret_val != AL_SUP_SUCCESS) {
		return ret_val;
	}
	return succ_val;
}

static gint
al_sup_folder_pop3_op(GSList * account_list,
		      GSList ** folder_append_list_ptr,
		      GSList ** folder_update_list_ptr,
		      gboolean move_flag,
		      gint total, const gchar * srcfolder,
		      PrefsAccount * prefs_account,
		      const gchar * destfolder, gint * cur_ptr, GSList ** folder_del_list_ptr)
{
	GSList *cur = NULL;
	gint ret_val = FOLDER_FAIL;
	gint succ_val = AL_SUP_SUCCESS;
	GSList *list_retr_struct = NULL;
	gboolean remote = FALSE;
	GSList *to_retr = NULL;


	if (move_flag == FALSE) {
		/* check if the list sent by UI has headers or fully downloaded
		 * mails Should be done in case of copying
		 */
		remote = al_sup_remote_mails(account_list);
		if (remote == TRUE) {
			/*if headers to be retrieved */
			list_retr_struct = al_sup_retr_msglist(account_list);
		}
	} else {
		/*for moving we need to separate list of headers and
		 * fully downloaded and fetch all headers
		 */
		to_retr = al_sup_unavailable_list(account_list);
		if (to_retr != NULL) {
			list_retr_struct = al_sup_retr_msglist(to_retr);
		}
	}


	if ((list_retr_struct != NULL)&&( prefs_account != NULL )) {
		gl_cur_account = prefs_account;
		UNSET_CB_FLAG;
		pop3_recv_mail(prefs_account, NULL, FOLDER_OP_COPY, FALSE, list_retr_struct);
		SET_CB_FLAG;

		/*if operation is cancelled */
		if (gl_al_folder_cancel == TRUE) {
			if (move_flag == TRUE) {

				/*update information about already downloaded
				 * headers to UI. Use folder_del_list_ptr
				 * list for move operation
				 */
				*folder_del_list_ptr =
				    al_sup_update_available_list(account_list,
								 *folder_del_list_ptr);
			} else {

				/*update downloaded mail info to UI */
				*folder_update_list_ptr =
				    al_sup_update_available_list(account_list,
								 *folder_update_list_ptr);
			}
		}
		gl_cur_account = NULL;
	}

	/*If any message is not unavailable in server handle it */
	if (gl_unavailable_list != NULL) {
		succ_val = AL_SUP_NOT_AVAILABLE;
		for (cur = gl_unavailable_list; cur != NULL; cur = cur->next) {
			folder_delete_list_msg(cur->data, TRUE);

			if (move_flag == TRUE) {

				/*If message to be deleted from UI as it is not
				 * available on server
				 */
				*folder_update_list_ptr =
				    folder_append_header_list(*folder_update_list_ptr,
							      cur->data, NULL, prefs_account,
							      TRUE, NULL, NULL);
			} else {

				/*If message not available on server */
				*folder_del_list_ptr =
				    folder_append_header_list(*folder_del_list_ptr,
							      cur->data, NULL, prefs_account,
							      TRUE, NULL, NULL);
			}
		}
	}

	/*If operation is cancelled free and return */
	if (gl_al_folder_cancel == TRUE) {
		while (gl_unavailable_list != NULL) {
			g_free(gl_unavailable_list->data);
			gl_unavailable_list =
			    g_slist_remove(gl_unavailable_list, gl_unavailable_list->data);
		}
		gl_unavailable_list = NULL;
		procmsg_free_retr_data_list(list_retr_struct);
		g_slist_free(to_retr);
		return succ_val;
	}

	/*all messages available now. start moving or copying now */
	ret_val = al_sup_copy_move_pop_msg(move_flag, srcfolder, destfolder,
					   prefs_account, account_list, remote,
					   folder_append_list_ptr, cur_ptr, total,
					   folder_update_list_ptr, folder_del_list_ptr);

	if ((ret_val != FOLDER_SUCCESS) && (succ_val != AL_SUP_NOT_AVAILABLE)) {
		succ_val = AL_SUP_FAIL;
	}
	while (gl_unavailable_list != NULL) {
		g_free(gl_unavailable_list->data);
		gl_unavailable_list = g_slist_remove(gl_unavailable_list,
						     gl_unavailable_list->data);
	}
	gl_unavailable_list = NULL;
	procmsg_free_retr_data_list(list_retr_struct);
	g_slist_free(to_retr);

	return succ_val;
}

static
GSList *al_sup_unavailable_list(GSList * full_list)
{
	GSList *list = NULL;
	GSList *cur = NULL;
	gchar *file_path = NULL;
	gchar *folder_name = NULL;
	gchar *arch_path = NULL;
	gint msgnum = 0;

	for (cur = full_list; cur != NULL; cur = cur->next) {
		folder_name = get_fold_name(cur->data);
		if (g_strrstr(ARCHIVE, folder_name) != NULL) {
			arch_path = folder_get_archive_path();
			msgnum = get_msgid_num(cur->data);
			file_path = g_strdup_printf("%s%c%d", arch_path, G_DIR_SEPARATOR, msgnum);
			g_free(arch_path);
			arch_path = NULL;
		} else {
			file_path = folder_get_message_file_path(cur->data);
		}
		g_free(folder_name);
		folder_name = NULL;

		if (is_file_entry_exist(file_path) == FALSE) {
			list = g_slist_append(list, cur->data);
		}
		g_free(file_path);
		file_path = NULL;
	}
	return list;
}

GSList *al_sup_retr_msglist(GSList * list)
{
	GSList *return_list = NULL;
	RetrMsgData *retr_msg = NULL;
	GSList *cur = NULL;

	for (cur = list; cur != NULL; cur = cur->next) {
		retr_msg = g_new0(RetrMsgData, 1);
		retr_msg->msgid = g_strdup(cur->data);
		return_list = g_slist_append(return_list, retr_msg);
	}

	return return_list;
}

static
gint al_sup_copy_move_pop_msg(gboolean move_flag, const gchar * srcfolder,
			      const gchar * destfolder, PrefsAccount * prefs_account,
			      GSList * account_list, gboolean remote,
			      GSList ** folder_append_list_ptr, gint * cur_ptr,
			      gint total, GSList ** folder_update_list_ptr,
			      GSList ** folder_del_list_ptr)
{
	gint ret_val = 0;
	gchar *dest_msgid = NULL;
	gint succ_val = 0;
	GSList *cur = NULL;

	for (cur = account_list; cur != NULL; cur = cur->next) {

		if (gl_al_folder_cancel == FALSE) {
			(*cur_ptr)++;
			if (move_flag == TRUE) {
				ret_val = folder_move_msg(srcfolder, destfolder,
							  cur->data, &dest_msgid);
			} else {
				ret_val = folder_copy_msg(srcfolder, destfolder,
							  cur->data, &dest_msgid);
			}

			if (ret_val == FOLDER_SUCCESS) {
				ret_val = folder_update_unread(cur->data, dest_msgid, move_flag);
				/* update the remote copy list */
				if (move_flag == TRUE) {
					folder_delete_list_msg(cur->data, TRUE);
				}

				*folder_append_list_ptr =
				    folder_append_header_list(*folder_append_list_ptr,
							      cur->data, destfolder, prefs_account,
							      FALSE, folder_update_list_ptr,
							      dest_msgid);
			}

			FOLDER_PROGRESS_CB(*cur_ptr, total);

			if (ret_val != FOLDER_SUCCESS) {
				succ_val = ret_val;
			}
			(void) g_main_context_iteration(NULL, FALSE);
		} else {

			/*If operation is cancelled */
			if (remote == FALSE)
				break;
			if (move_flag == TRUE) {
				/*update list about downloaded headers */
				*folder_del_list_ptr =
				    al_sup_add_update_list(cur->data, *folder_del_list_ptr);
			} else {
				/*update list for copy operation abt downloaded hdrs */
				*folder_update_list_ptr =
				    al_sup_add_update_list(cur->data, *folder_update_list_ptr);
			}
		}
	}

	return succ_val;
}

static
GSList *al_sup_add_update_list(gchar * msgid, GSList * list)
{

	gchar *folder_name = NULL;
	gchar *file_path = NULL;
	gint msgnum = 0;
	gchar *arch_path = NULL;
	MsgHeader *header = NULL;

	if (msgid == NULL)
		return list;

	folder_name = get_fold_name(msgid);
	if (g_strrstr(ARCHIVE, folder_name) != NULL) {
		arch_path = folder_get_archive_path();
		msgnum = get_msgid_num(msgid);
		file_path = g_strdup_printf("%s%c%d", arch_path, G_DIR_SEPARATOR, msgnum);
		g_free(arch_path);
	} else {
		file_path = folder_get_message_file_path(msgid);
	}
	g_free(folder_name);

	if (is_file_entry_exist(file_path) == TRUE) {
		/*means header is downloaded. tell UI about this
		 * in case operation was cancelled before copying/moving
		 * was completed
		 */
		header = g_new0(MsgHeader, 1);
		header->msgid = g_strdup(msgid);
		header->msgtag.fullydownloaded = TRUE;
		header->msgtag.bodydownloaded = TRUE;
		list = g_slist_append(list, header);
	}
	g_free(file_path);

	return list;
}

static
gboolean al_sup_remote_mails(GSList * full_list)
{
	gchar *file_path = NULL;
	gchar *folder_name = NULL;
	gint msgnum = 0;
	gchar *arch_path = NULL;
	gboolean remote = FALSE;

	folder_name = get_fold_name(full_list->data);
	if (g_strrstr(ARCHIVE, folder_name) != NULL) {
		arch_path = folder_get_archive_path();
		msgnum = get_msgid_num(full_list->data);
		file_path = g_strdup_printf("%s%c%d", arch_path, G_DIR_SEPARATOR, msgnum);
		g_free(arch_path);
	} else {
		file_path = folder_get_message_file_path(full_list->data);
	}
	g_free(folder_name);
	if (is_file_entry_exist(file_path) == FALSE) {
		remote = TRUE;
	}
	g_free(file_path);

	return remote;
}

static
GSList *al_sup_update_available_list(GSList * full_list, GSList * hdr_list)
{
	GSList *cur = NULL;
	gchar *file_path = NULL;
	gchar *folder_name = NULL;
	gint msgnum = 0;
	gchar *arch_path = NULL;
	MsgHeader *header = NULL;

	for (cur = full_list; cur != NULL; cur = cur->next) {
		folder_name = get_fold_name(cur->data);
		if (g_strrstr(ARCHIVE, folder_name) != NULL) {
			arch_path = folder_get_archive_path();
			msgnum = get_msgid_num(cur->data);
			file_path = g_strdup_printf("%s%c%d", arch_path, G_DIR_SEPARATOR, msgnum);
			g_free(arch_path);
		} else {
			file_path = folder_get_message_file_path(cur->data);
		}
		g_free(folder_name);

		if (is_file_entry_exist(file_path) == TRUE) {
			header = g_new0(MsgHeader, 1);
			header->msgid = g_strdup(cur->data);
			header->msgtag.fullydownloaded = TRUE;
			header->msgtag.bodydownloaded = TRUE;
			hdr_list = g_slist_append(hdr_list, header);
		}
		g_free(file_path);
	}

	return hdr_list;
}

gboolean
al_sup_folder_rem_local(const gchar * account, GSList ** folder_delete_list,
			const gchar * folder, const gchar * curdata,
			gint * ret_val, gboolean * pop_flag)
{
	GSList *file_list = NULL;
	PrefsAccount *prefs_account = NULL;
	gboolean local_copy_flag = FALSE;
	gboolean flag = FALSE;

	*ret_val = FOLDER_SUCCESS;
	flag = *pop_flag;

	prefs_account = (PrefsAccount *) account_get_account_cb(account);
	/* archival addition */
	if (strcmp(folder, ARCHIVE) == 0) {
		file_list = g_slist_append(file_list, g_strdup(curdata));
	} else {
		file_list = folder_get_msgid_all_files(folder, curdata, NULL, TRUE);
	}
	if (file_list != NULL) {
		*pop_flag = folder_check_pop_header(file_list);
	}

	if (file_list != NULL) {
		if (prefs_account != NULL) {
			if (prefs_account->receive_settings.protocol == PROTO_POP3) {
				if ((flag == FALSE) || (folder_is_msg_remote(curdata) == FALSE)) {

					*ret_val = folder_remove_msgs(folder, file_list);
				}
			} else {
				*ret_val = folder_remove_msgs(folder, file_list);

			}
		} else {
			*ret_val = folder_remove_msgs(folder, file_list);

		}
	}


	/* this has local copy */
	if ((file_list != NULL) && (*pop_flag == FALSE)) {

		if (prefs_account == NULL) {

			/* this has only local copy */
			local_copy_flag = TRUE;
			*folder_delete_list =
			    folder_append_header_list(*folder_delete_list, curdata,
						      NULL, prefs_account, TRUE, NULL, NULL);

		} else if ((folder_is_msg_remote(curdata) == FALSE) ||
			   (strcmp(folder, INBOX_DIR) != 0)) {


			*folder_delete_list =
			    folder_append_header_list(*folder_delete_list, curdata,
						      NULL, prefs_account, TRUE, NULL, NULL);

			/* this has only local copy */
			local_copy_flag = TRUE;
		}
		slist_free_strings(file_list);

		g_slist_free(file_list);
	}
	return local_copy_flag;
}
static gchar *al_sup_get_imap_mimepart(const gchar * msgid,
				       const gchar * enc_filename, const gchar * default_font)
{
	MimeInfo *sub_mimeinfo = NULL;
	MimeInfo *partinfo = NULL;
	gint pos = -1;
	gint level = 0;
	glong fpos = 0;
	GSList *file_pos_list = NULL;
	gchar *out_filename = NULL;
	FILE *fp = NULL;
	FILE *outfp = NULL;

	fp = fopen(enc_filename, "r");

	if (fp == NULL) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		return NULL;
	}
	sub_mimeinfo = (MimeInfo *) procmime_scan_message(fp, pos, level, fpos, file_pos_list);
	partinfo = sub_mimeinfo;

	while (partinfo != NULL) {
		if (IS_MHTML_MESSAGE(partinfo) == TRUE) {
			break;
		}
		partinfo = partinfo->children;
	}
	if (partinfo == NULL) {
		procmime_mimeinfo_free_all(sub_mimeinfo);
		fclose(fp);
		return NULL;
	}
	out_filename = procmime_get_tmp_mhtml_file(msgid, NULL);
	if(!out_filename) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		procmime_mimeinfo_free_all(sub_mimeinfo);
		fclose(fp);
		return NULL;
	}

	outfp = fopen(out_filename, "w+");

	if (outfp == NULL) {
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else
			engine_errno = ENGINE_ERROR_FILE_ERROR;
		g_free(out_filename);
		fclose(fp);
		fclose(outfp);
		procmime_mimeinfo_free_all(sub_mimeinfo);
		return NULL;
	}

	if (al_sup_get_mhtml_contents(partinfo, fp, outfp, FALSE,
				      FALSE, default_font) == AL_SUP_FAIL) {

		g_free(out_filename);
		out_filename = NULL;
	}
	fclose(fp);
	fclose(outfp);
	procmime_mimeinfo_free_all(sub_mimeinfo);

	return out_filename;
}

gint
al_sup_rem_accountwise(GSList * sorted_list,
		       GSList ** folder_delete_list_ptr,
		       GSList ** folder_update_list_ptr,
		       gint total, const gchar * folder, gboolean flag)
{
	GSList *cur = NULL;
	gboolean local_copy_flag = TRUE;
	GSList *cur_account_list = NULL;
	gchar *account = NULL;
	gint ret_val = FOLDER_FAIL;
	gint tmp_ret = FOLDER_SUCCESS;
	gint succ_val = FOLDER_SUCCESS;
	gchar *prev_account = NULL;
	PrefsAccount *prefs_account = NULL;
	gboolean pop_hdr_flag = FALSE;

	for (cur = sorted_list; cur != NULL; cur = cur->next) {

		if (check_msgid(cur->data) == FALSE) {
			engine_errno = ENGINE_ERROR_INVALID_MSGID;
			continue;
		}
		account = get_account_name(cur->data);
		pop_hdr_flag = flag;
		local_copy_flag =
		    al_sup_folder_rem_local(account, folder_delete_list_ptr, folder,
					    cur->data, &ret_val, &pop_hdr_flag);

		if (local_copy_flag == FALSE) {

			if (flag == TRUE) {
				/* remote deletion */
				if (prev_account == NULL) {
					prev_account = g_strdup(account);
				}
				if (strcmp(account, prev_account) != 0) {
					tmp_ret = al_sup_del_remote_list_free(prev_account,
									      folder_delete_list_ptr,
									      cur_account_list,
									      total);

					cur_account_list = NULL;
					prev_account = g_strdup(account);
				}
				cur_account_list = g_slist_append(cur_account_list,
								  g_strdup(cur->data));

			} else {
				/* the message does not have local copy.
				   so update the status */
				prefs_account = (PrefsAccount *) account_get_account_cb(account);
				*folder_delete_list_ptr =
				    folder_append_header_list(*folder_delete_list_ptr,
							      cur->data, NULL, prefs_account,
							      TRUE, NULL, NULL);
				folder_delete_list_msg(cur->data, TRUE);
				folder_message_del_list(cur->data, UNREAD);

			}
		} else {
			folder_delete_list_msg(cur->data, TRUE);
			folder_message_del_list(cur->data, UNREAD);

		}
		g_free(account);

		if ((ret_val != FOLDER_SUCCESS) || (tmp_ret != FOLDER_SUCCESS)) {
			succ_val = FOLDER_FAIL;
		}
	}
	if (cur_account_list != NULL) {
		/* for the last list,previous is same as current account */
		tmp_ret = al_sup_del_remote_list_free(prev_account,
						      folder_delete_list_ptr, cur_account_list,
						      total);

	}

	if ((succ_val != FOLDER_SUCCESS) || (tmp_ret != FOLDER_SUCCESS)) {
		return FOLDER_FAIL;
	}
	return FOLDER_SUCCESS;
}

static gint al_sup_del_remote_list_free(gchar * prev_account,
					GSList ** folder_delete_list_ptr,
					GSList * cur_account_list, gint total)
{
	PrefsAccount *prevprefs_account = NULL;
	gint tmp_ret = FOLDER_SUCCESS;

	prevprefs_account = (PrefsAccount *) account_get_account_cb(prev_account);
	tmp_ret = al_sup_folder_rem_msg(prevprefs_account,
					cur_account_list, folder_delete_list_ptr, total);

	slist_free_strings(cur_account_list);
	g_slist_free(cur_account_list);
	g_free(prev_account);
	return tmp_ret;
}

GSList *al_sup_get_all_attach_list(MimeInfo * mimeinfo,
				   gboolean first_partflag,
				   gint * attach_count, const gchar * msgid, gboolean section_flag)
{
	MimeInfo *temp_mimeinfo = NULL;
	FolderAttachInfo *folder_attachinfo = NULL;
	GSList *temp_list = NULL;
	GSList *ret_list = NULL;


	for (temp_mimeinfo = mimeinfo; temp_mimeinfo != NULL; temp_mimeinfo = temp_mimeinfo->next) {
		if (temp_mimeinfo->content_type != NULL) {

			if ((g_ascii_strcasecmp(temp_mimeinfo->content_type,
						MULTIPART_MIXED) == 0) ||
			    (g_ascii_strcasecmp(temp_mimeinfo->content_type,
						MULTIPART_REPORT) == 0)) {

				temp_list = al_sup_get_all_attach_list(temp_mimeinfo->children,
								       first_partflag, attach_count,
								       msgid, section_flag);

				/* add the child list to main list */
				ret_list = g_slist_concat(ret_list, temp_list);
				continue;
			}
		}
		if (first_partflag == TRUE) {

			if ((mimeinfo->mime_type == MIME_TEXT) ||
			    (mimeinfo->mime_type == MIME_TEXT_HTML) ||
			    (mimeinfo->mime_type == MIME_MULTIPART)) {
				/* This is body part,so dont add as attachment */
				first_partflag = FALSE;
				continue;
			}
		}
		if (temp_mimeinfo->is_downloaded == FALSE)
			continue;

		if (section_flag == TRUE) {

			folder_attachinfo = g_new0(FolderAttachInfo, 1);

			folder_attachinfo->msgid = g_strdup(msgid);

			if (temp_mimeinfo->section_number != NULL) {

				folder_attachinfo->section_number =
				    g_strdup(temp_mimeinfo->section_number);
			}
			(*attach_count)++;

			ret_list = g_slist_append(ret_list, folder_attachinfo);
		} else {
			ret_list = g_slist_append(ret_list, GINT_TO_POINTER(temp_mimeinfo->fpos));
		}
	}
	return ret_list;
}

gint al_sup_rem_part(FILE * infp, FILE * outfp,
		     glong fpos, GSList * rem_fpos_list, MimeInfo * mimeinfo, const gchar * msgid)
{

	gchar buf[BUFFSIZE];
	gchar *boundary = NULL;
	gint boundary_len = 0;
	MimeInfo *partinfo = NULL;
	gint ret_val = AL_SUP_SUCCESS;
	gboolean retract = FALSE;
	glong pos = 0;
	GSList *attach_list = NULL;
	gint attach_count = 0;


	memset(buf, '\0', BUFFSIZE);
	rewind(infp);
	if ((fpos == mimeinfo->fpos) && (mimeinfo->next == NULL)) {

		procheader_copy_headers(infp, outfp);
		return AL_SUP_SUCCESS;
	}
	if ((g_ascii_strcasecmp(mimeinfo->content_type,
				MULTIPART_MIXED) == 0) ||
	    (g_ascii_strcasecmp(mimeinfo->content_type, MULTIPART_REPORT) == 0)) {

		if (mimeinfo->children == NULL) {
			return AL_SUP_FAIL;
		}

		/* there is no attachment,but has body which cant be displayed */
		if ((fpos == mimeinfo->children->fpos) && (mimeinfo->children->next == NULL)) {

			procheader_copy_headers(infp, outfp);
			return AL_SUP_SUCCESS;
		}
		/* what are we trying to delete?? */
		if (mimeinfo->children->next == NULL) {
			return AL_SUP_FAIL;
		}

		attach_list = al_sup_get_all_attach_list(mimeinfo, TRUE,
							 &attach_count, msgid, FALSE);

		/* there is only one attachment */
		/* or if there are multiple attachments and all of them
		 * should be purged */
		if (((mimeinfo->children->next->fpos == fpos) &&
		     (mimeinfo->children->next->next == NULL)) ||
		    (g_slist_length(attach_list) == g_slist_length(rem_fpos_list))) {
			/* change the main content type */
			ret_val =
			    al_sup_get_mhtml_contents(mimeinfo->children, infp,
						      outfp, TRUE, TRUE, NULL);
			g_slist_free(attach_list);
			return ret_val;
		}
	}
	/* Iterate through the file and filter out the purged attachments */
	while (fgets(buf, sizeof(buf), infp) != NULL) {

		/* we may have skipped the starting point of attachment 
		 * in the fgets above so go back to pos
		 */
		if (retract == TRUE) {
			if (fseek(infp, pos, SEEK_SET) < 0) {
				g_slist_free(attach_list);
				return AL_SUP_FAIL;
			}
		}

		/* Once gone back set it to FALSE otherwise buf will not be
		 * put in to the file
		 */
		retract = FALSE;
		pos = 0;

		/* Check if the current file position is starting point of attachment
		 * that we are interested in purging
		 */
		if (g_slist_find(rem_fpos_list, GINT_TO_POINTER(ftell(infp))) != NULL) {

			/*If yes find the correspoding mimeinfo and set retract to 
			 * TRUE to prevent writing the content into file
			 */
			partinfo = al_sup_get_matching_mimeinfo(mimeinfo, ftell(infp));
			retract = TRUE;
			pos = ftell(infp);

			if (partinfo && partinfo->parent && partinfo->parent->boundary) {
				boundary = partinfo->parent->boundary;
				boundary_len = strlen(boundary);
			}

			if (boundary == NULL) {
				return AL_SUP_FAIL;
			}

			/* skip the removed part */

			while (fgets(buf, sizeof(buf), infp) != NULL) {
				if (IS_BOUNDARY(buf, boundary, boundary_len)) {

					/* If boundary is reached, this has to be put to file
					 * as this can be file end boundary so set retract is
					 *  FALSE
					 */
					pos = 0;
					retract = FALSE;

					/*But if boundary is not end of file boundary with "--"
					 * and the mime information after the boundary is an 
					 * attachment that needs to be purged set retract to 
					 * TRUE as its contents should not  be written to file
					 */
					if (g_slist_find(rem_fpos_list,
							 GINT_TO_POINTER(ftell(infp))) != NULL) {
						pos = ftell(infp);
						retract = TRUE;
					}
					break;
				}
			}
		}
		if (retract == FALSE)
			fputs(buf, outfp);
	}
	g_slist_free(attach_list);
	return AL_SUP_SUCCESS;
}

static
MimeInfo *al_sup_get_matching_mimeinfo(MimeInfo * mimeinfo, glong pos)
{
	MimeInfo *partinfo = NULL;

	partinfo = mimeinfo;

	while (partinfo != NULL) {
		if (partinfo->fpos == pos) {
			return partinfo;
		}

		partinfo = procmime_mimeinfo_next(partinfo);
	}

	return NULL;
}


gint al_sup_retr_list(GSList * account_list, const gchar * account)
{
	GSList *tmp = NULL;
	gint ret = AL_SUP_SUCCESS;
	gboolean imap_flag = FALSE;
	PrefsAccount *prefs_account = NULL;
	gint cur = 1;
	gint total = 0;
	gchar *msgid = NULL;
	gboolean complete_download = FALSE;
	Session *session = NULL;
	RetrMsgData *retr_msg = NULL;


	imap_flag = FALSE;
	prefs_account = account_get_account(account);

	if (prefs_account == NULL) {
		return AL_SUP_FAIL;
	}
	if (prefs_account->receive_settings.protocol == PROTO_IMAP4) {
		imap_flag = TRUE;
	}
	total = g_slist_length(account_list);

	if (imap_flag == TRUE) {

		for (tmp = account_list; tmp != NULL; tmp = tmp->next) {

			engine_errno = ENGINE_ERROR_DEFAULT;

			retr_msg = (RetrMsgData *) tmp->data;

			msgid = retr_msg->msgid;

			complete_download = retr_msg->flag;

			if (imap_retr_imap_msg(msgid, prefs_account,
					       complete_download, &session, &cur, total)
			    != IMAP_SUCCESS) {

				ret = AL_SUP_FAIL;
			}
			if (session == NULL) {
				break;
			}
			if (imap_is_session_cancelled() == TRUE) {
				break;
			}

			if (get_engine_errno() == ENGINE_ERROR_CONNECT_SERVER) {
				break;
			}
			cur++;

		}
		if (session != NULL) {
			imap_session_complete(session);
		}
		return ret;
	}
	pop3_recv_mail(prefs_account, msgid, FOLDER_OP_NONE, FALSE, account_list);
	return ret;
}
void al_sup_imap_set_seen(const gchar * msgid, PrefsAccount * account)
{
	GSList *seen_list = NULL;
	if (account->user_settings.passwd == NULL) {
		folder_message_add_list(msgid, SEEN);
		return;
	}

	if (folder_is_msg_remote(msgid) == TRUE) {
#if CL_USE
		cl_iap_set_type(TRUE);	/* set for timed out connection
					   to suppress connection ui */
#endif
		seen_list = g_slist_append(seen_list, g_strdup(msgid));
		UNSET_CB_FLAG;
		imap_expunge_msg(account, seen_list, NULL, SEEN, NULL, 0, NULL);
		SET_CB_FLAG;
#if CL_USE
		cl_iap_set_type(FALSE);
#endif
		slist_free_strings(seen_list);
		g_slist_free(seen_list);
	}
}

static
gchar *al_sup_get_email_attach_path(FILE * fp, MimeInfo * mimeinfo)
{
	gchar buf[BUFFSIZE];
	MimeInfo *match_mimeinfo = NULL;
	gchar *boundary = NULL;
	MimeInfo *mime = NULL;
	gchar *temp_path = NULL;
	FILE *outfp = NULL;
	gint boundary_len = 0;


	if ((fp == NULL) || (mimeinfo == NULL)) {
		return NULL;
	}

	rewind(fp);
	memset(buf, '\0', BUFFSIZE);

	/* get mime info of the entire file */
	mime = procmime_scan_message(fp, -1, 0, 0, NULL);

	/* get mimeinfo that matches with input mimeinfo */
	match_mimeinfo = al_sup_get_matching_mimeinfo(mime, mimeinfo->fpos);

	if ((match_mimeinfo == NULL) || (match_mimeinfo->parent == NULL)
	    || (match_mimeinfo->parent->boundary == NULL)) {
		procmime_mimeinfo_free_all(mime);
		rewind(fp);
		return NULL;
	}

	/* get the parent boundary of input mimeinfo */
	boundary = g_strdup(match_mimeinfo->parent->boundary);
	boundary_len = strlen(boundary);
	procmime_mimeinfo_free_all(mime);

	/* get the message header information */
	if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) {
		engine_errno = ENGINE_ERROR_FILE_ERROR;
		g_free(boundary);
		rewind(fp);
		return NULL;
	}

	/*go to the header info part */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (buf[0] == '\r' || buf[0] == '\n') {
			break;
		}
	}

	memset(buf, '\0', BUFFSIZE);
	temp_path = get_tmp_file();
	outfp = fopen(temp_path, "w+");
	if (outfp == NULL) {
		g_free(boundary);
		send_engine_error_to_ui(errno);
		rewind(fp);
		g_free(temp_path);
		return NULL;
	}

	/*copy email attachment contents till boundary is reached */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (IS_BOUNDARY(buf, boundary, boundary_len)) {
			break;
		}

		strcrchomp(buf);
		fputs(buf, outfp);
		memset(buf, '\0', sizeof(buf));
	}

	fclose(outfp);
	g_free(boundary);
	rewind(fp);
	return temp_path;
}
