
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#define DBUS_API_SUBJECT_TO_CHANGE

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <unistd.h>
#include <log-functions.h>
#include "dbus_utils.h"
#include "dbus_utils_def.h"
#include "dbus_utils_xml.h"
#include "callbacks.h"
#include "engine_dbus.h"
#include "al.h"
#include "account.h"
#include "smtp_errorcodes.h"
#include "engine_errno.h"
#include "manager.h"
#include "defs.h"
#include "xml_cache.h"
#include "email_gconf.h"

#include "utils.h"
#include "imap.h"
#include "scheduling.h"
#include "folder.h"
#include "notify.h"

#if CL_USE
#include "cl.h"
#include "iap_confserv.h"
#endif

static DBusConnection *connection = NULL;
static gboolean connection_flag = FALSE;
static GMainLoop *gl_loop = NULL;

static gint engine_init_status = ENGINE_ERROR_DEFAULT;

extern EmailMgrCallback *email_mgr_cb;	/* Defined in manager.c file */
extern gint emailMgrState;	/* Defined in manager.c file */
extern gint engine_errno;	/* utils.c file */

extern gboolean gl_callback_enable_flag;	/* defined in utils.c */
static gboolean gl_normal_dbus_handler_busy = FALSE;
static gboolean gl_send_recv_dbus_handler_busy = FALSE;
extern gboolean gl_email_mgr_waiting;	/* Defined in manager.c file */


/*static functions */

osso_pack_return_t engine_dbus_smtp_send_complete(void);

static osso_pack_return_t
engine_dbus_send_pack_str_to_remote(gchar * pack_str, const gchar * func, gboolean free_pack_str);
static osso_pack_return_t engine_dbus_editor_open_mail(const gchar * recv_string, gchar ** reply);
static osso_pack_return_t engine_dbus_folder_operation_status(gint status, gchar ** reply);

static osso_pack_return_t engine_dbus_send_selected_mails(const gchar * recv_str, gchar ** reply);

static osso_pack_return_t
engine_dbus_recv_msg_multiple_attachments(const gchar * recv_str, gchar ** reply);

/*Folder Related Static functions */

static osso_pack_return_t engine_dbus_folder_get_size(const gchar * recv_string, gchar ** reply);
/**
   This method is used to get the list of existing prefs account 

   @param   recv_string  Packed string contains the functionId:
   @param  reply returned with NULL or string contains callback information 
   of all folders list 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
static osso_pack_return_t
engine_dbus_prefs_account_get_list(const gchar * recv_str, gchar ** reply);

/* Statesave read/write functions */

static osso_pack_return_t engine_dbus_statesave_read(const gchar * recv_string, gchar ** reply);

static osso_pack_return_t engine_dbus_statesave_write(const gchar * recv_string, gchar ** reply);

/* MMC callback handler */
static osso_pack_return_t
engine_dbus_mmc_hlr_volume_mounted_cb(GnomeVFSVolumeMonitor *
				      volume_monitor, GnomeVFSVolume * volume);

static osso_pack_return_t
engine_dbus_mmc_hlr_volume_unmounted_cb(GnomeVFSVolumeMonitor *
					volume_monitor, GnomeVFSVolume * volume);

/**
  Function to get mimeinfo of multipart attachment

  @param recv_str xml string contains request to get iap list
  @param  reply outparmeter with reply
  @return OSSO_PACK_OK on success and OSSO_PACK_ERROR on failure
*/
static osso_pack_return_t
engine_dbus_recv_multipart_attachment(const gchar * recv_str, gchar ** reply);

/* connectivity Layer  Functions */
#if CL_USE
static osso_pack_return_t engine_dbus_get_iapinfo_list(const gchar * recv_string, gchar ** reply);

static osso_pack_return_t engine_dbus_set_iapinfo(const gchar * recv_str, gchar ** reply);
/**
 Function to disconnect IAP connection

 @param recv_str Request for disconnection IAP
 @param reply outparam contaning the response

 @return OSSO_PACK_OK on success and OSSO_PACK_ERROR on failure
*/
static osso_pack_return_t engine_dbus_disconnect_iap(const gchar * recv_str, gchar ** reply);
#endif

static osso_pack_return_t engine_dbus_ui_ready(const gchar * recv_str, gchar ** reply);

static osso_pack_return_t engine_dbus_remove_notifications(const gchar * recv_str, gchar ** reply);


/**
  This method is callback used to send the list messages to 
  be added or deleted after a successful folder operation 
  @param  list_type    Type of the list 
  @param  msghdr_list    List of the message headers 
  @return on successful operation returns OSSO_PACK_OK. Otherwise 
   returns errorcode 
*/
static gint engine_dbus_al_folder_list(gint list_type, const GSList * msghdr_list);

/**
 This function is callback used to send MsgInfo to UI.
 @param msg_info  MsgInfo of the message
 @param attach_tag attachment identification tag.
 @return on successful operation returns OSSO_PACK_OK. Otherwise 
  returns errorcode 
*/
static gint engine_dbus_al_recv_attachment_cb(const MsgInfo * msg_info, gint attach_tag);

/**
  This method is callback used to send the message to 
  be added or deleted after a successful folder operation 
  @param  msg    Message to be displayed 
  @return on successful operation returns OSSO_PACK_OK. Otherwise 
   returns errorcode 
*/
static gint engine_dbus_al_folder_set_msg(gint msg_number, const gchar * format, ...);

/**
   This method is used to remove the all attachments of 
   messages in the list 

   @param   recv_string  Packed string contains the
                         functionId:GSlist of attachments
   @param   reply returned with NULL or string contains callback information
    of result of the operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
static osso_pack_return_t
engine_dbus_folder_remove_all_attachment(const gchar * recv_str, gchar ** reply);

/**
  Function to cancel folder operations

  @param recv_str Packed string contains the function Identifier
  @param reply outparam
  @return on successful operation returns OSSO_PACK_OK. Otherwise
  returns errorcode
*/
static osso_pack_return_t engine_dbus_folder_cancel(const gchar * recv_str, gchar ** reply);

/**
  Function to set temporary passwd
  @param recv_str Packed string contains the function Identifier
  @param reply outparam
  @return on successful operation returns OSSO_PACK_OK. Otherwise
  returns errorcode
*/
static osso_pack_return_t engine_dbus_set_tmp_passwd(const gchar * recv_str, gchar ** reply);

/**
   This method is used to remove the attachments of 
   message in the folders other than Inbox 
   @param   recv_string  Packed string contains the
                         functionId:GSlist of attachments
   @param   reply returned with NULL or string contains callback information
    of result of the operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
static osso_pack_return_t engine_dbus_rem_msg_attachments(const gchar * recv_str, gchar ** reply);

/**
  Function to check if engine is ready

  @param recv_str Packed string contains the function Identifier
  @param reply outparam

  @return on successful operation returns OSSO_PACK_OK. Otherwise
  returns errorcode
*/
static osso_pack_return_t engine_dbus_is_engine_ready(const gchar * recv_str, gchar ** reply);

/**
  Function to start send/receive process

  @param recv_str Packed string contains the function Identifier
  @param reply outparam

  @return on successful operation returns OSSO_PACK_OK. Otherwise
  returns errorcode
*/
static osso_pack_return_t engine_dbus_send_recv_mail(const gchar * recv_str, gchar ** reply);

/*Bug ID: 15172*/
/**
  This message is called before composing a draft mail to get
  msg_id of the mail to be saved
  
  @recv_str - containing the xml str received from UI
  @reply - outparam with msgid
  @return OSSO_PACK_OK on success and errorcode on failure
*/
static osso_pack_return_t engine_dbus_send_get_msg_id(const gchar * recv_str, gchar ** reply);

/**
  Function to call packing of msginfo list

  @param msginfo_list List of msginfo's
  @return OSSO_PACK_OK on success and errorcode on failure
*/
static osso_pack_return_t engine_dbus_attach_list(GSList * msginfo_list);

/**
  Function to send UI list of messages that need to be removed

  @param msgid_list - List of mesgIDs
  @return OSSO_PACK_OK on success and errorcode on failure
*/
static gint engine_dbus_rx_delete_list(GSList * msgid_list);

/**
  Function to retrieve msg content for msgs sent by UI

  @param data Function ID and message IDs of msgs to be retr
  @param reply Outparam
  @return on successful operation returns OSSO_PACK_OK. Otherwise
  returns errorcode
*/
static osso_pack_return_t engine_dbus_retr_msgs(const gchar * data, gchar ** reply);

/**
  Function closes engine process if UI exits and there is no scheduling 

  @param none 
  @return none 
*/
static void engine_dbus_exit_on_check(void);


/**
  Function closes engine process if UI exits and there is no scheduling 

  @param none 
  @return none 
*/
static void engine_dbus_exit_on_check(void);

static EmailDBusFuncs engineDbusFuncList[] = {
	{UI_IS_READY, engine_dbus_ui_ready},
        {REMOVE_NOTIFICATIONS, engine_dbus_remove_notifications},

	/* Send Mail Functionality */
	{SEND_MAIL, engine_dbus_send_mail},
	{CANCEL_SEND_MAIL, engine_dbus_cancel_send_mail},
	{SEND_OUTBOX_MAIL, engine_dbus_send_outbox_mail},
	{SEND_SAVE_DRAFT_MAIL, engine_dbus_send_save_draft_mail},
	{SEND_SELECTED_MAILS, engine_dbus_send_selected_mails},
	{SEND_GET_MSG_ID, engine_dbus_send_get_msg_id},

	/* Receive Mail Functionality */
	{RECV_MAIL, engine_dbus_recv_mail},
	{RECV_MSG_BODY, engine_dbus_recv_msg_body},
	{RECV_MSG_BODY_CB, engine_dbus_recv_msg_body},
	{UPDATE_SEEN_FLAG, engine_dbus_update_seen_flag},
	{RECV_MSG_ATTACHMENT, engine_dbus_recv_msg_attachment},
	{RECV_MULTIPART_BODY, engine_dbus_recv_multipart_body},
	{CANCEL_RECV_MAIL, engine_dbus_cancel_recv_mail},
	{RECV_MULTIPART_ATTACHMENT, engine_dbus_recv_multipart_attachment},
	{RECV_MULTIPLE_ATTACHMENTS, engine_dbus_recv_msg_multiple_attachments},
	{SEND_RECV_MAIL, engine_dbus_send_recv_mail},

	/* Account Management Functionality */
	{GET_ACCOUNT_LIST, engine_dbus_get_account_list},
	{GET_ACCOUNT_SETTINGS, engine_dbus_get_account_settings},
	{SET_ACCOUNT_SETTINGS, engine_dbus_set_account_settings},
	{DELETE_ACCOUNT, engine_dbus_delete_account},
	{RENAME_ACCOUNT, engine_dbus_rename_account},
	{ACCOUNT_CHECK_STATUS, engine_dbus_account_check_status},
	{PREFS_ACCOUNT_GET_LIST, engine_dbus_prefs_account_get_list},

	/* Connectivity related Functionality */
#if CL_USE
	{GET_IAPINFO_LIST, engine_dbus_get_iapinfo_list},
	{SET_IAPINFO_SETTINGS, engine_dbus_set_iapinfo},
	{UI_DISCONNECT_IAP, engine_dbus_disconnect_iap},
#endif

	/* Folder Management Functionality */
	{FOLDER_CREATE, engine_dbus_folder_create},
	{FOLDER_DELETE, engine_dbus_folder_delete},
	{FOLDER_RENAME, engine_dbus_folder_rename},
	{FOLDER_MOVE_MSG, engine_dbus_folder_move_msg},
	{FOLDER_MOVE_MSGS, engine_dbus_folder_move_msgs},
	{FOLDER_REMOVE_MSG, engine_dbus_folder_remove_msg},
	{FOLDER_REMOVE_MSGS, engine_dbus_folder_remove_msgs},
	{FOLDER_REMOVE_ALL_MSGS, engine_dbus_folder_remove_all_msg},
	{FOLDER_GET_LIST, engine_dbus_folder_get_list},
	{POPULATE_FOLDER, engine_dbus_populate_folder},
	{FOLDER_COPY_MSG, engine_dbus_folder_copy_msg},
	{FOLDER_COPY_MSGS, engine_dbus_folder_copy_msgs},
	{FOLDER_GET_SIZE, engine_dbus_folder_get_size},
	{FOLDER_REMOVE_ATTACHMENT, engine_dbus_folder_remove_attachment},
	{FOLDER_REMOVEALL_ATTACHMENT, engine_dbus_folder_remove_all_attachment},
	{FOLDER_CANCEL, engine_dbus_folder_cancel},

	/* Statesave Functionality */
	{STATESAVE_READ, engine_dbus_statesave_read},
	{STATESAVE_WRITE, engine_dbus_statesave_write},

	/* Other Functionality */
	{GET_ENGINE_ERRNO, engine_dbus_get_engine_errno},
	{GET_ENGINE_INIT_STATUS, engine_dbus_get_engine_init_status},
	{HTML_TO_TEXT_CONVERTER, engine_dbus_html_to_text_converter},
	{AUTOSAVE_MAIL, engine_dbus_autosave_mail},
	{GET_MSG_HEADER_LIST, engine_dbus_get_msg_header_list},
	{EDITOR_OPEN_MAIL, engine_dbus_editor_open_mail},
	{SET_TMP_PASSWD, engine_dbus_set_tmp_passwd},
	{REM_MSG_ATTACHMENTS, engine_dbus_rem_msg_attachments},
	{ENGINE_CHECK_STATUS, engine_dbus_is_engine_ready},
	{AL_RETR_MSGS, engine_dbus_retr_msgs},
	{NULL, NULL}
};


/**
   This fuction is used to send packed to data to through Dbus connecttion 

   @param  pack_str  string to be sent to remote 
   @param  func  from which function this method called ( Debug purpose) 
   @param  free_pack_str to free pack_str set this value to TRUE  
   @return on successful sending data returns OSSO_PACK_OK. In case of
           error returns OSSO_PACK_ERROR 
*/

static osso_pack_return_t
engine_dbus_send_pack_str_to_remote(gchar * pack_str, const gchar * func, gboolean free_pack_str)
{
	osso_pack_return_t rv = OSSO_PACK_OK;

        osso_log(LOG_DEBUG, "%s starts with func=%s", __FUNCTION__, func);

	if ((pack_str == NULL)) {
		osso_log(LOG_INFO, "pack str is NULL \n");
		return OSSO_PACK_ERROR;
	}

	if (engine_dbus_connect_to_remote(pack_str) == FALSE) {
		if (func) {
			osso_log(LOG_ERR, "Unable to connect remote %s", func);
		} else {
			osso_log(LOG_ERR, "Unable to connect remote ");
		}
		rv = OSSO_PACK_ERROR;
	}
	if (free_pack_str == TRUE) {
		g_free(pack_str);
	}
	return rv;
}

/**
   This Method is to called to initialise the engine module 
 
   @param    none
   @return   none 
*/

void engine_dbus_init(void)
{
	if (engine_dbus_al_register_engine() == OSSO_PACK_ALLOC_FAIL) {
		return;
	}
	engine_dbus_server_thread();
	return;
}

/**
   This Method is to check UI Dbus connection is exist or not
 
   @param    none
   @return   In case of UI is running return TRUE otherwise FALSE
*/

gboolean engine_dbus_ui_is_available(void)
{
	gint rv = FALSE;
	if ((connection != NULL) && (connection_flag == TRUE)) {
		rv = TRUE;
	}
	osso_log(LOG_INFO, "In engine_dbus_ui_is_available rv is : %d", rv);
	return rv;
}


/**
   This fuction registers the call back functions with other modules
   like SMTP, IMAP, POP3, Account Management 

   @param none 
   @return on successful registartion returns OSSO_PACK_OK. In case of
    error returns OSSO_PACK_ERROR or Memory allocation failure returns 
    OSSO_PACK_ALLOC_FAIL
*/

osso_pack_return_t engine_dbus_al_register_engine(void)
{
	gint succ = 0;

	SmtpCallback *dbus_smtpCallback = NULL;
	Pop3Callback *dbus_pop3Callback = NULL;
	Imap4Callback *dbus_imap4Callback = NULL;
	AcctMgmtCallback *dbus_acctCallback = NULL;
	MmcHlrCallback *dbus_mmc_hlrCallback = NULL;
	AlCallback *dbus_al_callback = NULL;

	/* Assign the SMTP callbacks */

	dbus_smtpCallback = g_new0(SmtpCallback, 1);
	if (dbus_smtpCallback == NULL) {
		goto endshere;
	}
	dbus_smtpCallback->smtp_param_value_missing_cb = engine_dbus_smtp_param_value_missing;
	dbus_smtpCallback->smtp_set_msg_cb = engine_dbus_smtp_set_msg;
	dbus_smtpCallback->smtp_set_percentage_cb = engine_dbus_smtp_set_percentage;
	dbus_smtpCallback->smtp_update_folder_view_cb = engine_dbus_smtp_update_folder_view;
	dbus_smtpCallback->smtp_update_outbox_mail_count =
	    engine_dbus_smtp_update_outbox_mails_sent;
	dbus_smtpCallback->smtp_send_complete_cb = engine_dbus_smtp_send_complete;
#ifdef CL_USE
	dbus_smtpCallback->smtp_get_iap_conf_info = iap_smtpconf_get_iap_info;
#endif
	/* Assign the POP3 Callbacks */

	dbus_pop3Callback = g_new0(Pop3Callback, 1);
	if (dbus_pop3Callback == NULL) {
		goto endshere;
	}

	dbus_pop3Callback->pop3_param_value_missing_cb = engine_dbus_pop3_param_value_missing;
	dbus_pop3Callback->pop3_set_msg_cb = engine_dbus_pop3_set_msg;
	dbus_pop3Callback->pop3_set_percentage_cb = engine_dbus_pop3_set_percentage;

	dbus_pop3Callback->pop3_recv_list_cb = engine_dbus_imap4_msg_headers;
	dbus_pop3Callback->pop3_recv_complete_cb = engine_dbus_pop3_recv_complete;
	dbus_pop3Callback->pop3_recv_body_complete_cb = engine_dbus_pop3_recv_body_complete;
	dbus_pop3Callback->pop3_delete_list_cb = engine_dbus_rx_delete_list;
	dbus_pop3Callback->pop3_send_progress_cb = engine_dbus_al_folder_set_msg;


	/* Assign the IMAP4 Callbacks */

	dbus_imap4Callback = g_new0(Imap4Callback, 1);
	if (dbus_imap4Callback == NULL) {
		goto endshere;
	}

	dbus_imap4Callback->imap4_param_value_missing_cb = engine_dbus_imap4_param_value_missing;
	dbus_imap4Callback->imap4_msg_headers_cb = engine_dbus_imap4_msg_headers;
	dbus_imap4Callback->imap4_set_msg_cb = engine_dbus_imap4_set_msg;
	dbus_imap4Callback->imap4_del_list_cb = engine_dbus_rx_delete_list;

	/* Assign the Account Management Callbacks */

	dbus_acctCallback = g_new0(AcctMgmtCallback, 1);
	if (dbus_acctCallback == NULL) {
		goto endshere;
	}
	/* Assign the Mmc Hlr callback       */
	dbus_mmc_hlrCallback = g_new0(MmcHlrCallback, 1);
	if (dbus_mmc_hlrCallback == NULL) {
		goto endshere;
	}
	dbus_mmc_hlrCallback->mmc_hlr_volume_mounted_cb = engine_dbus_mmc_hlr_volume_mounted_cb;

	dbus_mmc_hlrCallback->mmc_hlr_volume_unmounted_cb = engine_dbus_mmc_hlr_volume_unmounted_cb;

	dbus_al_callback = g_new0(AlCallback, 1);
	if (dbus_al_callback == NULL) {
		goto endshere;
	}
	dbus_al_callback->al_error_code_cb = engine_dbus_engine_error_code;
	dbus_al_callback->al_folder_msg_cb = engine_dbus_al_folder_set_msg;
	dbus_al_callback->al_folder_list_cb = engine_dbus_al_folder_list;
	dbus_al_callback->al_recv_attachment_cb = engine_dbus_al_recv_attachment_cb;

	succ = al_register_to_engine(dbus_smtpCallback, dbus_pop3Callback,
				     dbus_imap4Callback, dbus_acctCallback,
				     dbus_mmc_hlrCallback, dbus_al_callback);
       #if CL_USE
	iap_smtpconf_readlist();
#endif
	engine_init_status = al_get_engine_errno();

	/*CHECK_STATUS ( ( succ == 0 ), "Register Engine Failed", 
	   return OSSO_PACK_ERROR ); */
	osso_log(LOG_INFO, " al_register to engine is done ..\n");
	return OSSO_PACK_OK;

      endshere:

	g_free(dbus_smtpCallback);
	g_free(dbus_pop3Callback);
	g_free(dbus_imap4Callback);
	g_free(dbus_acctCallback);
	g_free(dbus_mmc_hlrCallback);

	osso_log(LOG_ERR, " Memory Alloc failed \n");
	engine_errno = ENGINE_ERROR_NO_MEM;
	return OSSO_PACK_ALLOC_FAIL;
}


/**
   This fuction unregisters the call back functions of different modules
   (SMTP, POP3, etc )and frees the global memory owned by these modules.
 
   @param    none
   @return   void
*/

void engine_dbus_al_unregister_engine(void)
{
	osso_log(LOG_DEBUG, "\nOSSO unregister engine called\n");
	al_engine_cleanup_memory();
	g_main_loop_quit(gl_loop);
	g_main_loop_unref(gl_loop);
}

/* 
 * SMTP Callback Handling Functionality
 */

/**
   This fuction registerd as a call back function to SMTP module. 
   SMTP Module invokes this function when engine wants to send status 
   or missing parameter.

   @param  param_number Missing parameter or status 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_smtp_param_value_missing(const ParamMissing param_number)
{
	gchar *str = NULL;
	if (param_number == 0) {
		return OSSO_PACK_INVALID;
	}
	str = osso_email_utils_concat_str_int(SMTP_PARAM_VALUE_MISSING, param_number);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to SMTP module and
   invokes this function when engine wants to send status or error code. 

   @param  msg_number Error code or status 
   @param  server_ip  IP Address of the server 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t
engine_dbus_smtp_set_msg(const gint msg_number, const gchar * ipaddr, const gchar * acc_name)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, (gint *) & msg_number);

	if (ipaddr != NULL) {
		osso_email_xml_pack_str(root, doc, XML_STRING, (gchar *) & ipaddr);
	}

	if (acc_name != NULL) {
		osso_email_xml_pack_str(root, doc, XML_STRING, (gchar *) & acc_name);
	}
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(SMTP_SET_MSG, xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to SMTP module and calls
   this method when engine wants to give percentage of mail is sent.

   @param  percent  Percentage of mail sent. 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_smtp_set_percentage(const gint percent)
{
	gchar *str = NULL;

	if (percent == 0) {
		return OSSO_PACK_INVALID;
	}

	str = osso_email_utils_concat_str_int(SMTP_SET_PERCENTAGE, percent);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to SMTP module. SMTP Module
   invokes this function when engine wants update the mails sent from outbox.

   @param  totalMails Mails exist in Outbox 
   @param  sentMails  Mails sent from outbox 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t
engine_dbus_smtp_update_outbox_mails_sent(const gint totalMails, const gint sentMails)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	if (totalMails == 0) {
		return OSSO_PACK_INVALID;
	}

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, OUTBOXMAILSCOUNT, (gint *) & totalMails);
	osso_email_xml_pack_int(root, doc, OUTBOXMAILSSENT, (gint *) & sentMails);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(SMTP_UPDATE_OUTBOX_MAILS_SENT, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}


/**
   This fuction registerd as a call back function to SMTP module. 
   SMTP Module invokes this function when engine wants to update the status
   of mail in folder.

   @param  msgheader  Message Header information of the mail  
   @param  operation  on TRUE Message id needs added to Outbox, on FALSE 
    removed from outbox   
   @param  status On TRUE UI displays Email status change conformation note 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t
engine_dbus_smtp_update_folder_view(MsgHeader * msgheader, gint operation, gboolean status)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	if (msgheader == NULL) {
		osso_log(LOG_INFO, "Msg Header is NULL ");
		return OSSO_PACK_INVALID;
	}
	if (operation < 0) {
		return OSSO_PACK_INVALID;
	}

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_msgheader(root, doc, XML_MSG_HEADER, msgheader);
	osso_email_xml_pack_int(root, doc, XML_INT, &operation);
	osso_email_xml_pack_bool(root, doc, XML_BOOL, &status);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(SMTP_UPDATE_FOLDER_VIEW, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to SMTP module. SMTP Module
   invokes this function when engine wants to to UI about completion of send 
   operation 

   @param  None 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_smtp_send_complete(void)
{
	gchar *str = NULL;

	str = osso_email_utils_concat_strings(SMTP_SEND_COMPLETE, NULL);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/*
 * POP3 Callback Handling Functionality
 */

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to update the missing parameter .

   @param  param_number Missing parameter required for POP3 receiving 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_pop3_param_value_missing(ParamMissing param_number)
{
	gchar *str = NULL;
	if (param_number == 0) {
		return OSSO_PACK_INVALID;
	}
	str = osso_email_utils_concat_str_int(POP3_PARAM_VALUE_MISSING, param_number);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to send errorcode or status of 
   receiving mail .

   @param  msg_number Errorcode or status
   @param  format     Extra information for the status 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

gint engine_dbus_pop3_set_msg(gint msg_number, const gchar * format, ...)
{
	gchar *str = NULL;

	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	va_list ap;
	gint int_val = 0;
	gchar *str_val = NULL;
	gchar *fmt = NULL;

	if (msg_number == 0) {
		return OSSO_PACK_INVALID;
	}
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, &msg_number);

	if (format != NULL) {
		fmt = (gchar *) format;
		va_start(ap, format);
		while (*fmt) {
			switch (*fmt++) {
			case 's':
				str_val = va_arg(ap, char *);
				osso_email_xml_pack_str(root, doc, XML_STRING, &str_val);
				break;
			case 'd':
				int_val = va_arg(ap, int);
				osso_email_xml_pack_int(root, doc, XML_INT, &int_val);
				break;
			}
		}
		va_end(ap);
	}
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(POP3_SET_MSG, xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to end percentage of email received.

   @param  percent  Percentage of email received 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_pop3_set_percentage(gint percent)
{
	gchar *str = NULL;

	if (percent == 0) {
		return OSSO_PACK_INVALID;
	}
	str = osso_email_utils_concat_str_int(POP3_SET_PERCENTAGE, percent);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to send  received mail header information.

   @param  msgheader Message Header information of received mail 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_pop3_recv_single(MsgHeader * msgheader)
{
	gchar *list_string = NULL;
	gchar *str = NULL;

	if (msgheader == NULL) {
		osso_log(LOG_INFO, "Msg Header is NULL ");
		return OSSO_PACK_INVALID;
	}

	osso_email_utils_pack_msgHeader(msgheader, &list_string);
	str = osso_email_utils_concat_strings(POP3_RECV_SINGLE, list_string);
	g_free(list_string);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to send status of  received mail complete. 

   @param  None 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_pop3_recv_complete(void)
{
	osso_pack_return_t rv = OSSO_PACK_ERROR;
	GString *converted_str = NULL;

	converted_str = g_string_new(NULL);
	CHECK_STATUS((converted_str == NULL), "Memory Alloc Failed ", return OSSO_PACK_INVALID);
	g_string_printf(converted_str, "%s%s", POP3_RECV_COMPLETE, seperator);
	rv = engine_dbus_send_pack_str_to_remote(converted_str->str, __FUNCTION__, FALSE);
	g_string_free(converted_str, TRUE);
	return rv;
}

/**
   This fuction registerd as a call back function to POP3 module. POP3 Module
   invokes this function when engine wants to send complete information of recv mail.

   @param  mimeinfo  Mime information of received mail
   @param  msgid     Message id of received mail 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t
engine_dbus_pop3_recv_body_complete(MimeInfo * mimeinfo, const gchar * msgid, gboolean remote_copy)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	CHECK_STATUS((msgid == NULL), "msgid is NULL ", return OSSO_PACK_INVALID);

	DBUS_XML_PACK_INIT;
	if (mimeinfo != NULL) {
		osso_email_xml_pack_mimeinfo(root, doc, XML_MIME_INFO, mimeinfo);
	}
	osso_email_xml_pack_str(root, doc, XML_MSG_ID, &msgid);
	osso_email_xml_pack_bool(root, doc, XML_BOOL, &remote_copy);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(POP3_RECV_BODY_COMPLETE, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}


/*
 * IMAP4 Callback Handling Functionality
 */

/**
   This fuction registerd as a call back function to IMAP4 module. IMAP4 Module
   invokes this function when engine wants to send status from imap4 module .

   @param  param_number Status or Error code 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_imap4_param_value_missing(ParamMissing param_number)
{
	gchar *str = NULL;

	str = osso_email_utils_concat_str_int(IMAP4_PARAM_VALUE_MISSING, param_number);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to IMAP4 module. IMAP4 Module
   invokes this function when engine wants to send received mail information .

   @param  msginfo_list GSList of msgheaders of received mail information 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_imap4_msg_headers(GSList * msginfo_list)
{
	gchar *list = NULL;
	gchar *str = NULL;
	if (msginfo_list == NULL) {
		osso_log(LOG_INFO, "MsgInfo List arg NULL ");
		return OSSO_PACK_INVALID;
	}

	osso_email_utils_pack_from_gslist_to_msginfo(msginfo_list, &list);

	if (list != NULL) {
		str = osso_email_utils_concat_strings(IMAP4_MSG_HEADERS, list);
		xmlFree(list);
	}
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction registerd as a call back function to IMAP module. IMAP4 Module
   invokes this function when engine wants to send errorcode or status of 
   receiving mail.

   @param  status  error code/status of the imap4 module 
   @param  recvcount  mails received  
   @param  totalcount  total number of mails in imap4 module 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

gint engine_dbus_imap4_set_msg(gint msg_number, const gchar * format, ...)
{
	gchar *str = NULL;

	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	va_list ap;
	gint int_val = 0;
	gchar *str_val = NULL;
	gchar *fmt = NULL;

	if (msg_number == 0) {
		return OSSO_PACK_INVALID;
	}
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, &msg_number);

	if (format != NULL) {
		fmt = (gchar *) format;
		va_start(ap, format);
		while (*fmt) {
			switch (*fmt++) {
			case 's':
				str_val = va_arg(ap, char *);
				osso_email_xml_pack_str(root, doc, XML_STRING, &str_val);
				break;
			case 'd':
				int_val = va_arg(ap, int);
				osso_email_xml_pack_int(root, doc, XML_INT, &int_val);
				break;
			}
		}
		va_end(ap);
	}
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(IMAP4_SET_MSG, xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

/**
   This fuction is called when UI requests (through DBUS) for the latest error 
   occured in Engine Module. 

   @param  recv_string contains the request string( Fn Indentifier).  
   @param  reply_str contains engine error code packed in xml format.  
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_get_engine_errno(const gchar * recv_string, gchar ** reply)
{
	gchar *str = NULL;
	gint errorcode = 0;

        osso_log(LOG_DEBUG, "%s called", __FUNCTION__);

	errorcode = al_get_engine_errno();
	str = osso_email_utils_concat_str_int(GET_ENGINE_ERRNO_CALLBACK, errorcode);
	CHECK_STATUS((str == NULL), "Packstr failed", return OSSO_PACK_ERROR);
	if (reply != NULL) {
		*reply = str;
	}
	return OSSO_PACK_OK;
}


/**
   This fuction is called when UI requests (through DBUS) to get the status of
   engine registration to othe other modules.

   @param  recv_string contains the request string( Fn Indentifier).  
   @param  reply_str contains engine inita status packed in xml format.  
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/


osso_pack_return_t engine_dbus_get_engine_init_status(const gchar * recv_string, gchar ** reply)
{
	gchar *str = NULL;
        osso_log(LOG_DEBUG, "%s called", __FUNCTION__);
	str = osso_email_utils_concat_str_int(GET_ENGINE_INIT_STATUS_CALLBACK, engine_init_status);
	CHECK_STATUS((str == NULL), "Packstr failed", return OSSO_PACK_ERROR);
	if (reply != NULL) {
		*reply = str;
	}
	return OSSO_PACK_OK;
}

/**
   This fuction registerd as a call back function to Account Management module. 
   Account Mgnt Module invokes this function when engine wants to send to status   to UI.

   @param  error_code Error code occured in Acct Mgmt module 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_engine_error_code(gint error_code)
{
	gchar *str = NULL;
	if (error_code == 0) {
		return OSSO_PACK_INVALID;
	}
        osso_log(LOG_DEBUG, "%s called", __FUNCTION__);
	str = osso_email_utils_concat_str_int(ACCT_ERROR_CODE, error_code);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}
static gboolean watchdog_timeout_cb(gpointer userdata)
{
	osso_log(LOG_ERR,"Engine exiting by watchdog timer\n");
	exit(0);
}
static void watchdog_engine_exit()
{
	g_timeout_add(2000,watchdog_timeout_cb,NULL);
}
/**
   This Method is called when Engine module receives a DBus message 
   from UI. Then it checks for connection and received message is signal 
   or contains data.  When it receives Disconnected signal, it changes 
   email state to Limited initialisation.

   @param    message     Message contains signal or data
   @param    user_data   user data
   @return   status of DBus Message handler 
*/

static DBusHandlerResult
engine_message_handler(DBusConnection * connection, DBusMessage * message, void *user_data)
{
	DBusMessageIter iter;
	DBusMessage *reply = NULL;
	const gchar *str = NULL;
	gchar *word = NULL;
	gchar *reply_string = NULL;
	const gchar *attachments_dir = NULL;
	gint idx = 0, status = 0;

	gchar word1[MEM_ALLOC_SIZE1];

	if ((connection == NULL) || (message == NULL)) {
		osso_log(LOG_INFO, "Invalid values received for connection" "or message\n");

		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	if (dbus_message_is_signal(message, dbus_message_get_interface(message), "Disconnected")) {
                osso_log(LOG_INFO, "Disconnected signal received \n");
		watchdog_engine_exit();
		connection = NULL;
		connection_flag = FALSE;
		osso_email_gconf_get_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, &status);
		if (status == SEND_RECV_PROGRESS) {
			osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS,
							     SEND_RECV_DONE);
		}
		al_cleanup_sessions();
		if ((emailMgrState == OSSO_STATE_INIT_COMPLETE) &&
		    (gl_send_recv_dbus_handler_busy == FALSE)) {
			osso_log(LOG_INFO, "cleanup");
			attachments_dir = get_tem_attach_dir();
			remove_dir_recursive(attachments_dir);
			osso_email_gconf_unset(EMAIL_GCONF_UI_CLOSED);
			engine_dbus_exit_on_check();
		} else {
			osso_log(LOG_DEBUG, "\nEngine is busy in send/recv\n");
		}
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	dbus_message_iter_init(message, &iter);
	dbus_message_iter_get_basic(&iter, &str);

	if (str == NULL) {
		osso_log(LOG_INFO, " Dbus Message received invalid string \n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	word = (gchar *) strstr(str, seperator);
	if (word == NULL) {
		osso_log(LOG_INFO, "Got a NULL value in the string \n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	memset(word1, '\0', MEM_ALLOC_SIZE1);
	strncpy(word1, str, word - str);

	for (idx = 0; engineDbusFuncList[idx].hname != NULL; idx++) {
		if (g_ascii_strcasecmp(word1, engineDbusFuncList[idx].hname) == 0) {
                        osso_log(LOG_DEBUG, "%s message received", word1);
			gl_normal_dbus_handler_busy = TRUE;
			engineDbusFuncList[idx].msghandler(str, &reply_string);
			gl_normal_dbus_handler_busy = FALSE;
			break;
		}
	}


	if (reply_string == NULL) {
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	reply = dbus_message_new_method_return(message);
	if (!(dbus_message_append_args(reply, DBUS_TYPE_STRING, &reply_string, DBUS_TYPE_INVALID))) {
		dbus_message_unref(reply);
		g_free(reply_string);
		osso_log(LOG_INFO, "Could not append the arguments\n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	g_free(reply_string);
	if (!(dbus_connection_send(connection, reply, NULL))) {
		dbus_message_unref(reply);
		osso_log(LOG_ERR, "Unable to send the reply as a callback\n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	dbus_connection_flush(connection);
	dbus_message_unref(reply);
	return DBUS_HANDLER_RESULT_HANDLED;
}


/**
   This Method is called when Engine module receives a DBus message from UI.
   Then it checks for connection and received message is signal or contains 
   data.  When it receives Disconnected signal, it changes email state to
   Limited initialisation.
 
   @param    connection  Connection established with UI Dbus module
   @param    message     Message contains signal or data
   @param    user_data   user data
   @return   status of DBus Message handler 
 */

static DBusHandlerResult
engine_send_recv_message_handler(DBusConnection * connection,
				 DBusMessage * message, void *user_data)
{
	DBusMessageIter iter;
	const gchar *str = NULL;
	gchar *word = NULL;
	gchar *reply_string = NULL;
	gint idx = 0, status = 0;
	const gchar *attachments_dir = NULL;

	gchar word1[MEM_ALLOC_SIZE1];

	if ((connection == NULL) || (message == NULL)) {
		osso_log(LOG_INFO, "Invalid values received for connection" "or message\n");
		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	}

	if (dbus_message_is_signal(message, dbus_message_get_interface(message), "Disconnected")) {
		watchdog_engine_exit();
		osso_log(LOG_INFO, " Disconnected signal received \n");
		connection = NULL;
		connection_flag = FALSE;
		osso_email_gconf_get_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, &status);
		if (status == SEND_RECV_PROGRESS) {
			osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS,
							     SEND_RECV_DONE);
		}
		al_cleanup_sessions();

		if ((emailMgrState == OSSO_STATE_INIT_COMPLETE) &&
		    (gl_normal_dbus_handler_busy == FALSE)) {

			attachments_dir = get_tem_attach_dir();
			remove_dir_recursive(attachments_dir);
			osso_log(LOG_INFO, " Disconnected2 signal received \n");
			osso_email_gconf_unset(EMAIL_GCONF_UI_CLOSED);
			engine_dbus_exit_on_check();
		} else {
			osso_log(LOG_DEBUG, "\n Engine busy\n");
		}
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	dbus_message_iter_init(message, &iter);
	dbus_message_iter_get_basic(&iter, &str);

	if (str == NULL) {
		osso_log(LOG_INFO, " Dbus Message received invalid string \n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	word = (gchar *) strstr(str, seperator);
	if (word == NULL) {
		osso_log(LOG_INFO, "Got a NULL value in the string \n");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	memset(word1, '\0', MEM_ALLOC_SIZE1);
	strncpy(word1, str, word - str);

	for (idx = 0; engineDbusFuncList[idx].hname != NULL; idx++) {
		if (g_ascii_strcasecmp(word1, engineDbusFuncList[idx].hname) == 0) {
                        osso_log(LOG_DEBUG, "%s message received", word1);
			gl_send_recv_dbus_handler_busy = TRUE;
			engineDbusFuncList[idx].msghandler(str, &reply_string);
			gl_send_recv_dbus_handler_busy = FALSE;
			break;
		}
	}

	return DBUS_HANDLER_RESULT_HANDLED;
}


/**
   This fuction is used to send mail. This function extracts the
   msginfo and message body from the packed string.

   @param  recv_string Packed string contains the 
                       function Identifier:msginfo:body 
   @param  reply  Nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_send_mail(const gchar * recv_str, gchar ** reply)
{
	MsgInfo *msginfo = NULL;
	gchar *msg_to_send = NULL;
	gchar *temp = NULL;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

        osso_log(LOG_DEBUG, "[%s] starts", __FUNCTION__);

	GET_DATA_FROM_RECV_STRING(recv_str, temp, return OSSO_PACK_ERROR);

	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem Alloc Failed for msginfo struct");
		return OSSO_PACK_ALLOC_FAIL;
	}

	DBUS_XML_UNPACK_INIT(temp, return OSSO_PACK_ERROR);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_INFO) == 0) {
				osso_email_xml_unpack_msginfo(node, doc, XML_MSG_INFO, msginfo);
			}
			if (g_ascii_strcasecmp(node->name, XML_MESSAGE) == 0) {
				osso_email_xml_unpack_str_base64(node, doc, XML_MESSAGE,
								 &msg_to_send);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	al_send_mail(msginfo, msg_to_send);
	xmlFree(msg_to_send);
	procmsg_msginfo_free(msginfo);
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);
        osso_log(LOG_DEBUG, "[%s] ends", __FUNCTION__);
	return OSSO_PACK_OK;
}

/**
   This fuction is used to send mails in outbox folder. This function 
   extracts the msginfo and message body from the packed string.

   @param  recv_string Packed string contains the 
           function Identifier:accountname:msgid:keep connection alive
   @param  reply Nothing will be returned 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_send_outbox_mail(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gchar *account = NULL, *msgid = NULL;
	gboolean keepconnalive = 0;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_STRING) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_STRING, &account);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
			if (g_ascii_strcasecmp(node->name, XML_BOOL) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &keepconnalive);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	al_send_outbox_mail(account, msgid, keepconnalive);
	xmlFree(msgid);
	xmlFree(account);
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);
	return OSSO_PACK_OK;
}


/**
   This fuction is used to cancel the list of mails. 

   @param  recv_string Packed string contains the function Id:msgId list
   @param  reply  Nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_cancel_send_mail(const gchar * recv_str, gchar ** reply)
{
	GSList *msgid_list = NULL;
	gchar *message = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	osso_email_utils_unpack_string_in_gslist(message, &msgid_list);

	al_send_cancel_mails(msgid_list);

	if (msgid_list != NULL) {
		slist_free_strings(msgid_list);
		g_slist_free(msgid_list);
	}
	return OSSO_PACK_OK;
}

/**
   This fuction is used to save a mail in draft folder.

   @param  recv_string Packed string contains the 
    func_identifer:msginfo:message:flag
   @param reply contains response info for this method, 
    which will be send to UI process. 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/


osso_pack_return_t engine_dbus_send_save_draft_mail(const gchar * recv_str, gchar ** reply)
{
	MsgInfo *msginfo = NULL;
	gchar *msg_to_send = NULL;
	gchar *message = NULL;
	gboolean flag = FALSE;
	/*bug ID:15172 */
	gchar *old_id = NULL;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	if (reply == NULL) {
		osso_log(LOG_INFO, "Invalid arguments ");
		return OSSO_PACK_ERROR;
	}

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem allocation fail for MsgInfo struct ");
		return OSSO_PACK_ALLOC_FAIL;
	}
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_INFO) == 0) {
				osso_email_xml_unpack_msginfo(node, doc, XML_MSG_INFO, msginfo);
			}
			if (g_ascii_strcasecmp(node->name, XML_MESSAGE) == 0) {
				osso_email_xml_unpack_str_base64(node, doc, XML_MESSAGE,
								 &msg_to_send);
			}
			if (g_ascii_strcasecmp(node->name, XML_BOOL) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &flag);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSGID) == 0) {
				osso_email_xml_unpack_str_base64(node, doc, XML_MESSAGE, &old_id);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	al_send_save_draft_mail(msginfo, msg_to_send, flag, old_id);

	xmlFree(msg_to_send);
	procmsg_msginfo_free(msginfo);
	g_free(old_id);

	return OSSO_PACK_OK;
}

/**
   This fuction is used to send the list of mails from the given folder.

   @param  recv_string Packed string contains the funcid:
    xmlData<folder:GsList of msgids>
   @param  reply  nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
static osso_pack_return_t engine_dbus_send_selected_mails(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	GSList *msgid_list = NULL;
	gint f_type = 0;

	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	if (reply == NULL) {
		osso_log(LOG_INFO, "Invalid arguments ");
		return OSSO_PACK_ERROR;
	}

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_INT) == 0) {
				osso_email_xml_unpack_int(node, doc, XML_INT, &f_type);
				continue;
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID_LST) == 0) {
				osso_email_xml_unpack_msgid_list(node, doc, XML_MSG_ID_LST,
								 &msgid_list);
				continue;
			}
		}
	}

	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	al_send_selected_mails(f_type, msgid_list);
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);
	return OSSO_PACK_OK;
}

/**
   This fuction is used to receive the list of attachments of a mail .

   @param  recv_string Packed string contains list of mimeinfo,list attach tags
   @param  reply  nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
static osso_pack_return_t
engine_dbus_recv_msg_multiple_attachments(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	GSList *mimeinfo_list = NULL;
	GSList *attach_tag_list = NULL;
	GSList *msg_info_list = NULL;
	gchar *msgid = NULL;
	GSList *cur_node = NULL;
	gint index = 0;

	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	if (reply == NULL) {
		osso_log(LOG_INFO, "Invalid arguments ");
		return OSSO_PACK_ERROR;
	}

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MIME_INFO_LST) == 0) {
				osso_email_xml_unpack_mimeinfo_list(node, doc,
								    XML_MIME_INFO_LST,
								    &mimeinfo_list);
				continue;
			}
			if (g_ascii_strcasecmp(node->name, XML_ATTACH_TAG_LST) == 0) {
				osso_email_xml_unpack_int_list(node, doc, XML_ATTACH_TAG_LST,
							       &attach_tag_list);
				continue;
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
				continue;
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	msg_info_list = al_recv_multiple_attachments(msgid, mimeinfo_list, attach_tag_list);
	if (msg_info_list == NULL) {
		debug_print("\n Error in decoding attachments");
	}

	for (cur_node = mimeinfo_list; cur_node != NULL; cur_node = cur_node->next, index++) {
		procmime_mimeinfo_free_all((MimeInfo *) cur_node->data);

	}
	g_slist_free(mimeinfo_list);
	g_slist_free(attach_tag_list);
	g_free(msgid);

	engine_dbus_attach_list(msg_info_list);

	/* freeing the list after sending */
	while (msg_info_list) {
		procmsg_msginfo_free((MsgInfo *) msg_info_list->data);
		msg_info_list = g_slist_remove(msg_info_list, msg_info_list->data);
	}
	msg_info_list = NULL;

	return OSSO_PACK_OK;
}

/*
 * Receiving Functionality
 */

/**
   This fuction is used to Receive a mail for the specified account 

   @param  recv_string Packed string contains the function Id:account name
   @param  reply nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_recv_mail(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	osso_log(LOG_INFO, "al_recv_mail \n");
	al_recv_mail(message, FALSE);
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);
	return OSSO_PACK_OK;
}
/**
   This fuction is used to update the seen flag for imap unread mail,
   after the mail is read.

   @param  recv_string Packed string contains the function Id:account name
   @param  reply nothing will be returned
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/
osso_pack_return_t
engine_dbus_update_seen_flag(const gchar * recv_str, gchar ** reply)
{
	gchar *msgid = NULL;
	gchar *message = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	if (reply == NULL) {
		osso_log(LOG_INFO, "Invalid arguments ");
		return OSSO_PACK_ERROR;
	}

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
				break;
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	al_imap_set_seen(msgid);
	xmlFree(msgid);
	return OSSO_PACK_OK;

	
}
/**
   This fuction is used to receive a message body of the specified mail(msgid)

   @param  recv_string Packed string contains the function Identifier:msgid
   @param  reply  returned with packed mimeinfo of given msgid
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_recv_msg_body(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gchar *mime_message = NULL;
	gchar *msgid = NULL;
	MimeInfo *mimeinfo = NULL;
	gchar *pack_str = NULL;
	gboolean complete_download = FALSE;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gchar *signer_email_id = NULL;
	gchar *recip_email_id = NULL;
	int smime_type = 0;
	gboolean cb_flag = FALSE;
	gboolean read_flag = FALSE;
	gchar *account_name = NULL;
	PrefsAccount *account = NULL;
	gboolean edit_flag = FALSE;
	gboolean html_flag = FALSE;
	guint cert_id = 0;
	gchar *password = NULL;
	gboolean local_file = FALSE;
	gboolean remote_copy = FALSE;
	gchar *file_path = NULL;


	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	node = cur->xmlChildrenNode;
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
			if (g_ascii_strcasecmp(node->name, XML_BOOL) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &complete_download);
			}

			/*smime changes */
			if (g_ascii_strcasecmp(node->name, SENDER_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, SENDER_ID, &signer_email_id);
			}

			if (g_ascii_strcasecmp(node->name, RECIP_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, RECIP_ID, &recip_email_id);
			}

			if (g_ascii_strcasecmp(node->name, XML_INT) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_INT, &smime_type);
			}
			if (g_ascii_strcasecmp(node->name, CB_FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, NULL, &cb_flag);
			}
			if (g_ascii_strcasecmp(node->name, READ_FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, NULL, &read_flag);
			}
			if (g_ascii_strcasecmp(node->name, PASS) == 0) {
				osso_email_xml_unpack_str(node, doc, PASS, &password);
			}
			if (g_ascii_strcasecmp(node->name, CERT_ID) == 0) {
				osso_email_xml_unpack_int(node, doc, XML_INT, &cert_id);
			}
			if (g_ascii_strcasecmp(node->name, EDIT_FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &edit_flag);
			}
			if (g_ascii_strcasecmp(node->name, HTML_FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &html_flag);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	local_file = FALSE;
	file_path = folder_get_message_file_path(msgid);
	if (is_file_entry_exist(file_path) == TRUE) {
		local_file = TRUE;
	}

	mimeinfo = al_recv_msg_body(msgid, complete_download, signer_email_id,
				    recip_email_id, smime_type, read_flag,
				    edit_flag, html_flag, cert_id, password);
	if (msgid != NULL) {
		remote_copy = folder_is_msg_remote(msgid);
	}

	if (cb_flag == FALSE) {
		if (mimeinfo == NULL) {
			account_name = (gchar *) get_account_name(msgid);

			if (account_name != NULL) {
				account = (PrefsAccount *) account_get_account_cb(account_name);
				g_free(account_name);
			}
			if (account == NULL) {
				engine_dbus_pop3_recv_body_complete(mimeinfo, msgid, remote_copy);

			} else if ((account->receive_settings.protocol != PROTO_POP3) ||
				   (local_file == TRUE)) {
				engine_dbus_pop3_recv_body_complete(mimeinfo, msgid, remote_copy);
			} else if (is_file_entry_exist(file_path) == FALSE) {
				/* could not download the message */
				engine_dbus_pop3_recv_body_complete(mimeinfo, msgid, remote_copy);
			}

		} else {
			engine_dbus_pop3_recv_body_complete(mimeinfo, msgid, remote_copy);
		}
		procmime_mimeinfo_free_all(mimeinfo);
		g_free(msgid);
	} else {

		if (mimeinfo == NULL) {
			osso_log(LOG_INFO, "Received NULL value for mimeinfo \n");
		} else {
			mime_message = osso_email_utils_pack_mimeInfo(mimeinfo);
			procmime_mimeinfo_free_all(mimeinfo);
		}

		if (mime_message != NULL) {
			pack_str =
			    osso_email_utils_concat_strings(RECV_MSG_BODY_CALLBACK,
							    (gchar *) mime_message);
			xmlFree(mime_message);
		} else {
			pack_str = osso_email_utils_concat_strings(RECV_MSG_BODY_CALLBACK, "NULL");
		}
		g_free(msgid);
		*reply = pack_str;
	}
	g_free(file_path);
	return OSSO_PACK_OK;
}

/**
   This fuction is used to receive get attachment of specified mail

   @param  recv_string Packed string contains the functionId:mimeinfo:msgid
   @param  reply  returned with packed msginfo of the attachment
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_recv_msg_attachment(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	MimeInfo *mimeinfo = NULL;
	MsgInfo *msginfo = NULL;
	gchar *msgid = NULL;
	gchar *msginfo_string = NULL, *pack_str = NULL;
	gint succ = 0;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	mimeinfo = g_new0(MimeInfo, 1);
	if (mimeinfo == NULL) {
		osso_log(LOG_ERR, " Memeory Alloc failed for MimeInfo struct");
		return OSSO_PACK_ALLOC_FAIL;
	}
	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	node = cur->xmlChildrenNode;
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MIME_INFO) == 0) {
				osso_email_xml_unpack_mimeinfo(node, doc, XML_MIME_INFO, mimeinfo);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	if (mimeinfo != NULL) {
		msginfo = al_recv_msg_attachment(msgid, mimeinfo);
		osso_email_utils_mimeinfo_free_all(mimeinfo);
	}
	xmlFree(msgid);
	if (msginfo == NULL) {
		osso_log(LOG_INFO, "Unable to receive message attachment\n");
		pack_str = osso_email_utils_concat_strings(REMOVE_ATTACHMENT_CALLBACK, NULL);
                if (reply != NULL) {
                       *reply = pack_str;
                }
		return OSSO_PACK_ERROR;
	}

	succ = osso_email_utils_pack_msgInfo(msginfo, &msginfo_string);
	procmsg_msginfo_free(msginfo);
	if (succ != OSSO_PACK_OK) {
		osso_log(LOG_INFO, "Unable to pack data properly \n");
		return OSSO_PACK_ERROR;
	}
	pack_str =
	    osso_email_utils_concat_strings(RECV_MSG_ATTACHMENT_CALLBACK, (gchar *) msginfo_string);
	xmlFree(msginfo_string);
	CHECK_STATUS((pack_str == NULL), "Packstr fail", return OSSO_PACK_ERROR);
	*reply = pack_str;
	return OSSO_PACK_OK;
}


/**
   This fuction is used to receive get multipart of the message 

   @param  recv_string Packed string contains the 
    function Id:mimeinfo:msgid
   @param  reply  returned with filename which contains
    multipart body of mail 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_recv_multipart_body(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	MimeInfo *mimeinfo = NULL;
	gchar *msgid = NULL, *pack_str = NULL, *filename = NULL;
	gchar *default_font = NULL;

	xmlNodePtr cur = NULL, node = NULL;
	xmlNodePtr root = NULL;
	xmlDocPtr doc = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;
	osso_pack_return_t rv = OSSO_PACK_ERROR;
	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:mimeinfo:msgid
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	mimeinfo = g_new0(MimeInfo, 1);
	if (mimeinfo == NULL) {
		osso_log(LOG_ERR, " Memeory Alloc failed for MimeInfo struct");
		return OSSO_PACK_ALLOC_FAIL;
	}
	DBUS_XML_UNPACK_INIT(message, return rv);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MIME_INFO) == 0) {
				osso_email_xml_unpack_mimeinfo(node, doc, XML_MIME_INFO, mimeinfo);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
			if (g_ascii_strcasecmp(node->name, XML_STRING) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_STRING, &default_font);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	filename = al_recv_multipart_body(mimeinfo, msgid, default_font);

	xmlFree(msgid);
	osso_email_utils_mimeinfo_free_all(mimeinfo);
	if (filename != NULL) {
		DBUS_XML_PACK_INIT;
		osso_email_xml_pack_str(root, doc, XML_STRING, &filename);
		xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
		DBUS_XML_PACK_UN_INIT;
		g_free(filename);
	}

	pack_str = osso_email_utils_concat_strings(RECV_MULTIPART_BODY_CALLBACK, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);

	*reply = pack_str;
	return OSSO_PACK_OK;
}

/**
   This fuction is used to cancel the recciving mail

   @param  recv_string Packed string contains the function Identifier
   @param  reply  return nothing
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_cancel_recv_mail(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	al_cancel_recv_mail(message);

	return OSSO_PACK_OK;
}
static osso_pack_return_t engine_dbus_folder_cancel(const gchar * recv_str, gchar ** reply)
{
	/*gchar *message = NULL;

	   GET_DATA_FROM_RECV_STRING (recv_str, message, return OSSO_PACK_ERROR);
	 */
	al_folder_cancel_mail();

	return OSSO_PACK_OK;
}
static osso_pack_return_t
engine_dbus_recv_multipart_attachment(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	MimeInfo *mimeinfo = NULL;
	gchar *msgid = NULL;
	gchar *default_font = NULL;
	gint attach_flag = 0;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	mimeinfo = g_new0(MimeInfo, 1);
	if (mimeinfo == NULL) {
		osso_log(LOG_ERR, " Memeory Alloc failed for MimeInfo struct");
		return OSSO_PACK_ALLOC_FAIL;
	}
	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	node = cur->xmlChildrenNode;
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MIME_INFO) == 0) {
				osso_email_xml_unpack_mimeinfo(node, doc, XML_MIME_INFO, mimeinfo);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
			if (g_ascii_strcasecmp(node->name, ATTA_TAG) == 0) {
				osso_email_xml_unpack_int(node, doc, ATTA_TAG, &attach_flag);
			}
			if (g_ascii_strcasecmp(node->name, XML_STRING) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_STRING, &default_font);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	if (mimeinfo != NULL) {
		al_recv_multipart_attachment(msgid, mimeinfo, attach_flag, default_font);
		osso_email_utils_mimeinfo_free_all(mimeinfo);
	}
	xmlFree(msgid);
	return OSSO_PACK_OK;
}

/* 
 * Folder Management function
 */

/**
   This fuction is used to pack folder status with folder callback header

   @param  status status of folder operation performed 
   @param  reply  contains foldercallbackid:status on success
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static osso_pack_return_t engine_dbus_folder_operation_status(gint status, gchar ** reply)
{
	if (reply == NULL) {
		return OSSO_PACK_ERROR;
	}
	*reply = osso_email_utils_concat_str_int(FOLDER_OPERATION_CALLBACK, status);
	if (*reply == NULL) {
		osso_log(LOG_INFO, "Pack sting failed ");
		return OSSO_PACK_ERROR;
	}
	return OSSO_PACK_OK;
}

/**
   This method is used to create new folder  

   @param  recv_str Packed string contains the functionId:foldername
   @reply  param returned NULL or string contains callback information 
    of folder creation 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_create(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *name = NULL;
	gchar *path = NULL;
	gchar *message = NULL;
	FolderType type = F_LOCAL;

	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &name);
			} else if (g_ascii_strcasecmp(node->name, PATH) == 0) {
				osso_email_xml_unpack_str(node, doc, PATH, &path);
			} else if (g_ascii_strcasecmp(node->name, TYPE) == 0) {
				osso_email_xml_unpack_int(node, doc, TYPE, &type);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_create(name, path, type);
	if (succ) {
		osso_log(LOG_ERR, "Unable to create folder\n");
	}
	xmlFree(name);
	xmlFree(path);
	return engine_dbus_folder_operation_status(succ, reply);
}


/**
   This method is used to delete the existing folder

   @param  recv_str Packed string contains the functionId:foldername
   @param  reply returned NULL or string contains callback information of 
    folder deletion 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_delete(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *name = NULL;
	gchar *message = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:name
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &name);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_delete(name);
	if (succ) {
		osso_log(LOG_ERR, "Unable to delete folder\n");
	}
	xmlFree(name);

	return engine_dbus_folder_operation_status(succ, reply);
}


/**
   This method is used to rename the existing folder

   @param  recv_str  Packstring contains the functionId:oldname:newname
   @param  reply returned   NULL or string contains callback information of 
    folder to be renamed 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_rename(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *oldname = NULL;
	gchar *newname = NULL;
	gchar *message = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_OLD_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_OLD_NAME, &oldname);
			} else if (g_ascii_strcasecmp(node->name, FOLDER_NEW_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NEW_NAME, &newname);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_rename(oldname, newname);
	if (succ) {
		osso_log(LOG_ERR, "Unable to rename folder\n");
	}
	xmlFree(oldname);
	xmlFree(newname);
	return engine_dbus_folder_operation_status(succ, reply);
}


/**
   This method is used to move the message from one folder to another folder

   @param   recv_string  Packed string contains the 
    functionId:src folder:des folder:msgid
   @param  replay returned   NULL or string contains callback 
    information of message  to be moved
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_move_msg(const gchar * recv_str, gchar ** reply)
{
	gchar *srcfolder = NULL;
	gchar *destfolder = NULL;
	gchar *msgid = NULL;
	gchar *message = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:srcfolder:destfolder:msgid
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, SRC_FOLDER)
			    == 0) {
				osso_email_xml_unpack_str(node, doc, SRC_FOLDER, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, DES_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, DES_FOLDER, &destfolder);
			} else if (g_ascii_strcasecmp(node->name, MSGID) == 0) {
				osso_email_xml_unpack_str(node, doc, MSGID, &msgid);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_move_msg(srcfolder, destfolder, msgid);

	if (succ) {
		osso_log(LOG_ERR, "Unable to move msg from src to dest folder\n");
	}
	xmlFree(srcfolder);
	xmlFree(destfolder);
	xmlFree(msgid);

	return engine_dbus_folder_operation_status(succ, reply);
}


/**
   This method is used to move the messages from one folder to another folder

   @param   recv_string  Packed string contains the 
    functionId:src folder:des folder:GSList of msgids
   @param  reply returned   NULL or string contains callback information 
   of messages to be moved
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_move_msgs(const gchar * recv_str, gchar ** reply)
{
	gchar *srcfolder = NULL;
	gchar *destfolder = NULL;
	gchar *message = NULL;
	GSList *msgid_list = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:srcfolder:destfolder:msg_string
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, SRC_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, SRC_FOLDER, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, DES_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, DES_FOLDER, &destfolder);
			} else if (g_ascii_strcasecmp(node->name, XML_MSG_ID_LST) == 0) {
				osso_email_xml_unpack_msgid_list(node, doc, XML_MSG_ID_LST,
								 &msgid_list);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_move_msgs(srcfolder, destfolder, msgid_list);

	if (succ) {
		osso_log(LOG_ERR, "Unable to move msgs from src to dest folder\n");
	}
	xmlFree(srcfolder);
	xmlFree(destfolder);
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);
	return OSSO_PACK_OK;
}

/**
   This method is used to remove the message from the existing folder

   @param   recv_string  Packed string contains the
    functionId:folder:msgid:flag
   @param  reply returned NULL or string contains callback information 
    of message to be removed locally or in the server also
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_remove_msg(const gchar * recv_str, gchar ** reply)
{
	gchar *folder = NULL;
	gchar *msgid = NULL;
	gchar *message = NULL;
	gboolean flag = FALSE;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:folder:msgid
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &folder);
			} else if (g_ascii_strcasecmp(node->name, MSGID) == 0) {
				osso_email_xml_unpack_str(node, doc, MSGID, &msgid);
			} else if (g_ascii_strcasecmp(node->name, FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, FLAG, &flag);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_remove_msg(folder, msgid, flag);
	if (succ) {
		osso_log(LOG_ERR, "Unable to remove msg from folder\n");
	}
	xmlFree(folder);
	xmlFree(msgid);

	return engine_dbus_folder_operation_status(succ, reply);
}

/**
   This method is used to remove the messages from the existing folder

   @param   recv_string  Packed string contains the
    functionId:folder:msgid list :flag
   @param  reply returns NULL  or string contains callback information 
    of messages to be removed locally or in the server also
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_remove_msgs(const gchar * recv_str, gchar ** reply)
{
	gchar *srcfolder = NULL;
	gchar *message = NULL;
	gboolean flag = 0;
	GSList *msgid_list = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:srcfolder:msg_string
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, XML_MSG_ID_LST) == 0) {
				osso_email_xml_unpack_msgid_list(node, doc, XML_MSG_ID_LST,
								 &msgid_list);
			} else if (g_ascii_strcasecmp(node->name, FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, FLAG, &flag);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_remove_msgs(srcfolder, msgid_list, flag, TRUE);

	if (succ) {
		osso_log(LOG_ERR, " Unable to  remove msgs from folder = %s \n", srcfolder);
	}
	xmlFree(srcfolder);
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);

	return OSSO_PACK_OK;
}


/**
   This method is used to remove the all messages from the existing folder

   @param   recv_string  Packed string contains the
                         functionId:folder:flag
   @param  reply returns with NULL or string contains callback information 
    of all messages to be removed locally or in the server also
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_remove_all_msg(const gchar * recv_str, gchar ** reply)
{
	gchar *srcfolder = NULL;
	gchar *message = NULL;
	gboolean flag = FALSE;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;
	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:folder
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, FOLDER_NAME)
			    == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, FLAG, &flag);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_remove_all_msg(srcfolder, flag);
	if (succ) {
		osso_log(LOG_ERR, "Unable to remove all msgs from folder\n");
	}
	xmlFree(srcfolder);

	return OSSO_PACK_OK;
}


/**
   This method is used to get the list of existing folders in Mail directory

   @param   recv_string  Packed string contains the functionId:
   @param  reply returned with NULL or string contains callback information 
   of all folders list 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_get_list(const gchar * recv_str, gchar ** reply)
{

	xmlChar *xmlbuff = NULL;
	gint buffersize = 0;
	GSList *list = NULL;
	xmlNodePtr root = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;
	gchar *pack_str = NULL;

	succ = al_folder_get_list(&list);
	if (succ) {
		osso_log(LOG_INFO, "Successfully got the folder list\n");
	}

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, FOLDER_STATUS, &succ);
	osso_email_xml_pack_folder_list(root, doc, XML_FOLDER_LST, list);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	pack_str = osso_email_utils_concat_strings(FOLDER_GET_LIST_CALLBACK, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Packstr fail", return OSSO_PACK_ERROR);
	if (reply != NULL) {
		*reply = pack_str;
	}
	return OSSO_PACK_OK;
}


/**
   This method is used to get the list of messages in the folder 

   @param   recv_string  Packed string contains the functionId:foldername
   @param  reply returned with NULL or string contains callback information 
   of all meesages list for the specified folder
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_populate_folder(const gchar * recv_str, gchar ** reply)
{
	gint index = 0;
	gchar *message = NULL;
	gchar *name = NULL;
	GSList *list = NULL;
	gchar *msg_string = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gchar *pack_str = NULL;

	/*for cache changes */
	gchar *maildir = NULL;
	gchar *arch_path = NULL;

	/*for bug: 17713 */
	gboolean is_last = TRUE;
	gchar *xml_cache = NULL;


	/*
	 * Skip the func_identifier which is in the beginning the recv_str.
	 * Format of the recv_str is: func_identifer:name:index
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);


	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &name);
			} else if (g_ascii_strcasecmp(node->name, INDEX) == 0) {
				osso_email_xml_unpack_int(node, doc, INDEX, &index);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	if (g_ascii_strcasecmp(name, ARCHIVE) == 0) {
		arch_path = al_folder_get_archive_path();
	}

	maildir = g_strdup_printf("%s/%s", get_mail_dir(), name);

	/* checking if cache file is up to date. For this time stamp on
	 * cache.xml and time on folder is used
	 */
	if (xml_cache_is_up_to_date(maildir, arch_path)) {
		osso_log(LOG_INFO, "cache for %s up-to-date: using cache", maildir);
		*reply = xml_cache_get(maildir, index);
		if (!*reply) {
			osso_log(LOG_WARNING, "could not get valid cache for %s", maildir);
		} else {
			g_free(maildir);
			g_free(name);
			g_free(arch_path);
			return OSSO_PACK_OK;
		}
	}


	if (index == 0) {
		xml_cache = g_strdup_printf("%s/%s", maildir, XML_CACHE_FILE);
		xml_cache_remove(xml_cache);
		g_free(xml_cache);
	}

	list = al_populate_folder(name, index);
	xmlFree(name);

	if (list == NULL) {
		osso_log(LOG_INFO, "No Messages to populate in folder\n");
	} else {
		osso_email_utils_pack_populate_folder_to_gslist(list, &msg_string);
		procmsg_msgheader_list_free(list);
	}
	pack_str = osso_email_utils_concat_strings(POPULATE_FOLDER_CALLBACK, (gchar *) msg_string);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ",;
	    );
	if (msg_string != NULL) {
		xmlFree(msg_string);
		is_last = FALSE;
	}
	if (reply != NULL) {
		osso_log(LOG_INFO, "cache for %s is not up-to-date, updating now", maildir);

		/*update cache for mails from index to (index + NUM_MAILS) or
		 * maximum available mails
		 */
		xml_cache_update(maildir, pack_str, strlen(pack_str), index, is_last);
		*reply = pack_str;
	}
	g_free(maildir);
	g_free(arch_path);
	return OSSO_PACK_OK;
}


/**
   This method is used to copy the message from one folder to another folder 

   @param   recv_string  Packed string contains the
    functionId:src folder: des folder:msgid
   @param   reply returned with NULL or string contains callback information 
    of result of the operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_copy_msg(const gchar * recv_str, gchar ** reply)
{
	gchar *srcfolder = NULL;
	gchar *destfolder = NULL;
	gchar *msgid = NULL;
	gchar *message = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:srcfolder:destfolder:msgid
	 */

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, SRC_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, SRC_FOLDER, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, DES_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, DES_FOLDER, &destfolder);
			} else if (g_ascii_strcasecmp(node->name, MSGID) == 0) {
				osso_email_xml_unpack_str(node, doc, MSGID, &msgid);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_folder_copy_msg(srcfolder, destfolder, msgid);

	if (succ) {
		osso_log(LOG_INFO, "Unable to copy msg from src to destfolder\n");
	}
	xmlFree(srcfolder);
	xmlFree(destfolder);
	xmlFree(msgid);
	return engine_dbus_folder_operation_status(succ, reply);
}


/**
   This method is used to copy the messages from one folder to another folder

   @param   recv_string  Packed string contains the
   functionId:src folder: des folder:GSList of msgids
   @param   reply returned with NULL or string contains callback information 
   of result of the operation 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_copy_msgs(const gchar * recv_str, gchar ** reply)
{
	/*gint succ = 0; */
	gchar *srcfolder = NULL;
	gchar *destfolder = NULL;
	gchar *message = NULL;
	GSList *msgid_list = NULL;

	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: funcid:srcfolder:destfolder:msg_string
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, SRC_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, SRC_FOLDER, &srcfolder);
			} else if (g_ascii_strcasecmp(node->name, DES_FOLDER) == 0) {
				osso_email_xml_unpack_str(node, doc, DES_FOLDER, &destfolder);
			} else if (g_ascii_strcasecmp(node->name, XML_MSG_ID_LST) == 0) {
				osso_email_xml_unpack_msgid_list(node, doc, XML_MSG_ID_LST,
								 &msgid_list);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	al_folder_copy_msgs(srcfolder, destfolder, msgid_list);

	xmlFree(srcfolder);
	xmlFree(destfolder);
	slist_free_strings(msgid_list);
	g_slist_free(msgid_list);
	return OSSO_PACK_OK;
}

/**
   This method is used to get the size of user defined folder 

   @param   recv_string  Packed string contains the functionId:src folder
   @param   reply returned with NULL or string contains folder size 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static osso_pack_return_t engine_dbus_folder_get_size(const gchar * recv_str, gchar ** reply)
{
	gint fsize = 0;
	gchar *srcfolder = NULL;
	gchar *message = NULL;

	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, FOLDER_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, FOLDER_NAME, &srcfolder);
			}
		}
	}
	fsize = al_get_folder_size(srcfolder);
	g_free(srcfolder);
	return engine_dbus_folder_operation_status(fsize, reply);
}


/**
   This method is used to remove the attachments of a mail

   @param   recv_string  Packed string contains the
                         functionId:GSlist of attachments
   @param   reply returned with NULL or string contains callback information
    of result of the operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_folder_remove_attachment(const gchar * recv_str, gchar ** reply)
{
	GSList *folder_attach_list = NULL;
	GSList *out_list = NULL;
	GSList *in_list = NULL;
	GSList *cur = NULL;
	gchar *message = NULL, *pack_str = NULL;
	gchar *msg_string = NULL;
	FolderAttachInfo *fold_info = NULL;

	/*
	 * Skip the func_identifier which is in the beginning the recv_str.
	 * Format of the recv_str is: func_identifer:GSList (FolderAttachInfo)
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	/*osso_email_utils_unpack_folderattach_in_gslist ( message, 
	   &folder_attach_list ); */

	osso_email_utils_unpack_msgid_list_in_gslist(message, (GSList **) & in_list);
	gchar *msgid = in_list->data;
	cur = in_list->next;
	for (; cur != NULL; cur = cur->next) {
		fold_info = g_new0(FolderAttachInfo, 1);
		fold_info->msgid = g_strdup(msgid);
		fold_info->section_number = g_strdup(cur->data);
		folder_attach_list = g_slist_append(folder_attach_list, fold_info);
	}

	if (folder_attach_list == NULL) {
		osso_log(LOG_INFO, " Unpack failed : attachment list null \n");
		pack_str = osso_email_utils_concat_strings(REMOVE_ATTACHMENT_CALLBACK, NULL);
		if (reply != NULL) {
			*reply = pack_str;
		}
		return OSSO_PACK_ERROR;
	}
	out_list = al_folder_remove_attachment(folder_attach_list);
	/* TODO free the folder attachment list & in_list */
	/*osso_email_utils_folder_attach_free ( folder_attach_list ); */

	if (out_list == NULL) {
		osso_log(LOG_ERR, "Unable to remove folder attachment\n");
	} else {
		osso_email_utils_pack_msgid_to_gslist(out_list, &msg_string);
		/* free the string */
		slist_free_strings(out_list);
		g_slist_free(out_list);
	}


	pack_str =
	    osso_email_utils_concat_strings(REMOVE_ATTACHMENT_CALLBACK, (gchar *) msg_string);
	CHECK_STATUS((pack_str == NULL), "Pack str failed",;);
	xmlFree(msg_string);
	if (reply != NULL) {
		*reply = pack_str;
	}

	return OSSO_PACK_OK;
}

/* 
 * Account Management Function
 */

/**
   This method is used to get the information of account list

   @param   recv_string  Packed string contains the
                         functionId:
   @reply  NULL or string contains callback information of account list 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_get_account_list(const gchar * recv_str, gchar ** reply)
{
	GSList *ac_list = NULL;
	gchar *msg_string = NULL;
	gchar *pack_str = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	CHECK_STATUS((recv_str == NULL), "Pack str failed ", return rv);
	al_get_account_list(&ac_list);
	if (ac_list == NULL) {
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb)) {
			email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb(FALSE);
		}
	}
	if (ac_list != NULL) {
		osso_email_utils_pack_from_gslist_to_string(ac_list, &msg_string);
		slist_free_strings(ac_list);
		g_slist_free(ac_list);
		if (msg_string == NULL) {
			osso_log(LOG_INFO, " pack  account list failed \n");
			return rv;
		}
	}
	pack_str = osso_email_utils_concat_strings(GET_ACCOUNT_LIST_CALLBACK, (gchar *) msg_string);
	if (msg_string != NULL) {
		xmlFree(msg_string);
	}
	CHECK_STATUS((pack_str == NULL), "pack str failed ", return rv);
	if (reply != NULL) {
		*reply = pack_str;
	}
	return OSSO_PACK_OK;
}





/**
   This method is used to get the specified account status 

   @param   recv_string  Packed string contains the
                         functionId:accountname
   @param   reply returned with NULL or string contains the callback information 
            for this account releted operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_account_check_status(const gchar * recv_str, gchar ** reply)
{
	gint status = 0;
	gchar *account = NULL;
	gchar *message = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, ACCOUNT_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, ACCOUNT_NAME, &account);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	status = al_check_accnt_status(account);
	xmlFree(account);

	pack_str = osso_email_utils_concat_str_int(ACCOUNT_CHECK_STATUS_CALLBACK, status);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);
	*reply = pack_str;
	return OSSO_PACK_OK;
}

static osso_pack_return_t engine_dbus_prefs_account_get_list(const gchar * recv_str, gchar ** reply)
{

	xmlChar *xmlbuff = NULL;
	gint buffersize = 0;
	GSList *list = NULL;
	xmlNodePtr root = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;
	gchar *pack_str = NULL;

	succ = al_prefs_account_get_list(&list);
	if (succ) {
		osso_log(LOG_INFO, "Successfully got the folder list\n");
	}

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, &succ);
	osso_email_xml_pack_prefsaccount_list(root, doc, XML_PREFS_ACCOUNT_LST, list);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	pack_str =
	    osso_email_utils_concat_strings(PREFS_ACCOUNT_GET_LIST_CALLBACK, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Packstr fail", return OSSO_PACK_ERROR);
	if (reply != NULL) {
		*reply = pack_str;
	}
	return OSSO_PACK_OK;
}


/**
   This method is used to get the existing account information 

   @param   recv_string  Packed string contains the
                         functionId:accountname:type
   @param   reply  returns the NULL string contains callback 
    account information.
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_get_account_settings(const gchar * recv_str, gchar ** reply)
{
	SettingType type;
	PrefsAccount *prefs_account = NULL;

	gint succ = 0;
	gchar *account = NULL;
	gchar *message = NULL;
	gchar *prefs_string = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	/*
	 * Skip the func_identifier which is in the beginning the recv_str.
	 * Format of the recv_str is: func_identifer:account:type
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return FALSE);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, ACCOUNT_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, ACCOUNT_NAME, &account);
			} else if (g_ascii_strcasecmp((gchar *) node->name, TYPE)
				   == 0) {
				osso_email_xml_unpack_int(node, doc, TYPE, &type);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_get_account_settings(account, type, &prefs_account);
	xmlFree(account);
	if (!succ) {
		osso_log(LOG_ERR, " Unable to get account settings\n");
		return rv;
	}

	osso_email_utils_pack_prefsAccount(prefs_account, &prefs_string);
	account_free_account(prefs_account);

	if (prefs_string == NULL) {
		osso_log(LOG_INFO, " Packing prefs account failed \n");
		return rv;
	}

	pack_str =
	    osso_email_utils_concat_strings(GET_ACCOUNT_SETTINGS_CALLBACK, (gchar *) prefs_string);
	xmlFree(prefs_string);
	CHECK_STATUS((pack_str == NULL), "pack str failed ", return rv);
	*reply = pack_str;
	return OSSO_PACK_OK;
}


/**
   This method is used to set the specified information for given account

   @param   recv_string  Packed string contains the
                         functionId:accountname:type:prefsAccount
   @param   reply return nothing for this operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_set_account_settings(const gchar * recv_str, gchar ** reply)
{
	SettingType type;
	PrefsAccount *prefs_account = NULL;
	gint succ = 0;
	gchar *message = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;
	 gboolean initialise_scheduling = FALSE;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:type:prefs_account
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, TYPE) == 0) {
				osso_email_xml_unpack_int(node, doc, TYPE, &type);
			} else if (g_ascii_strcasecmp((gchar *) node->name, XML_PREFS_ACCOUNT) == 0) {
				prefs_account = g_new0(PrefsAccount, 1);
				if (prefs_account != NULL) {
					osso_email_xml_unpack_prefsaccount(node, doc,
									   XML_PREFS_ACCOUNT,
									   prefs_account);
				}
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	if (prefs_account == NULL) {
		return rv;
	}
	/*Check whether a new account is created ,if no account
          present previously */
        if (!is_account_list_valid()) {
               initialise_scheduling = TRUE;
         }

	succ = al_set_account_settings(type, prefs_account);
	if (!succ) {
		osso_log(LOG_INFO, " Unable to set account settings\n");
		succ = OSSO_PACK_ERROR;
	} else {
		if (initialise_scheduling && (email_mgr_cb != NULL)
		    && (email_mgr_cb->email_mgr_update_scheduling != NULL)) {
			email_mgr_cb->email_mgr_update_scheduling();
		}
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb)) {
			email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb(TRUE);
		}

		succ = OSSO_PACK_OK;
	}
	account_free_account(prefs_account);

	pack_str = osso_email_utils_concat_str_int(SET_ACCOUNT_CALLBACK, succ);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);
	*reply = pack_str;

	return succ;
}


/**
   This method is used to delete the specified account from account list 

   @param   recv_string  Packed string contains the
                         functionId:accountname
   @param   reply returned with NULL or string contains the callback information 
            for this account releted operation
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_delete_account(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *account = NULL;
	gchar *message = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	GSList *ac_list = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, ACCOUNT_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, ACCOUNT_NAME, &account);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_delete_account(account);
	if (!succ) {
		osso_log(LOG_ERR, "Unable to delete account\n");
	}
	xmlFree(account);
	al_get_account_list(&ac_list);
	if (ac_list == NULL) {
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb)) {
			email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb(FALSE);
		}
		/*If There are no account stop scheduling */
	        if (email_mgr_cb && email_mgr_cb->email_mgr_update_scheduling) {
	                email_mgr_cb->email_mgr_update_scheduling();
                }
	} else {
		slist_free_strings(ac_list);
		g_slist_free(ac_list);
	}
	pack_str = osso_email_utils_concat_str_int(DELETE_ACCOUNT_CALLBACK, succ);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);
	*reply = pack_str;
	return OSSO_PACK_OK;
}


/**
   This method is used to rename the account name  

   @param   recv_string  Packed string contains the
                         functionId:oldname:newname
   @param   reply returned with NULL or string contains the callback 
    information for this account releted operation
   @return TRUE on success or FALSE on failure
*/

gint engine_dbus_rename_account(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *message = NULL;
	gchar *oldname = NULL;
	gchar *newname = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	/* 
	 * Skip the func_identifier which is in the beginning the recv_str.  
	 * Format of the recv_str is: func_identifer:account_name
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, OLD_ACCOUNT_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, OLD_ACCOUNT_NAME, &oldname);
			} else if (g_ascii_strcasecmp((gchar *) node->name, NEW_ACCOUNT_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, NEW_ACCOUNT_NAME, &newname);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	succ = al_rename_account(oldname, newname);
	if (!succ) {
		osso_log(LOG_ERR, "Fail to rename account=%s to %s \n", oldname, newname);
	}
	xmlFree(oldname);
	xmlFree(newname);
	pack_str = osso_email_utils_concat_str_int(RENAME_ACCOUNT_CALLBACK, succ);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);
	*reply = pack_str;
	return OSSO_PACK_OK;
}

/**
   This method is used free the data 

   @param   data   string contains the data to be freed  
   @return nothing  
*/

static void test_data_free(gchar * data)
{
	g_free(data);
}

/**
   This method is called new connection is requested 

   @param   server  Server name 
   @param  Dbus connection for the server 
   @param   User data to be passed to the process  
   @return TRUE on succes or FALSe on failure 
*/

static gint
send_recv_connection_callback(DBusServer * server,
			      DBusConnection * send_recv_connection, void *user_data)
{
	gint succ = 0;

	dbus_connection_ref(send_recv_connection);
	dbus_connection_setup_with_g_main(send_recv_connection, NULL);

	succ =
	    dbus_connection_add_filter(send_recv_connection,
				       engine_send_recv_message_handler,
				       NULL, (DBusFreeFunction) test_data_free);
	if (!succ) {
		osso_log(LOG_INFO, "server, Unable to add filter \n");
		return FALSE;
	}
	return succ;
}

/**
   This method is called new connection is requested 

   @param   server  Server name 
   @param  Dbus connection for the server 
   @param   User data to be passed to the process  
   @return TRUE on succes or FALSe on failure 
*/

static gint
new_connection_callback(DBusServer * server, DBusConnection * new_connection, void *user_data)
{
	gint succ = 0;

	dbus_connection_ref(new_connection);
	dbus_connection_setup_with_g_main(new_connection, NULL);

	succ =
	    dbus_connection_add_filter(new_connection, engine_message_handler,
				       NULL, (DBusFreeFunction) test_data_free);
	if (!succ) {
		osso_log(LOG_INFO, "server, Unable to add filter \n");
		return FALSE;
	}
	return succ;
}



/**
   This method is used to establish to start the server thread 

   @return TRUE on success or FALSE
*/

gboolean engine_dbus_server_thread(void)
{
	DBusError error;
	DBusServer *server = NULL, *send_recv_server = NULL;
	GMainContext *ctx_1 = NULL;

	osso_log(LOG_INFO, " engine_dbus_server_thread starting \n");
	dbus_error_init(&error);

	server = dbus_server_listen(DBUS_UNIX_SOCKET_CLNT_PATH, &error);
	if (server == NULL) {
		osso_log(LOG_ERR, "server, Unable to start server %s\n", error.message);
		dbus_error_free(&error);
		return FALSE;
	}
	send_recv_server = dbus_server_listen(DBUS_UNIX_SOCKET_SEND_RECV_PATH, &error);
	if (send_recv_server == NULL) {
		osso_log(LOG_ERR, "server, Unable to start server %s\n", error.message);
		dbus_error_free(&error);
		return FALSE;
	}

	dbus_server_set_new_connection_function(server, (DBusNewConnectionFunction)
						new_connection_callback, NULL, NULL);
	dbus_server_setup_with_g_main(server, NULL);

	dbus_server_set_new_connection_function(send_recv_server, (DBusNewConnectionFunction)
						send_recv_connection_callback, NULL, NULL);
	dbus_server_setup_with_g_main(send_recv_server, NULL);
	ctx_1 = g_main_context_default();
	gl_loop = g_main_loop_new(ctx_1, FALSE);
	g_main_loop_run(gl_loop);
	return TRUE;
}


/**
   This method is called when engine module wants to send mesaage to 
   UI dbus module.
   
   @param message  contains the information send to UI DBus module

   @return TRUE on successful connction or FALSE  
*/

gboolean engine_dbus_connect_to_remote(gchar * msg_to_send)
{
	DBusMessage *message = NULL;
	DBusError error;
	guint ret, succ = 0;
	char *env_string = NULL;
	static gboolean check_connection = FALSE;
	gint status = 0;

	/*DBusDispatchStatus status;  */

	if (msg_to_send == NULL) {
		osso_log(LOG_INFO, " Received NULL value for msg_to_send\n");
		return FALSE;
	}
	if (osso_email_gconf_get_int(EMAIL_GCONF_UI_CLOSED, &status) == FALSE) {
		osso_log(LOG_INFO, "\nERROR in reading GConf\n");
		return FALSE;
	}
	if (status != 0) {
		osso_log(LOG_INFO, "\nEngine dbus: UI is not UP\n");
		return FALSE;
	}
	dbus_error_init(&error);
	if ((connection_flag == FALSE) || (connection == NULL)) {
		osso_log(LOG_INFO, " Opening New connection \n");
		connection = dbus_connection_open(DBUS_UNIX_SOCKET_SERV_PATH, &error);
		if (connection == NULL) {
			osso_log(LOG_ERR, "could not open connection: %s\n", error.message);
			dbus_error_free(&error);
			return FALSE;
		} else {
			connection_flag = dbus_connection_get_is_connected(connection);
			connection_flag = TRUE;
		}
	}
	connection_flag = dbus_connection_get_is_connected(connection);
	if ((connection_flag == FALSE) && (connection == NULL)) {
		connection = dbus_connection_open(env_string, &error);
		if (connection == NULL) {
			osso_log(LOG_ERR, "could not open connection: %s\n", error.message);
			dbus_error_free(&error);
			return FALSE;
		}
	}
	check_connection = dbus_connection_get_is_connected(connection);
	if ((connection != NULL) && (check_connection == TRUE)) {
		dbus_connection_setup_with_g_main(connection, NULL);
		message =
		    dbus_message_new_method_call(EMAIL_DBUS_SERVICE,
						 EMAIL_DBUS_PATH,
						 EMAIL_DBUS_INTERFACE, EMAIL_DBUS_METHOD);
		if (message == NULL) {
			osso_log(LOG_ERR, "Could not allocate DBUS message\n");
			return FALSE;
		}
		succ = dbus_message_append_args(message, DBUS_TYPE_STRING,
						&msg_to_send, DBUS_TYPE_INVALID);
		if (!succ) {
			osso_log(LOG_ERR, "Unable to append args to message \n");
			return FALSE;
		}

		succ = dbus_connection_send(connection, message, &ret);

		if (!succ) {
			osso_log(LOG_ERR, "%s%d: Sending of message Failed\n", __FILE__, __LINE__);
		}
		dbus_connection_flush(connection);
		if (message != NULL) {
			dbus_message_unref(message);
		}
	}

	return TRUE;
}

/**
   This method is called when ui process wants to convert the text data 
   in a file to html data

   @param recv_string  contains the information of
                       funcid:filename:charset
   @param reply returned with  NULL on failure or string contains the html data 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_html_to_text_converter(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gchar *filename = NULL;
	gchar *charset = NULL;
	gchar *convertedtext = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_FILE_NAME) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_FILE_NAME, &filename);
			}
			if (g_ascii_strcasecmp(node->name, XML_CHARSET) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_CHARSET, &charset);
			}
			if (g_ascii_strcasecmp(node->name, XML_STRING) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_STRING, &convertedtext);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	al_html_to_text_converter(filename, charset, &convertedtext);
	xmlFree(filename);
	xmlFree(charset);

	pack_str =
	    osso_email_utils_concat_strings(HTML_TO_TEXT_CONVERTER_CALLBACK,
					    (gchar *) convertedtext);
	xmlFree(convertedtext);
	CHECK_STATUS((pack_str == NULL), "Pack str failed", return rv);
	*reply = pack_str;
	return OSSO_PACK_OK;
}


/*\@}*/
/**
 * \defgroup EmailMgr Email Manager
 */
/*\@{*/

/*\@}*/

/**
   This Metod executed when Email Manager wants to inform Email UI application
   about the window id 

   @param  windowid  Window id of message to be opened.   
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

gint engine_dbus_tasknav_window_id(const gchar * windowid)
{
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	gint buffersize = 0, succ = 0;
	gchar *pack_str = NULL;

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_str(root, doc, XML_STRING, &windowid);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	pack_str = osso_email_utils_concat_strings(TASKNAV_WINDOW_ID, (gchar *) xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Pack str failed",;);
	xmlFree(xmlbuff);

	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s\n", __FUNCTION__);
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

/**
   This fuction is called by email manager. When Email Mgr wants to 
   invoke UI to open editior window(Send As Functionality), this function 
   get called.

   @param  msginfo Message Info struct contains the information suppiled
           by File Manager or Browser
   @param  message Body of the message
   @param reply returned with  NULL on failure or contains the packed data 

   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_editor_create(MsgInfo * msginfo, gchar * message, gchar ** ret_str)
{
	gint succ = 0;

	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	gchar *pack_str = NULL;
	int buffersize = 0;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	CHECK_STATUS((msginfo == NULL), "NULL MsgInfo arg ", return rv);

	DBUS_XML_PACK_INIT;

	osso_email_xml_pack_msginfo(root, doc, XML_MSG_INFO, msginfo);
	if (message != NULL) {
		osso_email_xml_pack_str(root, doc, XML_MESSAGE, &message);
	}
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	DBUS_XML_PACK_UN_INIT;

	pack_str = osso_email_utils_concat_strings(EDITOR_CREATE, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	if (pack_str == NULL) {
		return rv;
	}
	if (ret_str != NULL) {
		*ret_str = g_strdup(pack_str);
	}
	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s\n", __FUNCTION__);
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}


/**
   This fuction is called by email manager. When Email Mgr wants to 
   invoke UI to open/delete specific mail. 

   @param  msgid  message id to be opened by UI 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_tasknav_msg_operation(const gchar * msgid, gint option)
{
	gint succ = 0;
	gchar *pack_str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_str(root, doc, XML_MSG_ID, &msgid);
	osso_email_xml_pack_int(root, doc, XML_INT, &option);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	DBUS_XML_PACK_UN_INIT;

	pack_str = osso_email_utils_concat_strings(TASKNAV_MSG_OPERATION, (gchar *) xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Pack str failed",;
	    );
	xmlFree(xmlbuff);

	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s\n", __FUNCTION__);
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

/**
   This fuction is called by email manager. When Email Mgr wants to 
   invoke UI to open inbox window (Request given by Tasknav).

   @param type whether to open inbox 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_open_inbox(gint type)
{
	gint succ = 0;
	gchar *pack_str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, XML_INT, &type);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	pack_str = osso_email_utils_concat_strings(TASKNAV_OPEN_INBOX, xmlbuff);
	CHECK_STATUS((pack_str == NULL), "Pack str failed",;
	    );
	xmlFree(xmlbuff);

	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s\n", __FUNCTION__);
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

/**
   This function is called when UI wants to  backup or restores 
   the autosave mail from temp folder.  
   @param recv_string contains funcid:msgid:operation in xml format 
   @param reply nothing is populated in this field 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_autosave_mail(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	gchar *msgid = NULL, *message = NULL;
	gchar *pack_str = NULL;
	gint option = 0;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			} else if (g_ascii_strcasecmp((gchar *) node->name, XML_INT)
				   == 0) {
				osso_email_xml_unpack_int(node, doc, XML_INT, &option);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_autosave_mail(msgid, option);
	if (succ) {
		osso_log(LOG_ERR, "fail in autosave \n");
	}
	xmlFree(msgid);

	pack_str = osso_email_utils_concat_str_int(AUTOSAVE_MAIL_CALLBACK, succ);
	CHECK_STATUS((pack_str == NULL), "Packstr failed", return rv);
	if (reply != NULL) {
		*reply = pack_str;
	}
	return OSSO_PACK_OK;
}


/**
   This method is used to get the list of messages in the folder
 
    @param   recv_str  Packed string contains the functionId:foldername
    @reply  NULL or string contains callback information of all meesages
     list for the specified folder
    @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
 **/

osso_pack_return_t engine_dbus_get_msg_header_list(const gchar * recv_str, gchar ** reply)
{
	GSList *out_list = NULL;
	GSList *in_list = NULL;
	gchar *message = NULL, *pack_str = NULL;
	gchar *msg_string = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	osso_email_utils_unpack_msgid_list_in_gslist(message, (GSList **) & in_list);
	if (in_list == NULL) {
		return rv;
	}
	out_list = al_get_msg_header(in_list);
	while (in_list) {
		g_free(in_list->data);
		in_list = g_slist_remove(in_list, in_list->data);
	}

	if (out_list == NULL) {
		osso_log(LOG_INFO, "Trying  to get a non-existing messages\n");
	} else {

		osso_email_utils_pack_msgheader_to_gslist(out_list, &msg_string);
	}

	while (out_list) {
		procmsg_msgheader_free((MsgHeader *) out_list->data);
		out_list = g_slist_remove(out_list, out_list->data);
	}

	pack_str =
	    osso_email_utils_concat_strings(GET_MSG_HEADER_LIST_CALLBACK, (gchar *) msg_string);
	CHECK_STATUS((pack_str == NULL), "Pack str failed",;);
	xmlFree(msg_string);
	if (reply != NULL) {
		*reply = pack_str;
	}
	return OSSO_PACK_OK;
}

/**
   This function is called when UI open/close a message for editing purpose.
   So that when send/receive operation performed mails opened for 
   editing will not be sent.
   @param recv_string contains funcid:msgid:operation in xml format 
   @param reply nothing is populated in this field 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static osso_pack_return_t engine_dbus_editor_open_mail(const gchar * recv_str, gchar ** reply)
{
	gboolean succ = FALSE;
	gchar *msgid = NULL, *message = NULL;
	gint option = 0;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);

	DBUS_XML_UNPACK_INIT(message, return rv);
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			} else if (g_ascii_strcasecmp(node->name, XML_INT) == 0) {
				osso_email_xml_unpack_int(node, doc, XML_INT, &option);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = al_editor_open_mail(msgid, option);
	if (succ == FALSE) {
		osso_log(LOG_INFO, "fail in add msgid in editor open list \n");
	}
	xmlFree(msgid);
	return OSSO_PACK_OK;
}


/**
 * \defgroup statesaving support from engine dbus 
 * 
 * @{
 *
 * This functions provide support for state saving operations from UI 
 */

/**
   This Method executed when UI wants to save its statesave information

   @param  recv_string  This message contains the function identifier   
           followed by statesave data
   @param  reply  returned NULL or status information
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static gint engine_dbus_statesave_write(const gchar * recv_string, gchar ** reply)
{
	gchar *message = NULL;
	GET_DATA_FROM_RECV_STRING(recv_string, message, return FALSE);

	if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_statesave_cb != NULL)) {
		email_mgr_cb->email_mgr_statesave_cb(&message, OSSO_EMAIL_STATESAVE_WRITE);
	} else {
		return FALSE;
	}
	return TRUE;
}

/**
   This Method executed when UI wants to get its statesave information

   @param  recv_string  This message contains the function identifier   
           followed by statesave data
   @param  reply  returned NULL or statesave information 
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static gint engine_dbus_statesave_read(const gchar * recv_string, gchar ** reply)
{
	gchar *stateinfo = NULL;
	gchar *pack_str = NULL;

	if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_statesave_cb != NULL)) {
		email_mgr_cb->email_mgr_statesave_cb(&stateinfo, OSSO_EMAIL_STATESAVE_READ);
	}

	pack_str = osso_email_utils_concat_strings(STATESAVE_READ_CALLBACK, stateinfo);
	g_free(stateinfo);
	CHECK_STATUS((pack_str == NULL), "Pack str failed", return FALSE);

	*reply = pack_str;
	return TRUE;
}

/**
   This Method executed from email manager to request the UI to save its
   statesave information. 

   @param  void  
   @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

osso_pack_return_t engine_dbus_statesave_signal(void)
{
	gchar *pack_str = NULL;
	gint succ = 0;
	osso_pack_return_t rv = OSSO_PACK_ERROR;
	pack_str = osso_email_utils_concat_str_int(STATESAVE_SIGNAL, TRUE);
	CHECK_STATUS((pack_str == NULL), "Pack str failed", return FALSE);
	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s", __FUNCTION__);
		return rv;
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

/*
 * Mmc Hlr Callback Handling Functionality
 */

/**
    This fuction registerd as a call back function to Mmc Hlr.
    MmcHlr invokes this function when mmc is mounted on the device
 
    @param volume_monitor pointer to the GnomeVFSVolumeMonitor information
    @param volume  pointer to the GnomeVFSVolume information
    @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
*/

static osso_pack_return_t
engine_dbus_mmc_hlr_volume_mounted_cb(GnomeVFSVolumeMonitor *
				      volume_monitor, GnomeVFSVolume * volume)
{
	gchar *pack_str = NULL;
	MmcStatus mmc_status = MMC_ERROR;
	gchar *path = NULL;
	gchar *mmc_path = NULL;
	GnomeVFSURI *mmc_uri;

	path = gnome_vfs_volume_get_activation_uri(volume);
	mmc_uri = gnome_vfs_uri_new(path);
	g_free(path);
	mmc_path = (gchar *) gnome_vfs_uri_get_path(mmc_uri);
	if (strcmp(mmc_path, MMC_PATH) != 0) {
		g_free(mmc_path);
		return OSSO_PACK_ERROR;
	}
	update_mmc_flag(TRUE, mmc_path);
	g_free(mmc_path);
	mmc_status = MMC_CONNECTED;
	pack_str = osso_email_utils_concat_str_int(MMC_HLR, mmc_status);
	if (pack_str == NULL) {
		return OSSO_PACK_ERROR;
	}

	if (engine_dbus_connect_to_remote(pack_str) == FALSE) {
		osso_log(LOG_INFO, "Unable connect remote mmc hlr volume mounted");
		g_free(pack_str);
		return OSSO_PACK_ERROR;
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

/**
    This fuction registerd as a call back function to Mmc Hlr.
    MmcHlr invokes this function when mmc is unmounted on the device

    @param volume_monitor pointer to the GnomeVFSVolumeMonitor information
    @param volume  pointer to the GnomeVFSVolume information
    @return on successful operation returns OSSO_PACK_OK. Otherwise 
    returns errorcode 
    @return on success returns TRUE or FAIL on fail
 */

static osso_pack_return_t
engine_dbus_mmc_hlr_volume_unmounted_cb(GnomeVFSVolumeMonitor *
					volume_monitor, GnomeVFSVolume * volume)
{
	gchar *pack_str = NULL;
	/*MmcStatus mmc_status = MMC_ERROR; */
	gchar *path = NULL;
	gchar *mmc_path = NULL;
	GnomeVFSURI *mmc_uri;
	MmcStatus mmc_status = MMC_DISCONNECTED;

	path = gnome_vfs_volume_get_activation_uri(volume);
	mmc_uri = gnome_vfs_uri_new(path);
	g_free(path);
	mmc_path = (gchar *) gnome_vfs_uri_get_path(mmc_uri);
	if (strcmp(mmc_path, MMC_PATH) != 0) {
		g_free(mmc_path);
		return OSSO_PACK_ERROR;
	}
	update_mmc_flag(TRUE, mmc_path);
	g_free(mmc_path);
	pack_str = osso_email_utils_concat_str_int(MMC_HLR, mmc_status);
	if (pack_str == NULL) {
		return OSSO_PACK_ERROR;
	}

	if (engine_dbus_connect_to_remote(pack_str) == FALSE) {
		osso_log(LOG_INFO, "Unable connect remote mmc hlr volume mounted");
		g_free(pack_str);
		return OSSO_PACK_ERROR;
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}

#if CL_USE
/**
  This function gets the iapinfo list infomation 

  @param  recv_str xml string contains request to get iap list 
  @param  reply outparmeter with reply
  @return on successful sending iap list returns OSSO_PACK_OK. In case of
   error returns OSSO_PACK_ERROR 
 */

static osso_pack_return_t engine_dbus_get_iapinfo_list(const gchar * recv_str, gchar ** reply)
{
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	gint buffersize = 0;
	gint succ = 0;
	gchar *pack_str = NULL;
	GSList *iaplist = NULL;
	succ = cl_get_iap_info_list(&iaplist);
	iap_confsmtp_get_serv_info_list(&iaplist);
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, &succ);
	osso_email_xml_pack_iapinfo_list(root, doc, XML_IAPINFO_LST, iaplist);
	cl_iap_info_list_free(iaplist);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;
	
	pack_str = osso_email_utils_concat_strings(GET_IAPINFO_LIST_CALLBACK, xmlbuff);
	xmlFree(xmlbuff);
	CHECK_STATUS((pack_str == NULL), "concat strings fail", return OSSO_PACK_ERROR);
	*reply = pack_str;
	return OSSO_PACK_OK;
}

/*
  This function used to set iap info

  @param  recv_str xml string contains iap info to be set
  @param  reply outparmeter with reply
  @return on successful returns OSSO_PACK_OK. In case of error 
  returns OSSO_PACK_ERROR 
*/

static osso_pack_return_t engine_dbus_set_iapinfo(const gchar * recv_str, gchar ** reply)
{
	gint succ = 0;
	IapInfo *iapinfo = NULL;
	gchar *message = NULL;
	gchar *pack_str = NULL;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	osso_pack_return_t rv = OSSO_PACK_ERROR;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return rv);
	DBUS_XML_UNPACK_INIT(message, return rv);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp((gchar *) node->name, XML_IAPINFO) == 0) {
				iapinfo = g_new0(IapInfo, 1);
				osso_email_xml_unpack_iapinfo(node, doc, XML_IAPINFO, iapinfo);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;

	succ = iap_smtpconf_set_iap_info(iapinfo);
	pack_str = osso_email_utils_concat_str_int(SET_IAPINFO_SETTINGS_CALLBACK, succ);
	*reply = pack_str;
	return OSSO_PACK_OK;
}

#endif


/**
  This function is called when engine recived DBus message about UI ready 

  @param  recv_str xml string contains status of UI is ready 
  @param  reply outparmeter with reply
  @return  OSSO_PACK_OK
 */
static osso_pack_return_t engine_dbus_ui_ready(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gint ui_status = 0;
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	ui_status = atoi(message);
	if (ui_status == TRUE) {
		emailMgrState = OSSO_STATE_INIT_COMPLETE;
		osso_email_gconf_unset(EMAIL_GCONF_UI_CLOSED);
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_disable_timer != NULL)) {
			email_mgr_cb->email_mgr_disable_timer();
		}
	} else {
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_quit_loop != NULL)) {
			email_mgr_cb->email_mgr_quit_loop();
		}
	}
	return OSSO_PACK_OK;
}


/**
 * This method removes osso email notifications from task nav.
 *
 * @param  recv_str - no input parameter
 * @param  reply - no output parameter
 * @return OSSO_PACK_OK
 */
static osso_pack_return_t engine_dbus_remove_notifications(const gchar * recv_str, gchar ** reply)
{
        notify_remove_and_close_all();
        return OSSO_PACK_OK;
}

#if CL_USE
static osso_pack_return_t engine_dbus_disconnect_iap(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	al_cleanup_sessions();
	gl_email_mgr_waiting = FALSE;
	/* disconnect the current iap */
	if (osso_mgr_schdg_has_task() == FALSE) {
		cl_iap_disconnect(NULL);
	}
	return OSSO_PACK_OK;
}
#endif

static osso_pack_return_t
engine_dbus_folder_remove_all_attachment(const gchar * recv_str, gchar ** reply)
{
	GSList *out_list = NULL;
	GSList *in_list = NULL;
	gchar *message = NULL, *pack_str = NULL;
	gchar *msg_string = NULL;

	/*
	 * Skip the func_identifier which is in the beginning the recv_str.
	 * Format of the recv_str is: func_identifer:GSList (FolderAttachInfo)
	 */
	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	osso_email_utils_unpack_msgid_list_in_gslist(message, (GSList **) & in_list);

	if (in_list == NULL) {
		osso_log(LOG_INFO, " Unpack failed : attachment list null \n");
		pack_str = osso_email_utils_concat_strings(REMOVEALL_ATTACHMENT_CALLBACK, NULL);
		if (reply != NULL) {
			*reply = pack_str;
		}
		return OSSO_PACK_ERROR;
	}
	out_list = al_folder_remove_all_attachment(in_list);

	if (out_list == NULL) {
		osso_log(LOG_ERR, "Unable to remove folder attachment\n");
	} else {
		osso_email_utils_pack_msgid_to_gslist(out_list, &msg_string);
		/* free the string */
		slist_free_strings(out_list);
		g_slist_free(out_list);
	}
	slist_free_strings(in_list);
	g_slist_free(in_list);

	pack_str =
	    osso_email_utils_concat_strings(REMOVEALL_ATTACHMENT_CALLBACK, (gchar *) msg_string);

	CHECK_STATUS((pack_str == NULL), "Pack str failed",;
	    );

	xmlFree(msg_string);
	if (reply != NULL) {
		*reply = pack_str;
	}

	return OSSO_PACK_OK;
}
static gint engine_dbus_al_folder_list(gint list_type, const GSList * msghdr_list)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	if (msghdr_list == NULL) {
		osso_log(LOG_INFO, "MsgInfo List arg NULL ");
		return OSSO_PACK_INVALID;
	}
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, FOLDER_NAME, &list_type);
	osso_email_xml_pack_msgheader_list(root, doc, XML_MSG_HEADER_LST, (gchar *) msghdr_list);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;
	str = osso_email_utils_concat_strings(AL_FOLDER_LST, xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

static gint engine_dbus_al_recv_attachment_cb(const MsgInfo * msginfo, gint attach_tag)
{
	gchar *str = NULL;
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;
	if (msginfo == NULL) {
		osso_log(LOG_INFO, " Received NULL values for msginfo\n");
		return OSSO_PACK_INVALID;
	}
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_msginfo(root, doc, XML_MSG_INFO, (MsgInfo *) msginfo);
	osso_email_xml_pack_int(root, doc, ATTA_TAG, &attach_tag);
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);
	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(RECV_ATTACHMENT_CALLBACK, (gchar *) xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}

static gint engine_dbus_al_folder_set_msg(gint msg_number, const gchar * format, ...)
{
	gchar *str = NULL;

	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *xmlbuff = NULL;
	int buffersize = 0;

	va_list ap;
	gint int_val = 0;
	gchar *str_val = NULL;
	gchar *fmt = NULL;

	if (msg_number == 0) {
		return OSSO_PACK_INVALID;
	}
	DBUS_XML_PACK_INIT;
	osso_email_xml_pack_int(root, doc, STATUS, &msg_number);

	if (format != NULL) {
		fmt = (gchar *) format;
		va_start(ap, format);
		while (*fmt) {
			switch (*fmt++) {
			case 's':
				str_val = va_arg(ap, char *);
				osso_email_xml_pack_str(root, doc, XML_STRING, &str_val);
				break;
			case 'd':
				int_val = va_arg(ap, int);
				osso_email_xml_pack_int(root, doc, XML_INT, &int_val);
				break;
			}
		}
		va_end(ap);
	}
	xmlDocDumpFormatMemory(doc, &xmlbuff, &buffersize, 1);

	DBUS_XML_PACK_UN_INIT;

	str = osso_email_utils_concat_strings(AL_FOLDER_MSG, xmlbuff);
	xmlFree(xmlbuff);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));

}

osso_pack_return_t engine_dbus_send_app_killer_signal(const gint exit_signal)
{
	gchar *str = NULL;
	str = osso_email_utils_concat_str_int(APP_KILLER_SIGNAL, exit_signal);
	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}
static osso_pack_return_t engine_dbus_set_tmp_passwd(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gchar *account_name = NULL;
	gchar *passwd = NULL;
	gboolean cancel = FALSE;
	xmlNodePtr cur = NULL;
	xmlNodePtr node = NULL;
	xmlDocPtr doc = NULL;
	gint succ = 0;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_STRING) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_STRING, &account_name);
			} else if (g_ascii_strcasecmp(node->name, MSGID) == 0) {
				osso_email_xml_unpack_str(node, doc, MSGID, &passwd);
			} else if (g_ascii_strcasecmp(node->name, XML_BOOL) == 0) {
				osso_email_xml_unpack_bool(node, doc, XML_BOOL, &cancel);
			}
		}
	}

	DBUS_XML_UNPACK_UN_INIT;
	succ = al_set_temp_passwd(account_name, passwd, cancel);
	if (succ < 0) {
		osso_log(LOG_INFO, "Unable to set tmp passwd\n");
	}
	xmlFree(account_name);
	xmlFree(passwd);
	return OSSO_PACK_OK;
}
static osso_pack_return_t engine_dbus_rem_msg_attachments(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	gchar *msgid = NULL;
	gint succ = 0;
	GSList *file_pos_list = NULL;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	node = cur->xmlChildrenNode;
	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_INT_LST) == 0) {
				osso_email_xml_unpack_int_list(node, doc, XML_INT_LST,
							       &file_pos_list);
			}
			if (g_ascii_strcasecmp(node->name, XML_MSG_ID) == 0) {
				osso_email_xml_unpack_str(node, doc, XML_MSG_ID, &msgid);
			}
		}
	}
	DBUS_XML_UNPACK_UN_INIT;
	succ = al_remove_msg_attachments(msgid, file_pos_list);

	return engine_dbus_folder_operation_status(succ, reply);

}
static osso_pack_return_t engine_dbus_is_engine_ready(const gchar * recv_str, gchar ** reply)
{
	gint status = 0;
	gchar *pack_str = NULL;

	osso_pack_return_t rv = OSSO_PACK_ERROR;

	status = 0xffff;
	pack_str = osso_email_utils_concat_str_int(ENGINE_STATUS_CALLBACK, status);
	CHECK_STATUS((pack_str == NULL), "Pack str failed ", return rv);

	return (engine_dbus_send_pack_str_to_remote(pack_str, __FUNCTION__, TRUE));
}
static osso_pack_return_t engine_dbus_send_recv_mail(const gchar * recv_str, gchar ** reply)
{
	gchar *message = NULL;
	GET_DATA_FROM_RECV_STRING(recv_str, message,;);

	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	osso_log(LOG_INFO, "al_recv_mail \n");
	if ((message != NULL) && (strlen(message) != 0)) {
		g_print("\nal_sned_recv_mail \n");
		al_send_recv_mail(message, FALSE);
	} else {
		al_update_mail_list();
	}
	if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_update_outbox_status_cb != NULL)) {
		email_mgr_cb->email_mgr_update_outbox_status_cb();
	}
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);
	return OSSO_PACK_OK;
}

static osso_pack_return_t engine_dbus_send_get_msg_id(const gchar * recv_str, gchar ** reply)
{
	MsgInfo *msginfo = NULL;
	gchar *message = NULL;
	gchar *pack_str = NULL;
	gboolean send_flag = FALSE;

	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;

	GET_DATA_FROM_RECV_STRING(recv_str, message, return OSSO_PACK_ERROR);
	if (reply == NULL) {
		osso_log(LOG_INFO, "Invalid arguments ");
		return OSSO_PACK_ERROR;
	}

	DBUS_XML_UNPACK_INIT(message, return OSSO_PACK_ERROR);

	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem allocation fail for MsgInfo struct ");
		return OSSO_PACK_ALLOC_FAIL;
	}

	for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {
		if (node->type == XML_ELEMENT_NODE) {
			if (g_ascii_strcasecmp(node->name, XML_MSG_INFO) == 0) {
				osso_email_xml_unpack_msginfo(node, doc, XML_MSG_INFO, msginfo);
			}
			if (g_ascii_strcasecmp(node->name, SEND_FLAG) == 0) {
				osso_email_xml_unpack_bool(node, doc, SEND_FLAG, &send_flag);
			}
		}
	}

	DBUS_XML_UNPACK_UN_INIT;

	al_send_get_msg_id(msginfo, send_flag);

	if ((msginfo != NULL) && (msginfo->header != NULL)) {
		pack_str =
		    osso_email_utils_concat_strings(SEND_GET_MSG_ID_CALLBACK,
						    msginfo->header->msgid);
		CHECK_STATUS((pack_str == NULL), "Packstr failed ",;
		    );
	}

	procmsg_msginfo_free(msginfo);
	*reply = pack_str;
	return OSSO_PACK_OK;
}

static gint engine_dbus_rx_delete_list(GSList * msgid_list)
{
	gchar *str = NULL;
	gchar *pack_str = NULL;

	if (msgid_list == NULL) {
		osso_log(LOG_INFO, "MsgInfo List arg NULL ");
		return OSSO_PACK_INVALID;
	}
	osso_email_utils_pack_msgid_to_gslist(msgid_list, &str);
	pack_str = osso_email_utils_concat_strings(RX_DEL_LST, (gchar *) str);
	xmlFree(str);
	return (engine_dbus_send_pack_str_to_remote(pack_str, __FUNCTION__, TRUE));
}

static osso_pack_return_t engine_dbus_attach_list(GSList * msginfo_list)
{
	gchar *list = NULL;
	gchar *str = NULL;



	osso_email_utils_pack_from_gslist_to_msginfo(msginfo_list, &list);

	str = osso_email_utils_concat_strings(ATTACH_LIST_CB, list);
	xmlFree(list);

	return (engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE));
}
static osso_pack_return_t engine_dbus_retr_msgs(const gchar * data, gchar ** reply)
{
	GSList *ret_list = NULL;
	xmlNodePtr cur = NULL, node = NULL;
	xmlDocPtr doc = NULL;
	gchar *temp = NULL;
	gboolean copy = FALSE;
        gint status = 0;
        gchar *str = NULL;

	GET_DATA_FROM_RECV_STRING(data, temp, return OSSO_PACK_ERROR);

	DBUS_XML_UNPACK_INIT(temp, return OSSO_PACK_ERROR)

	    for (node = cur->xmlChildrenNode; node != NULL; node = node->next) {

		if (node->type == XML_ELEMENT_NODE) {

			if (g_ascii_strcasecmp((gchar *) node->name, XML_RETR_LST) == 0) {

				osso_email_xml_unpack_retr_list(node, doc, XML_RETR_LST, &ret_list);
			}else if (g_ascii_strcasecmp(node->name, XML_BOOL) == 0) {
                                osso_email_xml_unpack_bool(node, doc, XML_BOOL, &copy);
                        }
			
		}
	}

	DBUS_XML_UNPACK_UN_INIT;
	
	if (al_retr_msgs(ret_list) == 0) {
                status = OSSO_PACK_OK;
        }else {
                status = OSSO_PACK_INVALID;
        }

        /*send retrieved callback*/
        if(copy){
                str = osso_email_utils_concat_str_int(AL_RETR_MSGS_STATUS, 0);
                engine_dbus_send_pack_str_to_remote(str, __FUNCTION__, TRUE);

        }
        procmsg_free_retr_data_list(ret_list);
        return status;

}
void engine_dbus_backup_handle(void)
{
	al_cleanup_sessions();
}
osso_pack_return_t engine_dbus_backup_event(void)
{
	gchar *pack_str = NULL;
	gint succ = 0;
	osso_pack_return_t rv = OSSO_PACK_ERROR;
	pack_str = osso_email_utils_concat_str_int(BACKUP_START_EVENT, rv);
	CHECK_STATUS((pack_str == NULL), "Pack str failed", return rv);
	if ((succ = engine_dbus_connect_to_remote(pack_str)) == FALSE) {
		osso_log(LOG_ERR, "Unable to connect remote %s", __FUNCTION__);
		return rv;
	}
	g_free(pack_str);
	return OSSO_PACK_OK;
}
static void engine_dbus_exit_on_check(void)
{
	if ((osso_mgr_schdg_has_task() == FALSE) && (gl_email_mgr_waiting == FALSE)) {
		engine_dbus_al_unregister_engine();
		if ((email_mgr_cb != NULL) && (email_mgr_cb->email_mgr_quit_loop != NULL)) {
			email_mgr_cb->email_mgr_quit_loop();
		} else {

			emailMgrState = OSSO_STATE_INIT_LIMITED;
		}
	} else {
		emailMgrState = OSSO_STATE_INIT_LIMITED;

	}
	gl_email_mgr_waiting = FALSE;
}

/**
 * This function returns the busyness of the send recv connection 
 *
 * @return gboolean send recv queue status
 **/
gboolean send_recv_dbus_handler_busy(void)
{
	return gl_send_recv_dbus_handler_busy;
}

