
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#define DBUS_API_SUBJECT_TO_CHANGE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <dbus/dbus-glib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libosso.h>
#include <log-functions.h>
#include <locale.h>
#include <hildon/hildon-notification.h>
#include <libnotify/notify.h>

#include <mcheck.h> /* XXX DEBUG */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "intl.h"
#include "common.h"
#include "utils.h"
#include "defs.h"
#include "engine_dbus.h"
#include "al.h"
#include "procmsg.h"
#include "account.h"
#include "callbacks.h"
#include "manager.h"
#include "scheduling.h"
#include "dbus_utils.h"
#include "email_gconf.h"

gint emailMgrState = OSSO_STATE_UN_INIT;
EmailMgrCallback *email_mgr_cb = NULL;
gboolean gl_email_mgr_waiting = FALSE;

static osso_context_t *osso_email_mgr_ctx = NULL;
static gint manager_timeout_id = 0;
static GMainLoop *loop = NULL;

/**
  This function is callback function for handling the events 
  like shutdown, low memory signals.
  
  @ param state contains the information of signal   
  @ data  additional data given by application while registering
  @ return nothing
*/

static void osso_email_hw_cb_f(osso_hw_state_t * state, gpointer data);

static osso_return_t osso_email_mgr_statesave_operation_cb(gchar ** statesaveinfo, gint operation);

static osso_return_t osso_email_mgr_send_recieve(osso_context_t * context);

static osso_return_t osso_email_mgr_update_outbox_status(void);

static osso_return_t osso_initialize_email_application(osso_context_t * ctx);

static osso_return_t osso_initialize_engine_process(void);

static osso_return_t osso_initialize_ui_process(const gint limited, const gchar * arg);

static osso_return_t osso_email_mgr_check_ui_available(void);

static osso_return_t osso_email_mgr_register_with_engine_dbus(void);

static osso_return_t osso_email_mgr_unregister_with_engine_dbus(void);

static osso_return_t osso_email_mgr_open_msg(osso_context_t * ctx, const gchar * args);

static osso_return_t osso_email_mgr_delete_msg(osso_context_t * ctx, const gchar * args);

static osso_return_t osso_email_mgr_open_inbox(const gint type);

static osso_return_t osso_email_mgr_files(osso_context_t * context, const gchar * list);

static osso_return_t osso_email_mgr_url(osso_context_t * context, const gchar * url);

static osso_return_t osso_email_mgr_mailto(osso_context_t * context, const gchar * mailto);

static osso_return_t
osso_email_mgr_parse_mailto_str(const gchar * mailto, MsgHeader * msgheader, gchar ** body);
static void osso_email_mgr_decode_content(gchar * content);

static gboolean osso_email_sync_timeout(gpointer data);

static gboolean osso_email_mgr_disable_timer(void);

static osso_return_t osso_email_mgr_get_folder_list(osso_rpc_t * ret_rpc_t);

static osso_return_t osso_email_mgr_get_standard_folders(osso_rpc_t * ret_rpc_t);

static osso_return_t
osso_email_mgr_rpc_callback_handler(const gchar * interface,
				    const gchar * method,
				    GArray * arguments, gpointer data, osso_rpc_t * retval);

static osso_return_t osso_email_mgr_tasknav_send_recv_focus(guint status);

static void osso_email_mgr_exit_cb(gboolean flag, gpointer userdata);

static void osso_email_mgr_quit_loop(void);

static void osso_email_alwayson_send_recv(void);

/**
  This function registers the callback function to handle the signals for 
  backup start & finish 
  @return none.
 */
static void osso_email_mgr_backup_reg(void);

/**
  This is the callback function to handle the signals for 
  backup start & finish 
  @return DBUS_HANDLER_RESULT_NOT_YET_HANDLED always.
 */
static DBusHandlerResult
backup_signal_cb(DBusConnection * connection, DBusMessage * message, void *user_data);
static osso_return_t osso_uninitialize_engine_process(void);

/* checking double frees */
static void
my_check (enum mcheck_status status)
{
        if (status == MCHECK_FREE) {
                g_warning ("mcheck: double free detected; ignoring");
        } else {
                g_warning ("mcheck: nothing we can do");
                abort (); /* terminating... nothing we can do */
        }
}

/**
  This main function gets the context from AFW and initialises
  email application
  @return OSSO_OK on success or OSSO_ERROR on failure.
 */

int main(int argc, char **argv)
{
        /* DEBUG XXX */
        if(mcheck(my_check) != 0) {
                osso_log(LOG_ERR, "mcheck won't work");
        } else {
                osso_log(LOG_INFO, "mcheck enabled");
        }



#ifdef ENABLE_NLS
	bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset(PACKAGE, "UTF-8");
	textdomain(PACKAGE);
#endif
	setlocale(LC_ALL, "");

	emailMgrState = OSSO_STATE_UN_INIT;

	osso_email_mgr_ctx = osso_initialize(EMAIL_SERVICE,	/*EMAIL_APP */
					     EMAIL_APPLICATION_VERSION, TRUE, NULL);

	if (osso_email_mgr_ctx == NULL) {
		osso_log(LOG_ERR, "AFW can't initialise Email Manager \n");
		return OSSO_ERROR;
	}
	/* GConf Initialisation for email */
	osso_email_gconf_init(argc, argv);

	if (osso_hw_set_event_cb(osso_email_mgr_ctx, NULL,	/*all states */
				 *(osso_hw_cb_f *) osso_email_hw_cb_f, NULL) != OSSO_OK) {
		osso_deinitialize(osso_email_mgr_ctx);
		osso_log(LOG_ERR, " Email Manager osso hw set cb register failed..");
		return OSSO_ERROR;
	}

	if (osso_rpc_set_default_cb_f_with_free(osso_email_mgr_ctx,
						osso_email_mgr_rpc_callback_handler, NULL,
						osso_rpc_free_val) != OSSO_OK) {
		osso_deinitialize(osso_email_mgr_ctx);
		osso_log(LOG_ERR, " Email Manager rpc cb register failed ..");
		return OSSO_ERROR;
	}
	osso_email_mgr_backup_reg();

        if(!notify_init("Osso-email")) {
                osso_log(LOG_ERR, "notify_init failed");
                return OSSO_ERROR;
        }

	if (emailMgrState == OSSO_STATE_UN_INIT) {
		if (osso_initialize_email_application(osso_email_mgr_ctx) != OSSO_OK) {
			osso_log(LOG_ERR, "Email Manager init failed..\n");
			emailMgrState = OSSO_STATE_UN_INIT;
		} else {
			emailMgrState = OSSO_STATE_INIT_LIMITED;
		}
	}

	loop = g_main_loop_new(NULL, FALSE);
	g_main_loop_run(loop);
	osso_deinitialize(osso_email_mgr_ctx);
	osso_uninitialize_engine_process();
	return OSSO_OK;
}

static void osso_email_hw_cb_f(osso_hw_state_t * state, gpointer data)
{

	if (state->shutdown_ind == TRUE) {
		osso_log(LOG_DEBUG, "device is about to shutdown \n");
		engine_dbus_send_app_killer_signal(SIGNAL_SHUTDOWN_IND);
	} else if (state->save_unsaved_data_ind == TRUE) {
		osso_log(LOG_DEBUG, "save unsaved data indication received \n");
		engine_dbus_send_app_killer_signal(SIGNAL_SAVE_UNSAVED_DATA_IND);
	} else if (state->memory_low_ind == TRUE) {
		osso_log(LOG_DEBUG, "memory low indication received \n");
		engine_dbus_send_app_killer_signal(SIGNAL_MEMORY_LOW_IND);
	} else if (state->system_inactivity_ind == TRUE) {
		osso_log(LOG_DEBUG, "system inactivity indication received \n");
		engine_dbus_send_app_killer_signal(SIGNAL_SYSTEM_INACTIVITY_IND);
	}
	if (state->sig_device_mode_ind == OSSO_DEVMODE_FLIGHT) {
		disable_device_mode_flag();
	} else {
		enable_device_mode_flag();
	}

}

/**
 * This function is called to check if the engine is free to perform sendrecv
 * for autoretrival
 * @param userdata gpointer
 * @return gboolean TRUE if the function has to be called again
 * false if not
 **/
static gboolean idle_auto_retrival_cb(gpointer userdata) {
	if(send_recv_dbus_handler_busy()) {
		return TRUE;
	} else {
		osso_email_alwayson_send_recv();
		return FALSE;
	}
}
			
			
/**
 This function is called to do send/recv of messages upon signal received from alarm.
 @param void.
 @return void
*/
static void osso_email_alwayson_send_recv(void)
{
	osso_log(LOG_INFO, "osso_email_alwayson_send_recv \n");
	
#if CL_USE
	if(emailMgrState == OSSO_STATE_INIT_LIMITED || emailMgrState == OSSO_STATE_UN_INIT)
        	cl_iap_connect(NULL, TRUE);
#endif
	if(send_recv_dbus_handler_busy()) {
		g_idle_add(idle_auto_retrival_cb,NULL);
	}
	
	osso_mgr_schdg_send_recv (NULL);
	if ((emailMgrState == OSSO_STATE_UN_INIT) || (emailMgrState == OSSO_STATE_INIT_LIMITED)) {
		 osso_uninitialize_engine_process();
		 osso_deinitialize(osso_email_mgr_ctx);
		 exit(0);
	}
}

/**
  This function is a callback function for rpc handler.
  This function get called when ever Dbus message comes from the other
  applications(like Task Navigator, Browser etc.)

  @param  intaeface  Which contains interface name
  @param  method  Which contains corresponding method to execute
  @param  arguments Contains arguments in GAarry struct
  @param  data    generic data
  @param  retval  contains the return value after executing this method
  @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t
osso_email_mgr_rpc_callback_handler(const gchar * interface,
				    const gchar * method,
				    GArray * arguments, gpointer data, osso_rpc_t * retval)
{
	osso_return_t rv = OSSO_OK;
	osso_log(LOG_DEBUG, " osso_email_mgr_rpc_callback_handler starts %s\n", method);

	if ((method == NULL) || (interface == NULL)) {
		osso_log(LOG_DEBUG, "Invalid input params received \n");
		rv = OSSO_ERROR;
		goto endshere;
	}
	if (emailMgrState == OSSO_STATE_UI_INIT_PROGRESS) {
		osso_log(LOG_DEBUG, "rpc cb handler: UI init  progress \n");
		return OSSO_OK;
	}

	if (!strcmp(method, SEND_AND_RECEIVE_METHOD)) {
		/* perform send receive operation  */
		rv = osso_email_mgr_send_recieve(osso_email_mgr_ctx);
	} else if (!strcmp(method, ACTIVATE_MAIL_METHOD)) {
		/* open a mail */
		rv = OSSO_ERROR;
		if ((g_array_index(arguments, osso_rpc_t, 0).type)
		    != DBUS_TYPE_STRING) {
			goto endshere;
		}
		rv = osso_email_mgr_open_msg(osso_email_mgr_ctx,
					     g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, NEW_MAIL_METHOD)) {
		/* open a new new mail */
		rv = osso_email_mgr_open_msg(osso_email_mgr_ctx, NULL);
	} else if ((!strcmp(method, OPEN_INBOX_METHOD)) || (!strcmp(method,"mime_open"))) {
		/* open inbox */
		if (osso_email_mgr_check_ui_available() != OSSO_OK) {
			if (osso_initialize_ui_process(0, NULL) != OSSO_OK) {
				rv = OSSO_ERROR;
				goto endshere;
			}
		} else {
			rv = osso_email_mgr_open_inbox(OSSO_TASKNAV_OPEN_INBOX);
		}
	} else if (!strcmp(method, SEND_FILES_METHOD)) {
		/* perform send as files to UI */
		if (g_array_index(arguments, osso_rpc_t, 0).type != DBUS_TYPE_STRING) {
			rv = OSSO_ERROR;
			goto endshere;
		}
		rv = osso_email_mgr_files(osso_email_mgr_ctx,
					  g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, SEND_URL_METHOD)) {
		/* perform send url to UI */
		if (g_array_index(arguments, osso_rpc_t, 0).type != DBUS_TYPE_STRING) {
			rv = OSSO_ERROR;
			goto endshere;
		}
		rv = osso_email_mgr_url(osso_email_mgr_ctx,
					g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, SEND_MAILTO_METHOD)) {
		/* send mailto argumets to UI */
		if (g_array_index(arguments, osso_rpc_t, 0).type != DBUS_TYPE_STRING) {
			rv = OSSO_ERROR;
			goto endshere;
		}
		rv = osso_email_mgr_mailto(osso_email_mgr_ctx,
					   g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, STATE_SAVE_METHOD)) {
		/* send statesave signal */
		rv = engine_dbus_statesave_signal();
	} else if (!strcmp(method, OPEN_HIT_METHOD)) {
		/* Request comes from Global search  to open mail */
		if (g_array_index(arguments, osso_rpc_t, 0).type != DBUS_TYPE_STRING) {
			rv = OSSO_ERROR;
			goto endshere;
		}
		rv = osso_email_mgr_open_msg(osso_email_mgr_ctx,
					     g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, DELETE_HIT_METHOD)) {
		/* Request comes from Global search to delete mail */
		if (g_array_index(arguments, osso_rpc_t, 0).type != DBUS_TYPE_STRING) {
			rv = OSSO_ERROR;
			goto endshere;
		}
		rv = osso_email_mgr_delete_msg(osso_email_mgr_ctx,
					       g_array_index(arguments, osso_rpc_t, 0).value.s);
	} else if (!strcmp(method, GET_FOLDERS_LIST_METHOD)) {
		/* Request comes from Global search to get folder list */
		rv = osso_email_mgr_get_folder_list(retval);
		return rv;
	} else if (!strcmp(method, GET_STANDARD_FOLDERS_METHOD)) {
		/* Request comes from Global search Email Plugin */
		rv = osso_email_mgr_get_standard_folders(retval);
		return rv;
	} else if (!strcmp(method, TOP_APPLICATION)) {
		/* Window topping Request from AF */
		if (osso_email_mgr_check_ui_available() != OSSO_OK) {
			if (osso_initialize_ui_process(0, NULL) != OSSO_OK) {
				rv = OSSO_ERROR;
				goto endshere;
			}
		}
		return rv;
	} 
	else if (!strcmp(method, START_SEND_RECV)) {
		/* Request comes from alarm for send/recv */
		osso_email_alwayson_send_recv();
		return OSSO_OK;
	}
      endshere:
	if (retval != NULL) {
		retval->type = DBUS_TYPE_INT32;
		retval->value.i = rv;
	}
	return rv;
	
}

/**
  This function is registered to Engine as a callback function. When no
  accounts are defined this method sends a DBus message to set dim the
  Send Recv button in TaskNav window.
  @param  state set to 0 for dimming
  @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_tasknav_send_recv_focus(guint status)
{
	gint rv = OSSO_OK;

	rv = osso_rpc_run_with_defaults(osso_email_mgr_ctx,
					TASKNAV_APP,
					OSSO_TASKNAV_SEND_RECV_FOCUS, NULL,
					DBUS_TYPE_INT32, status, DBUS_TYPE_INVALID);
	if (rv != OSSO_OK) {
		osso_log(LOG_ERR, "Error in send send recv focus to tasknav");
	}
	return rv;
}

/**
   This function is registered to Engine as a callback function. When
   statesave information is saved/retrived this function is called by engine.

   @param statesaveinfo contains state save information
   @param operation Read/Write operation to be performed for state save
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_statesave_operation_cb(gchar ** statesaveinfo, gint operation)
{
        osso_state_t state;
        osso_return_t ret;

	if (statesaveinfo == NULL) {
		osso_log(LOG_INFO, "statesave info NULL");
		return OSSO_ERROR;
	}
	switch (operation) {
	case OSSO_EMAIL_STATESAVE_WRITE:
                state.state_size = strlen(*statesaveinfo) + 1;
                state.state_data = (gpointer)*statesaveinfo;
                ret = osso_state_write(osso_email_mgr_ctx, &state);
                if (ret != OSSO_OK)
                {
			osso_log(LOG_ERR, "osso_state_write error");
			return OSSO_ERROR;
		}
		break;

	case OSSO_EMAIL_STATESAVE_READ:
                state.state_data = NULL; /* will be dynamically allocated */
                state.state_size = 0;    /* will be read from the state file */
                ret = osso_state_read(osso_email_mgr_ctx, &state);
                if (ret != OSSO_OK)
                {
                        osso_log(LOG_ERR, "osso_state_read error");
                        return OSSO_ERROR;
                }
                *statesaveinfo = state.state_data;
                g_assert ((*statesaveinfo)[state.state_size - 1] == 0 &&
                          strlen(*statesaveinfo) == state.state_size - 1);
		break;

	default:
		osso_log(LOG_ERR, "Invalid state save operation");
		return OSSO_ERROR;
		break;
	}
	return OSSO_OK;
}

/**
   This function is called to perform the Send/Receive operation from other
   applications (like Task Navigator). This method perform send receive
   operation on all accounts and update the outbox status to task navigator.

   @param osso  email application context
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_send_recieve(osso_context_t * context)
{
	GSList *ac_list = NULL;
	PrefsAccount *def_account = NULL;

	gl_email_mgr_waiting = TRUE;
	al_get_account_list(&ac_list);
	if (ac_list == NULL) {
		osso_email_mgr_tasknav_send_recv_focus(FALSE);
	} else {
		slist_free_strings(ac_list);
		g_slist_free(ac_list);
		/* perform Send receive operation on default acoount */
		if ((def_account = account_get_default_account()) != NULL) {
			g_print("\nSend rx start message 1\n");
			al_send_recv_mail(def_account->user_settings.account_name, TRUE);
			osso_email_mgr_update_outbox_status();
		}
	}
	gl_email_mgr_waiting = FALSE;
	/* UI process not running close the engine also */
	if ((emailMgrState == OSSO_STATE_UN_INIT) || (emailMgrState == OSSO_STATE_INIT_LIMITED)) {
#if CL_USE
		cl_iap_disconnect(NULL);
#endif
		osso_uninitialize_engine_process();
		osso_deinitialize(osso_email_mgr_ctx);
		exit(0);
	}
	return OSSO_OK;
}

/**
   This function is called for the last update of the outbox
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_update_outbox_status(void)
{
	osso_return_t rv = OSSO_OK;

	rv = osso_rpc_run_with_defaults(osso_email_mgr_ctx,
					TASKNAV_APP,
					OSSO_TASKNAV_UPDATE_STATUS,
					NULL,
					DBUS_TYPE_UINT32,
					(unsigned int) time(NULL), DBUS_TYPE_INVALID);
	if (rv != OSSO_OK) {
		osso_log(LOG_ERR, " Couldn't send the Dbus message \n");
	}
	return rv;
}


/**
   This function open a specific message, when the user clicks on a particular
   message in the Task Navigator. Subsequent to the closing of this message,
   this message will be removed from the Task Navigator window pane.

   @param osso The library context as returned by osso_initialize.
   @param id the unique id of the mail.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/

static osso_return_t osso_email_mgr_open_msg(osso_context_t * ctx, const gchar * args)
{
	if (osso_email_mgr_check_ui_available() != OSSO_OK) {
		if (osso_initialize_ui_process(OSSO_TASKNAV_OPEN_MSG, args)
		    != OSSO_OK) {
			return OSSO_ERROR;
		}
	} else {
		engine_dbus_tasknav_msg_operation(args, OSSO_TASKNAV_OPEN_MSG);
	}
	return OSSO_OK;
}

/**
   This function delete a specific message. When request comes from other
   application (like Global Search) to delete the message, Engine will
   asks UI to delete. In case of UI is not availble, engine delete the
   message locally and in the server also.

   @param osso The library context as returned by osso_initialize.
   @param id the unique id of the mail.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/

static osso_return_t osso_email_mgr_delete_msg(osso_context_t * ctx, const gchar * args)
{
	gchar *folder_name = NULL;
	osso_return_t rv = OSSO_OK;

	folder_name = get_fold_name(args);
	if (al_folder_remove_msg(folder_name, args, FALSE) != FOLDER_SUCCESS) {
		rv = OSSO_ERROR;
	}
	g_free(folder_name);
	return rv;
}


/**
   This function is called to open inbox folder in Email UI application.

   @param args the method type of open inbox
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/

static osso_return_t osso_email_mgr_open_inbox(const gint type)
{
	osso_return_t rv = OSSO_OK;

	if (osso_email_mgr_check_ui_available() != OSSO_OK) {
		rv = osso_initialize_ui_process(type, NULL);
	} else {
		rv = engine_dbus_open_inbox(type);
	}
	return rv;
}

/*\@}*/
/**
 * \defgroup EmailMgr Email Manager
 */
/*\@{*/

/*\@}*/

/**
   This function is called when the Email application needs to be
   initialized. This  can be called when the user clicks the Email
   Application Icon from the Home, or when SendAs is done by other
   application, or request comes from the Task Navigator

   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */
static osso_return_t osso_initialize_email_application(osso_context_t * ctx)
{
	osso_log(LOG_DEBUG, "osso_initialize_email_application starts\n");
	if (osso_initialize_engine_process() != OSSO_OK) {
		return OSSO_ERROR;
	}
	return OSSO_OK;
}

/**
   This function is for intializing the Engine process.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */
static osso_return_t osso_initialize_engine_process(void)
{
	osso_return_t rv = OSSO_OK;

	osso_log(LOG_DEBUG, "osso_initialize_engine_process starts \n");
	if ((rv = engine_dbus_al_register_engine()) != OSSO_OK) {
		osso_log(LOG_INFO, "Error in al register to engine \n");
	}
	osso_email_mgr_register_with_engine_dbus();
	/*Intialise scheduling */
	osso_mgr_schdg_initializescheduler();
	osso_mgr_schdg_notify_changes();
	engine_dbus_server_thread();
	return rv;
}

/**
   This function is for unintializing the Engine process.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */
osso_return_t osso_uninitialize_engine_process(void)
{
	if (emailMgrState != OSSO_STATE_UN_INIT) {
		engine_dbus_al_unregister_engine();
		osso_email_mgr_unregister_with_engine_dbus();
		/* engine_dbus_server_thread_stop(); */
		emailMgrState = OSSO_STATE_UN_INIT;
		return OSSO_OK;
	}
	return OSSO_ERROR;
}


/**
   This function is for intializing the UI process.

   @param limited used to initialise UI application from command line with
    specific option like Open Inbox etc
   @param args addtional arguments for above option
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/
static osso_return_t osso_initialize_ui_process(const gint limited, const gchar * args)
{
	GString *cmd_line = NULL;
	gchar **argv = NULL;
	gint argp = 0;
	gint rc = 0;
	gchar *escaped_args = NULL;

	osso_log(LOG_DEBUG, " Initialise UI process here....\n");
	if(emailMgrState == OSSO_STATE_UI_INIT_PROGRESS)
	{
		osso_log(LOG_ERR, " ui init again..\n");
		return OSSO_OK;
	}
		


	cmd_line = g_string_new(NULL);
	if (cmd_line == NULL) {
		osso_log(LOG_ERR, " Memory Alloc failed \n");
		return OSSO_ERROR;
	}

	if (limited == 0) {
		g_string_append_printf(cmd_line, "%s", EMAIL_UI_PROCESS);
	} else {
		g_string_append_printf(cmd_line, "%s %d", EMAIL_UI_PROCESS, limited);
		if (args != NULL) {
			escaped_args = gnome_vfs_escape_path_string(args);
			if(escaped_args != NULL) {
				g_string_append_printf(cmd_line, " %s", escaped_args);
				g_free(escaped_args);
			} else {
				g_string_append_printf(cmd_line, " %s", args);
			}
		}
	}
	osso_log(LOG_DEBUG, " args = %s \n", cmd_line->str);

	rc = g_shell_parse_argv(cmd_line->str, &argp, &argv, NULL);
	g_string_free(cmd_line, TRUE);

	if (!rc) {
		osso_log(LOG_ERR, " Invalid args to launch UI process..\n");
		g_strfreev(argv);
		return OSSO_ERROR;
	}
	osso_log(LOG_DEBUG, " Starting UI Application \n");

	emailMgrState = OSSO_STATE_UI_INIT_PROGRESS;
	rc = g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL);
	g_strfreev(argv);
	if (rc == FALSE) {
		emailMgrState = OSSO_STATE_INIT_LIMITED;
		osso_log(LOG_ERR, " Fail : UI Process not created ..\n");
		return OSSO_ERROR;
	} else {
		manager_timeout_id = g_timeout_add(UI_START_TIMEOUT, osso_email_sync_timeout, NULL);
	}
	return OSSO_OK;
}

/**
   This function is to check whether UI is Up and running.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_check_ui_available(void)
{

	gint status = 0;

	if ((emailMgrState != OSSO_STATE_INIT_COMPLETE) &&
	    (emailMgrState != OSSO_STATE_UI_INIT_PROGRESS)) {
		return OSSO_ERROR;
	}
	if (osso_email_gconf_get_int(EMAIL_GCONF_UI_CLOSED, &status) == FALSE) {
		return OSSO_OK;

	}

	if (status == 1) {
		osso_log(LOG_DEBUG, "\nUI has closed\n");
		gl_email_mgr_waiting = TRUE;
		return OSSO_ERROR;
	}
	osso_log(LOG_DEBUG, "\nUI is already running\n");

	return OSSO_OK;
}

/**
   This function is called to initialise tasknavigator callback functions

   @param  osso The library context as returned by osso_initialize.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_register_with_engine_dbus(void)
{
	if (email_mgr_cb != NULL) {
		return OSSO_OK;
	}
	osso_log(LOG_DEBUG, " Register Email Manager Call back functions \n");
	email_mgr_cb = g_new0(EmailMgrCallback, 1);
	if (email_mgr_cb == NULL) {
		osso_log(LOG_ERR, "emailmanager callback Mem Alloc failed ..\n");
		return OSSO_ERROR;
	}
	email_mgr_cb->email_mgr_update_outbox_status_cb = osso_email_mgr_update_outbox_status;

	email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb = osso_email_mgr_tasknav_send_recv_focus;

	email_mgr_cb->email_mgr_update_scheduling = osso_mgr_schdg_update_scheduling;

	email_mgr_cb->email_mgr_statesave_cb = osso_email_mgr_statesave_operation_cb;

	email_mgr_cb->email_mgr_disable_timer = osso_email_mgr_disable_timer;

	email_mgr_cb->email_mgr_quit_loop = osso_email_mgr_quit_loop;

	return OSSO_OK;
}


/**
   This function is called to uninitialise tasknavigator callback functions

   @param  osso    The library context as returned by osso_initialize.
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_unregister_with_engine_dbus(void)
{
	if (email_mgr_cb == NULL) {
		return OSSO_ERROR;
	}
	email_mgr_cb->email_mgr_tasknav_send_recv_focus_cb = NULL;
	email_mgr_cb->email_mgr_update_scheduling = NULL;
	email_mgr_cb->email_mgr_statesave_cb = NULL;
	g_free(email_mgr_cb);
	return OSSO_OK;
}

/**
 * \defgroup SendAs File Manager, Browser Interaction
 */
/*\@{*/

/**
   This function is called when DBus SendAs Files request comes from other
   applications (like File Manager). This method sends the file list to UI.

   @param  osso The library context as returned by osso_initialize.
   @param  list GSList of attachments to be send
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_files(osso_context_t * context, const gchar * files)
{
	MsgInfo *msginfo = NULL;
	gchar **list = NULL;
	const gchar *cur = NULL;
	gint i = 0;
	gint rv = OSSO_OK;
	AttachInfo *attinfo = NULL;
	gchar *packed_string = NULL;
	gchar *outbuf = NULL;

	MGR_CHECK_STATUS((context == NULL), "Context NULL arg", return OSSO_ERROR);
	MGR_CHECK_STATUS((files == NULL), "Attach list NULL arg", return OSSO_ERROR);
	if (strncmp(files, FILE_MGR_ARG, strlen(FILE_MGR_ARG)) != 0) {
		osso_log(LOG_INFO, " Inavlid Files list \n");
		return OSSO_ERROR;
	}
	cur = files;
	cur += strlen(FILE_MGR_ARG);

	/* Populate the attachment list in msginfo */
	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem Alloc Failed for msginfo \n");
		return OSSO_ERROR;
	}

	list = g_strsplit(cur, ",", 0);

	for (i = 0; list[i] != NULL; i++) {
		attinfo = g_new0(AttachInfo, 1);
		if (attinfo != NULL) {
			attinfo->type = DIRECT;
			attinfo->filepath = g_strdup(list[i]);
			msginfo->attach_list = g_slist_append(msginfo->attach_list, attinfo);
		}
	}
	g_strfreev(list);

	if (osso_email_mgr_check_ui_available() != OSSO_OK) {
		engine_dbus_editor_create(msginfo, NULL, &packed_string);
		osso_email_utils_pack_str_to_base64((guchar *) packed_string, &outbuf);
		rv = osso_initialize_ui_process(OSSO_EMAIL_FILE_MGR, outbuf);
		g_free(outbuf);
		g_free(packed_string);
	} else {
		engine_dbus_editor_create(msginfo, NULL, NULL);
	}
	procmsg_msginfo_free(msginfo);	/* ??? Can we use engine commonlib */
	return rv;
}

/**
   This function is called when DBus Send Url request comes from other
   applications (like Browser). This method sends url link to UI.

   @param  osso The library context as returned by osso_initialize.
   @param  url  Link which is to be send
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_url(osso_context_t * context, const gchar * url)
{
	MsgInfo *msginfo = NULL;
	gchar *message = NULL;
	const gchar *cur = NULL;
	gchar *packed_string = NULL;
	gchar *outbuf = NULL;
	gint rv = OSSO_OK;

	MGR_CHECK_STATUS((context == NULL), "Context NULL arg", return OSSO_ERROR);
	MGR_CHECK_STATUS((url == NULL), "URL is NULL arg", return OSSO_ERROR);

	if (strncmp(url, URL_ARG, strlen(URL_ARG)) != 0) {
		osso_log(LOG_INFO, " Invalid url \n");
		return OSSO_ERROR;
	}
	cur = url;
	cur += strlen(URL_ARG);

	/* Populate the attachment list in msginfo */
	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem Alloc Failed for msginfo \n");
		return OSSO_ERROR;
	}
	if (cur != NULL) {
		message = g_strdup(cur);
	}

	if (osso_email_mgr_check_ui_available() != OSSO_OK) {
		engine_dbus_editor_create(msginfo, message, &packed_string);
		osso_email_utils_pack_str_to_base64((guchar *) packed_string, &outbuf);
		rv = osso_initialize_ui_process(OSSO_EMAIL_URL, outbuf);
		g_free(outbuf);
		g_free(packed_string);
	} else {
		engine_dbus_editor_create(msginfo, message, NULL);
	}
	g_free(message);
	procmsg_msginfo_free(msginfo);	/* ??? Can we use engine commonlib */
	return rv;
}

/**
   This function is called when DBus mailto request comes from other
   applications (like Browser). This method parses the mailto string and
   sends the information to UI application.

   @param  osso The library context as returned by osso_initialize.
   @param  mailto Contains the list mailto arguments in RFC standards
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/

static osso_return_t osso_email_mgr_mailto(osso_context_t * context, const gchar * mailto)
{
	MsgInfo *msginfo = NULL;
	gchar *message = NULL;
	const gchar *cur = NULL;
	gchar *packed_string = NULL;
	gchar *outbuf = NULL;
	gint rv = OSSO_OK;

	MGR_CHECK_STATUS((context == NULL), "Context NULL arg", return OSSO_ERROR);
	MGR_CHECK_STATUS((mailto == NULL), "mailto arg NULL", return OSSO_ERROR);

	if (strncmp(mailto, MAIL_TO_ARG, strlen(MAIL_TO_ARG)) == 0) {
		cur = mailto;
		cur += strlen(MAIL_TO_ARG);
	}
	else if (strncmp(mailto, MAILTO, strlen(MAILTO)) == 0) {
		cur = mailto;
	}
	else {
		osso_log(LOG_INFO, " Invalid mailto received: '%s'\n", mailto);
		return OSSO_ERROR;
	}

	/* Populate the attachment list in msginfo */
	msginfo = g_new0(MsgInfo, 1);
	if (msginfo == NULL) {
		osso_log(LOG_ERR, "Mem Alloc Failed for msginfo \n");
		return OSSO_ERROR;
	}
	msginfo->header = g_new0(MsgHeader, 1);
	if (msginfo->header == NULL) {
		procmsg_msginfo_free(msginfo);
		osso_log(LOG_ERR, "Mem Alloc Failed for msginfo \n");
		return OSSO_ERROR;
	}

	/* get the arguments like cc, bcc  subject, message from url */
	osso_email_mgr_parse_mailto_str(cur, msginfo->header, &message);
	if (osso_email_mgr_check_ui_available() != OSSO_OK) {
		engine_dbus_editor_create(msginfo, message, &packed_string);
		osso_email_utils_pack_str_to_base64((guchar *) packed_string, &outbuf);
		rv = osso_initialize_ui_process(OSSO_EMAIL_MAIL_TO, outbuf);
		g_free(outbuf);
		g_free(packed_string);
	} else {
		engine_dbus_editor_create(msginfo, message, NULL);
	}
	g_free(message);
	procmsg_msginfo_free(msginfo);	/* Can we use engine common lib */
	return rv;
}


/**
  This function is called to parse the mailto string and populate the
  the data in message header

  @param mailto string received from other application
  @param msgheader populated with mailto parameters
  @param body populated with body part in mailto string
  @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
 */

static osso_return_t
osso_email_mgr_parse_mailto_str(const gchar * mailto, MsgHeader * msgheader, gchar ** body)
{

	const char *QUESTION_STR = "?";
	const char *AMPERSAND_STR = "&";
	const char *EQUALTO_AMPERSEND = "=&";

	GString *to = NULL;
	gchar *bcc = NULL;
	gchar *cc = NULL;
	gchar *subject = NULL;
	gchar *header = NULL;
	gchar *content = NULL;
	gchar *buf = NULL;
	const gchar *ptr = NULL;
	gint len = 0;
	gint clen = 0;

	MGR_CHECK_STATUS((mailto == NULL), "mailto arg NULL", return OSSO_ERROR);
	MGR_CHECK_STATUS((msgheader == NULL), "msgheader NULL", return OSSO_ERROR);
	MGR_CHECK_STATUS((body == NULL), "body arg NULL", return OSSO_ERROR);

	if (g_ascii_strncasecmp(mailto, MAILTO, strlen(MAILTO)) != 0) {
		osso_log(LOG_DEBUG, "Invalid mailto string \n");
		return OSSO_ERROR;
	}

	buf = g_strndup(mailto, strlen(mailto));
	if (buf == NULL) {
		osso_log(LOG_ERR, "mailto buf : g_strndup failed \n");
		return OSSO_ERROR;
	}

	/* Parse recipients (everything after ':' until '?' or end of sting). */
	/* Skip the length of "mailto:" */

	ptr = buf + 7;
	len = strcspn(ptr, QUESTION_STR);

	if (to == NULL) {
		to = g_string_new(NULL);
		if (to == NULL) {
			osso_log(LOG_ERR, "mailto : Memory Alloc failed \n");
			g_free(buf);
			return OSSO_ERROR;
		}
	}

	/* Add the recipients to the TO list */
	if (len > 0) {
		content = g_strndup(ptr, len);
		osso_email_mgr_decode_content(content);
		g_string_append(to, content);
		g_free(content);
	}

	ptr += len;
	if (*ptr == QUESTION) {
		ptr++;

		while (*ptr) {
			len = strcspn(ptr, EQUALTO_AMPERSEND);

			/* If the string is not formed properly  ignore. */

			if (ptr[len] != EQUALTO) {
				break;
			}
			header = (gchar *) ptr;

			/* Append NULL termination to the end */
			header[len] = '\0';
			ptr += len + 1;

			clen = strcspn(ptr, AMPERSAND_STR);
			content = g_strndup(ptr, clen);
			osso_email_mgr_decode_content(content);

			if (!g_ascii_strncasecmp(header, TO_STR, strlen(TO_STR))) {
				g_string_append_printf(to, ";%s", content);
			} else if (!g_ascii_strncasecmp(header, CC_STR, strlen(CC_STR))) {
				cc = content;
				content = NULL;
			} else if (!g_ascii_strncasecmp(header, BCC_STR, strlen(BCC_STR))) {
				bcc = content;
				content = NULL;
			} else if (!g_ascii_strncasecmp(header, SUBJECT_STR, strlen(SUBJECT_STR))) {
				g_free(subject);
				/* Validate the subject according to UTF8 format */
				if (g_utf8_validate(content, -1, NULL)) {
					subject = content;
					content = NULL;
				}

			} else if (!g_ascii_strncasecmp(header, BODY_STR, strlen(BODY_STR))) {
				g_free(*body);
				if (g_utf8_validate(content, -1, NULL)) {
					*body = content;
					content = NULL;
				}
			} else if (!g_ascii_strncasecmp(header, ATTACH_STR, strlen(ATTACH_STR))) {
				/*Change file url to absolute path */

				if (!g_ascii_strncasecmp(content, FILE_COLEN, strlen(FILE_COLEN))) {
					/* Parse the Absolute URL that we receive here??? Does
					 * Email Mgr  need to support
					 */
					// url = parse_url (content);
				}
			}
			g_free(content);

			ptr += clen;
			if (*ptr == AMPERSAND) {
				ptr++;
				/* check for AMP(amp;) */
				if (!strncmp(ptr, AMP, strlen(AMP))) {
					ptr += 4;
				}
			}
		}
	}

	if (to != NULL) {
		if (to->str != NULL) {
			msgheader->to = g_strdup(to->str);
		}
		g_string_free(to, TRUE);
	}
	msgheader->cc = cc;
	msgheader->bcc = bcc;
	msgheader->subject = subject;
	g_free(buf);
	return OSSO_OK;
}

/**
   This function is called to decode the content

   @param content contains the decoded data
   @return OSSO_OK if the application can be launched, OSSO_ERROR if not.
*/

static void osso_email_mgr_decode_content(gchar * content)
{

	gchar *source = NULL, *dest = NULL;
	if (content == NULL) {
		return;
	}
	source = dest = (gchar *) content;
	do {
		if (*source == '%' && source[1] && source[2]) {
			*dest++ = (TRNSLDIGIT(source[1]) << 4) + TRNSLDIGIT(source[2]);
			source += 2;
		} else {
			*dest++ = *source;
		}
	} while (*source++);
}


/**
   This function is called when UI is unable to start and time out
   has occured.

   @param data contains data at the time of timer creation
   @return FALSE so that timer will not start again
*/

static gboolean osso_email_sync_timeout(gpointer data)
{
	if (emailMgrState == OSSO_STATE_UI_INIT_PROGRESS) {
		emailMgrState = OSSO_STATE_INIT_LIMITED;
	}
	return FALSE;
}

/**
   This function is called engine received notification about UI
   startup nd to cancel the timer.

   @return TRUE
*/

static gboolean osso_email_mgr_disable_timer(void)
{
        if(manager_timeout_id > 0)
		g_source_remove(manager_timeout_id);
	return TRUE;
}

/**
  This function is called to get the email standard folders and populate them
  to osso_rpt_t struct.

  @param ret_rpc_t populated with folder names with ':' as separator
  @return OSSO_OK on success or  OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_get_standard_folders(osso_rpc_t * ret_rpc_t)
{
	gchar *out = NULL;
	if (ret_rpc_t == NULL) {
		return OSSO_ERROR;
	}

	out = g_strdup_printf("%s:%s:%s:%s:%s", _("mcen_me_folder_inbox"),
			      _("mcen_me_folder_outbox"), _("mcen_me_folder_drafts"),
			      _("mcen_me_folder_sent"), _("mcen_me_folder_archive"));

	ret_rpc_t->type = DBUS_TYPE_STRING;
	ret_rpc_t->value.s = out;
	return OSSO_OK;
}

/**
  This function is called to get the email folders and populate them
  to osso_rpt_t struct.

  @param ret_rpc_t populated with folder names with '/' as separator
  @return OSSO_OK on success or  OSSO_ERROR if not.
 */

static osso_return_t osso_email_mgr_get_folder_list(osso_rpc_t * ret_rpc_t)
{
	GSList *flist = NULL, *cur = NULL;
	GString *fl_str = NULL;
	Folder *folder = NULL;
	gboolean first = FALSE;

	al_folder_get_list(&flist);
	if (flist == NULL) {
		return OSSO_ERROR;
	}
	cur = flist;
	fl_str = g_string_new(FALSE);
	for (cur = flist; cur != NULL; cur = cur->next) {
		folder = cur->data;
		if ((folder == NULL) || (folder->name == NULL)) {
			continue;
		}
		if (strchr(folder->name, DIR_SEPRTR) != NULL) {
			continue;
		}
		if (first != FALSE) {
			g_string_append_c(fl_str, DIR_SEPRTR);
		}
		first = TRUE;
		if (strcmp(folder->name, INBOX_DIR) == 0) {
			g_string_append(fl_str, "MAND:");
			g_string_append(fl_str, _("mcen_me_folder_inbox"));
		} else if (strcmp(folder->name, OUTBOX_DIR) == 0) {
			g_string_append(fl_str, "MAND:");
			g_string_append(fl_str, _("mcen_me_folder_outbox"));
		} else if (strcmp(folder->name, DRAFT_DIR) == 0) {
			g_string_append(fl_str, "MAND:");
			g_string_append(fl_str, _("mcen_me_folder_drafts"));
		} else if (strcmp(folder->name, SENT_DIR) == 0) {
			g_string_append(fl_str, "MAND:");
			g_string_append(fl_str, _("mcen_me_folder_sent"));
		} else if (strcmp(folder->name, ARCHIVE) == 0) {
			g_string_append(fl_str, "MAND:");
			g_string_append(fl_str, _("mcen_me_folder_archive"));
		} else {
			g_string_append(fl_str, "USER:");
			g_string_append(fl_str, folder->name);
		}
	}
	if (ret_rpc_t) {
		ret_rpc_t->type = DBUS_TYPE_STRING;
		ret_rpc_t->value.s = fl_str->str;
		g_string_free(fl_str, FALSE);
	}
	return OSSO_OK;
}

/**
   This function is called when exit signal is received by email application.

   @param  flag TRUE incase exit signal FALSE incase of exit if possible
   @param  userdata  data which is supplied for callback registartion
   @return nothing
*/

static void osso_email_mgr_exit_cb(gboolean flag, gpointer userdata)
{
	if (emailMgrState == OSSO_STATE_INIT_COMPLETE) {
		engine_dbus_send_app_killer_signal(SIGNAL_EXIT);
	} else {
		osso_email_mgr_quit_loop();
	}
	return;
}

/**
  This function is called to quit from the g_main_loop
  @return nothing
*/

static void osso_email_mgr_quit_loop(void)
{
	if (loop != NULL) {
		g_main_loop_quit(loop);
		g_main_loop_unref(loop);
	}
	/* to kill enprocess on receiving kill signal */
	exit(-1);
}


static void osso_email_mgr_backup_reg(void)
{
	DBusConnection *sess_bus = NULL;
	gchar *match_rule = NULL;

	sess_bus = osso_get_dbus_connection(osso_email_mgr_ctx);
	match_rule = g_strdup_printf("type='signal', interface='com.nokia.backup'");
	dbus_bus_add_match(sess_bus, match_rule, NULL);
	dbus_connection_add_filter(sess_bus, backup_signal_cb, NULL, NULL);
	dbus_connection_flush(sess_bus);
	g_free(match_rule);
}
static DBusHandlerResult
backup_signal_cb(DBusConnection * connection, DBusMessage * message, void *user_data)
{
	if (dbus_message_is_signal(message, "com.nokia.backup", "backup_start")) {
		osso_log(LOG_DEBUG, "Backup START signal received\n");
		engine_dbus_backup_handle();
		engine_dbus_backup_event();
	} else if (dbus_message_is_signal(message, "com.nokia.backup", "backup_finish")) {
		osso_log(LOG_DEBUG, "Backup STOP signal received\n");
	} else {
		osso_log(LOG_DEBUG, "Unknown Backup signal received\n");
	}
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

