
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include "manager.h"
#include "scheduling.h"
#include "al.h"
#include "dbus_utils_def.h"
#include "account.h"
#include "email_gconf.h"

#include <alarm_event.h>

#if CL_USE
#include "cl.h"
#endif


extern gint emailMgrState;
static guint sch_notify_id = 0;
static void osso_mgr_schdg_remove_timer(void);
static glong get_alarm_id(void);

/**
This function is called whenever timer for scheduling expires

@return TRUE to rerun the timer
        FALSE to stop the timer.
*/
gboolean osso_mgr_schdg_send_recv(gpointer data)
{
	GSList *tmp_acc_list = NULL;
	GSList *account_list = NULL;
	gboolean connect_via_wlan = FALSE;

	osso_log(LOG_INFO, "[%s] starts", __FUNCTION__);

	/*Get all the configured accounts */
	al_get_account_list(&account_list);
	if (!account_list) {
		osso_log(LOG_INFO, " Account list is NULL!\n");
		return FALSE;
	}

	if (!osso_email_gconf_get_bool(EMAIL_SCHEDULING_VIA_WLAN, &connect_via_wlan)) {
		osso_log(LOG_DEBUG, "error getting the value");
		return FALSE;
	}
#if CL_USE
	if (connect_via_wlan) {
		if (!active_con_is_wlan())
			return FALSE;
	}
#endif
	tmp_acc_list = account_list;
	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_PROGRESS);
	/*send and receive mails for each account */
	while (tmp_acc_list != NULL) {
#if CL_USE
		cl_iap_set_type(TRUE);	/* set for timed out connection */
#endif
		osso_log(LOG_INFO, "send/recv done by auto-update");
		al_recv_mail(tmp_acc_list->data, TRUE);
#if CL_USE
		cl_iap_set_type(FALSE);
#endif
		tmp_acc_list = tmp_acc_list->next;
	}

	osso_email_gconf_set_sendrecv_status(EMAIL_GCONF_SENDRECV_STATUS, SEND_RECV_DONE);

	/*free the accountlist */
	g_slist_foreach(account_list, (GFunc) g_free, NULL);
	g_slist_free(account_list);

#if CL_USE
	if ((emailMgrState == OSSO_STATE_UN_INIT) || (emailMgrState == OSSO_STATE_INIT_LIMITED)) {
		cl_iap_disconnect(NULL);
	}
#endif
	osso_log(LOG_INFO, "[%s] ends", __FUNCTION__);

	return TRUE;
}

/*Get the alarm_id for the alarm event for email.*/
static glong get_alarm_id()
{
	glong alarm_id = 0;
	gchar *alarm_id_str = NULL;

	osso_log(LOG_INFO, "[%s] starts", __FUNCTION__);
	/*Get the alarm_id for the alarm event for email. */
	if (!osso_email_gconf_get_str(EMAIL_ALARM_ID, &alarm_id_str)) {
		osso_log(LOG_DEBUG, "error getting the value");
		return 0;
	}
	if (!alarm_id_str)
		return 0;
	alarm_id = (glong) atol(alarm_id_str);
	g_free(alarm_id_str);
	return alarm_id;
}

/**
This function is called to initialize  scheduling process.

@return void.
*/
void osso_mgr_schdg_initializescheduler(void)
{
	gint scheduling_interval = 0;
	gint time_period = 0;

	osso_log(LOG_INFO, "[%s] starts", __FUNCTION__);

	/*Check if there are any configured account.If not return without
	   starting the  timer
	 */
	if (!is_account_list_valid()) {
		osso_log(LOG_INFO, " Account list is NULL!\n");
		return;
	}

	/*Get the timer interval value if scheduling is on */
	if (!osso_email_gconf_get_int(EMAIL_SCHEDULING_INTERVAL, &scheduling_interval)) {
		osso_log(LOG_DEBUG, "error getting the value");
		return;
	}
	if (scheduling_interval < 0) {
		osso_log(LOG_INFO, "Auto retrieve is not set");
		return;
	}
	/*GConf is still not set with any value. */
        if (scheduling_interval == 0) {
                scheduling_interval = 15;
        }

	glong alarm_id = 0;
	gchar *alarm_id_str = NULL;
	if (get_alarm_id() > 0) {
		osso_log(LOG_INFO, "schgd already init");
		/*alarm daemon is already running */
		return;
	}
	time_period = scheduling_interval;
	alarm_event_t event = {
		.alarm_time = time(NULL) + (time_period * SECS_IN_MIN),
		.recurrence = time_period,
		.recurrence_count = -1,
		.snooze = 0,
		.message = NULL,
		.sound = NULL,
		.icon = NULL,
		.dbus_interface = DBUS_EMAIL_INTERFACE,
		.dbus_service = DBUS_EMAIL_SERVICE,
		.dbus_path = DBUS_EMAIL_OBJ_PATH,
		.dbus_name = START_SEND_RECV,
		.exec_name = NULL,
		.flags = ALARM_EVENT_NO_DIALOG |
		    ALARM_EVENT_NO_SNOOZE |
		    ALARM_EVENT_RUN_DELAYED | ALARM_EVENT_CONNECTED | ALARM_EVENT_ACTIVATION
	};
	alarm_id = alarm_event_add(&event);
	alarm_id_str = g_strdup_printf("%ld", alarm_id);
	if (!osso_email_gconf_set_str(EMAIL_ALARM_ID, alarm_id_str))
		osso_log(LOG_DEBUG, "Error setting the interval value");
	g_free(alarm_id_str);

	osso_log(LOG_INFO, "[%s] ends", __FUNCTION__);
	return;
}

/**
 This function is used to uninitialise scheduling.
 @return void.
*/
void osso_mgr_schdg_uninit(void)
{
	GConfClient *gc = NULL;

	osso_mgr_schdg_remove_timer();
	gc = gconf_client_get_default();
	gconf_client_notify_remove(gc, sch_notify_id);

	return;
}

/**
This function is called to change scheduling process if any modification 
is made in account management for e.g deletion of all accounts/creation of first new account
or in scheduling settings through UI.

@return void.
*/
void osso_mgr_schdg_update_scheduling()
{
	/*Remove any previous timer */
	osso_mgr_schdg_remove_timer();

	/*Reinitialises the scheduler */
	osso_mgr_schdg_initializescheduler();

	return;
}

/**
 This function is used to check whether scheduling is in progress.
 
 @return void
*/
gboolean osso_mgr_schdg_has_task(void)
{
	return FALSE;
}

/**
 This is a gconf notifier function.Which is called when ever the scheduling gconf
 values change.
 
 @param client is GConfClient pointer.
 @param id is the 

 @return.
*/
static void osso_mgr_schdg_update_scheduling_change(GConfClient * client,
						    guint id, GConfEntry * entry, gpointer data)
{
	/*upadte scheduling for values change */
	osso_mgr_schdg_update_scheduling();

	return;
}

/**
 This function is used to register the gconf client to the 
 scheduling changes.
 
 @return void.
*/
void osso_mgr_schdg_notify_changes(void)
{
	GConfClient *gconf_client = NULL;
	GError *error = NULL;

	gconf_client = gconf_client_get_default();

	/* add gconf dir to watch list */
	gconf_client_add_dir(gconf_client, EMAIL_SCHEDULING_PATH, GCONF_CLIENT_PRELOAD_ONELEVEL,
			     &error);
	if (error != NULL) {
		osso_log(LOG_ERR, "Scheduling status couldn't watch");
		g_error_free(error);
		return;
	}
	/*set the callback function to be called if scheduling value changes */
	sch_notify_id = gconf_client_notify_add(gconf_client, EMAIL_SCHEDULING_PATH,
						osso_mgr_schdg_update_scheduling_change, NULL, NULL,
						NULL);
	return;
}

/**
Remove the scheduling timer in situations where  there are no accounts
@return void
*/
static void osso_mgr_schdg_remove_timer()
{
	glong alarm_id = 0;
	alarm_id = get_alarm_id();
	if (alarm_id > 0)
		alarm_event_del(alarm_id);

	if (!osso_email_gconf_set_str(EMAIL_ALARM_ID, "0"))
		osso_log(LOG_DEBUG, "Error setting the interval value");
	return;
}
