
/*
 * LibSylph -- E-Mail client library
 * Copyright (C) 1999-2006 Hiroyuki Yamamoto
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>

#include "config.h"
#include "common.h"
#include "defs.h"
#include "prefs_account.h"
#include "account.h"
#include "utils.h"
#include "procmsg.h"
#include "procheader.h"
#include "folder.h"
#include "pop.h"
#include "pop3_msgs.h"
#include "md5.h"
#include "md5_hmac.h"
#include "base64.h"
#include "pop_recv.h"
#include "ssl.h"
#include "engine_errno.h"
#include "email_gconf.h"

extern GSList *gl_unavailable_list;
/*defined in pop_recv.c */

extern Pop3Callback *gl_pop3_callback_ptr;
/* defined in pop_recv.c */
static GSList *gl_pop3_delete_list = NULL;

static gint pop3_greeting_recv(Pop3Session * session, const gchar * msg);

static gint pop3_getauth_user_send(Pop3Session * session);
static gint pop3_getauth_pass_send(Pop3Session * session);
static gint pop3_getauth_apop_send(Pop3Session * session);
static gint pop3_getauth_cram_md5_send(Pop3Session * session);
static gint pop3_getrange_stat_send(Pop3Session * session);
static gint pop3_getrange_stat_recv(Pop3Session * session, const gchar * msg);
static gint pop3_getsize_list_send(Pop3Session * session);
static gint pop3_getsize_list_recv(Pop3Session * session, const gchar * data, guint len);
static gint pop3_retr_send(Pop3Session * session);
static gint pop3_retr_recv(Pop3Session * session, const gchar * data, guint len);

static gint pop3_delete_send(Pop3Session * session);
static gint pop3_delete_recv(Pop3Session * session);
static void pop3_gen_send(Pop3Session * session, const gchar * format, ...);

static void pop3_session_destroy(Session * session);

static Pop3State pop3_lookup_next(Pop3Session * session);

static Pop3ErrorValue pop3_ok(Pop3Session * session, const gchar * msg);

static gint pop3_session_recv_msg(Session * session, const gchar * msg);

static gint pop3_session_recv_data_finished(Session * session, guchar * data, guint len);
static gint pop3_cram_md5_recv(Pop3Session * session, const gchar * data);

static void pop3_handle_sizelimit_error(Pop3Session * session);

static gint pop3_save_msg(const gchar * file, const gchar * data, guint len);

static gint pop3_send_auth(Pop3Session * pop3_session);

static gint pop3_top_recv(Pop3Session * session, const gchar * data, guint len);

/**
 This function gets the temporary password for the connection,
 by waiting till user enters the password.   
 @param session pointer to the pop3 session 
 @return success(PS_PASSWD_FAIL) or PS_PASSWD_CANCEL on user 
 cancelling the operation. 
*/
static gint pop3_connect_tmp_passwd(Pop3Session * pop3_session);

#if USE_SSL
static gint pop3_stls_send(Pop3Session * session);

static gint pop3_stls_recv(Pop3Session * session);
#endif

/**
 This function send uidl command   
 @param session pointer to the pop3 session 
 @return success(PS_SUCCESS) 
*/
static gint pop3_getrange_uidl_send(Pop3Session * session);
/**
 This function handles uidl command response,to check any new mail & 
 updates the uidl table  
 @param session pointer to the pop3 session 
 @param data pointer to the pop3 response 
 @param len length of response 
 @return success(PS_SUCCESS) 
*/
static gint pop3_getrange_uidl_recv(Pop3Session * session, const gchar * data, guint len);


/**
 This function reads uidl table saved in file,prepares a hash table 
 of uidl.If msgid is passed,returns the matching uidl entry. 
 updates the uidl table.
 @param ac_prefs pointer to the account information 
 @param msgid pointer to the msgid to fetch/delete,this can be NULL 
 @param uidl_ptr, OUT parameter,is the uidl entry found in the table
 for the message number 
 @return uidl hash table 
*/
static GHashTable *pop3_get_uidl_table(PrefsAccount * ac_prefs,
				       const gchar * msgid, gchar ** uidl_ptr);

/**
 This function will be called to get the sequence number on 
 the server for which uidl is present. 
 @param  pointer to the pop3_session 
 @return PS_SUCCESS if sequence number on the server is found or POP3_FAIL 
*/
static gint pop3_sync_msg(Pop3Session * session);

/**
  This function gets file path of message
  @param msgid - message ID
  @return gchar* - file path
*/
static gboolean pop3_get_header_file_path(gint msgnum,
					  gchar * account_name, gchar ** file_path_ptr);

/**
  This function gets the sequence number of the 
  message on the server,for the message id in the list 
  @param session    pointer to the session strcuture 
  @return    POP3_SUCCESS on successfully searching the list,
             POP3_FAIL on error 
*/
static gint pop3_sync_msg_list(Pop3Session * session);

/**
 This function will be called to create a new pop3 session.
 @param Account pointer to the account structure
 @param msgid pointer to the messageid,incase of message 
 specific operation. 
 @return session pointer or NULL if input pointer is NULL
*/

/*@null@*/
Session *pop3_session_new(PrefsAccount * account, const gchar * msgid)
{
	Pop3Session *session = NULL;
	GHashTable *uidl_table = NULL;
	gchar *uidl = NULL;
	gchar *file_path = NULL;

	CHECK_RETURN_VAL_IF_FAIL(account != NULL, NULL);

	uidl_table = pop3_get_uidl_table(account, msgid, &uidl);

	if ((msgid != NULL) && (uidl == NULL)) {
		/* this is due to the Synchronisation problem */
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
				gl_pop3_callback_ptr->
				    pop3_set_msg_cb(POP3_MSG_DELETE_ERROR, "%s", msgid);
			}
		}
		set_engine_errno(ENGINE_ERROR_POP_MSG_DELETE);
		file_path = folder_get_message_file_path(msgid);
		gl_unavailable_list = g_slist_append(gl_unavailable_list, g_strdup(msgid));
		pop3_remove_header_file(file_path);
		g_free(file_path);
		folder_delete_list_msg(msgid, TRUE);
		folder_message_del_list(msgid, UNREAD);

		return NULL;
	}
	session = g_new0(Pop3Session, 1);
	if (session == NULL) {
		return NULL;
	}
	session_init(SESSION(session));

	SESSION(session)->type = SESSION_POP3;

	SESSION(session)->recv_msg = pop3_session_recv_msg;
	SESSION(session)->recv_data_finished = pop3_session_recv_data_finished;
	SESSION(session)->send_data_finished = NULL;

	SESSION(session)->destroy = pop3_session_destroy;

	session->state = POP3_READY;
	session->ac_prefs = account;
	session->current_time = time(NULL);
	session->error_val = PS_SUCCESS;
	session->size_error = 0;
	session->error_msg = NULL;
	session->msgid = NULL;
	session->get_msg_num = 0;
	session->delete_msg_num = 0;
	session->uidl = uidl;
	session->uidl_table = uidl_table;
	session->uidl_is_valid = FALSE;
	session->retr_list = NULL;
	session->retr_count = 0;
	session->prog_count = 0;
	session->prog_cur_msg = 0;
	return SESSION(session);
}

/**
 This function will be called on receiving greeting message.
 @param session pointer to the pop3 session.
 @param msg pointer to the response received
 @return  success(0) always
*/

static gint pop3_greeting_recv(Pop3Session * session, const gchar * msg)
{
	session->state = POP3_GREETING;
	session->greeting = g_strdup(msg);
	return PS_SUCCESS;
}

/**
 This function will be called for sending STLS command to server.
 @param session pointer to the pop3 session.
 @return  success(0) 
*/
#if USE_SSL
static gint pop3_stls_send(Pop3Session * session)
{
	session->state = POP3_STLS;
	pop3_gen_send(session, "STLS");
	return PS_SUCCESS;
}

/**
 This function will be called for receiving STLS command response. 
 @param session pointer to the pop3 session.
 @return  POP3_SUCCESS(0) if tls session starts successfully or 
    else POP3_FAIL(-1)
*/
static gint pop3_stls_recv(Pop3Session * session)
{
	if (session_start_tls(SESSION(session)) < 0) {
		session->error_val = PS_SOCKET;
		return POP3_FAIL;
	}
	return PS_SUCCESS;
}
#endif
/**
 This function will be called for sending user command to server.
 @param session pointer to the pop3 session.
 @return  POP3_SUCCESS(0) on successful sending,POP3_FAIL(-1) if 
    no userid found 
*/
static gint pop3_getauth_user_send(Pop3Session * session)
{
	CHECK_RETURN_VAL_IF_FAIL(session->ac_prefs->user_settings.userid != NULL, POP3_FAIL);

	session->state = POP3_GETAUTH_USER;
	pop3_gen_send(session, "USER %s", session->ac_prefs->user_settings.userid);
	return PS_SUCCESS;
}

/**
 This function will be called for sending password to server.
 @param session pointer to the pop3 session.
 @return  success(0) on successful sending,POP3_FAIL(-1) if no password found 
*/

static gint pop3_getauth_pass_send(Pop3Session * session)
{

	session->state = POP3_GETAUTH_PASS;
	pop3_gen_send(session, "PASS %s", session->passwd);
	return PS_SUCCESS;
}

/**
 This function will be called for sending apop command to server.
 @param session pointer to the pop3 session.
 @return  POP3_SUCCESS(0) on successful sending,POP3_FAIL(-1)
    if no userid/password or time stamp is found 
*/
static gint pop3_getauth_apop_send(Pop3Session * session)
{
	gchar *start = NULL, *end = NULL;
	gchar *apop_str = NULL;
	SMD5 *md5;
	gchar *md5sum;

	CHECK_RETURN_VAL_IF_FAIL(session->ac_prefs->user_settings.userid != NULL, POP3_FAIL);

	session->state = POP3_GETAUTH_APOP;

	if ((start = strchr(session->greeting, '<')) == NULL) {
		log_message("Required APOP timestamp not found in greeting\n");
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}

	if ((end = strchr(start, '>')) == NULL || end == start + 1) {
		log_message("Timestamp syntax error in greeting\n");
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}
	if ((end + 1) == NULL) {
		log_message("Timestamp syntax error in greeting\n");
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}
	*(end + 1) = '\0';

	apop_str = g_strconcat(start, session->passwd, NULL);
	md5 = s_gnet_md5_new((guchar *)apop_str, strlen(apop_str));
	md5sum = s_gnet_md5_get_string(md5);
	/*md5_hex_digest(md5sum, (unsigned char *) apop_str);*/

	pop3_gen_send(session, "APOP %s %s", session->ac_prefs->user_settings.userid, md5sum);
	
	g_free(md5sum);
	s_gnet_md5_delete(md5);
	g_free(apop_str);
	
	return PS_SUCCESS;
}

/**
 This function will be called for sending AUTH command to server.
 @param session pointer to the pop3 session.
 @return POP3_SUCCESS(0) on successful sending,POP3_FAIL(-1)
*/
static gint pop3_getauth_cram_md5_send(Pop3Session * session)
{
	pop3_gen_send(session, POP3_AUTH_CMD_FORMAT, POP3_CRAM_MD5_AUTH);
	session->state = POP3_GETAUTH_CRAM_MD5;
	return POP3_SUCCESS;
}

/**
 This function will be called for sending STAT command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_getrange_stat_send(Pop3Session * session)
{
	session->state = POP3_GETRANGE_STAT;
	pop3_gen_send(session, "STAT");
	return PS_SUCCESS;
}

/**
 This function will be called for receiving STAT response from server.
 @param Session pointer to the pop3 session.
 @param msg pointer to the response received
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on 
   response errors
*/

static gint pop3_getrange_stat_recv(Pop3Session * session, const gchar * msg)
{
	if (sscanf(msg, "%d %d", &session->count, &session->total_bytes) != 2) {
		log_message("Pop3 protocol error\n");
		debug_print("\n SESSION COUNT:%d\n\n", session->count);
		session->error_val = PS_PROTOCOL;
		return POP3_FAIL;
	} else {
		if (session->count == 0) {
			session->uidl_is_valid = TRUE;
			/* STAT Response is 0 and list is marked for Deletion */
			if( session->delete_all == TRUE ) {
				pop3_write_retr_del_list(session->retr_list, FALSE);
			}
		} else {
			session->msg = g_new0(Pop3MsgInfo, session->count + 1);
			if (session->msg == NULL) {
				return POP3_FAIL;
			}
			session->cur_msg = 1;

		}
		if ((session->msgid != NULL) && (session->count == 0)) {
			session->error_val = PS_DELERR;
			return POP3_FAIL;

		}
	}
	return PS_SUCCESS;
}

/**
 This function will be called for sending LIST command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_getsize_list_send(Pop3Session * session)
{
	session->state = POP3_GETSIZE_LIST;
	pop3_gen_send(session, "LIST");
	return PS_SUCCESS;
}

/**
 This function will be called for receiving LIST response from server.
 @param session pointer to the pop3 session.
 @param data pointer to the response received
 @param len length of the data received.
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on 
    response errors
*/

static gint pop3_getsize_list_recv(Pop3Session * session, const gchar * data, guint len)
{
	gchar buf[POPBUFSIZE];
	gint buf_len;
	guint num, size;
	const gchar *p = data;
	const gchar *lastp = data + len;
	const gchar *newline = NULL;

	memset(buf, '\0', POPBUFSIZE);

	/* if message id is present,response is for only one message */
	while (p < lastp) {
		if ((newline = memchr(p, '\r', lastp - p)) == NULL) {
			return POP3_FAIL;
		}
		buf_len = MIN(newline - p, sizeof(buf) - 1);
		memcpy(buf, p, buf_len);
		buf[buf_len] = '\0';

		p = newline + 1;
		if (p < lastp && *p == '\n') {
			p++;
		}

		if (sscanf(buf, "%u %u", &num, &size) != 2) {
			session->error_val = PS_PROTOCOL;
			return POP3_FAIL;
		}

		if (num > 0 && num <= session->count) {
			session->msg[num].size = size;
		}
		if (num > 0 && num < session->cur_msg) {
			session->cur_total_bytes += size;
		}
	}
	debug_print("THE COUNT is %d\tTHE TOTAL SIZE IS:%d\n", session->count,
		    session->cur_total_bytes);
	debug_print("THE actual size is:%d", session->total_bytes);
	return PS_SUCCESS;
}

/**
 This function will be called for sending RETR command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_retr_send(Pop3Session * session)
{
	session->state = POP3_RETR;
	/* The message body has to be retrieved for a message number only */
	pop3_gen_send(session, "RETR %d", session->cur_msg);
	return PS_SUCCESS;
}

/**
 This function will be called on receiving response to RETR command.
 @param session pointer to the pop3 session.
 @param data pointer to the data received
 @param len length of the data received.
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) if data
    could not be saved.
*/

static gint pop3_retr_recv(Pop3Session * session, const gchar * data, guint len)
{
	gchar *file = NULL;
	gint drop_ok = 0;
	gint msg_num = 0;
	gchar *inbox_folder = NULL;
	gchar *msg_id = NULL;

	inbox_folder = g_strdup_printf("%s%c%s", get_mail_dir(), G_DIR_SEPARATOR, INBOX_DIR);

	if (session->msgid != NULL) {

		file = g_strdup_printf("%s%c%s", inbox_folder, G_DIR_SEPARATOR, session->msgid);

	} else if (session->retr_list != NULL) {

		file = g_strdup_printf("%s%c%s%c%s%c%u", inbox_folder,
				       G_DIR_SEPARATOR,
				       session->ac_prefs->user_settings.
				       account_name, MSGID_SEPARATOR, INBOX_DIR,
				       MSGID_SEPARATOR, session->msg[session->cur_msg].recv_num);

	} else {
		file =
		    pop3_get_new_msg_file(session, session->ac_prefs->user_settings.
					  account_name, inbox_folder, &msg_num);
	}
	if (file == NULL) {
		g_free(inbox_folder);
		return POP3_FAIL;
	}
	if (pop3_recv_check_state() == FALSE) {
		g_free(file);
		g_free(inbox_folder);
		session->error_val = PS_IOERR;
		return POP3_FAIL;
	}
	if (pop3_save_msg(file, data, len) == POP3_FAIL) {
		g_free(file);
		g_free(inbox_folder);
		session->error_val = PS_IOERR;
		return POP3_FAIL;
	}


	if (session->retr_list != NULL) {
		msg_id = folder_get_msgid_from_path(file);
		msg_num = session->msg[session->cur_msg].recv_num;

	} else if (session->msgid == NULL) {
		msg_id = g_strdup_printf("%s%c%s%c%d",
					 session->ac_prefs->user_settings.
					 account_name, MSGID_SEPARATOR,
					 INBOX_DIR, MSGID_SEPARATOR, msg_num);
	} else {
		msg_id = g_strdup_printf("%s", session->msgid);
		msg_num = get_msgid_num(session->msgid);
	}
	drop_ok = pop3_get_msg_header(session, file, msg_num, msg_id);

	if (pop3_remove_header_file(file) == FALSE) {
		debug_print("header file not removed\n");
	}
	if (drop_ok < 0) {
		unlink(file);
		g_free(file);
		g_free(inbox_folder);
		if (session->error_val == PS_SUCCESS) {
			session->error_val = PS_IOERR;
		}
		return POP3_FAIL;
	}

	g_free(file);
	g_free(inbox_folder);

	session->cur_total_bytes += session->msg[session->cur_msg].size;
	session->cur_total_recv_bytes += session->msg[session->cur_msg].size;
	session->cur_total_num++;

	session->msg[session->cur_msg].received = TRUE;
	session->msg[session->cur_msg].recv_time =
	    drop_ok == 1 ? RECV_TIME_KEEP : session->current_time;

	session->msg[session->cur_msg].rx_msgid = msg_id;
	session->msg[session->cur_msg].recv_num = msg_num;
	return PS_SUCCESS;
}

/**
 This function will be called for sending DELE command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_delete_send(Pop3Session * session)
{
	session->state = POP3_DELETE;
	if (session->get_msg_num != 0) {
		pop3_gen_send(session, "DELE %d", session->get_msg_num);
	} else if (session->delete_msg_num != 0) {
		pop3_gen_send(session, "DELE %d", session->delete_msg_num);
	} else {
		pop3_gen_send(session, "DELE %d", session->cur_msg);
	}
	return PS_SUCCESS;
}

/**
 This function will be called for receiving response to DELE command.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_delete_recv(Pop3Session * session)
{
	gchar *msgid = NULL;

	/* update message list */
	if (session->msgid != NULL) {
		/* update the remote list */
		folder_delete_list_msg(session->msgid, TRUE);
	}

	if ((session->ac_prefs->receive_settings.leave_msg == TRUE)
	    || (session->retr_list != NULL)) {
		/* update the remote list */
		msgid = g_strdup_printf("%s%c%s%c%u",
					session->ac_prefs->user_settings.
					account_name, MSGID_SEPARATOR, INBOX_DIR,
					MSGID_SEPARATOR, session->msg[session->cur_msg].recv_num);

		folder_delete_list_msg(msgid, TRUE);
		g_free(msgid);
	}
	session->msg[session->cur_msg].deleted = TRUE;

	return PS_SUCCESS;
}

/**
 This function will be called for sending LOGOUT command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

gint pop3_logout_send(Pop3Session * session)
{
	session->state = POP3_LOGOUT;
	pop3_gen_send(session, "QUIT");
	return PS_SUCCESS;
}

/**
 This function will be called for sending TOP command to server.
 @param session pointer to the pop3 session.
 @return  success(0) always
*/

static gint pop3_top_send(Pop3Session * session)
{
	session->state = POP3_TOP;
	if (session->delete_msg_num != 0) {
		pop3_gen_send(session, POP3_TOP_CMD_FORMAT, session->delete_msg_num);
	} else {
		pop3_gen_send(session, POP3_TOP_CMD_FORMAT, session->cur_msg);
	}
	return PS_SUCCESS;
}

/**
 This function will be called on receiving response to RETR command.
 @param session pointer to the pop3 session.
 @param data pointer to the data received
 @param len length of the data received.
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) if data 
  could not be saved.
*/

static gint pop3_top_recv(Pop3Session * session, const gchar * data, guint len)
{
	gint drop_ok;
	gint msgnum = 0;

	if (pop3_recv_check_state() == FALSE) {
		debug_print("\n The session has been cancelled");
		if (session->delete_all == TRUE) {
			pop3_logout_send(session);
		}
		return POP3_FAIL;
	}
	/* if the operation is only to be deleted,check the msgid only */
	if ((session->delete_msg_num != 0) || (session->delete_all == TRUE)) {
		/* Get the header information */
		drop_ok = pop3_get_header_info(session, data, len, &msgnum);

		if (drop_ok < 0) {
			return POP3_FAIL;
		}
		return PS_SUCCESS;
	}
	/* Get the header information */
	drop_ok = pop3_get_header_info(session, data, len, &msgnum);

	if (drop_ok < 0) {
		return POP3_FAIL;
	}
	session->msg[session->cur_msg].recv_num = msgnum;
	session->msg[session->cur_msg].received = TRUE;

	session->cur_total_num++;

	return PS_SUCCESS;
}

/**
 This function will be called on receiving response to AUTH command.
 @param session pointer to the pop3 session.      
 @param data pointer to the data received
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) 
 if authentication could not proceed .
*/
static gint pop3_cram_md5_recv(Pop3Session * session, const gchar * data)
{
	gchar *response = NULL;
	gchar *response64 = NULL;
	guchar *challenge = NULL;
	gint challengelen;
	gchar hexdigest[33];

	if ((data[0] != '+') || (data[1] != ' ')) {
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}
	challenge = g_malloc0(strlen(data + 2) + 1);
	if (challenge == NULL) {
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}

	/* Get the challenge from the server response */
	challengelen = base64_decode(challenge, data + 2, -1);
	challenge[challengelen] = '\0';
	md5_hex_hmac(hexdigest, challenge, challengelen,
		     (const unsigned char *) session->passwd, strlen(session->passwd));
	g_free(challenge);

	/* Get the md5 digest of the challenge */
	response = g_strdup_printf("%s %s", session->ac_prefs->user_settings.userid, hexdigest);

	response64 = g_malloc0((strlen(response) + 3) * 2 + 1);
	if (response64 == NULL) {
		g_free(response);
		session->error_val = PS_AUTHFAIL;
		return POP3_FAIL;
	}
	base64_encode(response64, (const guchar *) response, strlen(response));
	g_free(response);
	/* send the response to the server */
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, response64);
	g_free(response64);
	session->state = POP3_GETAUTH_CRAM_MD5_RECV;
	return POP3_SUCCESS;
}

/**
 This function sends the commands to server,by writing to socket.
 @param session pointer to the pop3 session.
 @param format Comamnd format(syntax)
 @return  none
*/

static void pop3_gen_send(Pop3Session * session, const gchar * format, ...)
{
	gchar buf[POPBUFSIZE + 1];
	va_list args;

	va_start(args, format);
	g_vsnprintf(buf, sizeof(buf) - 2, format, args);
	va_end(args);

	if (!strncasecmp(buf, "PASS ", 5)) {
		log_message("POP3> PASS ********\n");
	} else {
		log_message("POP3> %s\n", buf);
	}

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
}

/**
 This function will be called for destroying pop3session
 @param session pointer to the pop3 session.
 @return  none
*/
static void pop3_session_destroy(Session * session)
{
	Pop3Session *pop3_session = POP3_SESSION(session);
	gint n = 0;

	g_return_if_fail(session != NULL);
	g_free(pop3_session->greeting);
	g_free(pop3_session->error_msg);
	g_free(pop3_session->msgid);
	g_free(pop3_session->uidl);
	for (n = 1; n <= pop3_session->count; n++) {
		g_free(pop3_session->msg[n].uidl);
	}
	g_free(pop3_session->msg);

	if (pop3_session->uidl_table) {
		hash_free_strings(pop3_session->uidl_table);
		g_hash_table_destroy(pop3_session->uidl_table);
	}
}

/**
 This function will be called for receiving message and writing to a file.
 @param session pointer to the pop3 session.
 @param data pointer to the response received
 @param len length of the data received.
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on file errors
*/

gint pop3_write_msg_to_file(FILE * fp, const gchar * data, guint len)
{
	const gchar *prev = NULL, *cur = NULL;


	/* +------------------+----------------+--------------------------+ *
	 * ^data              ^prev            ^cur             data+len-1^ */

	prev = data;
	while ((cur = memchr(prev, '\r', len - (prev - data))) != NULL) {
		if ((cur > prev && fwrite(prev, cur - prev, 1, fp) < 1)
		    || fputc('\n', fp) == EOF) {
			return POP3_FAIL;
		}

		if (cur == data + len - 1) {
			prev = cur + 1;
			break;
		}

		if (*(cur + 1) == '\n') {
			prev = cur + 2;
		} else {
			prev = cur + 1;
		}

		if (prev - data < len - 1 && *prev == '.' && *(prev + 1) == '.') {
			prev++;
		}

		if (prev - data >= len) {
			break;
		}
	}

	if (prev - data < len && fwrite(prev, len - (prev - data), 1, fp) < 1) {
		return POP3_FAIL;
	}
	if (data[len - 1] != '\r' && data[len - 1] != '\n') {
		if (fputc('\n', fp) == EOF) {
			return POP3_FAIL;
		}
	}
	return POP3_SUCCESS;
}

/**
 This function is called for moving to the next state in the state machine
 @param Session pointer to the pop3 session.
 @return  next pop3 session state
*/

static Pop3State pop3_lookup_next(Pop3Session * session)
{
	Pop3MsgInfo *msg = NULL;
	PrefsAccount *ac = session->ac_prefs;
	gint size = 0;
	gboolean size_limit_over = FALSE;
	gboolean allow_size_check = TRUE;


	/* do not check size limit, for the headers */
	if ((session->ac_prefs->receive_settings.recv_type == HEADERS) ||
	    (CHECK_CB_FLAG == FALSE) || (session->msgid != NULL)
	    || (session->retr_list != NULL)) {
		allow_size_check = FALSE;
	}
	for (;;) {
		msg = &session->msg[session->cur_msg];
		size = msg->size;
		size_limit_over =
		    ((ac->receive_settings.enable_size_limit) &&
		     (allow_size_check) &&
		     (ac->receive_settings.size_limit > 0) &&
		     (size > ac->receive_settings.size_limit * 1024));

		if (size_limit_over == TRUE) {

			pop3_handle_sizelimit_error(session);

			if (session->msgid == NULL) {
				/* Download only header since message exceeds the limit */
				pop3_top_send(session);
				return POP3_TOP;
			}

			log_message("POP3: Skipping message %d (%d bytes)\n",
				    session->cur_msg, size);
		}
		if (session->retr_list != NULL) {
			if (session->msg[session->cur_msg].skip_download == FALSE) {
				break;
			}

		} else if ((size != 0) && (msg->received == FALSE)
			   && (size_limit_over == FALSE)) {
			break;

		}
		session->cur_total_bytes += size;

		if ((session->cur_msg == session->count)
		    || (session->msgid != NULL)) {
			pop3_logout_send(session);
			return POP3_LOGOUT;
		} else {
			session->cur_msg++;
		}
	}
	if ((session->ac_prefs->receive_settings.recv_type == HEADERS) &&
	    (session->msgid == NULL) && (session->retr_list == NULL)) {
		session->prog_cur_msg++;
		pop3_top_send(session);
		return POP3_TOP;
	} else if (session->delete_all == TRUE) {
		pop3_top_send(session);
		return POP3_TOP;
	} else {
		session->prog_cur_msg++;
		pop3_retr_send(session);
		return POP3_RETR;
	}
}

/**
 This function will be called to process the response received
 @param msg pointer to the response received
 @return  enum value 0 if response is OK or POP3ErrorValue,
 to indicate the type of error in reponse 
*/

static Pop3ErrorValue pop3_ok(Pop3Session * session, const gchar * msg)
{
	Pop3ErrorValue ok;

	log_message("POP3< %s\n", msg);

	if (!strncmp(msg, "+OK", 3)) {
		ok = PS_SUCCESS;
	} else if (!strncmp(msg, "-ERR", 4)) {
		if (strstr(msg + 4, "lock") || strstr(msg + 4, "Lock") || strstr(msg + 4, "LOCK")
		    || strstr(msg + 4, "wait")) {
			log_message("mailbox is locked\n");
			debug_print("mailbox is locked\n");
			ok = PS_LOCKBUSY;
		} else if (strcasestr(msg + 4, "timeout")) {
			log_message("session timeout\n");
			debug_print("session timeout\n");
			ok = PS_ERROR;
		} else {

			switch (session->state) {
#if USE_SSL
			case POP3_STLS:
				log_message("can't start TLS session\n");
				ok = PS_NOTSUPPORTED;
				break;
#endif
			case POP3_GETAUTH_USER:
			case POP3_GETAUTH_CRAM_MD5:
				log_message("error occurred on authentication\n");
				debug_print("error occurred on authentication\n");
				ok = PS_AUTHFAIL;
				break;
			case POP3_GETAUTH_PASS:
			case POP3_GETAUTH_APOP:
			case POP3_GETAUTH_CRAM_MD5_RECV:
				if (session->ac_prefs->user_settings.passwd == NULL) {
					ok = pop3_connect_tmp_passwd(session);
				} else {
					ok = PS_AUTHFAIL;
				}
				break;

			case POP3_TOP:
				log_message("TOP command not supported\n");
				if (session->msgid != NULL) {
					ok = PS_DELERR;
				} else {
					debug_print("\nTOP command not supported\n");
					ok = PS_NOTSUPPORTED;
				}
				break;
			case POP3_RETR:
				if (session->msgid != NULL) {
					ok = PS_DELERR;
				} else {
					ok = PS_ERROR;
				}
				break;
			default:
				log_message("error occurred on POP3 session\n");
				debug_print("error occurred on Pop3 session\n");
				ok = PS_ERROR;
			}
		}

		g_free(session->error_msg);
		session->error_msg = g_strdup(msg);
		debug_print("POP3: %s\n", msg);
	} else {
		ok = PS_PROTOCOL;
	}
	session->error_val = ok;
	return ok;
}

/**
 This callback function will be called for receiving response from socket.
 @param Session pointer to the pop3 session.
 @param msg pointer to the response received
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on errors
*/

static gint pop3_session_recv_msg(Session * session, const gchar * msg)
{
	Pop3Session *pop3_session = POP3_SESSION(session);
	Pop3ErrorValue val = PS_SUCCESS;
	const gchar *body = NULL;
	gint ok = POP3_FAIL;

	if ((msg == NULL) || (session == NULL)) {
		return POP3_FAIL;
	}
	body = msg;
	if ((pop3_session->state != POP3_GETSIZE_LIST_RECV) &&
	    (pop3_session->state != POP3_GETAUTH_CRAM_MD5)) {
		val = pop3_ok(pop3_session, msg);
		if (val != PS_SUCCESS) {
			if ((val != PS_NOTSUPPORTED) && (val != PS_PASSWDFAIL) &&
			    (val != PS_PASSWD_CANCEL)) {
				pop3_session->state = POP3_ERROR;
				return POP3_FAIL;
			}
		}

		if (*body == '+' || *body == '-') {
			body++;
		}
		while (isalpha(*body)) {
			body++;
		}
		while (isspace(*body)) {
			body++;
		}
	}
	switch (pop3_session->state) {

	case POP3_READY:
	case POP3_GREETING:
		pop3_greeting_recv(pop3_session, body);

		pop3_session->passwd = pop3_session->ac_prefs->user_settings.passwd;
		if (pop3_session->ac_prefs->user_settings.passwd == NULL) {
			if ((ok = pop3_connect_tmp_passwd(pop3_session))
			    == PS_PASSWD_CANCEL) {
				pop3_logout_send(pop3_session);
				return POP3_SUCCESS;
			}
		}
#if USE_SSL
		if ((pop3_session->ac_prefs->advanced_settings.use_ssl == TRUE)
		    && (SESSION(pop3_session)->ssl_type == SSL_STARTTLS)) {
			pop3_stls_send(pop3_session);
		} else {
#endif
			if (pop3_send_auth(pop3_session) == POP3_FAIL) {
				return POP3_FAIL;
			}
#if USE_SSL
		}
#endif
		break;

#if USE_SSL
	case POP3_STLS:
		if (val == PS_NOTSUPPORTED) {
			return POP3_FAIL;
		}
		if (pop3_stls_recv(pop3_session) != PS_SUCCESS) {
			return POP3_FAIL;
		}

		if (pop3_session->ac_prefs->receive_settings.passwd_auth_type == AUTH_APOP) {
			if (pop3_getauth_apop_send(pop3_session) < 0) {
				return POP3_FAIL;
			}

		} else if (pop3_session->ac_prefs->receive_settings.
			   passwd_auth_type == AUTH_CRAM_MD5) {
			pop3_getauth_cram_md5_send(pop3_session);
		} else {
			pop3_getauth_user_send(pop3_session);
		}
		break;
#endif
	case POP3_GETAUTH_USER:
		pop3_getauth_pass_send(pop3_session);
		break;

	case POP3_GETAUTH_PASS:
	case POP3_GETAUTH_APOP:

		if (val == PS_NOTSUPPORTED) {
			return POP3_FAIL;
		}
		if (val == PS_PASSWDFAIL) {
			if (pop3_send_auth(pop3_session) == POP3_FAIL) {
				return POP3_FAIL;
			}
			return POP3_SUCCESS;
		} else if (val == PS_PASSWD_CANCEL) {
			pop3_logout_send(pop3_session);
			return POP3_SUCCESS;
		}
		if (pop3_session->msgid != NULL) {
			if (pop3_session->get_msg_num != 0) {
				POP3_SET_MSGBODY_SRV_CB(POP3_MSG_CONNECT_SUCCESS,
							pop3_session->ac_prefs,
							pop3_session->msgid);
			}

		} else {
			POP3_SET_MSG_SRV_CB(POP3_MSG_CONNECT_SUCCESS, pop3_session->ac_prefs);
		}
		pop3_free_old_list();
		pop3_getrange_stat_send(pop3_session);
		break;

	case POP3_GETAUTH_CRAM_MD5:
		if (pop3_cram_md5_recv(pop3_session, msg) < 0) {
			return POP3_FAIL;
		}
		break;
	case POP3_GETAUTH_CRAM_MD5_RECV:
		if (pop3_session->msgid != NULL) {
			/* Inform the higher layer that server is up */
			POP3_SET_MSGBODY_SRV_CB(POP3_MSG_CONNECT_SUCCESS,
						pop3_session->ac_prefs, pop3_session->msgid);
			pop3_getrange_uidl_send(pop3_session);

		} else {
			/* Inform the higher layer that server is up */
			POP3_SET_MSG_SRV_CB(POP3_MSG_CONNECT_SUCCESS, pop3_session->ac_prefs);
			pop3_getrange_stat_send(pop3_session);
		}
		break;
	case POP3_GETRANGE_STAT:
		if (pop3_getrange_stat_recv(pop3_session, body) < 0) {
			return POP3_FAIL;
		}
		if (pop3_session->count <= 0) {
			pop3_logout_send(pop3_session);
		} else {
			pop3_getrange_uidl_send(pop3_session);
		}
		break;

	case POP3_GETSIZE_LIST:
		pop3_session->state = POP3_GETSIZE_LIST_RECV;
		session_recv_data(session, 0, ".\r\n");
		break;
	case POP3_GETRANGE_UIDL:
		if (val == PS_NOTSUPPORTED) {
			pop3_session->error_val = PS_SUCCESS;
			pop3_getsize_list_send(pop3_session);
		} else {
			pop3_session->state = POP3_GETRANGE_UIDL_RECV;
			session_recv_data(session, 0, ".\r\n");
		}
		break;

	case POP3_RETR:
		pop3_session->state = POP3_RETR_RECV;
		session_recv_data(session, 0, ".\r\n");
		break;

	case POP3_TOP:
		if (val == PS_NOTSUPPORTED) {
			pop3_logout_send(pop3_session);
		} else {
			pop3_session->state = POP3_TOP_RECV;
			session_recv_data(session, 0, ".\r\n");
		}
		break;

	case POP3_DELETE:
		pop3_delete_recv(pop3_session);
		if ((pop3_session->cur_msg == pop3_session->count)
		    || (pop3_session->msgid != NULL)) {
			pop3_logout_send(pop3_session);
		} else {
			pop3_session->cur_msg++;
			if (pop3_lookup_next(pop3_session) == POP3_ERROR) {
				return POP3_FAIL;
			}
		}
		break;
	case POP3_LOGOUT:
		session_disconnect(session);
		break;
	case POP3_ERROR:
	default:
		return POP3_FAIL;
	}

	return POP3_SUCCESS;
}

/**
 This is callback function called by socket,when receiving multiline 
 response is completed.
 @param Session pointer to the pop3 session.
 @param data pointer to the response received
 @param len length of data received
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on errors
*/

static gint pop3_session_recv_data_finished(Session * session, guchar * data, guint len)
{
	Pop3Session *pop3_session = POP3_SESSION(session);
	Pop3ErrorValue val = PS_SUCCESS;

	if ((session == NULL) || (data == NULL) || (!len)) {
		return POP3_FAIL;
	}
	switch (pop3_session->state) {

	case POP3_GETRANGE_UIDL_RECV:
		val = pop3_getrange_uidl_recv(pop3_session, data, len);

		if (val == PS_SUCCESS) {
			if (pop3_session->get_msg_num != 0) {
				pop3_retr_send(pop3_session);
			} else if (pop3_session->delete_msg_num != 0) {
				pop3_top_send(pop3_session);
			} else if ((pop3_session->new_msg_exist) ||
				   (pop3_session->retr_list != NULL)) {
				pop3_getsize_list_send(pop3_session);
			} else {
				pop3_logout_send(pop3_session);
			}
		} else {
			return POP3_FAIL;
		}
		break;
	case POP3_GETSIZE_LIST_RECV:
		val = pop3_getsize_list_recv(pop3_session, (const gchar *) data, len);
		if (val == PS_SUCCESS) {
			if (pop3_lookup_next(pop3_session) == POP3_ERROR) {
				return POP3_FAIL;
			}
		} else {
			return POP3_FAIL;
		}
		break;

	case POP3_RETR_RECV:
		if (pop3_retr_recv(pop3_session, (const gchar *) data, len) < 0) {
			return POP3_FAIL;
		}
		/* delete the messages on check */
		if ((pop3_session->ac_prefs->receive_settings.leave_msg == FALSE)
		    || (pop3_session->operation == FOLDER_OP_MOVE)) {
			pop3_delete_send(pop3_session);
		} else if ((pop3_session->cur_msg == pop3_session->count) ||
			   (pop3_session->msgid != NULL)) {
			pop3_logout_send(pop3_session);

		} else {
			pop3_session->cur_msg++;
			if (pop3_lookup_next(pop3_session) == POP3_ERROR) {
				return POP3_FAIL;
			}
		}
		break;

	case POP3_TOP_RECV:
		if (pop3_top_recv(pop3_session, (const gchar *) data, len) < 0) {
			return POP3_FAIL;
		}
		if ((pop3_session->delete_msg_num != 0)
		    || (pop3_session->delete_all == TRUE)) {
			pop3_delete_send(pop3_session);
			break;
		}
		if (pop3_session->cur_msg == pop3_session->count) {
			pop3_logout_send(pop3_session);
		} else {
			pop3_session->cur_msg++;
			if (pop3_lookup_next(pop3_session) == POP3_ERROR) {
				return POP3_FAIL;
			}
		}
		break;

	case POP3_ERROR:

	default:
		return POP3_FAIL;
	}

	return POP3_SUCCESS;
}

/**
 This function handles sizelimit error,sends error callback to UI.   
 @param session pointer to the pop3 session 
 @return nothing 
*/
static void pop3_handle_sizelimit_error(Pop3Session * session)
{
	if ((session->msgid == NULL)
	    && (session->size_error != PS_SIZELMTERR)) {

		POP3_SET_MSG_CB(POP3_MSG_SIZELIMIT_ERROR);
		session->size_error = PS_SIZELMTERR;
	}

}

/**
 This function saves message to file   
 @param file the name file to save data 
 @param data pointer to the data received
 @param len length of the data received.
 @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) if data 
    could not be saved.
*/
static gint pop3_save_msg(const gchar * file, const gchar * data, guint len)
{
	FILE *fp = NULL;

	if ((fp = fopen(file, "wb")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return POP3_FAIL;
	}

	if (change_file_mode_rw(fp, file) < 0) {
		FILE_OP_ERROR(file, "chmod");
	}

	if (pop3_write_msg_to_file(fp, data, len) < 0) {
		fclose(fp);
		unlink(file);
		return POP3_FAIL;
	} else if (fclose(fp) == EOF) {
		unlink(file);
		return POP3_FAIL;
	}
	return POP3_SUCCESS;
}

/**
 This function handles authentication   
 @param pop3_session pointer to the pop3 session 
 @return failure or success 
*/
static gint pop3_send_auth(Pop3Session * pop3_session)
{

	if (pop3_session->ac_prefs->receive_settings.passwd_auth_type == AUTH_APOP) {
		if (pop3_getauth_apop_send(pop3_session) < 0) {
			return POP3_FAIL;
		}
	} else if (pop3_session->ac_prefs->receive_settings.passwd_auth_type == AUTH_CRAM_MD5) {
		pop3_getauth_cram_md5_send(pop3_session);
	} else {
		pop3_getauth_user_send(pop3_session);
	}
	return POP3_SUCCESS;
}

static gint pop3_connect_tmp_passwd(Pop3Session * pop3_session)
{
	gboolean tmp_passwd_used = FALSE;
	gint msg_number = 0;
	gint ui_closed = 0;

	if (pop3_session->ac_prefs->user_settings.scheduled_op == TRUE) {
		return PS_AUTHFAIL;
	}
	if (pop3_session->passwd != NULL) {
		tmp_passwd_used = TRUE;
	}
	g_free(pop3_session->ac_prefs->user_settings.tmp_passwd);
	pop3_session->ac_prefs->user_settings.tmp_passwd = NULL;
	pop3_session->ac_prefs->user_settings.cancel = FALSE;
	if (gl_pop3_callback_ptr) {
		if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
			/* Inform the higher layer */
			if (tmp_passwd_used == TRUE) {
				msg_number = POP3_PASSWD_WRONG;
			} else {
				msg_number = POP3_MSG_PASSWD_FAIL;

			}
			gl_pop3_callback_ptr->pop3_set_msg_cb(msg_number, "%s",
							      pop3_session->ac_prefs->
							      user_settings.account_name);

		}
	}
	while (1) {
		(void) g_main_context_iteration(NULL, FALSE);

		if (pop3_session->ac_prefs->user_settings.tmp_passwd != NULL) {
			pop3_session->passwd = pop3_session->ac_prefs->user_settings.tmp_passwd;
			pop3_session->ac_prefs->user_settings.tmp_passwd = NULL;
			break;
		}
		if (pop3_session->ac_prefs->user_settings.cancel == TRUE) {
			return PS_PASSWD_CANCEL;
		}
		if(osso_email_gconf_get_int(EMAIL_GCONF_UI_CLOSED,&ui_closed)) {
			if(ui_closed) {
				break;
			}
	        } else if (pop3_recv_check_state() == FALSE) {
	                break;
	        }
	}
	return PS_PASSWDFAIL;
}
static GHashTable *pop3_get_uidl_table(PrefsAccount * ac_prefs,
				       const gchar * msgid, gchar ** uidl_ptr)
{
	GHashTable *table = NULL;
	gchar *path;
	FILE *fp;
	gchar buf[POPBUFSIZE];
	gchar uidl[POPBUFSIZE];
	time_t now;
	gint msgnum = 0;
	gint check_num = 0;
	gboolean check_for_entry = FALSE;

	table = g_hash_table_new(g_str_hash, g_str_equal);

	path = g_strdup_printf("%s%c%s%s", get_rc_dir(), G_DIR_SEPARATOR,
			       ac_prefs->user_settings.account_name, UID_EXT);
	if (msgid != NULL) {
		check_num = get_msgid_num(msgid);
		check_for_entry = TRUE;
	}

	if ((fp = fopen(path, "rb")) == NULL) {
		if (ENOENT != errno)
			FILE_OP_ERROR(path, "fopen");
		g_free(path);
		return table;
	}
	g_free(path);

	now = time(NULL);

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		strretchomp(buf);
		msgnum = 0;
		if (sscanf(buf, "%s\t%d", uidl, &msgnum) != 2) {
			if (sscanf(buf, "%s", uidl) != 1)
				continue;
		}
		if ((check_for_entry == TRUE) && (msgnum == check_num)) {
			*uidl_ptr = g_strdup(uidl);
		}
		g_hash_table_insert(table, g_strdup(uidl), GINT_TO_POINTER(msgnum));


	}

	fclose(fp);
	return table;
}

gint pop3_write_uidl_list(Pop3Session * session)
{
	gchar *path;
	FILE *fp;
	Pop3MsgInfo *msg;
	gint n;

	if (!session->uidl_is_valid)
		return 0;

	path = g_strdup_printf("%s%c%s%s", get_rc_dir(), G_DIR_SEPARATOR,
			       session->ac_prefs->user_settings.account_name, UID_EXT);

	if ((fp = fopen(path, "w")) == NULL) {
		FILE_OP_ERROR(path, "fopen");
		g_free(path);
		return -1;
	}

	for (n = 1; n <= session->count; n++) {
		msg = &session->msg[n];
		if (msg->uidl && msg->received && !msg->deleted)
			fprintf(fp, "%s\t%d\n", msg->uidl, msg->recv_num);
	}

	if (fclose(fp) == EOF)
		FILE_OP_ERROR(path, "fclose");
	g_free(path);

	return 0;
}

static gint pop3_getrange_uidl_send(Pop3Session * session)
{
	session->state = POP3_GETRANGE_UIDL;
	pop3_gen_send(session, "UIDL");
	return PS_SUCCESS;
}
static gint pop3_getrange_uidl_recv(Pop3Session * session, const gchar * data, guint len)
{
	gchar id[IDLEN + 1];
	gchar buf[BUFFSIZE];
	gint buf_len;
	gint num;
	gint recv_num;
	const gchar *p = data;
	const gchar *lastp = data + len;
	const gchar *newline;
	gint ret = PS_SUCCESS;

	while (p < lastp) {
		if ((newline = memchr(p, '\r', lastp - p)) == NULL)
			return -1;
		buf_len = MIN(newline - p, sizeof(buf) - 1);
		memcpy(buf, p, buf_len);
		buf[buf_len] = '\0';

		p = newline + 1;
		if (p < lastp && *p == '\n')
			p++;

		if (sscanf(buf, "%d %" Xstr(IDLEN) "s", &num, id) != 2)
			return -1;

		if (num <= 0 || num > session->count) {
			return -1;
		}

		session->msg[num].uidl = g_strdup(id);

		recv_num = (gint) g_hash_table_lookup(session->uidl_table, id);

		session->msg[num].recv_num = recv_num;

		if (recv_num != RECV_MSG_NONE) {
			session->msg[num].received = TRUE;
		} else {
			/* number of message to be retrieved */
			session->prog_count++;

		}


		if ((session->new_msg_exist == FALSE) && (recv_num == RECV_MSG_NONE)
		    && (session->retr_list == NULL)) {

			session->cur_msg = num;
			session->new_msg_exist = TRUE;
		}
	}
	if (session->msgid != NULL) {
		/* get the correct message number on server */
		ret = pop3_sync_msg(session);
	}
	if (session->retr_list != NULL) {
		/* get the correct message number on server */
		ret = pop3_sync_msg_list(session);
	}

	session->uidl_is_valid = TRUE;
	return ret;
}

static gint pop3_sync_msg(Pop3Session * session)
{
	gint num = 0;

	for (num = 1; num < session->count + 1; num++) {

		if (strcmp(session->msg[num].uidl, session->uidl) == 0) {
			break;
		}
	}
	if (num >= session->count + 1) {
		/* uidl does not exist */
		session->error_val = PS_DELERR;
		return POP3_FAIL;
	}
	if (session->get_msg_num != 0) {
		session->get_msg_num = num;

	} else if (session->delete_msg_num != 0) {
		session->delete_msg_num = num;

	}

	session->cur_msg = num;

	return PS_SUCCESS;
}

gboolean pop3_hash_entry_removed(gpointer key, gpointer value, gpointer user_data)
{
	gint n = 0;
	Pop3MsgInfo *msg = NULL;
	Pop3Session *session = (Pop3Session *) user_data;
	gint msg_num = (gint) value;
	gchar *file_path = NULL;
	gchar *msgid = NULL;

	if (msg_num == RECV_MSG_NONE) {
		/* This message is not downloaded */
		return FALSE;
	}

	if (pop3_get_header_file_path(msg_num,
				      session->ac_prefs->user_settings.
				      account_name, &file_path) == FALSE) {
		g_free(file_path);
		/* This is fully downloaded message,so no need to delete */
		return FALSE;
	}

	for (n = 1; n <= session->count; n++) {
		msg = &session->msg[n];
		/* This message is not yet received */
		if (msg->recv_num == RECV_MSG_NONE) {
			continue;
		}
		if (msg->uidl && msg->received && !msg->deleted) {
			if (strcmp(msg->uidl, key) == 0) {
				/* This uidl exists on server,dont delete */
				g_free(file_path);
				return FALSE;

			}
		}
	}
	/* delete the message file */
	if (unlink(file_path) != 0) {
		g_warning("Unable to delete file:%s", file_path);
	}
	g_free(file_path);

	/* free the entry from the table */
	g_free(key);

	/* this entry does not exist on server.So remove it from list */
	msgid = g_strdup_printf("%s%c%s%c%u",
				session->ac_prefs->user_settings.account_name,
				MSGID_SEPARATOR, INBOX_DIR, MSGID_SEPARATOR, msg_num);
	gl_pop3_delete_list = g_slist_append(gl_pop3_delete_list, msgid);
	folder_delete_list_msg(msgid, TRUE);
	folder_message_del_list(msgid, UNREAD);
	return TRUE;
}

static gboolean
pop3_get_header_file_path(gint msg_num, gchar * account_name, gchar ** file_path_ptr)
{
	gchar *file_path = NULL;
	gint ret_val = 0;


	file_path = g_strdup_printf("%s%c%s%c%s%c%s%c%u%c%s",
				    get_mail_dir(),
				    G_DIR_SEPARATOR, INBOX_DIR, G_DIR_SEPARATOR,
				    account_name, MSGID_SEPARATOR, INBOX_DIR,
				    MSGID_SEPARATOR, msg_num, DOT, MSG_HDR);

	if (is_file_entry_exist(file_path) == TRUE) {
		ret_val = TRUE;
	}
	*file_path_ptr = file_path;
	return ret_val;
}

void pop3_update_headers(Pop3Session * pop3_session)
{
	/* check &remove locally the headers, 
	 * that were deleted on server */

	if (!pop3_session->uidl_is_valid)
		return;

	if (pop3_session->msgid == NULL) {

		g_hash_table_foreach_remove(pop3_session->uidl_table,
					    pop3_hash_entry_removed, pop3_session);

		/* Indicate through callback the headers to be removed */
		if (POP3_HAS_CALLBACK) {
			if ((gl_pop3_callback_ptr->pop3_delete_list_cb != NULL) &&
			    (gl_pop3_delete_list != NULL)) {
				gl_pop3_callback_ptr->pop3_delete_list_cb(gl_pop3_delete_list);
			}
		}
		slist_free_strings(gl_pop3_delete_list);
		g_slist_free(gl_pop3_delete_list);
		gl_pop3_delete_list = NULL;
	}
}
static gint pop3_sync_msg_list(Pop3Session * session)
{
	gint num = 0;
	GSList *cur = NULL;
	gint msg_num = 0;
	RetrMsgData *retr_msg = NULL;
	gchar *file_path = NULL;


	for (num = 1; num < session->count + 1; num++) {
		session->msg[num].skip_download = TRUE;
	}
	/* check update for the list of mesasges to download */
	for (cur = session->retr_list; cur != NULL; cur = cur->next) {

		if (cur->data == NULL) {
			continue;
		}
		retr_msg = (RetrMsgData *) cur->data;

		if (retr_msg->msgid == NULL) {
			continue;
		}
		if (check_msgid(retr_msg->msgid) == FALSE) {
			continue;
		}
		msg_num = get_msgid_num(retr_msg->msgid);

		for (num = 1; num < session->count + 1; num++) {

			if (msg_num == session->msg[num].recv_num) {
				session->msg[num].skip_download = FALSE;
				break;
			}

		}
		if (num >= session->count + 1) {
			/* uidl does not exist */
			POP3_MSG_DELETE_ERROR_CB(retr_msg->msgid);
			gl_unavailable_list = g_slist_append(gl_unavailable_list,
							     g_strdup(retr_msg->msgid));
			file_path = folder_get_message_file_path(retr_msg->msgid);
			if (is_file_entry_exist(file_path) == TRUE) {
				if (unlink(file_path) != 0) {
					g_warning("Unable to delete file:%s", file_path);
				}

			} else {
				pop3_remove_header_file(file_path);

			}
			folder_delete_list_msg(retr_msg->msgid, TRUE);
			folder_message_del_list(retr_msg->msgid, UNREAD);
			g_free(file_path);

		}
	}

	return PS_SUCCESS;
}

