
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <glib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "config.h"
#include "socket.h"
#include "session.h"
#include "common.h"
#include "defs.h"
#include "prefs_account.h"
#include "account.h"
#include "utils.h"
#include "pop3_msgs.h"
#include "pop_recv.h"
#include "procmsg.h"
#include "procheader.h"
#include "folder.h"
#include "pop.h"
#include "ssl.h"
#include "engine_errno.h"
#include "smime_security.h"
#include <al_support.h>
#include "email_gconf.h"
#include "notify.h"

#ifdef CL_USE
#include "cl.h"
#endif

gint gl_cur_count = 0;
gint gl_total_count = 0;

/*@null@*/ GSList *gl_unavailable_list = NULL;

/*@null@*/ static Pop3RecvSession *pop3_recv_session_new(PrefsAccount *
							 account, const gchar * msgid);

static Pop3RecvState pop3_recv_session_do(Pop3RecvSession * session);

static void pop3_recv_start(Pop3RecvSession * session);

static gint pop3_recv_data_progressive(Session * session, guint cur_len,
				       guint total_len, gpointer data);

/*@null@*/ static MsgInfo *pop3_parse_msg(const gchar * file,
					  gint msg_num);

static void pop3_recv_session_destroy(Pop3RecvSession * session);

static void pop3_recv_put_error(Pop3RecvState istate, Pop3Session * pop3_session);

static gint pop3_recv_message(Session * session);

static gint pop3_recv_data_finished(Session * session, guint len, gpointer data);

static gboolean check_session_msgid(const gchar * msgid, gchar * retr_msgid);

static gint pop3_get_port(Pop3Session * pop3_session);

static gint pop3_handle_connect_error(Pop3RecvSession * session);

static gint pop3_set_msginfo(MsgInfo * msginfo, Pop3Session * session, const gchar * msgid);


static gint pop3_send_mimeinfo(const gchar * file, MsgHeader * msgheader, const gchar * msgid);

static MsgInfo *pop3_get_msginfo(const gchar * data, guint len, const gchar * path);

/*Function to send smime info*/
static gint pop3_send_smime_mimeinfo(const gchar * file,
				     MsgHeader * msgheader,
				     const gchar * msgid, gboolean saved,
				     gint recip_cert_state,
				     gint private_cert_state,
				     gboolean signature, gboolean trust,
				     gchar * cert_path, gint smime,
				     gint sender_cert_id, gint recip_cert_id);

static void pop3_recv_update_mail_status(guint total, guint cur);

/* globals */
/*@null@*/ Pop3Callback *gl_pop3_callback_ptr = NULL;

/*@null@*/ static Pop3RecvSession *gl_pop3_cur_session = NULL;
/*@null@*/ static GSList *gl_msg_list = NULL;
static gint recv_msg_count = 0;

/*@null@*/ static gboolean gl_pop3_queued_delete = FALSE;

/** Function to send pop message information to email UI
    @return - None
*/

static void pop3_recv_msg_headers_to_view(void)
{
	if (gl_pop3_callback_ptr->pop3_recv_list_cb) {
		if(gl_msg_list != NULL) {
			recv_msg_count += g_slist_length(gl_msg_list);
			gl_pop3_callback_ptr->pop3_recv_list_cb(gl_msg_list);
                        notify_mail_list(gl_msg_list);
			while (gl_msg_list) {
				procmsg_msginfo_free((MsgInfo *) gl_msg_list->data);
				gl_msg_list = g_slist_remove(gl_msg_list, gl_msg_list->data);
			}
			g_slist_free(gl_msg_list);
			gl_msg_list = NULL;
		}
	}
}

/** Function to update mail received status.
    @param total number of messages to be received. 
    @param total number of messages already received. 
    @return - None
*/
static void pop3_recv_update_mail_status(guint total, guint cur)
{
	osso_email_gconf_set_int(EMAIL_RECV_MAILS_PROGRESS_TOTAL, total);
	osso_email_gconf_set_int(EMAIL_RECV_MAILS_PROGRESS_CUR, cur);
}

/**
  This function will be called to register the callbacks.
  @param cb_ptr    pointer to the callback
  @return TRUE if successfully registered.FALSE otherwise
 */

gboolean pop3_init(Pop3Callback * cb_ptr)
{
	if (cb_ptr != NULL) {

		gl_pop3_callback_ptr = cb_ptr;

		return TRUE;
	} else {
		return FALSE;
	}

}

/**
  This function will be called by adaptation layer to start receiving mails.
  @param account    pointer to the account structure,in the global 
  account list.Global account list is maintained by the account management.
  @param msgid    message id of the message to be used 
  @param delete_flag    indicates whether message has to be deleted 
  @param queue_flag    indicates whether message to be deleted is queued 
  @param retr_list    list of messages o be retrieved from server 
  @return none
 */

void
pop3_recv_mail(PrefsAccount * account, const gchar * msgid,
	       FolderOpernType operation, gboolean queue_flag, GSList * retr_list)
{
	Pop3RecvSession *session = NULL;
	Pop3Session *pop3_session = NULL;
	gint index_msgnum = 0;
	gboolean delete_flag = FALSE;
	PrefsAccount *new_account = NULL;

	if ((account->receive_settings.recv_type != HEADERS) &&
	    (account->receive_settings.recv_type != BODY_ATTCH)) {

		POP3_CB_PARAM_MISSING(RECV_TYPE_MISSING);
		return;
	}

	if (gl_unavailable_list != NULL) {
		g_slist_foreach(gl_unavailable_list, (GFunc) g_free, NULL);
		gl_unavailable_list = NULL;
	}

	if (operation == FOLDER_OP_DEL) {
		delete_flag = TRUE;
	}

	if ((queue_flag == TRUE) && (delete_flag == FALSE)) {

		POP3_SET_MSG_CB(POP3_MSG_ID_ERROR);
		return;
	}

	if (msgid == NULL) {
		/* dont show progress bar at UI bottom for delete operation */
		if ((queue_flag == FALSE) && (delete_flag == FALSE)) {
			POP3_SET_MSG_SRV_CB(POP3_MSG_START, account);
		}
	} else {
		POP3_SET_MSGBODY_SRV_CB(POP3_MSG_START, account, msgid);
	}

	session = pop3_recv_session_new(account, msgid);

	/* There is error in creating the session */
	if (session == NULL) {
		POP3_SET_MSG_SRV_CB(POP3_MSG_COMPLETE, account);
		return;
	}

	gl_pop3_cur_session = session;

	pop3_session = POP3_SESSION(session->session);

	if (msgid != NULL) {
		gl_pop3_queued_delete = FALSE;
		index_msgnum = get_msgid_num(msgid);
		if (delete_flag == TRUE) {
			pop3_session->delete_msg_num = index_msgnum;
			if (queue_flag == TRUE) {
				gl_pop3_queued_delete = TRUE;
			}
		} else {
			pop3_session->get_msg_num = index_msgnum;
		}
		pop3_session->msgid = g_strdup(msgid);
		pop3_session->operation = operation;

	} else if (retr_list != NULL) {
		pop3_session->retr_list = retr_list;
		if (delete_flag == TRUE) {
			pop3_session->delete_all = TRUE;
		}
		if (queue_flag == TRUE) {
			pop3_session->queue_flag = TRUE;
		}
	}

	/* set the lock to indicate usage of this account */
	ACCNT_SET_FLAGS(account, ACCNT_RX);

	pop3_recv_start(session);

	/* reset the status flag for account.Account is now editable */
	ACCNT_RESET_FLAGS(account, ACCNT_RX);


	/* Indicate through callback the list of message headers */
	if (POP3_HAS_CALLBACK) {
		if (gl_pop3_callback_ptr->pop3_recv_list_cb != NULL) {
			pop3_recv_msg_headers_to_view();
		}

		/*If we have received even one mail editing login name and 
		 * IP address should be disabled for the account
		 */
		if ((recv_msg_count > 0) &&
		    (account->receive_settings.is_not_editable == FALSE)) {
			account->receive_settings.is_not_editable = TRUE;
			new_account = account_copy_prefs(account);
			if (account_set_settings
			    (new_account->user_settings.account_name, RECEIVE_SETTING,
			     new_account) == FALSE) {
				debug_print("Unable to set account settings\n");
			}
			account_free_account(new_account);
		}
	}
	recv_msg_count = 0;
	if(gl_msg_list) {
		while (gl_msg_list) {
			procmsg_msginfo_free((MsgInfo *) gl_msg_list->data);
			gl_msg_list = g_slist_remove(gl_msg_list, gl_msg_list->data);
		}
		g_slist_free(gl_msg_list);
		gl_msg_list = NULL;
	}
	POP3_SET_MSG_PROG_CB(POP3_MSG_COMPLETE, account);
}

/**
  This function is a callback function which is registered to socket
  @param session    pointer to the session
  @param len    length of the data
  @param data    data from the socket
  @return POP3_SUCCESS if session is valid else POP3_FAIL 
 */

static gint pop3_recv_data_finished(Session * session, guint len, gpointer data)
{
	Pop3Session *pop3_session = NULL;
	gint total_recv_bytes = 0;
	gint count = 0;

	pop3_session = POP3_SESSION(session);

	if (pop3_session->state != POP3_RETR &&
	    pop3_session->state != POP3_RETR_RECV &&
	    pop3_session->state != POP3_DELETE &&
	    pop3_session->state != POP3_TOP && pop3_session->state != POP3_TOP_RECV) {
		return POP3_SUCCESS;
	}
	if (session == NULL) {
		return POP3_FAIL;
	}

	/* if downloading only headers,indicate only the number of headers */
	if (pop3_session->ac_prefs->receive_settings.recv_type == HEADERS) {

		/* this is to reduce the number of callbacks */
		if ((pop3_session->retr_list != NULL)
		    && (pop3_session->delete_all == FALSE)) {

			count = g_slist_length(pop3_session->retr_list);
			pop3_recv_update_mail_status( count, ++pop3_session->retr_count);

		} else if (pop3_session->msgid == NULL) {
			/* Display message through call back */
			if ((POP3_HAS_CALLBACK)
			    && ((pop3_session->prog_cur_msg % 10 == 0)
				|| (pop3_session->prog_cur_msg == 1))) {
				pop3_recv_update_mail_status( pop3_session->prog_count, pop3_session->prog_cur_msg);
			}
		}
		return POP3_SUCCESS;

	}

	total_recv_bytes = pop3_session->cur_total_bytes;

	/* Display message through call back */
	if (POP3_HAS_CALLBACK) {
		pop3_recv_update_mail_status( pop3_session->prog_count, pop3_session->prog_cur_msg);
#if 0
		if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
			gl_pop3_callback_ptr->pop3_set_msg_cb(POP3_MSG_PROGRESS,
							      "%d %d %d %d",
							      pop3_session->
							      prog_cur_msg,
							      pop3_session->prog_count,
							      total_recv_bytes,
							      pop3_session->total_bytes);
		}
#endif
	}

	return POP3_SUCCESS;
}

/**
  This function is a dummy callback function which is registered to socket,
  @param session    pointer to the session
  @return POP3_SUCCESS(0) if session is valid else POP3_FAIL(-1)
 */

static gint pop3_recv_message(Session * session)
{
	/* Just to provide callbacks */
	if (session == NULL) {
		return POP3_FAIL;
	}
	return POP3_SUCCESS;
}

/**
  This function is used to initialise new pop3 receive session.
  @param account    pointer to the account structure
  @return session pointer,NULL on errors
 */
/*@null@*/
static Pop3RecvSession *pop3_recv_session_new(PrefsAccount * account, const gchar * msgid)
{
	Pop3RecvSession *session = NULL;

	if (account == NULL) {
		POP3_CB_PARAM_MISSING(ACCOUNT_NAME_MISSING);
		return NULL;
	}

	if (account->user_settings.account_name == NULL) {
		POP3_CB_PARAM_MISSING(ACCOUNT_NAME_MISSING);
		return NULL;
	}
	/* check the required parameters for the POP3 session */
	if (account->receive_settings.protocol != PROTO_POP3) {
		POP3_CB_PARAM_MISSING(RECV_PROTOCOL_MISSING);
		return NULL;

	}

	if (account->receive_settings.passwd_auth_type != AUTH_APOP &&
	    account->receive_settings.passwd_auth_type != AUTH_LOGIN &&
	    account->receive_settings.passwd_auth_type != AUTH_CRAM_MD5) {
		POP3_CB_PARAM_MISSING(PASSWD_AUTH_TYPE_MISSING);
		return NULL;
	}
	if (!account->receive_settings.recv_server) {
		POP3_CB_PARAM_MISSING(RECV_SERVER_MISSING);
		return NULL;
	}
	if (!account->user_settings.userid) {
		POP3_CB_PARAM_MISSING(USERID_MISSING);
		return NULL;
	}
	if (account->receive_settings.port == 0) {
		/* if the port number is missing just report the error.Donot return */
		POP3_CB_PARAM_MISSING(RECV_SERVER_PORT_MISSING);
	}
#if CL_USE
	debug_print(" Recv POP3 Mail through iap \n");
	if (cl_iap_connect(account->advanced_settings.iap_name, FALSE) != CL_OK) {
		debug_print("CL connection Error in imap \n");
		set_engine_errno(ENGINE_ERROR_IAP_CONNECT);
		return NULL;
	}
#endif
	session = g_new0(Pop3RecvSession, 1);
	if (session == NULL) {
		POP3_SET_MSG_CB(POP3_MSG_SESSION_ERROR);
		return NULL;
	} else {
		session->session = pop3_session_new(account, msgid);
		if (session->session == NULL) {
			g_free(session);
			session = NULL;
			if (get_engine_errno() != ENGINE_ERROR_POP_MSG_DELETE) {
				POP3_SET_MSG_CB(POP3_MSG_SESSION_ERROR);
			}
			return NULL;
		} else {
			session->session->data = session;
		}

	}

	return session;
}

/**
  This function is used to connect to socket and start receiving operation
  @param session    pointer to the pop3recv session
  @return session state
 */

static Pop3RecvState pop3_recv_session_do(Pop3RecvSession * session)
{
	Pop3Session *pop3_session = POP3_SESSION(session->session);
	gchar *server = NULL;
	gushort port = 0;


	server = pop3_session->ac_prefs->receive_settings.recv_server;

	port = pop3_get_port(pop3_session);

	if (session_connect(SESSION(pop3_session), server, port) < 0) {

		return (pop3_handle_connect_error(session));
	}
	/* wait for the socket to receive data */
	while ((session_is_connected(SESSION(pop3_session)) == TRUE) &&
	       (session->pop3_recv_state != POP3_RECV_CANCEL)) {
		(void) g_main_context_iteration(NULL, TRUE);
	}
	if (session->pop3_recv_state == POP3_RECV_SUCCESS) {

		switch (pop3_session->error_val) {

		case PS_SUCCESS:
			if (SESSION(pop3_session)->state == SESSION_ERROR) {
				if (pop3_session->state == POP3_READY) {
#if USE_SSL
					if (SESSION(pop3_session)->ssl_type != SSL_NONE) {
						session->pop3_recv_state = POP3_RECV_SSL_ERROR;
					} else {
#endif
						session->pop3_recv_state = POP3_RECV_CONNECT_ERROR;
#if USE_SSL
					}
#endif

				} else {
					session->pop3_recv_state = POP3_RECV_ERROR;
				}
			} else if (SESSION(pop3_session)->state == SESSION_EOF) {
				session->pop3_recv_state = POP3_RECV_EOF;
			} else {
				session->pop3_recv_state = POP3_RECV_SUCCESS;
			}
			break;

		case PS_AUTHFAIL:
			session->pop3_recv_state = POP3_RECV_AUTH_FAILED;
			break;
		case PS_PASSWD_CANCEL:
			session->pop3_recv_state = POP3_RECV_PASSWD_CANCEL;
			break;

		case PS_IOERR:
			session->pop3_recv_state = POP3_RECV_IO_ERROR;
			break;

		case PS_NOTSUPPORTED:
#if USE_SSL
			if (SESSION(pop3_session)->ssl_type != SSL_NONE) {
				session->pop3_recv_state = POP3_RECV_SSL_ERROR;
			} else {
#endif
				session->pop3_recv_state = POP3_RECV_NOSUPPORT;
#if USE_SSL
			}
#endif
			break;

		case PS_SOCKET:
#if USE_SSL
			if (SESSION(pop3_session)->ssl_type != SSL_NONE) {
				session->pop3_recv_state = POP3_RECV_SSL_ERROR;
			} else {
#endif
				session->pop3_recv_state = POP3_RECV_SOCKET_ERROR;
#if USE_SSL
			}
#endif
			break;

		case PS_LOCKBUSY:
			session->pop3_recv_state = POP3_RECV_LOCKED;
			break;

		case PS_DELERR:
			session->pop3_recv_state = POP3_RECV_DEL;
			break;

		case PS_PROTOCOL:
			session->pop3_recv_state = POP3_RECV_CONNECT_ERROR;
			break;
		default:
			session->pop3_recv_state = POP3_RECV_ERROR;
			break;
		}
	}

	return session->pop3_recv_state;
}

/**
  This function is used to start pop3 operation.This function calls 
  pop3_recv_session_do function
  @param session    pointer to the session structure
  @return none
 */

static void pop3_recv_start(Pop3RecvSession * session)
{
	Pop3Session *pop3_session = NULL;
	Pop3RecvState pop3_recv_state;

	pop3_session = POP3_SESSION(session->session);

	/* set the session parameters */
	session_set_recv_message_notify(session->session, pop3_recv_message);
	session_set_recv_data_progressive_notify
	    (session->session, pop3_recv_data_progressive, session);
	session_set_recv_data_notify(session->session, pop3_recv_data_finished, session);

	/* Connect POP3 session */
	pop3_recv_state = pop3_recv_session_do(session);

	/* this is to reduce the number of callbacks */
#if PROGRESS_CALLBACKS
	if (pop3_recv_state == POP3_RECV_SUCCESS) {

		/* call the registered function to indicate the number of messages */
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
				gl_pop3_callback_ptr->pop3_set_msg_cb
				    (POP3_MSG_RECV_SUCCESS,
				     "%d %d",
				     pop3_session->cur_total_num,
				     pop3_session->cur_total_recv_bytes);
			}
		}
	}
#endif
	if (pop3_recv_state != POP3_RECV_SUCCESS) {

		pop3_recv_put_error(pop3_recv_state, pop3_session);
	}
	if ((pop3_session->msgid == NULL)) {
		if ((pop3_recv_state == POP3_RECV_SUCCESS)
		    || (pop3_recv_state == POP3_RECV_CANCEL)) {
			pop3_write_uidl_list(pop3_session);
			pop3_update_headers(pop3_session);
			pop3_recv_update_mail_status( pop3_session->prog_count, pop3_session->prog_count);
		}
	}
	/* Destroy the session */
	pop3_recv_session_destroy(session);

}

/**
  This function is called to indicate pop3 operation progress.
  @param session    pointer to the session structure
  @param cur_len    length of the message received
  @param total_len    length of the total message
  @param data    pointer to the pop3recv session,passed as data
  @return POP3_SUCCESS(0) on success,POP3_FAIL(-1) on failure
 */

static gint
pop3_recv_data_progressive(Session * session, guint cur_len, guint total_len, gpointer data)
{

	gint total_recv_bytes = 0;
	gchar *total_size = NULL;
	gint disp = 0;


	Pop3RecvSession *pop3_recv_session = (Pop3RecvSession *) data;

	Pop3Session *pop3_session = POP3_SESSION(session);

	debug_print("Receive data PROGRESS called:%d,%d", cur_len, total_len);
	if ((pop3_recv_session == NULL) || (session == NULL)) {
		return POP3_FAIL;
	}
	if (pop3_session->state != POP3_RETR &&
	    pop3_session->state != POP3_RETR_RECV &&
	    pop3_session->state != POP3_TOP_RECV &&
	    pop3_session->state != POP3_TOP &&
	    pop3_session->state != POP3_DELETE && pop3_session->state != POP3_LOGOUT) {
		return POP3_SUCCESS;
	}

	/* if downloading only headers,indicate only the number of headers */
	if ((pop3_session->ac_prefs->receive_settings.recv_type == HEADERS) &&
	    (!pop3_session->msgid)) {
		return POP3_SUCCESS;

	}

	total_recv_bytes = pop3_session->cur_total_bytes + cur_len;

	if (total_recv_bytes > pop3_session->total_bytes) {
		total_recv_bytes = pop3_session->total_bytes;
	}

	Xstrdup_a(total_size, to_human_readable(pop3_session->total_bytes), return POP3_FAIL);
	if (pop3_session->total_bytes) {
		/* Indicate the percentage back to UI */
		disp = (total_recv_bytes * 100) / pop3_session->total_bytes;
		debug_print("PERCENTAGE:%d", disp);
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_set_percentage_cb != NULL) {
				gl_pop3_callback_ptr->pop3_set_percentage_cb(disp);
			}
		}
	}
	return POP3_SUCCESS;
}

/**
  This function is used to get message header information from 
  the message received.
  @param session   session of pop3 used 
  @param file    file name
  @param msgnum    message number
  @param msgid    pointer to the message id,used for mesasge information.
  This should not be freed by the calling function. 
  @return POP3_SUCCESS(0) on success,POP3_FAIL(-1) on memory errors
*/

gint pop3_get_msg_header(Pop3Session * session, const gchar * file, gint msg_num, gchar * msgid)
{

	/*variables for smime */
	gchar outfile[BSIZE];
	gchar detached[BSIZE];
	gboolean saved = FALSE;
	gchar *signer_email_id = NULL;
	gchar *recip_email_id = NULL;
	gint private_cert_state = 0;
	gint recip_cert_state = 0;
	gboolean signature = TRUE;
	gboolean trust = FALSE;
	gchar cert_path[BSIZE];
	gint sender_cert_id = 0;
	gint recip_cert_id = 0;
	gint retval = 0;
	gint smime = 0;
	gboolean remote_copy = 0;

	MsgInfo *msginfo = NULL;
	if (msgid == NULL) {
		return POP3_FAIL;
	}
	
	/* Get the message information from the header */
	msginfo = pop3_parse_msg(file, msg_num);

	if (msginfo == NULL) {
		g_free(msgid);
		return POP3_FAIL;
	}
	/* make sure the message is the one, we want */

	if (session->msgid != NULL) {

		if (check_session_msgid(session->msgid, msginfo->header->msgid) == FALSE) {

			session->error_val = PS_DELERR;
			/* Free the message information */
			procmsg_msginfo_free(msginfo);
			g_free(msgid);
			return POP3_FAIL;
		}
	}
	if (session->retr_list != NULL) {
		if (check_session_msgid(msgid, msginfo->header->msgid) == FALSE) {


			/* Free the message information */
			folder_delete_list_msg(msgid, TRUE);
			folder_message_del_list(msgid, UNREAD);
			unlink(file);
			gl_unavailable_list = g_slist_append(gl_unavailable_list, g_strdup(msgid));
			POP3_MSG_DELETE_ERROR_CB(msgid);
		} else {
			if (session->ac_prefs->receive_settings.leave_msg == TRUE) {
				remote_copy = TRUE;
			}
			POP3_MSG_DOWNLOAD_CB(msgid, remote_copy);
		}
		procmsg_msginfo_free(msginfo);
		return POP3_SUCCESS;
	}
	if ((pop3_set_msginfo(msginfo, session, msgid)) == POP3_FAIL) {
		/* Free the message information */
		procmsg_msginfo_free(msginfo);
		return POP3_FAIL;
	}
	/* If message body is downloaded,give mime information back */
	if (session->msgid != NULL) {

		if (msginfo->header->smime_type != SMIME_NONE) {
			signer_email_id = g_strdup(parse_from_address(msginfo->header->from));
			recip_email_id = g_strdup(session->ac_prefs->user_settings.email);

			if (0 != (retval = folder_prepare_process_smime(outfile,
									detached,
									cert_path,
									file, &saved,
									signer_email_id,
									recip_email_id,
									&recip_cert_state,
									&private_cert_state,
									&signature,
									&trust,
									&smime,
									&sender_cert_id,
									&recip_cert_id))) {
				debug_print("Unable to verify/decrypt S/MIME\n");
				POP3_SET_MSG_CB(retval);
				g_free(signer_email_id);
				g_free(recip_email_id);
				remove_smime_file(outfile);
				remove_smime_file(detached);
				remove_smime_file(cert_path);
				if (POP3_HAS_CALLBACK) {
					if (gl_pop3_callback_ptr->
					    pop3_recv_body_complete_cb != NULL)
						gl_pop3_callback_ptr->
						    pop3_recv_body_complete_cb(NULL,
									       msgid,
									       msginfo->header->
									       msgtag.remote_copy);
				}
				procmsg_msginfo_free(msginfo);
				return POP3_RECV_SUCCESS;
			}

			if (pop3_send_smime_mimeinfo(outfile, msginfo->header,
						     session->msgid, saved,
						     recip_cert_state,
						     private_cert_state, signature,
						     trust, cert_path, smime,
						     sender_cert_id, recip_cert_id) == POP3_FAIL) {

				/*Free the message information */
				procmsg_msginfo_free(msginfo);
				debug_print("pop3_send_smime_mimeinfo failed\n");
				g_free(signer_email_id);
				g_free(recip_email_id);
				remove_smime_file(outfile);
				remove_smime_file(detached);
				remove_smime_file(cert_path);
				return POP3_FAIL;
			}
			remove_smime_file(outfile);
			remove_smime_file(detached);
		}

		else if (pop3_send_mimeinfo(file, msginfo->header, session->msgid)
			 == POP3_FAIL) {

			/* Free the message information */
			procmsg_msginfo_free(msginfo);
			return POP3_FAIL;
		}

	} else {

		if (procmime_check_attachments(file, FALSE) == TRUE) {
			msginfo->header->flags.perm_flags |= MSG_ATTACH;
			msginfo->header->smime_type = folder_get_smime_type(msginfo->header, file);
		}
		/* Send message header to registered layer using callback */
		gl_msg_list = g_slist_append(gl_msg_list, msginfo);
		pop3_recv_msg_headers_to_view();
	}
	return POP3_SUCCESS;
}

/**
  This function is used to parse message file to get message information
  @param file    file name
  @param msg_num    message number
  @return pointer to the message information.This has to be freed by the 
  calling function.
 */

/*@null@*/ static MsgInfo *
pop3_parse_msg(const gchar * file, gint msg_num)
{
	struct stat s;
	MsgInfo *msginfo = NULL;
	MsgFlags flags;

	CHECK_RETURN_VAL_IF_FAIL(msg_num != 0, NULL);
	CHECK_RETURN_VAL_IF_FAIL(file != NULL, NULL);

	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;

	msginfo = procheader_parse_file(file, flags, TRUE);
	if (!msginfo) {
		return NULL;
	}

	if (stat(file, &s) < 0) {
		/* Free the message information */
		procmsg_msginfo_free(msginfo);
		return NULL;
	}
	if (!S_ISREG(s.st_mode)) {
		/* Free the message information */
		procmsg_msginfo_free(msginfo);
		return NULL;
	}
	msginfo->header->msgnum = (guint) msg_num;
	msginfo->header->size = s.st_size;
	msginfo->header->mtime = s.st_mtime;
	msginfo->header->msgtag.fullydownloaded = TRUE;

	return msginfo;
}

/**
  This function is used to cancel pop3 operation.
  @return none
 */

void pop3_recv_cancel(void)
{
	if (gl_pop3_cur_session != NULL) {
		gl_pop3_cur_session->pop3_recv_state = POP3_RECV_CANCEL;
	}
}

/**
  This function is used to destroy pop3 recv session.
  @param session    pointer to the session structure
  @param return none
 */

static void pop3_recv_session_destroy(Pop3RecvSession * session)
{
	/* set the global pointer to NULL */

	gl_pop3_cur_session = NULL;

	if (session == NULL) {
		return;
	}

	session_destroy(session->session);

	g_free(session);
	session = NULL;

}

/**
  This function is indicates the appropriate errors through the callbacks
  @param istate    the pop3 recv session state
  @param pop3_session    the pop3 recv session 
  @return none
 */

static void pop3_recv_put_error(Pop3RecvState istate, Pop3Session * pop3_session)
{
	gchar *log_msg = NULL;
	gchar *hfile = NULL;

	switch (istate) {

	case POP3_RECV_CONNECT_ERROR:
		log_msg = "Connection failed.";
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
				gl_pop3_callback_ptr->
				    pop3_set_msg_cb(POP3_MSG_CONNECT_ERROR, "%s",
						    pop3_session->ac_prefs->
						    receive_settings.recv_server);
			}
		}

		/* add to the list only if not iniated from queue */
		if ((gl_pop3_queued_delete == FALSE) && ((pop3_session->delete_msg_num != 0)
							 || (pop3_session->delete_all == TRUE))) {

			if (gl_pop3_queued_delete == FALSE) {
				if (pop3_session->delete_msg_num != 0) {
					folder_add_list(pop3_session->msgid, FALSE);
				} else if (pop3_session->delete_all == TRUE) {
					pop3_write_retr_del_list(pop3_session->retr_list, TRUE);
				}
			}
			set_engine_errno(ENGINE_ERROR_CONNECT_SERVER);
			break;

	case POP3_RECV_ERROR:
			log_msg = "Error occurred while processing mail.";
			POP3_SET_MSG_CB(POP3_MSG_RECV_ERROR);
			break;

	case POP3_RECV_IO_ERROR:
			log_msg = "Can't write file.";
			POP3_SET_MSG_CB(POP3_MSG_IO_ERROR);
			break;

	case POP3_RECV_SOCKET_ERROR:
			log_msg = "Socket error.";
			POP3_SET_MSG_CB(POP3_MSG_SOCKET_ERROR);
			break;

	case POP3_RECV_EOF:
			log_msg = "Connection closed by the remote host.";
			POP3_SET_MSG_CB(POP3_MSG_EOF);
			break;

	case POP3_RECV_LOCKED:
			log_msg = "Mailbox is locked.";
			POP3_SET_MSG_CB(POP3_MSG_LOCK_ERROR);
			break;

	case POP3_RECV_AUTH_FAILED:
			log_msg = "Authentication failed.";
			if (POP3_HAS_CALLBACK) {
				if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
					gl_pop3_callback_ptr->
					    pop3_set_msg_cb(POP3_MSG_AUTH_ERROR, "%s %s",
							    pop3_session->ac_prefs->
							    receive_settings.recv_server,
							    pop3_session->ac_prefs->
							    user_settings.account_name);
				}
			}
			break;

	case POP3_RECV_CANCEL:
			log_msg = "Incorporation cancelled\n";
			if (pop3_session->delete_all == TRUE) {
				/* If not logged out when delete operation 
				 * is not cancelled, delete not effected in server
				 */
				pop3_logout_send(pop3_session);
			}
			POP3_SET_MSG_CB(POP3_MSG_RECV_CANCEL);
			break;

	case POP3_RECV_NOSUPPORT:
			log_msg = "Functionality not supported\n";
			POP3_SET_MSG_CB(POP3_MSG_NOSUPPORT_ERROR);
			break;

	case POP3_RECV_PASSWD_CANCEL:
			log_msg = "Temporary password cancelled\n";
			break;

#if USE_SSL
	case POP3_RECV_SSL_ERROR:
			log_msg = "SSL Connection failed.";
			if (POP3_HAS_CALLBACK) {
				if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
					gl_pop3_callback_ptr->
					    pop3_set_msg_cb(POP3_MSG_SSL_ERROR, "%d", ssl_error);
				}
			}
			if (pop3_session->delete_msg_num != 0)
				folder_add_list(pop3_session->msgid, FALSE);
		} else if (pop3_session->delete_all == TRUE) {
			pop3_write_retr_del_list(pop3_session->retr_list, TRUE);
		}
		set_engine_errno(ENGINE_ERROR_CONNECT_SERVER);
		break;
#endif

	case POP3_RECV_DEL:
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_set_msg_cb != NULL) {
				gl_pop3_callback_ptr->
				    pop3_set_msg_cb(POP3_MSG_DELETE_ERROR, "%s",
						    pop3_session->msgid);
			}
		}
		/* This is the only way for synchronising with the server,
		 * if the UIDLs on server change.The UIDs are supposed to be 
		 * unique onserver but some servers are reassigning the UIDs 
		 * in some cases.
		 * This change is done after testing with Redhat server when 
		 * encountered problems  */
		hfile = folder_get_message_file_path(pop3_session->msgid);
		pop3_remove_header_file(hfile);
		folder_delete_list_msg(pop3_session->msgid, TRUE);
		folder_delete_list_msg(pop3_session->msgid, FALSE);
		folder_message_del_list(pop3_session->msgid, UNREAD);
		g_free(hfile);
		gl_unavailable_list = g_slist_append(gl_unavailable_list,
						     g_strdup(pop3_session->msgid));
		set_engine_errno(ENGINE_ERROR_POP_MSG_DELETE);
		break;

	default:
		break;
	}

	if (log_msg) {
		log_message("%s\n", log_msg);
	}
}

/**
  This function is called for each key value pair in UID hash table. 
  @param key hash table key
  @param value hash table value mapped for the key 
  @param msgid user data populated with max value in hash table 
  @return nothing 
*/
static void pop3_find_max_id(gpointer key, gpointer value, gpointer msgid)
{
	if ((gint )value > *(gint *)msgid) {
		*(gint *)msgid = (gint )value;
	}
}

/**
  This function is used to get new name for the message received
  @param account    account name
  @param destfolder     destination folder name
  @param msgnum_ptr    pointer to the message number
  @return file name
 */

gchar *pop3_get_new_msg_file( Pop3Session * pop3_session, const gchar * account_name,
			     const gchar * destfolder, gint * msgnum_ptr)
{

	gchar *destfile = NULL;
	gint msg_num = 0;
	gint maxid = 0;
	gchar *tmp_msgid = NULL;

	pop3_free_old_list();
	msg_num = folder_get_last_msg_num() ;
	if(pop3_session->uidl_table) {
		g_hash_table_foreach(pop3_session->uidl_table, pop3_find_max_id, &maxid);
	}
	if(maxid > msg_num) {
		msg_num = maxid;
	}
	msg_num++;

	debug_print("\n\nMSG NUMBER:%d", msg_num);

	if (is_dir_exist(destfolder) == FALSE) {
		debug_print("\nThe destination folder does not exist");
		return NULL;
	}

	tmp_msgid = g_strdup_printf("%s%c%s%c%d", account_name, MSGID_SEPARATOR,
				    INBOX_DIR, MSGID_SEPARATOR, msg_num);

	if (*(destfolder + strlen(destfolder) - 1) == G_DIR_SEPARATOR) {
		destfile = g_strdup_printf("%s%s", destfolder, tmp_msgid);
	} else {
		destfile = g_strdup_printf("%s%c%s", destfolder, G_DIR_SEPARATOR, tmp_msgid);

	}
	*msgnum_ptr = msg_num;

	g_free(tmp_msgid);

	return destfile;
}

/**
  This function checks the state of the cancel 
  @param none 
  @return TRUE if receive operation can continue 
 */
gboolean pop3_recv_check_state(void)
{
	if (gl_pop3_cur_session == NULL) {
		return FALSE;
	}

	if (gl_pop3_cur_session->pop3_recv_state == POP3_RECV_CANCEL) {
		return FALSE;
	} else {
		return TRUE;
	}
}

/**
  This function extracts the information from the headers 
  @param data    pointer to data
  @param len    message header length
  @param msgnum_ptr    message number pointer
  @return  POP3_SUCCESS(0) on receiving correctly,POP3_FAIL(-1) on errors
 */
gint pop3_get_header_info(Pop3Session * session, const gchar * data, guint len, gint * msgnum_ptr)
{
	MsgInfo *msginfo = NULL;
	gchar *file = NULL;
	gint msg_num;
	gchar *inbox_folder = NULL;
	gchar *path = NULL;
	gchar *dest_path = NULL;
	gchar *hfile = NULL;

	if (len == 0) {
		session->error_val = PS_IOERR;
		return POP3_FAIL;
	}

	path = get_tmp_file();

	msginfo = pop3_get_msginfo(data, len, path);

	if (msginfo == NULL) {
		session->error_val = PS_IOERR;
		g_free(path);
		return POP3_FAIL;
	}

	/* if this is delete operation,just check the message and return */
	if ((session->delete_msg_num != 0) || (session->delete_all == TRUE)) {
		if (session->delete_all == TRUE) {
			session->msgid = g_strdup_printf("%s%c%s%c%u",
							 session->ac_prefs->
							 user_settings.account_name,
							 MSGID_SEPARATOR, INBOX_DIR,
							 MSGID_SEPARATOR,
							 session->msg[session->cur_msg].recv_num);
		}
		hfile = folder_get_message_file_path(session->msgid);
		if (check_session_msgid(session->msgid, msginfo->header->msgid) == FALSE) {
			session->error_val = PS_DELERR;
			/* delete the local file */
			if (is_file_entry_exist(hfile) == TRUE) {
				if (session->operation != FOLDER_OP_MOVE) {
					unlink(hfile);
				}
			} else {
				pop3_remove_header_file(hfile);
			}

			if ((session->delete_all == TRUE)
			    && (session->queue_flag == FALSE)) {
				if (gl_pop3_callback_ptr->pop3_send_progress_cb != NULL) {
					gl_pop3_callback_ptr->
					    pop3_send_progress_cb(FOLDER_PROGRESS, "%d %d",
								  gl_cur_count, gl_total_count);
				}

				gl_cur_count = gl_cur_count + 1;
			}

			folder_delete_list_msg(session->msgid, TRUE);
			folder_message_del_list(session->msgid, UNREAD);
			g_free(hfile);
			procmsg_msginfo_free(msginfo);
			g_free(path);
			return POP3_FAIL;
		} else {
			/* delete the local file */
			if (is_file_entry_exist(hfile) == TRUE) {
				if (session->operation != FOLDER_OP_MOVE) {
					unlink(hfile);
				}
			} else {
				pop3_remove_header_file(hfile);
			}
			if ((session->delete_all == TRUE)
			    && (session->queue_flag == FALSE)) {
				if (gl_pop3_callback_ptr->pop3_send_progress_cb != NULL) {
					gl_pop3_callback_ptr->
					    pop3_send_progress_cb(FOLDER_PROGRESS, "%d %d",
								  gl_cur_count, gl_total_count);
				}

				gl_cur_count = gl_cur_count + 1;
			}
			folder_delete_list_msg(session->msgid, TRUE);
			folder_message_del_list(session->msgid, UNREAD);
			g_free(hfile);
			procmsg_msginfo_free(msginfo);
			g_free(path);
			g_free(session->msgid);
			session->msgid = NULL;
			return POP3_SUCCESS;

		}

	}
	if ((pop3_set_msginfo(msginfo, session, NULL)) == POP3_FAIL) {
		procmsg_msginfo_free(msginfo);
		g_free(path);
		return POP3_FAIL;

	}
	/* create new msgid for the mail & add to the list */
	inbox_folder = g_strdup_printf("%s%c%s", get_mail_dir(), G_DIR_SEPARATOR, INBOX_DIR);

	file =
	    pop3_get_new_msg_file(session, session->ac_prefs->user_settings.account_name,
				  inbox_folder, &msg_num);
	*msgnum_ptr = msg_num;
	g_free(inbox_folder);

	/* Set message id & account information */
	g_free(msginfo->header->msgid);
	msginfo->header->msgid = g_strdup_printf("%s%c%s%c%d",
						 session->ac_prefs->
						 user_settings.account_name,
						 MSGID_SEPARATOR, INBOX_DIR,
						 MSGID_SEPARATOR, msg_num);
	gl_msg_list = g_slist_append(gl_msg_list, msginfo);
	
	dest_path = g_strdup_printf("%s%c%s", file, DOT, MSG_HDR);
	move_file(path, dest_path, TRUE);
	folder_message_add_list(g_strdup(msginfo->header->msgid), UNREAD);
	folder_add_list(msginfo->header->msgid, TRUE);
	msginfo->header->msgtag.remote_copy = TRUE;
	
	if(g_slist_length(gl_msg_list) == POP3_DISPLAY_COUNT) {
		pop3_recv_msg_headers_to_view();
	}

	g_free(dest_path);
	g_free(path);
	g_free(file);
	return POP3_SUCCESS;
}

/**
  This function checks if the message is valid (if it has already been deleted)
  @param msgid    the message id(saved) of the received mail
  @param retr_msgid    the message id(server) of the received mail
  @return true if the message id in the list and received mail match
 */
static gboolean check_session_msgid(const gchar * msgid, gchar * retr_msgid)
{
	gchar *header_path = NULL;
	gchar *file = NULL;
	MsgFlags flags;
	MsgInfo *msginfo = NULL;

	/* it is required to check the message IDs of the mail since the 
	 * server reassign uids of the deleted mails */

	file = folder_get_message_file_path(msgid);
	/* no file,get hdr file */
	header_path = g_strdup_printf("%s%c%s", file, DOT, MSG_HDR);

	if (is_file_entry_exist(header_path) == TRUE) {
		g_free(file);
		file = header_path;

	} else if (is_file_entry_exist(file) == FALSE) {

		g_free(file);
		return FALSE;
	}

	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;

	msginfo = procheader_parse_file(file, flags, TRUE);
	if (!msginfo) {
		g_free(file);
		return FALSE;
	}
	if (msginfo->header->msgid == NULL) {
		g_free(file);
		procmsg_msginfo_free(msginfo);

		if (retr_msgid == NULL) {
			return TRUE;
		} else {
			return FALSE;
		}
	}
	if (retr_msgid == NULL) {
		g_free(file);
		procmsg_msginfo_free(msginfo);
		return FALSE;
	}
	if (strcmp(msginfo->header->msgid, retr_msgid) == 0) {
		g_free(file);
		procmsg_msginfo_free(msginfo);
		return TRUE;
	}
	g_free(file);
	procmsg_msginfo_free(msginfo);
	return FALSE;
}

/**
  This function cleansup all memory used by the module,
  so that the process can exit
  @param none 
  @return none 
 */
void pop3_cleanup(void)
{
	if (gl_pop3_cur_session != NULL) {
		pop3_recv_session_destroy(gl_pop3_cur_session);
		gl_pop3_cur_session = NULL;
	}
	if (gl_pop3_callback_ptr != NULL) {
		g_free(gl_pop3_callback_ptr);
		gl_pop3_callback_ptr = NULL;
	}
}

/**
  This function returns the server port number  
  @param pop3_session    pop3 session id 
  @return the port number on server  
 */
static gint pop3_get_port(Pop3Session * pop3_session)
{
	gushort port = 0;

#if USE_SSL

	if (pop3_session->ac_prefs->advanced_settings.use_ssl == TRUE) {

		if (pop3_session->ac_prefs->advanced_settings.use_tls == TRUE) {
			SESSION(pop3_session)->ssl_type = SSL_STARTTLS;
			port = pop3_session->ac_prefs->receive_settings.port;

		} else {
			SESSION(pop3_session)->ssl_type = SSL_TUNNEL;
			port = pop3_session->ac_prefs->receive_settings.port;

		}


	} else {

		SESSION(pop3_session)->ssl_type = SSL_NONE;
		/* check for the receive port number.Use default if no port number */
		port = (pop3_session->ac_prefs->receive_settings.port != 0) ?
		    pop3_session->ac_prefs->receive_settings.port : POP3_PORT;

	}

#else
	/* check for the receive port number.Use default if no port number */
	port = pop3_session->ac_prefs->receive_settings.port ?
	    pop3_session->ac_prefs->receive_settings.port : POP3_PORT;
#endif

	return port;
}

/**
  This function returns the connect error state  
  @param pop3_session    pop3 session id 
  @return the error state  
*/
static gint pop3_handle_connect_error(Pop3RecvSession * session)
{
	Pop3Session *pop3_session = NULL;

	pop3_session = POP3_SESSION(session->session);

#if USE_SSL
	if (SESSION(pop3_session)->ssl_type == SSL_TUNNEL) {
		session->pop3_recv_state = POP3_RECV_SSL_ERROR;
		return POP3_RECV_SSL_ERROR;
	} else {
		session->pop3_recv_state = POP3_RECV_CONNECT_ERROR;
		return POP3_RECV_CONNECT_ERROR;
	}
#else
	session->pop3_recv_state = POP3_RECV_CONNECT_ERROR;
	return POP3_RECV_CONNECT_ERROR;

#endif
}

/**
  This function sends mimeinformation through callback   
  @param file   the file to read 
  @param msgheader    the message header information
  @param msgid    the message id 
  @return the error state  
 */
static gint pop3_send_mimeinfo(const gchar * file, MsgHeader * msgheader, const gchar * msgid)
{
	FILE *fp = NULL;
	MimeInfo *mimeinfo = NULL;
	gint pos = -1;
	gint level = 0;
	glong fpos = 0;
	GSList *file_pos_list = NULL;
	gchar *decfilename = NULL;


	if ((file == NULL) || (msgheader == NULL) || (msgid == NULL)) {
		return POP3_FAIL;
	}
	fp = fopen(file, "r+");

	if (fp == NULL) {
		/* Give the error information back to higher layer */
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_recv_body_complete_cb != NULL) {
				gl_pop3_callback_ptr->pop3_recv_body_complete_cb(NULL, NULL, FALSE);
			}
		}
		return POP3_FAIL;
	}

	/* Send the information back to higher layer */
	mimeinfo = (MimeInfo *) procmime_scan_message(fp, pos, level, fpos, file_pos_list);

	fclose(fp);

	/* Decode the contents */
	if (mimeinfo != NULL) {
		decfilename = procmime_get_text_content(msgheader, FALSE, FALSE);

		procmime_check_and_set_filename(mimeinfo, decfilename);
	}
	/* Give the mime information back to higher layer */
	if (POP3_HAS_CALLBACK) {
		if (gl_pop3_callback_ptr->pop3_recv_body_complete_cb != NULL) {
			gl_pop3_callback_ptr->pop3_recv_body_complete_cb(mimeinfo,
									 msgid,
									 msgheader->
									 msgtag.remote_copy);
		}
	}
	/* Free the mime information */
	procmime_mimeinfo_free_all(mimeinfo);
	return POP3_SUCCESS;
}

/**
  This function sets the required flags in the message information 
  read from a file.  
  @param msginfo    the message information read from file. 
  @param session    pop3 session used 
  @param msgid    message id of the message 
  @return the error state  
 */
static gint pop3_set_msginfo(MsgInfo * msginfo, Pop3Session * session, const gchar * msgid)
{
	Folder *fold = NULL;

	if ((session == NULL) || (msginfo == NULL)) {
		return POP3_FAIL;
	}
	msginfo->header->msgtag.remote_copy = FALSE;

	if (msgid != NULL) {
		/* Free the one read from file */
		g_free(msginfo->header->msgid);
		/* Set message id & account information */
		msginfo->header->msgid = (gchar *) msgid;

		if (folder_message_add_list(msginfo->header->msgid, UNREAD) != FOLDER_SUCCESS) {
			return POP3_FAIL;
		}
		if (session->ac_prefs->receive_settings.leave_msg == TRUE) {

			if (folder_is_msg_remote(msgid) == FALSE) {
				/* Add to remote copy list */
				if (folder_add_list(msginfo->header->msgid, TRUE) != FOLDER_SUCCESS) {
					return POP3_FAIL;

				}
			}
			msginfo->header->msgtag.remote_copy = TRUE;
		}

	}
	/* Set other fields from the account information */
	msginfo->header->msgtag.protocol = session->ac_prefs->receive_settings.protocol;

	msginfo->header->msgtag.recvtype = session->ac_prefs->receive_settings.recv_type;

	/* set folder type */
	fold = folder_find_obj_by_name(INBOX_DIR);

	if (fold != NULL) {
		msginfo->header->msgtag.foldertype = fold->type;
	}

	return POP3_SUCCESS;
}

/**
  This function reads and returns the message information   
  @param data    received data 
  @param len    message header length
  @param path  Path in which file will be saved
  @return the pointer to message information,read from the data  
 */
static MsgInfo *pop3_get_msginfo(const gchar * data, guint len, const gchar * path)
{
	FILE *fp;
	MsgFlags flags;
	MsgInfo *msginfo = NULL;

	/* get the message information */
	fp = fopen(path, "w+");
	if (!fp) {
		FILE_OP_ERROR("str_open_as_stream", "my_tmpfile");
		return NULL;
	}
	if (pop3_write_msg_to_file(fp, data, len) < 0) {
		fclose(fp);
		return NULL;
	}
	flags.perm_flags = MSG_NEW | MSG_UNREAD;
	flags.tmp_flags = 0;

	rewind(fp);

	msginfo = procheader_parse_stream(fp, flags, TRUE);
	fclose(fp);
	return msginfo;
}

/** This function sends mimeinformation on an S/MIME message
 * back to UI through callback
 * @param file - File containing the message
 * @param msgheader - Msgheader information
 * @param msgid - ID of the message being opened
 * @param saved - Whether certificate needs to be saved
 * @param recip_cert_state - Our certificate state
 * @param private_cert_state - Sender certificate state
 * @param signature - Validity of signature
 * @param trust - whether signer certificate is trusted 
 * @param cert_path - Certificate path if cert to be stored
 * @param smime - Whether Signed encrypyted or both
 * @return error state
 */
static
    gint
pop3_send_smime_mimeinfo(const gchar * file, MsgHeader * msgheader,
			 const gchar * msgid, gboolean saved,
			 gint recip_cert_state, gint private_cert_state,
			 gboolean signature, gboolean trust,
			 gchar * cert_path, gint smime, gint sender_cert_id, gint recip_cert_id)
{
	FILE *fp = NULL;
	MimeInfo *mimeinfo = NULL;
	gint pos = -1;
	gint level = 0;
	glong fpos = 0;
	GSList *file_pos_list = NULL;

	debug_print("In function pop3_send_smime_mimeinfo\n");

	if ((file == NULL) || (msgheader == NULL) || (msgid == NULL)) {
		return POP3_FAIL;
	}

	fp = fopen(file, "r+");
	if (fp == NULL) {
		/* Give the error information back to higher layer */
		if (POP3_HAS_CALLBACK) {
			if (gl_pop3_callback_ptr->pop3_recv_body_complete_cb != NULL) {
				gl_pop3_callback_ptr->pop3_recv_body_complete_cb(NULL, NULL, FALSE);
			}
		}
		return POP3_FAIL;
	}
	/* Send the information back to higher layer */
	mimeinfo = (MimeInfo *) procmime_scan_message(fp, pos, level, fpos, file_pos_list);
	fclose(fp);
	/* Decode the contents */
	if (mimeinfo != NULL) {
		mimeinfo->filename = procmime_get_smime_text_content(msgheader, file);
		mimeinfo->trust = trust;
		mimeinfo->smime_type = smime;
		mimeinfo->saving = saved;
		mimeinfo->cert_path = cert_path;
		mimeinfo->cert_state_sender = private_cert_state;
		mimeinfo->cert_state_recip = recip_cert_state;
		mimeinfo->signature_verify = signature;
		mimeinfo->sender_cert_id = sender_cert_id;
		mimeinfo->recip_cert_id = recip_cert_id;

	}
	/* Give the mime information back to higher layer */
	if (POP3_HAS_CALLBACK) {
		if (gl_pop3_callback_ptr->pop3_recv_body_complete_cb != NULL) {
			gl_pop3_callback_ptr->pop3_recv_body_complete_cb(mimeinfo,
									 msgid,
									 msgheader->
									 msgtag.remote_copy);
		}
	}

	/* Free the mime information */
	procmime_mimeinfo_free_all(mimeinfo);
	return POP3_SUCCESS;
}

/**
  This function checks if the message is an S/MIME message.
  If it is an S/MIME message decrypts the message and returns
  the file path
  @param  msgid     msg Id of the mail in question
  return string which is the file path of decrypted message
  or the complete message
*/
gchar *return_file_if_smime(const gchar * msgid)
{
	gchar *file = NULL;
	MsgInfo *msg_info = NULL;
	gchar *smime_file = NULL;

	file = folder_get_message_file_path(msgid);

	if (file == NULL) {
		debug_print("%s %s : file path not found \n", __FILE__, __FUNCTION__);
		return NULL;
	}

	msg_info = folder_parse_msg(file);

	if (msg_info == NULL) {
		debug_print("%s %s : msg info NULL \n", __FILE__, __FUNCTION__);
		g_free(file);
		return NULL;
	}

	if ((msg_info->header->smime_type == SMIME_SIGN) ||
	    (msg_info->header->smime_type == SMIME_SIGN_ENCRYPT)) {
		smime_file = folder_process_smime_message(msgid);
		g_free(msg_info);
		g_free(file);
		if (smime_file == NULL) {
			debug_print("%s %s : msg info NULL \n", __FILE__, __FUNCTION__);
			return NULL;
		}
		return smime_file;
	} else {
		g_free(msg_info);
		return file;
	}
}


gboolean pop3_remove_header_file(const gchar * file)
{
	gchar *header_path = NULL;

	if (file == NULL) {
		return FALSE;
	}

	header_path = g_strdup_printf("%s%c%s", file, DOT, MSG_HDR);

	if (unlink(header_path) != 0) {
		debug_print("No header to unlink\n");
	}

	g_free(header_path);
	return TRUE;
}

/**
  This function scans folder 
  @param none 
  @return none
 */
void pop3_free_old_list(void)
{
	Folder *fold = NULL;
	gchar *folder_name = NULL;
	/* rescan the folder */
	folder_name = g_strdup_printf("%s", INBOX_DIR);
	fold = folder_find_obj_by_name(folder_name);
	g_free(folder_name);
	if (fold != NULL) {
		if (folder_scan(fold) == FOLDER_SCAN_ERROR) {
			return;
		}
	}
}

void pop3_send_complete_cb(PrefsAccount * account)
{
	POP3_SET_MSG_PROG_CB(POP3_MSG_COMPLETE, account);
}

void pop3_write_retr_del_list(GSList * delete_list, gboolean check_remotelist)
{
	gint length = 0;
	gint counter = 0;
	gchar *msgid = 0;
	RetrMsgData *retr = NULL;
	gchar *file_path = NULL;

	if (delete_list == NULL)
		return;
	length = g_slist_length(delete_list);

	for (counter = 0; counter < length; counter++) {

		retr = g_slist_nth_data(delete_list, counter);
		msgid = retr->msgid;
		if(check_remotelist == TRUE) {
			folder_add_list(msgid, FALSE);
		}
		file_path = folder_get_message_file_path(msgid);
		if (is_file_entry_exist(file_path) == TRUE) {
			unlink(file_path);
		} else {
			pop3_remove_header_file(file_path);
		}
		g_free(file_path);
		file_path = NULL;

	}


}
