
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999-2006 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "defs.h"

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>

#if (HAVE_WCTYPE_H && HAVE_WCHAR_H)
#  include <wchar.h>
#  include <wctype.h>
#endif

#include "quoted-printable.h"
#include "codeconv.h"
#include "utils.h"
#include "common.h"
#include "prefs.h"
#include "compose.h"
#include "smtp_errorcodes.h"
#include "prefs_account.h"
#include "base64.h"
#include "smtp.h"
#include "folder.h"
#include "htmltotextconv.h"

/**
    This function called to check the email receipant list in msg header.

    @param MsgHeader contains msg header information  
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint compose_check_for_valid_recipient(const MsgHeader * msgheader)
{
	debug_print("[%s][%s:%d]:starts here:\n", __FILE__, __FUNCTION__, __LINE__);
	if (msgheader == NULL) {
		return SMTP_MSG_HDR_NULL;
	}
	if (msgheader->to != NULL) {
		g_strstrip(msgheader->to);
	}
	if (msgheader->cc != NULL) {
		g_strstrip(msgheader->cc);
	}
	if (msgheader->bcc != NULL) {
		g_strstrip(msgheader->bcc);
	}
	if (!msgheader->to && !msgheader->cc && !msgheader->bcc) {
		return SMTP_RCPT_LIST_INVALID;
	} else {
		return SM_OK;
	}
}

/**
    This function called to check the user settings of  account.
  
    @param Prefs Account  contains account information  
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint compose_check_for_valid_account(const PrefsAccount * account)
{
	debug_print("[%s]:[[%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (account == NULL) {
		debug_print(" Prefs Account is null \n");
		return SMTP_ACCOUNT_NULL;
	}

	if ((account->send_settings.smtp_auth_type != SMTPAUTH_NO_AUTH) &&
	    (account->send_settings.smtp_auth_type != SMTPAUTH_LOGIN) &&
	    (account->send_settings.smtp_auth_type != SMTPAUTH_CRAM_MD5)) {
		debug_print("Invalid Authentication type \n");
		return SMTP_AUTHTYPE_INVALID;
	}

	if (account->user_settings.account_name == NULL) {
		debug_print("user account is null \n");
		return SMTP_ACCOUNT_NAME_INVALID;
	}

	if (account->user_settings.email == NULL) {
		debug_print("Email-id is null \n");
		return SMTP_EMAIL_INVALID;
	}

	if (account->send_settings.smtp_auth_type != SMTPAUTH_NO_AUTH) {
		if (account->send_settings.username == NULL) {
			debug_print(" user id is null \n");
			return SMTP_USERID_INVALID;
		}

		/*if ( account->send_settings.password  == NULL) {
		   debug_print ( " Password is null \n" );
		   return SMTP_PASSWD_INVALID;
		   } */
	}

	if (account->send_settings.smtp_server == NULL) {
		debug_print("Invalid smtp server(NULL) \n");
		return SMTP_SERVER_INVALID;
	}

	return SM_OK;
}


/**
   This function called to check the msg info and account information 
   suppilied send mail function.
  
   @param MsgInfo  contains msg header, attachment list etc.  
   @param Prefs Account contains account information  
   @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint compose_check_entries(const MsgInfo * msginfo, const PrefsAccount * prefs_account)
{

	gint retval = SM_OK;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if ((retval = compose_check_for_valid_recipient(msginfo->header)) != SM_OK) {
		return retval;
	}
	if ((retval = compose_check_for_valid_account(prefs_account)) != SM_OK) {
		return retval;
	}
	return retval;
}


/**
    This function writes the headers, message and attachment of a mail 
    into a file. 
  
    @param fp  File pointer where data to be written 
    @param MsgInfo  contains message information 
    @param attachment list  contains file attachments    
    @param composeinfo OUT filled with email receipant list 
    @param out_codeset  outgoing data char code set 
    @param encoding   encoding type 
    @return nothing
*/

void
compose_write_multipart_header(FILE * fp,
			       const MsgInfo * msginfo,
			       GSList * attach_list,
			       ComposeInfo * composeinfo,
			       const gchar * out_codeset, EncodingType encoding)
{
	if ((msginfo == NULL) || (fp == NULL) || (attach_list == NULL)) {
		return;
	}
	if (g_slist_length(attach_list) != 0) {
		// fputs ( "\nThis is a multi-part message in MIME format.\n", fp );
//      fprintf ( fp, "\n--%s\n", composeinfo->boundary );
		if ((msginfo->settings != NULL) && (msginfo->settings->type == EMAIL_HTML)) {
//           fprintf ( fp, "\n--%s1\n", composeinfo->boundary );
#if 0
			fprintf(fp, "Content-Type: text/html; charset=%s\n", out_codeset);
#endif
		} else {
#if 0
			fprintf(fp, "Content-Type: text/plain; charset=%s\n", out_codeset);
			fprintf(fp, "Content-Transfer-Encoding: %s\n",
				procmime_get_encoding_str(encoding));
			fputc('\n', fp);
#endif				// CHECK
		}
#if 0
		fprintf(fp, "Content-Transfer-Encoding: %s\n", procmime_get_encoding_str(encoding));
		fputc('\n', fp);
#endif
	}
	return;
}


/**
    This function writes the headers, message and attachment of a mail 
    into a file. 
  
    @param message IN  body of the message
    @param attachment list IN  contains file attachments    
    @param file IN  file name to store composed mail.    
    @param Prefs Account IN  contains account information  
    @param MsgInfo IN  contains msg header, attachment list etc.
    @param composeinfo OUT filled with email receipant list 
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint
compose_write_to_file(const gchar * message,
		      GSList * attach_list,
		      const gchar * file,
		      const PrefsAccount * prefs_account,
		      MsgInfo * msginfo, ComposeInfo * composeinfo)
{
	FILE *fp = NULL;
	size_t len = 0;
	gchar *chars = NULL;
	gchar *buf = NULL;
	gchar *canon_buf = NULL;
	const gchar *out_codeset = CS_UTF_8;
	const gchar *src_codeset = NULL;
	EncodingType encoding = ENC_7BIT;
	gint retval = SM_OK;
	gchar *outbuf = NULL;
	size_t outlen = 0;
	gint i = 0;
	gint l = 0;
	gchar *plain_text = NULL;
	gchar *plain_text_buf = NULL;
	gchar *tmp1 = NULL;
	gchar *tmp2 = NULL;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(prefs_account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);

	if (composeinfo->type != TRUE) {	/* compose as draft  */
		retval = compose_check_for_valid_account(prefs_account);
		if (retval != SM_OK) {
			debug_print(" [%s][%d] : Invalid account \n", __FUNCTION__, __LINE__);
			SMTP_SET_MSG_CB(retval,
					prefs_account->send_settings.smtp_server,
					prefs_account->user_settings.account_name);
			return retval;
		}
	}

	if ((fp = fopen(file, "wb")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		retval = SMTP_FILE_WRITE_ERROR;
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
		debug_print(" Errorcode(%d) : File Write Error ... \n", retval);
		return retval;
	}

	/* chmod for security */
	if (change_file_mode_rw(fp, file) < 0) {
		FILE_OP_ERROR(file, "chmod");
	}

	if (message != NULL) {
		/* get all composed text */
		chars = g_strdup(message);
		if (chars != NULL) {
			len = strlen(chars);
		}

		/* Check the src_codeset and out_codeset
		 * Determine the codeset of the characters and encoding scheme
		 */
		if (chars && is_ascii_str((guchar *) chars)) {
			buf = chars;
			chars = NULL;
			out_codeset = CS_UTF_8;
			encoding = ENC_7BIT;
		} else {
#if 0				/* This is puposely commented. Will be removed later */
			out_codeset = conv_get_outgoing_charset_str();
			if (!strcasecmp(out_codeset, CS_US_ASCII)) {
				out_codeset = CS_ISO_8859_1;
			}
			encoding = procmime_get_encoding_for_charset(out_codeset);
#endif
			out_codeset = CS_UTF_8;
			encoding = ENC_BASE64;

			/* Now determine the src_codeset */
			src_codeset = conv_get_current_charset_str();
			/* if current encoding is US-ASCII, set it the same as
			   outgoing one to prevent code conversion failure */
			if (!g_ascii_strcasecmp(src_codeset, CS_US_ASCII)) {
				src_codeset = out_codeset;
			}
			/* Convert the text in compose window from src_codeset 
			 * to out_codeset 
			 */
			buf = conv_codeset_strdup(chars, src_codeset, out_codeset);
			if (buf == NULL) {
				buf = chars;
				out_codeset = src_codeset;
				chars = NULL;
			}
		}

		if (chars) {
			g_free(chars);
		}
	}
	/* Now add \r\n at the end of the each line and last line also */
	if (buf) {
		canon_buf = canonicalize_str(buf);
		g_free(buf);
		buf = canon_buf;
	}

	/* write headers */
	if (compose_write_headers(fp, out_codeset,
				  encoding, prefs_account, msginfo, composeinfo) != SM_OK) {
		fclose(fp);
		if (unlink(file) != 0) {
			g_warning("unlink: failed to delete %s: %s", file, strerror(errno));
		}
		g_free(buf);
		retval = SMTP_WRITE_HEADERS_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	/* If the mail is having some attachment, 
	 * we need to write some more headers
	 */
	compose_write_multipart_header(fp, msginfo, attach_list,
				       composeinfo, out_codeset, encoding);

	/* Convert the contents of mail using the encoding scheme selected. */
	if (buf != NULL) {
		len = strlen(buf);
	}
	if (encoding == ENC_BASE64) {
		if (msginfo->settings->type == EMAIL_HTML) {
			if (html_to_text_convert_buffer(buf, out_codeset, &plain_text_buf) == TRUE) {
				tmp1 = base64_encode_buf(buf);
				tmp2 = base64_encode_buf(plain_text_buf);
				if (compose_write_html_mail(out_codeset, encoding,
							    tmp1, fp, file, retval, outlen,
							    tmp2) == FALSE) {
					g_free(buf);
					fclose(fp);
					g_free(tmp1);
					return SMTP_FILE_WRITE_ERROR;
				}
				g_free(tmp1);
			}
		} else {
			outbuf = g_malloc0(B64_BUFFSIZE + 1);
			if (outbuf == NULL) {
				fclose(fp);
				g_free(buf);
				SMTP_SET_MSG_CB(SMTP_INTERNAL_ERROR, NULL, NULL);
				return SMTP_INTERNAL_ERROR;
			}
			for (i = 0; i < len; i += B64_LINE_SIZE) {
				/*memset(&outbuf, '\0', B64_BUFFSIZE); */
				memset(outbuf, '\0', B64_BUFFSIZE);
				l = MIN(B64_LINE_SIZE, len - i);
				base64_encode(outbuf, (guchar *) (buf + i), l);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
			g_free(outbuf);
		}
	} else if (encoding == ENC_QUOTED_PRINTABLE) {
		if (len > 0) {
			outbuf = g_malloc0(len * 4);
			if (outbuf == NULL) {
				fclose(fp);
				g_free(buf);
				SMTP_SET_MSG_CB(SMTP_INTERNAL_ERROR, NULL, NULL);
				return SMTP_INTERNAL_ERROR;
			}
			qp_encode_line(outbuf, (guchar *) buf);
			outlen = strlen(outbuf);
			if (msginfo->settings->type == EMAIL_HTML) {
				/* composes multipart/alternative mail */

				if (html_to_text_convert_buffer(buf, out_codeset,
								&plain_text_buf) == TRUE) {
					qp_encode_line(plain_text_buf, (guchar *) plain_text_buf);
					if (compose_write_html_mail
					    (out_codeset, encoding, buf, fp, file, retval,
					     outlen, plain_text_buf) == FALSE) {
						g_free(buf);
						fclose(fp);
						return SMTP_FILE_WRITE_ERROR;
					}
				}
			} else {
				if (fwrite(outbuf, sizeof(gchar), outlen, fp) != outlen) {
					FILE_OP_ERROR(file, "fwrite");
					retval = SMTP_FILE_WRITE_ERROR;
					if (errno == ENOSPC) {
						send_engine_error_to_ui(errno);
					} else {
						SMTP_SET_MSG_CB(retval, NULL, NULL);
					}
					fclose(fp);
					unlink(file);
					g_free(outbuf);
					g_free(buf);
					return SMTP_FILE_WRITE_ERROR;
				}
			}
			g_free(outbuf);
		}
	} else {
		if (msginfo->settings->type == EMAIL_HTML) {
			/* composes multipart/alternative mail */
			if (html_to_text_convert_buffer(buf, out_codeset, &plain_text) == TRUE) {
				if (compose_write_html_mail(out_codeset, encoding,
							    buf, fp, file, retval,
							    outlen, plain_text) == FALSE) {
					fclose(fp);
					g_free(buf);
					return SMTP_FILE_WRITE_ERROR;
				}
			}
		} else {
			if (fwrite(buf, sizeof(gchar), len, fp) != len) {
				FILE_OP_ERROR(file, "fwrite");
				retval = SMTP_FILE_WRITE_ERROR;
				if (errno == ENOSPC) {
					send_engine_error_to_ui(errno);
				} else {
					SMTP_SET_MSG_CB(retval, NULL, NULL);
				}
				fclose(fp);
				unlink(file);
				g_free(outbuf);
				g_free(buf);
				return SMTP_FILE_WRITE_ERROR;
			}
		}
	}
	g_free(buf);

	/* Now write the inline images to file */
	if (g_slist_length(msginfo->image_list) != 0) {
		compose_write_inline_images(fp, msginfo, composeinfo);
	}
	/* Now write the contents of the attached files into the  file */
	if (g_slist_length(attach_list)) {
		compose_write_attach(fp, msginfo, msginfo->attach_list, composeinfo);
	}
	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		retval = SMTP_FILE_CLOSE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	/* Replace \r\n characters with \n and \0 */
	uncanonicalize_file_replace(file);

	return SM_OK;
}

/**
    This  function encodes the inline images and writes it into file 
    @param fp  file pointer to write the encoded data
    @param composeinfo contains boundary information 
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint compose_write_inline_images(FILE * fp, MsgInfo * msginfo, ComposeInfo * composeinfo)
{
	gchar *filepath = NULL;
	FILE *image_fp = NULL;
	EncodingType encoding;
	gint len = 0;
	gchar *attachmentname = NULL;
	gchar *content_type = NULL;
	GSList *cur = NULL;
	InlineImageInfo *imageInfo = NULL;

	debug_print("[%s:%d]:starts here :\n", __FUNCTION__, __LINE__);

	if (fp == NULL) {
		return SMTP_FILE_POINTER_NULL;
	}

	if (composeinfo == NULL) {
		return SMTP_INTERNAL_ERROR;
	}

	if ((msginfo == NULL) || (msginfo->image_list == NULL)) {
		return SMTP_ATTACH_LIST_NULL;
	}

	for (cur = msginfo->image_list; cur != NULL; cur = cur->next) {
		imageInfo = (InlineImageInfo *) cur->data;
		if ((imageInfo == NULL) || (imageInfo->image_path == NULL)) {
			continue;
		}
		filepath = imageInfo->image_path;

		if ((image_fp = fopen(filepath, "rb")) == NULL) {
			continue;
		}

		fprintf(fp, "\n--1%s\n", composeinfo->boundary);
		encoding = compose_get_attach_info(filepath, &attachmentname, &content_type);

		fprintf(fp, "Content-Type: %s;\n", content_type);
		fprintf(fp, "Content-Transfer-Encoding: %s\n", procmime_get_encoding_str(encoding));
		fprintf(fp, "Content-ID: %s\n", imageInfo->content_id);
		fprintf(fp, "Content-Location: %s\n", imageInfo->image_name);
		fprintf(fp, "Content-Disposition: inline; filename= \"%s\"\n\n",
			imageInfo->image_name);

		if (encoding == ENC_BASE64) {
			gchar inbuf[B64_LINE_SIZE], outbuf[B64_BUFFSIZE];
			while ((len = fread(inbuf, sizeof(gchar),
					    B64_LINE_SIZE, image_fp)) == B64_LINE_SIZE) {
				base64_encode(outbuf, (guchar *) inbuf, B64_LINE_SIZE);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
			if (len > 0 && feof(image_fp)) {
				base64_encode(outbuf, (guchar *) inbuf, len);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
		} else if (encoding == ENC_QUOTED_PRINTABLE) {
			gchar inbuf[BUFFSIZE], outbuf[BUFFSIZE * 4];
			while (fgets(inbuf, sizeof(inbuf), image_fp) != NULL) {
				qp_encode_line(outbuf, (guchar *) inbuf);
				fputs(outbuf, fp);
			}
		} else {
			gchar buf[BUFFSIZE];

			while (fgets(buf, sizeof(buf), image_fp) != NULL) {
				strcrchomp(buf);
				fputs(buf, fp);
			}
		}
		fclose(image_fp);
		g_free(attachmentname);
		g_free(content_type);
		attachmentname = NULL;
		filepath = NULL;
		content_type = NULL;
	}
	fprintf(fp, "\n--1%s--\n", composeinfo->boundary);
	return SM_OK;
}

/**
    This  function encodes the attachments and writes it into file 

    @param fp  file pointer to write the encoded data
    @param attachment list contains file attachments
    @param composeinfo contains boundary information 
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/


gint
compose_write_attach(FILE * fp, MsgInfo * msginfo, GSList * attach_list, ComposeInfo * composeinfo)
{
	gchar *filepath = NULL;
	FILE *attach_fp = NULL;
	gchar filename[BUFFSIZE];
	gint len = 0;
	EncodingType encoding;
	gchar *attachmentname = NULL, *content_type = NULL;
	GSList *cur = NULL;
	AttachInfo *attachinfo = NULL;

	debug_print("[%s][%s:%d]:starts here :\n", __FILE__, __FUNCTION__, __LINE__);
	if (fp == NULL) {
		return SMTP_FILE_POINTER_NULL;
	}
	if (attach_list == NULL) {
		return SMTP_ATTACH_LIST_NULL;
	}
	/* The loop is for each attached file
	 * Convert the contents of the file based on the encoding scheme
	 * and write it to the file. Attachinfo contains different types  
	 * of attachments and distingushed based on AttachInfo->type
	 */

	for (cur = attach_list; cur != NULL; cur = cur->next) {
		attachinfo = (AttachInfo *) cur->data;
		if (attachinfo == NULL) {
			continue;
		}
		filepath = NULL;
		attachmentname = NULL;
		switch (attachinfo->type) {
		case DIRECT:
		case ATT_VCARD:
			if (attachinfo->attachmentname != NULL) {
				filepath = g_strdup(attachinfo->filepath);
			}
			break;
		case EMAIL:
			if (attachinfo->email_path != NULL) {
				filepath = g_strdup(attachinfo->email_path);
			} else {
				filepath = folder_get_message_file_path(attachinfo->msgid);
			}
			if (msginfo->header->subject != NULL) {
				attachmentname = g_strdup(msginfo->header->subject);
			} else {
				attachmentname = g_strdup("No subject");
			}
			break;
		case MHTML:
			if (attachinfo->attachmentname != NULL) {
				filepath = g_strdup(attachinfo->filepath);
			}
			break;
		default:
			filepath = NULL;
			debug_print(" Invalid attachment \n");
			continue;
			break;
		}
		if ((filepath == NULL) || (attach_fp = fopen(filepath, "rb")) == NULL) {
			g_free(filepath);
			continue;
		}

		/* Message contains attachment */
		if ((msginfo != NULL) && (msginfo->header != NULL)) {
			msginfo->header->flags.perm_flags |= MSG_ATTACH;
		}
		fprintf(fp, "\n--%s\n", composeinfo->boundary);
		/* Content type is hardcoded because For email type 
		   content type should be  message/rfc822 */
		if ((attachinfo->type == EMAIL) || (attachinfo->type == MHTML)) {
			content_type = g_strdup("message/rfc822");
			encoding = ENC_7BIT;
		} else {
			encoding =
			    compose_get_attach_info(filepath, &attachmentname, &content_type);
		}
		attachmentname = g_strdup(attachinfo->attachmentname);
		compose_convert_header(filename, sizeof(filename), attachmentname, 12, FALSE);

		if (attachinfo->type == ATT_VCARD) {
			fprintf(fp, "Content-Type: %s; charset=\"UTF-8\"\n" " name=\"%s\"\n",
				content_type, filename);
		} else {
			fprintf(fp, "Content-Type: %s;\n" " name=\"%s\"\n", content_type, filename);
		}
		fprintf(fp, "Content-Disposition: attachment;\n" " filename=\"%s\"\n", filename);

		fprintf(fp, "Content-Transfer-Encoding: %s\n\n",
			procmime_get_encoding_str(encoding));

		if (encoding == ENC_BASE64) {
			gchar inbuf[B64_LINE_SIZE], outbuf[B64_BUFFSIZE];
			while ((len = fread(inbuf, sizeof(gchar),
					    B64_LINE_SIZE, attach_fp)) == B64_LINE_SIZE) {
				base64_encode(outbuf, (guchar *) inbuf, B64_LINE_SIZE);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
			if (len > 0 && feof(attach_fp)) {
				base64_encode(outbuf, (guchar *) inbuf, len);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
		} else if (encoding == ENC_QUOTED_PRINTABLE) {
			gchar inbuf[BUFFSIZE], outbuf[BUFFSIZE * 4];
			while (fgets(inbuf, sizeof(inbuf), attach_fp) != NULL) {
				qp_encode_line(outbuf, (guchar *) inbuf);
				fputs(outbuf, fp);
			}
		} else {
			gchar buf[BUFFSIZE];

			while (fgets(buf, sizeof(buf), attach_fp) != NULL) {
				strcrchomp(buf);
				fputs(buf, fp);
			}
		}
		fclose(attach_fp);
		/*If not draft mail then only unlink
		 * if draft mail UI unlinks the paths once the SAVE_COMPLETE
		 * callback comes*/
		if (composeinfo->type == FALSE)
			unlink(filepath);
		g_free(filepath);
		g_free(attachmentname);
		g_free(content_type);
		attachmentname = NULL;
		filepath = NULL;
		content_type = NULL;
	}
	fprintf(fp, "\n--%s--\n", composeinfo->boundary);
	return SM_OK;
}

/**
    This  function used to get Encoding type of attachment 

    @param filepath IN  Path of the atatchment 
    @param attachmentname  Name of the attachment 
    @param content_type IN  boundary information
    @return EncodingType( like ENC_BASE64 etc.) 
*/

EncodingType
compose_get_attach_info(const gchar * filepath, gchar ** attachmentname, gchar ** content_type)
{
	debug_print("[%s][%s:%d]:starts here :\n", __FILE__, __FUNCTION__, __LINE__);

	if (filepath == NULL) {
		return ENC_BASE64;
	}
	*attachmentname = g_strdup(g_basename(filepath));
	*content_type = (gchar *) gnome_vfs_get_mime_type(filepath);

	if (!*content_type) {
		return ENC_BASE64;
	} else if (!g_strncasecmp(*content_type, "text", 4)) {
		return procmime_get_encoding_for_file(filepath);
	} else {
		return ENC_BASE64;
	}
}

/**
   This function writes subject in to the file.  

   @param fp file pointer to write the encoded data
   @param Prefs Account   contains account information
   @param MsgInfo      contains msg header, attachment list etc.
   @param composeinfo  filled with email receipant list  
   @param aliasname    account contains alias name 
   @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint
compose_write_headers_receipant_list(FILE * fp,
				     const PrefsAccount * account,
				     MsgInfo * msginfo,
				     ComposeInfo * composeinfo, gchar ** aliasname)
{
	gchar *str = NULL;
	gchar buf[BUFFSIZE];
	gchar *name = NULL;

	IS_PARAM_NULL(fp, SMTP_FILE_POINTER_NULL, return SMTP_FILE_POINTER_NULL);
	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return SMTP_MSG_HDR_NULL);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);
	IS_PARAM_NULL(account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);

	memset(buf, '\0', BUFFSIZE);
	/* From */
	if (account->user_settings.alias_name != NULL) {
		compose_convert_header(buf, sizeof(buf),
				       account->user_settings.alias_name, strlen("From: "), TRUE);
		QUOTE_IF_REQUIRED(name, buf);

		if (aliasname != NULL) {
			*aliasname = g_strdup(name);
		}
		if ((account != NULL) && (account->user_settings.email != NULL)) {
			fprintf(fp, "From: %s <%s>\n", name, account->user_settings.email);
		}
	} else {
		if ((account != NULL) && (account->user_settings.email != NULL)) {
			fprintf(fp, "From: %s\n", account->user_settings.email);
		}
	}

	/* To */
	if (msginfo->header->to) {
		str = msginfo->header->to;
		PUT_RECIPIENT_HEADER("To", str);
	}

	/* Cc */
	if (msginfo->header->cc) {
		str = msginfo->header->cc;
		PUT_RECIPIENT_HEADER("Cc", str);
	}

	/* Bcc */
	if (msginfo->header->bcc) {
		str = msginfo->header->bcc;
		PUT_RECIPIENT_HEADER("Bcc", str);
	}

	if ((composeinfo->type != TRUE) && (composeinfo->to_list == NULL)) {
		g_free(*aliasname);
		return SMTP_RCPT_LIST_INVALID;
	}

	/* Copy to own Address */
	if ((msginfo->settings != NULL) && (msginfo->settings->copy_to_own_address == TRUE)) {
		if ((account != NULL) && (account->user_settings.email != NULL)) {
			composeinfo->to_list = address_list_append(composeinfo->to_list,
								   account->user_settings.email);
		}
	}
	return SM_OK;
}


/**
   This function writes subject in to the file.  

   @param fp IN  file pointer to write the encoded data
   @param MsgInfo      contains msg header, attachment list etc.
   @param buf    preallocated buffer of size BUFFSIZE 
   @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint compose_write_headers_subject(FILE * fp, MsgInfo * msginfo, gchar * buf)
{
	gchar *str = NULL;

	if ((fp == NULL) || (msginfo == NULL) || (msginfo->header == NULL)) {
		return -1;
	}

	str = msginfo->header->subject;
	if (str) {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose_convert_header(buf, BUFFSIZE, str, strlen("Subject: "), FALSE);
			fprintf(fp, "Subject: %s\n", buf);
		}
	}
	return SM_OK;
}


/**
   This function writes msginfo settings in to the file.  
  
   @param fp IN  file pointer to write the encoded data
   @param MsgInfo      contains msg header, attachment list etc.
   @param PrefsAccount contains account information
   @param aliasname    account contains alias name 
   @return nothing 
*/

void
compose_write_headers_msgsettings(FILE * fp,
				  MsgInfo * msginfo,
				  const PrefsAccount * account, const gchar * aliasname)
{
	if ((fp == NULL) || (msginfo == NULL) || (account == NULL)) {
		return;
	}

	if (msginfo->settings != NULL) {


		/* priority */
		if (msginfo->settings->priority == LOW) {
			fprintf(fp, "X-Priority: %d (Low)\n", msginfo->settings->priority);
		} else if (msginfo->settings->priority == HIGH) {
			fprintf(fp, "X-Priority: %d (High)\n", msginfo->settings->priority);
		} else {
			fprintf(fp, "X-Priority: %d (Normal)\n", NORMAL);
		}

		/* Read Receipt Request */
		if (msginfo->settings->read_receipt_request == TRUE) {
			if ((account != NULL) && (account->user_settings.email != NULL)) {
				if (aliasname != NULL) {
					fprintf(fp,
						"Disposition-Notification-To: %s <%s>\n",
						aliasname, account->user_settings.email);
				} else {
					fprintf(fp, "Disposition-Notification-To: %s\n",
						account->user_settings.email);
				}
				fprintf(fp, "Return-Receipt-To: %s\n",
					account->user_settings.email);
				fprintf(fp, "Return-Path: %s\n", account->user_settings.email);
			}
		}
	}
	return;
}

/**
   This function writes headers related to a message into a file

    @param fp   file pointer to write the encoded data
    @param charset   contains contains char set information 
    @param encoding encoding type 
    @param Prefs Account   contains account information
    @param MsgInfo   contains msg header, attachment list etc.
    @param composeinfo  filled with email receipant list  
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint
compose_write_headers(FILE * fp,
		      const gchar * charset,
		      EncodingType encoding,
		      const PrefsAccount * account, MsgInfo * msginfo, ComposeInfo * composeinfo)
{
	gchar buf[BUFFSIZE];
	gchar *str = NULL;
	gchar *alias = NULL;
	gint retval = SM_OK;

	debug_print("[%s]:starts here:\n", __FUNCTION__);

	IS_PARAM_NULL(fp, SMTP_FILE_POINTER_NULL, return SMTP_FILE_POINTER_NULL);
	IS_PARAM_NULL(charset, SMTP_CHARSET_NULL, return SMTP_CHARSET_NULL);
	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return SMTP_MSG_HDR_NULL);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);
	IS_PARAM_NULL(account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);

	if (composeinfo->type != TRUE) {	/* compose as draft  */
		if ((retval = compose_check_entries(msginfo, account)) != SM_OK) {
			return retval;
		}
	}
	memset(buf, '\0', BUFFSIZE);
	/* Date */
	compose_put_date(fp, msginfo);
	if (msginfo->header->date != NULL) {
		fprintf(fp, "Date: %s\n", msginfo->header->date);
	}
	/* To, Cc, Bcc */
	if ((retval = compose_write_headers_receipant_list(fp, account,
							   msginfo, composeinfo, &alias)) != SM_OK)
	{
		return retval;
	}

	/* Subject */
	compose_write_headers_subject(fp, msginfo, buf);

	/* Msg id */
	if ((composeinfo->type == TRUE) && (msginfo->header->msgid != NULL)) {
		compose_generate_msgid(msginfo, account, buf, sizeof(buf), TRUE);
	} else {
		compose_generate_msgid(msginfo, account, buf, sizeof(buf), FALSE);
	}

	fprintf(fp, "Message-Id: <%s>\n", buf);

	/* In-Reply-To */
	if (msginfo->header->inreplyto) {
		fprintf(fp, "In-Reply-To: <%s>\n", msginfo->header->inreplyto);
	}

	/* Reply-To */
	if (msginfo->header->replyto) {
		str = msginfo->header->replyto;
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose_convert_header(buf, sizeof(buf), str,
						       strlen("Reply-To: "), TRUE);
				fprintf(fp, "Reply-To: %s\n", buf);
			}
		}
	}

	/* Program version and system info */
	/* uname(&utsbuf); */
	if (composeinfo->to_list) {
		fprintf(fp, "X-Mailer: %s\n", PROG_VERSION);
	}

	/* This feild is used in saving draft mail */
	if (composeinfo->type == TRUE) {	/* draft mail  */
		fprintf(fp, "X-draft-edit: 1 \n");
	}

	if ((msginfo->settings) && (msginfo->settings->copy_to_own_address == TRUE)) {
		fprintf(fp, "X-Copy-to-own: %d\n", msginfo->settings->copy_to_own_address);
	}
	compose_write_headers_msgsettings(fp, msginfo, account, alias);

	/* MIME */
	fprintf(fp, "Mime-Version: 1.0\n");
	if (msginfo->attach_list != NULL) {
		composeinfo->boundary = generate_mime_boundary(NULL);
		fprintf(fp, "Content-Type: multipart/mixed; boundary=\"%s\"\n\n",
			composeinfo->boundary);
		fprintf(fp, "--%s\n", composeinfo->boundary);
	}
	if (g_slist_length(msginfo->image_list) != 0) {
		if (composeinfo->boundary == NULL) {
			composeinfo->boundary = generate_mime_boundary(NULL);
		}
		fprintf(fp, "Content-Type: multipart/related;\n"
			" boundary=\"1%s\"\n\n", composeinfo->boundary);
		fprintf(fp, "--1%s\n", composeinfo->boundary);
	} else {
		if ((msginfo->settings != NULL) && (msginfo->settings->type == EMAIL_HTML)) {
			/*
			   fprintf ( fp, "Content-Type: text/html; charset=%s\n", charset );
			 */
#if 0
			composeinfo->boundary = generate_mime_boundary(NULL);
			fprintf(fp,
				"Content-Type: multipart/alternative;\n boundary= %s;\n\n\n",
				composeinfo->boundary);
			fprintf(fp, "Content-Transfer-Encoding: %s",
				procmime_get_encoding_str(encoding));
#endif
		} else {
			fprintf(fp, "Content-Type: text/plain; charset=%s\n", charset);
			fprintf(fp, "Content-Transfer-Encoding: %s\n\n",
				procmime_get_encoding_str(encoding));
		}
	}


	/* separator between header and body */
#if 0
	fputs("\n", fp);
#endif

	g_free(alias);
	return SM_OK;
}

/**
   This function user to encode the information 

    @param dest    Destination buffer to save encoded info 
    @param len     length of Destination Buffer 
    @param src     Source buffer of the date to be encoded
    @param header_len Header length
    @param addr_field  TRUE or FALSE
    @return none 
*/

void
compose_convert_header(gchar * dest, gint len, gchar * src, gint header_len, gboolean addr_field)
{
	debug_print("[%s][%s]:starts here :\n", __FILE__, __FUNCTION__);
	if ((src == NULL) || (dest == NULL)) {
		return;
	}
	if (len < 1) {
		return;
	}
	memset(dest, 0, len);

	g_strchomp(src);
	conv_encode_header(dest, len, src, header_len, addr_field);
}

/**
    This function generates the New Message ID 

    @param msginfo    contains the message information 
    @param buf        New message id to be stored  
    @param len        Len of buffer 
    @param draft      TRUE if mail is draft mail
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint
compose_generate_msgid(MsgInfo * msginfo, const PrefsAccount * account,
		       gchar * buf, const gint len, gboolean draft)
{
	time_t t;
	gint mlen = 0;
	gchar *addr = NULL;
	gchar *id = NULL;
	gchar *msg_id = NULL;

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	t = time(NULL);
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	memset(buf, 0, len);

	if (draft == FALSE) {
		/* msgid needs to be changed based on request type */
		if ((msginfo->settings != NULL) &&
		    (msginfo->settings->options == SEND_NEXT_CONNECTION)) {
			mlen = g_snprintf(buf, len, "%1d%u%04x%1d",
					  SEND_NEXT_CONNECTION, (guint) t,
					  (guint) random(), OUTBOX_WAITING);
		} else if ((msginfo->settings != NULL)
			   && (msginfo->settings->options == SEND_ON_REQUEST)) {
			mlen = g_snprintf(buf, len, "%1d%u%04x%1d",
					  SEND_ON_REQUEST, (guint) t,
					  (guint) random(), OUTBOX_ON_REQUEST);
		} else {	/* Default is to send immediate */
			mlen = g_snprintf(buf, len, "%1d%u%04x%1d",
					  SEND_IMMEDIATE, (guint) t, (guint) random(),
					  OUTBOX_WAITING);
		}
		if (msginfo->header->msgid != NULL) {
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = NULL;
		}
		msginfo->header->msgid = g_strdup_printf("%s%c%s%c%s",
							 account->user_settings.account_name,
							 MSGID_SEPARATOR, OUTBOX_DIR,
							 MSGID_SEPARATOR, buf);
		debug_print("Message Id is for file  %s", buf);
	}

	else {
		msg_id = g_strdup(msginfo->header->msgid);
		id = strrchr(msg_id, MSGID_SEPARATOR);
		if (id != NULL) {
			id++;
		}
		mlen = g_snprintf(buf, len, "%s", id);
	}

	if ((account->user_settings.email != NULL)) {
		if (strchr(account->user_settings.email, '@')) {
			addr = g_strdup_printf("%s", account->user_settings.email);
		} else {
			addr = g_strdup_printf("%s@%s", account->user_settings.account_name,
					       get_domain_name());
		}
	} else {
		addr = g_strdup_printf("%s@%s", account->user_settings.account_name,
				       get_domain_name());
	}
	g_snprintf(buf + mlen, len - mlen, ".%d.%s", random(), addr);
	g_free(addr);
	g_free(msg_id);
	return SM_OK;
}

/**
   This function used to move message from src to dest

   @param src    Source file
   @param des    Destnation file
   @return 0 on success  or -1  on fail 
*/

gint compose_move_msg(const gchar * src, const gchar * dest)
{
	debug_print("[%s][%s:%d]:starts here :\n", __FILE__, __FUNCTION__, __LINE__);
	return move_file(src, dest, FALSE);
}

/**
   This function used to compute current asc time and store in msginfo struct

   @param fp         File pointer
   @param msginfo    contains the mail information
   @return none 
*/

void compose_put_date(FILE * fp, MsgInfo * msginfo)
{
	gchar buf[BUFFSIZE];

	struct tm *lt = NULL;
	time_t t;
	gchar day[4], mon[4];
	gint dd, hh, mm, ss, yyyy;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if ((msginfo == NULL) || (msginfo->header == NULL)) {
		debug_print(" Msg Info Invalid \n");
		return;
	}

	memset(buf, '\0', BUFFSIZE);
	memset(day, '\0', sizeof(day));
	memset(mon, '\0', sizeof(mon));

	t = time(NULL);
	lt = localtime(&t);

	sscanf(asctime(lt), "%3s %3s %d %d:%d:%d %d\n", day, mon, &dd, &hh, &mm, &ss, &yyyy);
	g_snprintf(buf, sizeof(buf), "%s, %d %s %d %02d:%02d:%02d %s",
		   day, dd, mon, yyyy, hh, mm, ss, tzoffset(&t));

	if (msginfo->header->date != NULL) {
		g_free(msginfo->header->date);
	}
	msginfo->header->date = g_strdup(buf);
	msginfo->header->date_t = t;
	msginfo->header->mtime = t;
}

/**
   This function used to free the compose Info struct

   @param composeinfo   contain email receipant list
   @return none 
*/

void compose_free_ComposeInfoStruct(ComposeInfo * composeinfo)
{
	if (composeinfo == NULL) {
		return;
	}
	slist_free_strings(composeinfo->to_list);
	g_slist_free(composeinfo->to_list);
	composeinfo->to_list = NULL;
	g_free(composeinfo->boundary);
	g_free(composeinfo);
	composeinfo = NULL;
}

/**
    This function is used to recompose mail resend mail and add resend 
    feilds to sent mail
    @param file   file name to store composed mail.
    @param sent_filename   file name to store composed mail.
    @param Prefs Account  contains account information
    @param MsgInfo contains msg header, attachment list etc
    @return SM_OK on success (or) SMTP modle errorcode on failure
*/

gint
compose_resend_mail(const gchar * file,
		    const gchar * sent_filename, const PrefsAccount * account, MsgInfo * msginfo)
{
	FILE *fp = NULL;
	FILE *send_fp = NULL;
	gchar *name = NULL;
	gchar *str = NULL;
	ComposeInfo *composeinfo = NULL;
	gint retval = SM_OK;
	gchar buf[BUFFSIZE];
	gchar *buf_msgid = NULL;

	debug_print("[%s][%d] : starts here :\n", __FILE__, __LINE__);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return SMTP_MSG_HDR_NULL);
	IS_PARAM_NULL(account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);
	IS_PARAM_NULL(sent_filename, SMTP_FILE_NAME_NULL, return SMTP_FILE_NAME_NULL);

	if ((retval = compose_check_for_valid_account(account)) != SM_OK) {
		debug_print(" [%s][%d] : Invalid account \n", __FUNCTION__, __LINE__);
		SMTP_PARAM_VALUE_MISSING_CB(retval);
		return retval;
	}

	if ((send_fp = fopen(sent_filename, "rb")) == NULL) {
		retval = SMTP_FILE_OPEN_ERROR;
		debug_print(" Errorcode(%d) : File Read Error \n", retval);
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	if ((fp = fopen(file, "wb")) == NULL) {
		retval = SMTP_FILE_WRITE_ERROR;
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
		debug_print(" Errorcode(%d) : File Write Error\n", retval);
		fclose(send_fp);
		return retval;
	}

	composeinfo = g_new0(ComposeInfo, 1);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);

	/* chmod for security */
	if (change_file_mode_rw(fp, file) < 0) {
		FILE_OP_ERROR(file, "chmod");
	}
	/* Resent-From */
	memset(buf, '\0', BUFFSIZE);
	if ((account->user_settings.alias_name) && (*account->user_settings.alias_name)) {
		compose_convert_header(buf, sizeof(buf),
				       account->user_settings.alias_name,
				       strlen("Resent-From: "), TRUE);
		QUOTE_IF_REQUIRED(name, buf);
		if ((account != NULL) && (account->user_settings.email != NULL)) {
			fprintf(fp, "Resent-From: %s <%s>\n", name, account->user_settings.email);
		}
	} else {
		if ((account != NULL) && (account->user_settings.email != NULL)) {
			fprintf(fp, "Resent-From: %s\n", account->user_settings.email);
		}
	}

	/* Resent-To */
	if (msginfo->header->to) {
		str = msginfo->header->to;
		PUT_RECIPIENT_HEADER("Resent-To", str);
	}

	/* Cc */
	if (msginfo->header->cc) {
		str = msginfo->header->cc;
		PUT_RECIPIENT_HEADER("Resent-Cc", str);
	}

	/* Bcc */
	if (msginfo->header->bcc) {
		str = msginfo->header->bcc;
		PUT_RECIPIENT_HEADER("Resent-Bcc", str);
	}

	/* Resent-Date */
	compose_put_date(fp, msginfo);
	if (msginfo->header->date != NULL) {
		fprintf(fp, "Resent-Date: %s\n", msginfo->header->date);
	}
	/* Resent-Msgid */
	memset(buf, '\0', sizeof(buf));
	compose_generate_msgid(msginfo, account, buf, sizeof(buf), FALSE);
	fprintf(fp, "Resent-Message-Id: <%s>\n", buf);
	buf_msgid = g_strdup_printf("Message-Id: <%s>\n", buf);

	/* Add the Original mail */
	memset(buf, '\0', sizeof(buf));
	while (fgets(buf, sizeof(buf), send_fp) != NULL) {
		if (!g_strncasecmp(buf, "Message-Id:", 11)) {
			strcrchomp(buf_msgid);
			fputs(buf_msgid, fp);
		} else {
			strcrchomp(buf);
			fputs(buf, fp);
			memset(buf, '\0', sizeof(buf));
		}
	}
	g_free(buf_msgid);
	fclose(send_fp);
	fclose(fp);
	compose_free_ComposeInfoStruct(composeinfo);
	return SM_OK;
}

gboolean
compose_write_html_mail(const gchar * out_codeset,
			EncodingType encoding, const gchar * buf, FILE * fp,
			const gchar * file, gint retval, gint outlen, gchar * plain_text)
{
	gchar *html_text = NULL;
	gint plain_text_len = 0;
	gchar *mime_boundary = NULL;
	mime_boundary = generate_mime_boundary(NULL);
	fprintf(fp, "Content-Type: multipart/alternative;\n boundary=\"%s\"\n\n\n", mime_boundary);
	html_text = g_strdup(buf);
	fprintf(fp, "--%s\n", mime_boundary);
	fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n", out_codeset);
	fprintf(fp, "Content-Transfer-Encoding: %s\n\n", procmime_get_encoding_str(encoding));
	plain_text_len = strlen(plain_text);
	if (fwrite(plain_text, sizeof(gchar), plain_text_len, fp) != plain_text_len) {
		FILE_OP_ERROR(file, "fwrite");
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
		unlink(file);
		g_free(html_text);
		g_free(plain_text);
		g_free(mime_boundary);
		return FALSE;
	}
	fprintf(fp, "\n\n--%s\n", mime_boundary);
	fprintf(fp, "Content-Type: text/html; charset=\"%s\"\n", out_codeset);
	fprintf(fp, "Content-Transfer-Encoding: %s\n\n", procmime_get_encoding_str(encoding));
	outlen = strlen(html_text);
	if (fwrite(html_text, sizeof(gchar), outlen, fp) != outlen) {
		FILE_OP_ERROR(file, "fwrite");
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
		unlink(file);
		g_free(html_text);
		g_free(plain_text);
		g_free(mime_boundary);
		return FALSE;
	}
	fprintf(fp, "\n--%s--\n", mime_boundary);
	g_free(mime_boundary);
	g_free(html_text);
	g_free(plain_text);

	return TRUE;
}

gchar *base64_encode_buf(const gchar * buf)
{
	gchar *outbuf = NULL;
	gchar *retbuf = NULL;
	gint i = 0;
	gint len = 0;
	gint l = 0;
	gchar *tmp = NULL;
	if (buf == NULL) {
		return NULL;
	}
	outbuf = g_malloc0(B64_BUFFSIZE + 1);
	len = strlen(buf);
	for (i = 0; i < len; i += B64_LINE_SIZE) {
		memset(outbuf, '\0', B64_BUFFSIZE);
		l = MIN(B64_LINE_SIZE, len - i);
		base64_encode(outbuf, (guchar *) (buf + i), l);
		if (retbuf == NULL) {
			retbuf = g_strdup_printf("%s\n", outbuf);
		} else {
			tmp = g_strdup_printf("%s%s\n", retbuf, outbuf);
			g_free(retbuf);
			retbuf = g_strdup(tmp);
			g_free(tmp);
		}
	}
	g_free(outbuf);
	return retbuf;
}
