
/**
 * This file is part of osso-email-engine
 *
 * Copyright (c) 2006 Nokia Corporation.
 * 
 * Contact: Dirk-Jan Binnema <dirk-jan.binnema@nokia.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "defs.h"

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>

#if (HAVE_WCTYPE_H && HAVE_WCHAR_H)
#  include <wchar.h>
#  include <wctype.h>
#endif

#include "quoted-printable.h"
#include "codeconv.h"
#include "utils.h"
#include "common.h"
#include "prefs.h"
#include "smtp_errorcodes.h"
#include "prefs_account.h"
#include "base64.h"
#include "smtp.h"
#include "folder.h"
#include "smime_security.h"
#include "compose_smime.h"
#include "htmltotextconv.h"


/**
This function writes the headers, message and attachment of a mail into a file. 
This function also makes a call to S/MIME functions in case encryption, 
signing or both is required before sending the message.

@param message IN  body of the message
@param attachment list IN  contains file attachments    
@param file IN  file name to store composed mail.    
@param Prefs Account IN  contains account information  
@param MsgInfo IN  contains msg header, attachment list etc.
@param composeinfo OUT filled with email receipant list 
@return 0 on success (or) positive errorcode on failure
*/

gint compose_write_smime_to_file(const gchar * message,
				 GSList * attach_list,
				 const gchar * file,
				 const PrefsAccount * prefs_account,
				 MsgInfo * msginfo, ComposeInfo * composeinfo)
{
	FILE *fp = NULL;
	FILE *filep = NULL;
	FILE *soutfilep = NULL;
	gchar sfile[BUFFSIZE + 1];
	gchar smime_output[BUFFSIZE + 1];
	size_t len = 0;
	gchar *chars = NULL;
	gchar *buf = NULL;
	gchar *canon_buf = NULL;
	const gchar *out_codeset = CS_UTF_8;
	const gchar *src_codeset = NULL;
	EncodingType encoding = ENC_7BIT;
	gchar *outbuf = NULL;
	size_t outlen = 0;
	gint retval = SM_OK;
	gchar in_buffer_smime[BUFFSIZE];
	gint i = 0;
	gint l = 0;
	gchar *plain_text = NULL;
	gchar *plain_text_buf = NULL;
	gchar *tmp1 = NULL;
	gchar *tmp2 = NULL;


	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(prefs_account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);

	if (composeinfo->type != TRUE) {	/* compose as draft  */
		retval = compose_check_for_valid_account(prefs_account);
		if (retval != SM_OK) {
			debug_print(" [%s][%d] : Invalid account \n", __FUNCTION__, __LINE__);
			SMTP_PARAM_VALUE_MISSING_CB(retval);
			return retval;
		}
	}

	if ((fp = fopen(file, "wb")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		retval = SMTP_FILE_WRITE_ERROR;
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval,
					prefs_account->send_settings.smtp_server,
					prefs_account->user_settings.account_name);
		}
		debug_print(" Errorcode(%d) : File Write Error ... \n", retval);
		return retval;
	}

	if (!(get_temp_smime_file(sfile, BUFFSIZE + 1))) {
		fclose(fp);
		retval = SMTP_INTERNAL_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}
	unlink(sfile);

	if ((filep = fopen(sfile, "wb")) == NULL) {
		FILE_OP_ERROR(sfile, "fopen");
		retval = SMTP_FILE_WRITE_ERROR;
		if (errno == ENOSPC) {
			send_engine_error_to_ui(errno);
		} else {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
		fclose(fp);
		debug_print(" Errorcode(%d) : File Write Error ... \n", retval);
		return retval;
	}

	/* chmod for security */
	if (change_file_mode_rw(fp, file) < 0) {
		FILE_OP_ERROR(file, "chmod");
	}

	if (change_file_mode_rw(filep, sfile) < 0) {
		FILE_OP_ERROR(sfile, "chmod");
	}

	if (message != NULL) {
		/* get all composed text */
		chars = g_strdup(message);
		if (chars != NULL) {
			len = strlen(chars);
		}

		/* Check the src_codeset and out_codeset
		 * Determine the codeset of the characters and encoding scheme
		 */
		if (chars && is_ascii_str((guchar *) chars)) {
			buf = chars;
			chars = NULL;
			out_codeset = CS_UTF_8;
			encoding = ENC_7BIT;
		} else {
			out_codeset = CS_UTF_8;
			encoding = ENC_BASE64;

			/* Now determine the src_codeset */
			src_codeset = conv_get_current_charset_str();
			/* if current encoding is US-ASCII, set it the same as
			   outgoing one to prevent code conversion failure */
			if (!g_ascii_strcasecmp(src_codeset, CS_US_ASCII)) {
				src_codeset = out_codeset;
			}
			/* Convert the text in compose window from src_codeset
			 * to out_codeset
			 */
			buf = conv_codeset_strdup(chars, src_codeset, out_codeset);
			if (buf == NULL) {
				buf = chars;
				out_codeset = src_codeset;
				chars = NULL;
			}
		}

		if (chars) {
			g_free(chars);
		}
	}
	/* Now add \r\n at the end of the each line and last line also */
	if (buf) {
		canon_buf = canonicalize_str(buf);
		g_free(buf);
		buf = canon_buf;
	}

	/* write headers */
	if (compose_write_headers_for_smime(fp, filep, out_codeset,
					    encoding, prefs_account, msginfo,
					    composeinfo) != SM_OK) {


		fclose(fp);
		fclose(filep);
		unlink(sfile);
		unlink(file);
		g_free(buf);
		retval = SMTP_WRITE_HEADERS_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	/* If the mail is having some attachment,
	 * we need to write some more headers
	 */
	compose_write_multipart_header(filep, msginfo, attach_list,
				       composeinfo, out_codeset, encoding);

	/* Convert the contents of mail using the encoding scheme selected. */
	if (buf != NULL) {
		len = strlen(buf);
	}
	if (encoding == ENC_BASE64) {
		if (msginfo->settings->type == EMAIL_HTML) {
			if (html_to_text_convert_buffer(buf, out_codeset, &plain_text_buf) == TRUE) {
				tmp1 = base64_encode_buf(buf);
				tmp2 = base64_encode_buf(plain_text_buf);
				if (compose_write_html_mail(out_codeset, encoding,
							    tmp1, filep, file, retval, outlen,
							    tmp2) == FALSE) {
					g_free(buf);
					fclose(fp);
					fclose(filep);
					g_free(tmp1);
					return SMTP_FILE_WRITE_ERROR;
				}
				g_free(tmp1);
			}
		} else {
			outbuf = g_malloc0(B64_BUFFSIZE + 1);
			if (outbuf == NULL) {
				fclose(filep);
				fclose(fp);
				g_free(buf);
				SMTP_SET_MSG_CB(SMTP_INTERNAL_ERROR, NULL, NULL);
				return SMTP_INTERNAL_ERROR;
			}
			for (i = 0; i < len; i += B64_LINE_SIZE) {
				memset(outbuf, '\0', B64_BUFFSIZE);
				l = MIN(B64_LINE_SIZE, len - i);
				base64_encode(outbuf, (guchar *) (buf + i), l);
				fputs(outbuf, filep);
				fputc('\n', filep);
			}
			g_free(outbuf);
		}
	} else if (encoding == ENC_QUOTED_PRINTABLE) {
		if (len > 0) {
			outbuf = g_malloc0(len * 4);
			if (outbuf == NULL) {
				fclose(filep);
				fclose(fp);
				g_free(buf);
				SMTP_SET_MSG_CB(SMTP_INTERNAL_ERROR, NULL, NULL);
				return SMTP_INTERNAL_ERROR;
			}
			qp_encode_line(outbuf, (guchar *) buf);
			outlen = strlen(outbuf);
			if (msginfo->settings->type == EMAIL_HTML) {
				/* composes multipart/alternative mail */

				if (html_to_text_convert_buffer(buf, out_codeset,
								&plain_text) == TRUE) {
					qp_encode_line(plain_text_buf, (guchar *) plain_text_buf);
					if (compose_write_html_mail(out_codeset, encoding,
								    buf, filep, file, retval,
								    outlen,
								    plain_text_buf) == FALSE) {
						g_free(buf);
						fclose(fp);
						return SMTP_FILE_WRITE_ERROR;
					}
				}
			} else {

				if (fwrite(outbuf, sizeof(gchar), outlen, filep) != outlen) {
					FILE_OP_ERROR(sfile, "fwrite");
					errno = SMTP_FILE_WRITE_ERROR;
					if (errno == ENOSPC) {
						send_engine_error_to_ui(errno);
					} else {
						SMTP_SET_MSG_CB(retval, NULL, NULL);
					}
					fclose(filep);
					fclose(fp);
					unlink(sfile);
					unlink(file);
					g_free(outbuf);
					g_free(buf);
					return SMTP_FILE_WRITE_ERROR;
				}
			}
			g_free(outbuf);
		}
	} else {
		if (msginfo->settings->type == EMAIL_HTML) {
			/* composes multipart/alternative mail */
			if (html_to_text_convert_buffer(buf, out_codeset, &plain_text) == TRUE) {
				if (compose_write_html_mail(out_codeset, encoding,
							    buf, filep, file, retval,
							    outlen, plain_text) == FALSE) {
					fclose(fp);
					fclose(filep);
					g_free(buf);
					return SMTP_FILE_WRITE_ERROR;
				}
			}
		} else {
			if (fwrite(buf, sizeof(gchar), len, filep) != len) {
				FILE_OP_ERROR(file, "fwrite");
				retval = SMTP_FILE_WRITE_ERROR;
				if (errno == ENOSPC) {
					send_engine_error_to_ui(errno);
				} else {
					SMTP_SET_MSG_CB(retval, NULL, NULL);
				}
				fclose(filep);
				fclose(fp);
				unlink(sfile);
				unlink(file);
				g_free(buf);
				return SMTP_FILE_WRITE_ERROR;
			}
		}
	}
	g_free(buf);

	/* Now write the inline images to file */
	if (g_slist_length(msginfo->image_list) != 0) {
		compose_write_inline_images(filep, msginfo, composeinfo);
	}

	/* Now write the contents of the attached files into the  file */
	if (g_slist_length(attach_list)) {
		compose_write_attach(filep, msginfo, msginfo->attach_list, composeinfo);
	}

	if (!(get_temp_smime_file(smime_output, BUFFSIZE + 1))) {
		fclose(fp);
		fclose(filep);
		retval = SMTP_INTERNAL_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	unlink(smime_output);

	if (fclose(filep) == EOF) {
		FILE_OP_ERROR(sfile, "fclose");
		unlink(sfile);
		fclose(fp);
		unlink(file);
		retval = SMTP_FILE_CLOSE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	switch (msginfo->header->smime_type) {

	case SMIME_SIGN:
		if ((retval = (sign_message(sfile, smime_output, msginfo->recip_certs_uid,
					    msginfo->password)))) {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
			return retval;
		}
		break;

	case SMIME_ENCRYPT:
		if ((retval = (encrypt_message(msginfo->recip_certs_uid,
					       msginfo->recip_algos, smime_output, sfile)))) {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
			return retval;
		}
		break;

	case SMIME_SIGN_ENCRYPT:
		if ((retval = (sign_encrypt(msginfo->recip_certs_uid, msginfo->recip_algos,
					    sfile, smime_output, msginfo->password)))) {
			SMTP_SET_MSG_CB(retval, NULL, NULL);
			return retval;
		}
		break;

	default:
		return retval;
		break;
	}

	if ((soutfilep = fopen(smime_output, "r")) == NULL) {
		FILE_OP_ERROR(smime_output, "fopen");
		debug_print(" Errorcode(%d) : File Write Error ... \n", retval);
		retval = SMTP_FILE_WRITE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	if (change_file_mode_rw(soutfilep, smime_output) < 0) {
		FILE_OP_ERROR(file, "chmod");
	}

	/*copy contents of signed mail into a file with RFC822 headers */

	memset(in_buffer_smime, '\0', sizeof(in_buffer_smime));
	while (fgets(in_buffer_smime, sizeof(in_buffer_smime), soutfilep) != NULL) {
		strcrchomp(in_buffer_smime);
		fputs(in_buffer_smime, fp);
		memset(in_buffer_smime, '\0', sizeof(in_buffer_smime));
	}

	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		fclose(soutfilep);
		retval = SMTP_FILE_CLOSE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	if (fclose(soutfilep) == EOF) {
		FILE_OP_ERROR(smime_output, "fclose");
		unlink(smime_output);
		retval = SMTP_FILE_CLOSE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}
	unlink(sfile);
	unlink(smime_output);

	/* Replace \r\n characters with \n and \0 */
	uncanonicalize_file_replace(file);

	return SM_OK;
}

/**
   This function writes headers related to a message into a file

    @param fp   file pointer to write the encoded data  
    @param charset   contains contains char set information
    @param encoding encoding type
    @param Prefs Account   contains account information
    @param MsgInfo   contains msg header, attachment list etc.
    @param composeinfo  filled with email receipant list
    @return 0 on success (or) positive errorcode on failure
*/

gint
compose_write_headers_for_smime(FILE * fp, FILE * filep,
				const gchar * charset,
				EncodingType encoding,
				const PrefsAccount * account,
				MsgInfo * msginfo, ComposeInfo * composeinfo)
{
	gchar buf[BUFFSIZE];
	gchar *str = NULL;
	gchar *alias = NULL;
	gint retval = SM_OK;

	debug_print("[%s]:starts here:\n", __FUNCTION__);

	IS_PARAM_NULL(fp, SMTP_FILE_POINTER_NULL, return SMTP_FILE_POINTER_NULL);
	IS_PARAM_NULL(filep, SMTP_FILE_POINTER_NULL, return SMTP_FILE_POINTER_NULL);
	IS_PARAM_NULL(charset, SMTP_CHARSET_NULL, return SMTP_CHARSET_NULL);
	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return SMTP_MSG_HDR_NULL);
	IS_PARAM_NULL(composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);
	IS_PARAM_NULL(account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);

	if (composeinfo->type != TRUE) {	/* compose as draft  */
		if ((retval = compose_check_entries(msginfo, account)) != SM_OK) {
			return retval;
		}
	}

	memset(buf, '\0', BUFFSIZE);
	/* Date */
	compose_put_date(fp, msginfo);
	if (msginfo->header->date != NULL) {
		fprintf(fp, "Date: %s\n", msginfo->header->date);
	}
	/* To, Cc, Bcc */
	if ((retval =
	     compose_write_headers_receipant_list(fp, account, msginfo, composeinfo,
						  &alias)) != SM_OK) {
		return retval;
	}

	/* Subject */
	compose_write_headers_subject(fp, msginfo, buf);

	/* Msg id */
	compose_generate_msgid(msginfo, account, buf, sizeof(buf), FALSE);
	fprintf(fp, "Message-Id: <%s>\n", buf);

	/* In-Reply-To */
	if (msginfo->header->inreplyto) {
		fprintf(fp, "In-Reply-To: <%s>\n", msginfo->header->inreplyto);
	}

	/* Reply-To */
	if (msginfo->header->replyto) {
		str = msginfo->header->replyto;
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose_convert_header(buf, sizeof(buf), str,
						       strlen("Reply-To: "), TRUE);
				fprintf(fp, "Reply-To: %s\n", buf);
			}
		}
	}

	/* Program version and system info */
	/* uname(&utsbuf); */
	if (composeinfo->to_list) {
		fprintf(fp, "X-Mailer: %s\n", PROG_VERSION);
	}

	/* This feild is used in saving draft mail */
	if (composeinfo->type == TRUE) {	/* draft mail  */
		fprintf(fp, "X-draft-edit: 1 \n");
	}

	if ((msginfo->settings) && (msginfo->settings->copy_to_own_address == TRUE)) {
		fprintf(fp, "X-Copy-to-own: %d\n", msginfo->settings->copy_to_own_address);
	}
	if (msginfo->attach_list != NULL) {
		composeinfo->boundary = generate_mime_boundary(NULL);
		fprintf(filep, "Content-Type: multipart/mixed;boundary=\"%s\"\n\n",
			composeinfo->boundary);
		fprintf(filep, "--%s\n", composeinfo->boundary);
	}
	if (g_slist_length(msginfo->image_list) != 0) {
		if (composeinfo->boundary == NULL) {
			composeinfo->boundary = generate_mime_boundary(NULL);
		}
		fprintf(filep, "Content-Type: multipart/related;\n"
			" boundary=\"1%s\"\n\n", composeinfo->boundary);
		fprintf(filep, "--1%s\n", composeinfo->boundary);
	} else {
		if ((msginfo->settings != NULL) && (msginfo->settings->type == EMAIL_HTML)) {
			debug_print("Dont put anything\n");
		} else {
			fprintf(filep, "Content-Type: text/plain; charset=%s\n", charset);
			fprintf(filep, "Content-Transfer-Encoding: %s\n",
				procmime_get_encoding_str(encoding));
			/* separator between header and body */
			fputs("\n", filep);
		}
	}

	compose_write_headers_msgsettings(fp, msginfo, account, alias);


	g_free(alias);
	return SM_OK;
}
