
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999-2006 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "send_message.h"
#include "compose.h"
#include "session.h"
#include "smtp.h"
#include "common.h"
#include "prefs_account.h"
#include "utils.h"
#include "callbacks.h"
#include "smtp_errorcodes.h"
#include "account.h"
#include "folder.h"
#include "compose_smime.h"
#include "engine_errno.h"

#if USE_SSL
#include "ssl.h"
extern gint ssl_error;		/* defined in ssl.c file */
#endif

#ifdef CL_USE
#include "cl.h"
#include "iap_confserv.h"
#endif


extern gint engine_errno;

static gint
send_message(const gchar * file,
	     const PrefsAccount * ac_prefs, GSList * to_list, gboolean display_progress);

static gint send_message_smtp(const PrefsAccount * ac_prefs, GSList * to_list, FILE * fp);

static gint send_recv_message(Session * session);

static gint
send_send_data_progressive(Session * session, const guint cur_len, const guint total_len);

static gint send_send_data_finished(Session * session, const guint len);

static void send_read_unread_update(const gchar * msgidpath, gboolean flag);

static PrefsAccount *send_get_prefs_account(const gchar * account);

static gboolean send_is_folder_match(const gchar * msgid, const gchar * foldername);

static gchar *send_newmsgid_with_outbox_options(const MsgInfo * msginfo);

static gint send_get_outbox_mails_count(const GSList * msglist, const gchar option);

static gint send_get_tmpfile(gchar * tmp, gint tmplen, const MsgInfo * msginfo);

static gint
smtp_compose_mail(const gchar * message,
		  MsgInfo * msginfo,
		  const PrefsAccount * prefs_account,
		  gchar * tmpfile, ComposeInfo ** composeinfo, gboolean composeType);

static gint
send_get_outbox_mails_list(const gchar * account,
			   const gchar * msgid, GSList ** msglist, gint * outboxcount);

static gint send_get_outboxmail_status(const gchar * msgid);

static gint send_get_msgid_file_size(gchar * msgid);

static gint send_get_msglist_file_size(const GSList * msgid_list);

static gboolean send_message_get_smtp_config_data(const PrefsAccount * account, Session * session);
static gchar *send_message_get_smtp_config_servername(const PrefsAccount * account);

/**
  This function changes the status of outbox mails to waiting/sending as
  case may be. 
  @param msgid - Message ID
  @param option - Whether outbox mail should be OUTBOX_WAITING/OUTBOX_SENDING
  @return gchar* - New message ID after status changes
*/
static gchar *send_change_outbox_status(gchar * msgid, const gchar option);

static gint send_get_sending_status(const gchar * msgid);

/**
  This function validates the to_list for the correct mail id form like xx@yy.z
  @param to_list to list of the mail
  @param file mail file
  
  @return TRUE on valid list, FALSE on invalid list
*/
static gboolean send_validate_to_list(const GSList * to_list, MsgHeader * header);

static SMTPSession *smtpsession = NULL;
gboolean smtpconnection_alive = FALSE;
static gchar *cur_send_msgid = NULL;
static GSList *open_msgid_list = NULL;

/* boolean to indicate if message has been sent or not */
static gboolean sent_message = FALSE;

struct _send_wa {
	long total_messages_size;
	long current_messages_size;
	gint cancel_status;
};

static struct _send_wa send_wa;

/** 
    This function called to create a temp file name in which composed 
	   mail stored temporarily. 

    @param tmp  file name is generated by this function 
    @param tmplen len of the tmp buffer 
    @param MsgInfo Message info structure contains the info of the mail 
    @return TRUE on success (or) FALSE on failure  
*/

static gint send_get_tmpfile(gchar * tmp, gint tmplen, const MsgInfo * msginfo)
{
	if ((tmp == NULL) || (msginfo == NULL)) {
		return FALSE;
	}
	/* Create a temporary file to write to compose the message */
	memset(tmp, '\0', tmplen);
	g_snprintf(tmp, tmplen - 1, "%s%ctmpmsg.%08x",
		   get_tmp_dir(), G_DIR_SEPARATOR, (gint) msginfo);

	debug_print("Temporary file to write msgs : %s \n", tmp);
	return TRUE;
}

/** 
    This function called to compose the mail for the given message information  

    @param message   body of the message 
    @param msginfo   contains msg header,attachment list,email settings  
    @param prefs_account  user account information 
    @param tmpfile  user account information 
    @param composeinfo  which is populated with mail receipant list 
    @param composeType  TRUE indicates it is  Draft mail. 
    @return SM_OK on success (or) smtp errorcode on failure  
*/
static gint smtp_compose_mail(const gchar * message,
			      MsgInfo * msginfo,
			      const PrefsAccount * prefs_account,
			      gchar * tmpfile, ComposeInfo ** composeinfo, gboolean composeType)
{
	gint retval = SM_OK;
	*composeinfo = g_new0(ComposeInfo, 1);
	IS_PARAM_NULL(*composeinfo, SMTP_INTERNAL_ERROR, return SMTP_INTERNAL_ERROR);
	(*composeinfo)->type = composeType;
	if (msginfo->header->smime_type != SMIME_NONE) {
		if ((retval = compose_write_smime_to_file(message,
							  msginfo->attach_list,
							  tmpfile,
							  prefs_account,
							  msginfo, *composeinfo)) != SM_OK) {
			debug_print(" Error : Fail to compose smime message \n");
			compose_free_ComposeInfoStruct(*composeinfo);
		}
	} else {
		if ((retval = compose_write_to_file(message,
						    msginfo->attach_list,
						    tmpfile,
						    prefs_account,
						    msginfo, *composeinfo)) != SM_OK) {
			debug_print(" Error : Fail to compose message \n");
			compose_free_ComposeInfoStruct(*composeinfo);
		}
	}
	return retval;
}

/** 
    This function called by Abstract layer to send a mail to SMTP server. 
    Using the compose functionality, it first composes the mail and then 
    sends the composed file to server. 

    @param message  body of the message 
    @param msginfo contains msg header,attachment list,email settings  
    @param prefs_account user account information 
    @return SM_OK on success (or) smtp errorcode on failure  
*/

gint send_mail(const gchar * message, MsgInfo * msginfo, PrefsAccount * prefs_account)
{
	gint retval = SM_OK;
	gchar tmp[MAXPATHLEN + 1];
	gchar *outpath = NULL;
	gchar *outpath_fail = NULL;
	gchar *sentpath = NULL;
	ComposeInfo *composeinfo = NULL;
	gchar *outbox_id = NULL;
	gchar *msg_id = NULL;
	smtpconnection_alive = FALSE;
	gint mailcount = 1;
	gboolean display_progress = TRUE;
	gchar *old_id = NULL;
	gchar *oldmsgid = NULL;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(prefs_account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);

	CHECK_ENTRIES;

	send_wa.cancel_status = 0;
	if (msginfo->header->msgid)
		old_id = g_strdup(msginfo->header->msgid);

	send_get_tmpfile(tmp, MAXPATHLEN + 1, msginfo);
	if ((retval = smtp_compose_mail(message, msginfo,
					prefs_account, tmp, &composeinfo, FALSE)) != SM_OK) {
		g_free(old_id);
		return retval;
	}

	msg_id = g_strdup(msginfo->header->msgid);

	outbox_id = send_newmsgid_with_outbox_options(msginfo);

	outpath = get_new_path(OUTBOX_DIR, prefs_account->user_settings.account_name, outbox_id);

	if (compose_move_msg(tmp, outpath) != SM_OK) {
		retval = SMTP_COPYTO_OUT_FOLDER_ERROR;
		debug_print("Error(%d):Mail move to outfolder Fail\n", retval);
		SMTP_SET_MSG_CB(retval, NULL, NULL);
	} else {
		debug_print("Mail Stored in out folder: %s \n", outpath);
		/* Update the status to UI that it has been moved to OUTBOX */
		g_free(msginfo->header->msgid);
		msginfo->header->msgid = g_strdup(outbox_id);

		msginfo->header->recv_priority = msginfo->settings->priority;
		msginfo->header->read_receipt_request = msginfo->settings->read_receipt_request;
		msginfo->header->copy_to_own_address = msginfo->settings->copy_to_own_address;
		msginfo->header->from = g_strdup(prefs_account->user_settings.email);
		msginfo->header->smime_type = msginfo->settings->smime_type;
		msginfo->header->msgtag.recvtype = prefs_account->receive_settings.recv_type;

		/*set the unread flag only if not opened from outbox */
		if (!old_id || !send_is_folder_match(old_id, OUTBOX_DIR)) {
			msginfo->header->flags.perm_flags |= MSG_UNREAD;
			send_read_unread_update(outpath, TRUE);
		} else if (is_message_unread(old_id)) {
			/*add the new msgid to the unread list */
			folder_message_del_list(old_id, UNREAD);
			folder_message_add_list(msginfo->header->msgid, UNREAD);
		}

		SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);
		/*send the callback only if the message is not opened from outbox */
		if (!old_id || !send_is_folder_match(old_id, OUTBOX_DIR))
			SMTP_SET_MSG_CB(SMTP_MAIL_MOVED_TO_OUTBOX_FOR_SENDING,
					msginfo->header->msgid, NULL);

		if (msginfo->settings != NULL) {
			if ((msginfo->settings->options == SEND_NEXT_CONNECTION) ||
			    (msginfo->settings->options == SEND_ON_REQUEST)) {
				goto endhere;
			} else {
				/* Store the current send msgid */
				cur_send_msgid = outbox_id;
				SMTP_UPDATE_OUTBOX_MAIL_COUNT(mailcount, mailcount);
				ACCNT_SET_FLAGS(prefs_account, ACCNT_SEND);

				send_wa.total_messages_size = get_file_size(outpath);
				send_wa.current_messages_size = 0;

				retval =
				    send_message(outpath, prefs_account,
						 composeinfo->to_list, display_progress);
				display_progress = FALSE;

				ACCNT_RESET_FLAGS(prefs_account, ACCNT_SEND);

				if ((retval != SM_OK) && (is_file_exist(outpath) == TRUE)) {
					g_free(outbox_id);
					if (retval == SMTP_CANCEL_MAIL) {
						outbox_id =
						    g_strdup_printf("%s%1d", msg_id,
								    OUTBOX_SUSPENDED);
					} else {
						outbox_id =
						    g_strdup_printf("%s%1d", msg_id, OUTBOX_FAILED);
					}

					outpath_fail =
					    get_new_path(OUTBOX_DIR,
							 prefs_account->user_settings.
							 account_name, outbox_id);

					if (compose_move_msg(outpath, outpath_fail) == SM_OK) {
						SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE,
									   FALSE);
						oldmsgid = msginfo->header->msgid;
						msginfo->header->msgid = g_strdup(outbox_id);
						SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE,
									   TRUE);
						/*add the new msgid to the unread list */
						if (is_message_unread(oldmsgid)) {
							folder_message_del_list(oldmsgid, UNREAD);
							folder_message_add_list(outbox_id, UNREAD);
						}
						g_free(oldmsgid);
					} else {
						send_read_unread_update(outpath, FALSE);
						send_read_unread_update(outpath_fail, TRUE);
						retval = SMTP_COPYTO_OUT_FOLDER_ERROR;
						SMTP_SET_MSG_CB(SMTP_COPYTO_OUT_FOLDER_ERROR,
								NULL, NULL);
					}
					debug_print(" Send Mail Response code  %d \n", retval);
				}
			}
		}
	}
	if (retval == SM_OK) {
		/* Move the file to sent items */
		send_read_unread_update(outpath, FALSE);
		sentpath = get_new_path(SENT_DIR,
					prefs_account->user_settings.account_name, msg_id);
		if (compose_move_msg(outpath, sentpath) != SM_OK) {
			debug_print(" Error : Mail moving to outfolder Fail\n");
			retval = SMTP_COPYTO_SENT_FOLDER_ERROR;
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		} else {
			/* update out box */
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
			/*update sent folder */
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = folder_get_msgid_from_path(sentpath);
			/*there is not concept of read/unread mails in sent folder */
			MSG_UNSET_FLAGS(msginfo->header->flags.perm_flags, MSG_UNREAD);
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);
			SMTP_SET_MSG_CB(SM_OK, NULL, NULL);
		}
		g_free(sentpath);
		sentpath = NULL;
	}

	SMTP_SEND_COMPLETE_CB();
	SMTP_SET_MSG_CB(SMTP_CHECK_FOR_OUT_MSGS, NULL, NULL);
      endhere:
	g_free(outpath);
	g_free(outpath_fail);
	compose_free_ComposeInfoStruct(composeinfo);
	g_free(msg_id);
	g_free(outbox_id);
	g_free(old_id);
	debug_print(" Send_mail_complete \n");
	return retval;
}

/**
    This function called by adaption layer to save a draft  mail.
    Using the compose functionality, it first composes the mail and then
    saves or sends the composed file to server.

    @param message  body of the message
    @param msginfo  contains msg header,attachment list,email settings
    @param prefs_account  user account information
    @param sendMail TRUE on send , FALSE on save
    @param old_msgid Message ID if the message is saving again
    @return SM_OK on success (or) smtp errorcode on failure
*/
gint
send_save_draft_mail(const gchar * message,
		     MsgInfo * msginfo,
		     const PrefsAccount * prefs_account, gboolean sendMail, gchar * old_id)
{
	gint retval = SM_OK;
	gchar tmp[MAXPATHLEN + 1];
	gchar *outpath = NULL;
	gchar *savepath = NULL;
	Folder *folder = NULL;
	ComposeInfo *composeinfo = NULL;
	gboolean display_progress = TRUE;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return SMTP_MESSAGE_INFO_NULL);
	IS_PARAM_NULL(msginfo->header, SMTP_MSG_HDR_NULL, return SMTP_MSG_HDR_NULL);
	IS_PARAM_NULL(prefs_account, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);
	smtpconnection_alive = FALSE;

	if (sendMail == TRUE) {
		CHECK_ENTRIES;
	}

	if (old_id != NULL && !msginfo->header->msgid) {
		folder = folder_get_msg_folder(old_id);
		/* Bug fix 4619 : Check added for NULL folder */
		if ((folder != NULL) && (strcmp(folder->name, ARCHIVE) == 0)) {
			folder = NULL;
		}
	} else if (msginfo->header->msgid) {	/*fix for 16141 */

		folder = folder_get_msg_folder(msginfo->header->msgid);
		/* Bug fix 4619 : Check added for NULL folder */
		if ((folder != NULL) && (strcmp(folder->name, ARCHIVE) == 0)) {
			folder = NULL;
		}
	}


	send_get_tmpfile(tmp, MAXPATHLEN + 1, msginfo);

	if ((retval = smtp_compose_mail(message, msginfo,
					prefs_account, tmp, &composeinfo, TRUE)) != SM_OK) {
		return retval;
	}

	if (msginfo->header->msgid != NULL) {
		savepath = g_strdup_printf("%s%c%s%c%s", get_mail_dir(),
					   G_DIR_SEPARATOR,
					   folder != NULL ? folder->name : DRAFT_DIR,
					   G_DIR_SEPARATOR, msginfo->header->msgid);
	}

	if (compose_move_msg(tmp, savepath) != SM_OK) {
		retval = SMTP_COPYTO_DRAFT_FOLDER_ERROR;
		debug_print("Errorcode(%d):Mail Storing in draft folder Fail\n", retval);
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		g_free(msginfo->header->msgid);
		msginfo->header->msgid = NULL;
		goto end_here;
	} else {
		send_read_unread_update(savepath, TRUE);
		if (old_id != NULL && folder != NULL) {
			/* if old_id is not NULL we must remove the old message. During this
			 * the number of messages in folder counter gets decremented but
			 * we should also increment since a new message is saved in draft
			 */
			if (folder_remove_msg(folder->name, old_id) == FOLDER_SUCCESS) {
				folder_increment_msg(folder->name);
				folder_message_del_list(old_id, UNREAD);
			}
		} else {
			folder_increment_msg(DRAFT_DIR);
		}

		if (sendMail == TRUE) {	/* Send Saved  Mail */
			/* Move Msg to outpath */
			outpath = get_new_path(OUTBOX_DIR,
					       prefs_account->user_settings.
					       account_name, msginfo->header->msgid);
			if (compose_move_msg(savepath, outpath) == SM_OK) {
				debug_print(" Mesage moved into outpath ...\n");
				if ((prefs_account) &&
				    (msginfo->settings->options == SEND_IMMEDIATE)) {
					retval =
					    send_outbox_mail(prefs_account->user_settings.
							     account_name,
							     msginfo->header->msgid, FALSE, 1, 1,
							     display_progress);
					display_progress = FALSE;
					send_read_unread_update(savepath, FALSE);
					send_read_unread_update(outpath, TRUE);
					SMTP_SEND_COMPLETE_CB();
				}
			}
		}
		SMTP_SET_MSG_CB(SMTP_SAVE_COMPLETE, msginfo->header->msgid, old_id);
	}

      end_here:
	g_free(savepath);
	g_free(outpath);
	compose_free_ComposeInfoStruct(composeinfo);
	debug_print(" Save message is complete \n");
	return retval;
}


/**
    This function called by adaption layer to send mails in the
    outbox folder.

    @param account  account name
    @param msgid    message id
    @param keepconnalive TRUE on connection alive , FALSE on 
     no connection alive
    @return SM_OK on success (or) smtp errorcode on failure
*/

gint
send_outbox_mail(const gchar * account,
		 const gchar * msgid,
		 gboolean keepconnalive,
		 gint mailsInOutBoxSent, gint mailsInOutBoxExist, gboolean display_progress)
{

	MsgInfo *msginfo = NULL;
	GSList *msginfo_list = NULL;
	GSList *cur = NULL;
	GSList *to_list = NULL;
	gchar *outpath = NULL;
	gchar *sentpath = NULL;
	gchar *newmsgid = NULL, *oldmsgid = NULL;
	gchar *newmsgid_path = NULL;
	PrefsAccount *prefs_account = NULL;
	gint retval = SM_OK, retval1 = SM_OK;
	gchar *new_state_id = NULL;
	gboolean no_connection = FALSE;
	gint send_status = 0;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if ((prefs_account = send_get_prefs_account(account)) == NULL) {
		return SMTP_ACCOUNT_NULL;
	}
	smtpconnection_alive = keepconnalive;

	if (msgid == NULL) {
		mailsInOutBoxSent = 0;
	}
	retval = send_get_outbox_mails_list(account, msgid, &msginfo_list, &mailsInOutBoxExist);
	if (retval != SM_OK) {
		procmsg_msginfo_list_free(msginfo_list);
		return retval;
	}

	for (cur = msginfo_list; cur != NULL; cur = cur->next) {

		msginfo = (MsgInfo *) cur->data;
		if ((msginfo == NULL) || (msginfo->header == NULL) ||
		    (msginfo->header->msgid == NULL)) {
			continue;
		}

		/* In case of mail is opened  for edit mode don't send it */
		if (send_editor_open_mail(msginfo->header->msgid, EDITOR_OPEN_MSG_EXISTS) == TRUE) {
			continue;
		}

		if (msgid == NULL) {
			/* Check added for whether mail is suspended/on request */
			retval = send_get_outboxmail_status(msginfo->header->msgid);
			send_status = send_get_sending_status(msginfo->header->msgid);
			if ((retval == OUTBOX_SUSPENDED) ||
			    (retval == OUTBOX_ON_REQUEST) ||
			    ((retval == OUTBOX_FAILED) && (send_status == SEND_ON_REQUEST))) {
				continue;
			}
			mailsInOutBoxSent++;
		}

		outpath = folder_get_message_file_path(msginfo->header->msgid);

		if (msginfo->header->to) {
			to_list = address_list_append(to_list, msginfo->header->to);
		}
		if (msginfo->header->cc) {
			to_list = address_list_append(to_list, msginfo->header->cc);
		}
		if (msginfo->header->bcc) {
			to_list = address_list_append(to_list, msginfo->header->bcc);
		}
		if (send_validate_to_list(to_list, msginfo->header) == FALSE) {
			engine_errno = ENGINE_RCPT_LIST_INVALID;
			continue;
		}
		/* Copy to own Address filed retriived from stored mail */
		if (msginfo->header->copy_to_own_address == TRUE) {
			if (prefs_account->user_settings.email != NULL) {
				to_list = address_list_append(to_list,
							      prefs_account->user_settings.email);
			}
		}
		/* Store the current send msgid */
		cur_send_msgid = msginfo->header->msgid;
		SMTP_UPDATE_OUTBOX_MAIL_COUNT(mailsInOutBoxExist, mailsInOutBoxSent);
		ACCNT_SET_FLAGS(prefs_account, ACCNT_SEND);
		if (procmime_check_attachments(outpath, FALSE) == TRUE) {
			msginfo->header->flags.perm_flags |= MSG_ATTACH;
		}

		if (send_wa.cancel_status != CANCEL) {
			/* Change mail status in outbox to sending */
			new_state_id = send_change_outbox_status(cur_send_msgid,
								 OUTBOX_SENDING | 0x30);
			if (new_state_id == NULL) {
				continue;
			}
			g_free(outpath);
			outpath = folder_get_message_file_path(g_strdup(new_state_id));
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(new_state_id);
			retval = send_message(outpath, prefs_account, to_list, display_progress);
			if (retval == SMTP_SERVER_INVALID || retval == SMTP_USERID_INVALID) {
				engine_errno = retval;
				/* BUG FIX 11311 */
				no_connection = TRUE;
			}
			if (engine_errno == ENGINE_ERROR_IAP_CONNECT || retval == SMTP_CANCEL_MAIL) {

				no_connection = TRUE;
			}
		} else {
			retval = SMTP_CANCEL_MAIL;
		}
		send_wa.current_messages_size += get_file_size(outpath);

		ACCNT_RESET_FLAGS(prefs_account, ACCNT_SEND);

		if (retval == SM_OK) {
			/* Move the file to sent items */
			sentpath = get_new_path(SENT_DIR,
						prefs_account->user_settings.
						account_name, msginfo->header->msgid);
			if (sentpath == NULL) {
				g_free(outpath);
				slist_free_strings(to_list);
				g_slist_free(to_list);
				to_list = NULL;
				continue;
			}

			if (compose_move_msg(outpath, sentpath) != SM_OK) {
				debug_print("Mail moving to sent folder Fail..\n");
				retval = SMTP_COPYTO_SENT_FOLDER_ERROR;
				SMTP_SET_MSG_CB(retval, NULL, NULL);
			} else {
				send_read_unread_update(outpath, FALSE);
				/* Notify to UI thatmsgid  moved out from OUTBOX to SENT ITEMS */
				SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
				/*update sent folder */
				g_free(msginfo->header->msgid);
				msginfo->header->msgid = folder_get_msgid_from_path(sentpath);
				msginfo->header->msgtag.recvtype =
				    prefs_account->receive_settings.recv_type;
				/*there is not concept of read/unread mails in sent folder */
				MSG_UNSET_FLAGS(msginfo->header->flags.perm_flags, MSG_UNREAD);
				SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);
			}
		} else {
			retval1 = send_get_outboxmail_status(msginfo->header->msgid);
			if (retval == SMTP_USERID_INVALID && retval1 == OUTBOX_SENDING) {
				newmsgid =
				    g_strdup_printf("%s%1d", msginfo->header->msgid,
						    OUTBOX_WAITING);
			} else if (send_wa.cancel_status == CANCEL) {
				newmsgid = g_strdup_printf("%s%1d", msginfo->header->msgid,
							   OUTBOX_SUSPENDED);
			} else if (retval1 == OUTBOX_WAITING) {
				newmsgid = g_strdup_printf("%s%1d", msginfo->header->msgid,
							   OUTBOX_ON_REQUEST);
			} else if (retval1 == OUTBOX_ON_REQUEST) {
				newmsgid = g_strdup_printf("%s%1d", msginfo->header->msgid,
							   OUTBOX_FAILED);
			} else if (retval1 == OUTBOX_SENDING) {
				newmsgid = g_strdup_printf("%s%1d", msginfo->header->msgid,
							   OUTBOX_FAILED);
			}

			if (newmsgid != NULL) {
				newmsgid_path = get_new_path(OUTBOX_DIR, account, newmsgid);
				if (compose_move_msg(outpath, newmsgid_path) == SM_OK) {
					SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
					oldmsgid = msginfo->header->msgid;
					msginfo->header->msgid = g_strdup(newmsgid);
					SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, TRUE);
					/*add the new msgid to the unread list */
					if (is_message_unread(oldmsgid)) {
						folder_message_del_list(oldmsgid, UNREAD);
						folder_message_add_list(newmsgid, UNREAD);
					}
					g_free(oldmsgid);
				}
				g_free(newmsgid);
				g_free(newmsgid_path);
			}
		}
		slist_free_strings(to_list);
		g_slist_free(to_list);
		g_free(outpath);
		g_free(sentpath);
		to_list = NULL;
		if (no_connection) {
			procmsg_msginfo_list_free(msginfo_list);
			if(!msgid) {
				SMTP_SEND_COMPLETE_CB();
			}
			return SMTP_CANCEL_MAIL;
		}

	}			/* end of for loop */
	procmsg_msginfo_list_free(msginfo_list);
	return SM_OK;
}

/**
    This function called by adaption layer to send a draft mails.
    @param msgid_list   list of message ids to be send
    @return SM_OK on success (or) smtp errorcode on failure
*/

gint send_draft_mails(const GSList * msgid_list)
{
	GSList *cur = NULL;
	gchar *outpath = NULL;
	gchar *inpath = NULL;
	gchar *account_name = NULL;
	gchar *msgid = NULL;
	gchar *mid = NULL;
	gchar *newmsgid = NULL;
	MsgInfo *msginfo = NULL;
	gint retval = SM_OK;
	gint msgidCount = 0;
	gint idx = 1;
	gboolean display_progress = TRUE;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (msgid_list == NULL) {
		return SMTP_SEND_ERROR;
	}

	msgidCount = g_slist_length((GSList *) msgid_list);
	send_wa.total_messages_size = send_get_msglist_file_size(msgid_list);
	send_wa.current_messages_size = 0;
	send_wa.cancel_status = 0;

	for (cur = (GSList *) msgid_list; cur != NULL; cur = cur->next, idx++) {
		msgid = (gchar *) cur->data;
		if (msgid == NULL) {
			continue;
		}
		/* Check for Valid message */
		if (send_is_folder_match(msgid, DRAFT_DIR) == FALSE) {
			continue;
		}
		/* Message is valid then proceed */
		account_name = get_account_name(msgid);
		if (account_name == NULL) {
			continue;
		}

		inpath = get_new_path(DRAFT_DIR, account_name, msgid);
		outpath = get_new_path(OUTBOX_DIR, account_name, msgid);

		msginfo = (MsgInfo *) folder_parse_msg(inpath);
		if (!msginfo || !(msginfo->header)) {
			g_free(inpath);
			g_free(outpath);
			procmsg_msginfo_free(msginfo);
			continue;
		}
		g_free(msginfo->header->msgid);
		msginfo->header->msgid = g_strdup(msgid);
		if (compose_move_msg(inpath, outpath) != SM_OK) {
			retval = SMTP_COPYTO_OUT_FOLDER_ERROR;
			debug_print("Errorcode(%d):Mail moving to outfolder Fail\n", retval);
			if (is_file_exist(inpath))
				SMTP_SET_MSG_CB(retval, NULL, NULL);
			procmsg_msginfo_free(msginfo);
		} else {
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
			procmsg_msginfo_free(msginfo);

			mid = strrchr(msgid, MSGID_SEPARATOR);
			mid++;
			newmsgid = g_strdup_printf("%s%c%s%c%s", account_name,
						   MSGID_SEPARATOR, OUTBOX_DIR,
						   MSGID_SEPARATOR, mid);
			if (newmsgid != NULL) {
				msginfo = (MsgInfo *) folder_parse_msg(outpath);
				if (msginfo && (msginfo->header)) {
					g_free(msginfo->header->msgid);
					msginfo->header->msgid = g_strdup(newmsgid);
					SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);
					procmsg_msginfo_free(msginfo);
					/* Send Mails with immediate option */
					if (mid[0] == (SEND_IMMEDIATE | 0x30)) {
						send_outbox_mail(account_name, newmsgid, 0, idx,
								 msgidCount, display_progress);
						display_progress = FALSE;
					}
				}
			}
			g_free(newmsgid);
		}
		g_free(account_name);
		g_free(inpath);
		g_free(outpath);
	}
	SMTP_SEND_COMPLETE_CB();
	if (engine_errno != ENGINE_RCPT_LIST_INVALID) {
		SMTP_SET_MSG_CB(SMTP_CHECK_FOR_OUT_MSGS, NULL, NULL);
	} else {
		engine_errno = 0;
	}
	return SM_OK;
}

/**
    This function called by adaption layer to  selected outbox mails. 

    @param msgid_list list of message ids selected
    @return SM_OK on success (or) smtp errorcode on failure
*/
gint send_outbox_mails(const GSList * msgid_list)
{
	gint msgidCount = 0;
	gchar *account = NULL, *prev_account = NULL;
	GSList *cur = NULL;
	gint idx = 1;
	GSList *newmsgid_list = NULL;
	gchar *msg_id = NULL;
	GSList *msg = NULL;
	gint count = 1;
	gboolean display_progress = TRUE, waiting_msg_changed = FALSE;
	gint retval = SM_OK, returned_error = 0;

	if (msgid_list == NULL) {
		return SMTP_SEND_ERROR;
	}

	msgidCount = g_slist_length((GSList *) msgid_list);
	send_wa.total_messages_size = send_get_msglist_file_size(msgid_list);
	send_wa.current_messages_size = 0;
	send_wa.cancel_status = 0;

	for (msg = (GSList *) msgid_list; msg != NULL; msg = msg->next, count++) {
		if (msg->data == NULL) {
			continue;
		}
		/* changing status of outbox mails to waiting */
		msg_id = send_change_outbox_status(msg->data, OUTBOX_WAITING | 0x30);
		if (msg_id != NULL) {
			newmsgid_list = g_slist_append(newmsgid_list, msg_id);
		} else {
			continue;
		}

	}
	for (cur = (GSList *) newmsgid_list; cur != NULL;idx++) {
		if (cur->data == NULL) {
			continue;
		}
		account = get_account_name(cur->data);
		prev_account = account;
		retval = send_outbox_mail(account, cur->data, 0, idx, msgidCount, display_progress);
		returned_error = engine_errno;
		display_progress = FALSE;
		if (retval == SM_OK) {
			g_free(account);
			cur = cur->next;
			continue;
		}
		if (returned_error == SMTP_SERVER_INVALID || returned_error == SMTP_USERID_INVALID) {
			/*Checking whether messages belongs to the same account,in which the error ocuured.
			   If so, then the messages status is changed to FAILED.
			   If the error is user invalid,then status is changed to waiting and once the username 
			   is entered the messages are resent */
			while (cur && cur->data
			       && (strcmp(prev_account, (account = get_account_name(cur->data))) ==
				   0)) {
				if (returned_error == SMTP_SERVER_INVALID) {
					msg_id =
					    send_change_outbox_status(cur->data,
								      OUTBOX_FAILED | 0x30);
					g_free(msg_id);
				}
				g_free(account);
				account = NULL;
				cur = cur->next;
			}
			g_free(account);
			engine_errno = 0;
			//if msg belongs to a different account,then send it.
			continue;
		} else if (retval == SMTP_MSG_HDR_NULL){
			waiting_msg_changed = TRUE;
			cur = cur->next;
			g_free(account);
			account = NULL;
			continue;
		} else if (engine_errno == 0) {
			g_free(account);
			break;
		}
		cur = cur->next;
	}
	SMTP_SEND_COMPLETE_CB();
	if ((retval == SM_OK && engine_errno != ENGINE_RCPT_LIST_INVALID) || waiting_msg_changed) {
		SMTP_SET_MSG_CB(SMTP_CHECK_FOR_OUT_MSGS, NULL, NULL);
	} else {
		engine_errno = 0;
		if (retval != SM_OK) {
#if 0				/*need not chnage the state of mails which are in waiting. */
			for (msg = (GSList *) cur; msg && msg->data; msg = msg->next) {
				/* changing status of outbox mails to failed */
				msg_id = send_change_outbox_status(msg->data, OUTBOX_FAILED | 0x30);
			}
#endif
		}
	}
	return SM_OK;
}

static
gchar *send_change_outbox_status(gchar * msgid, const gchar option)
{
	gchar *newmsgid = NULL;
	gchar *msgid_path = NULL;
	gchar *account_name = NULL;
	gchar *newmsgid_path = NULL;
	Folder *msgfolder = NULL;
	MsgInfo *msginfo = NULL;


	if (msgid == NULL) {
		return NULL;
	}

	/* If the message is already in that state return same message ID */
	if (msgid[strlen(msgid) - 1] == option)
		return g_strdup(msgid);

	msgfolder = folder_get_msg_folder(msgid);
	if (msgfolder == NULL) {
		return NULL;
	}

	if (strncmp(OUTBOX_DIR, msgfolder->name, strlen(OUTBOX_DIR)) != 0) {
		return NULL;
	}

	account_name = get_account_name(msgid);
	msgid_path = folder_get_message_file_path(msgid);
	msginfo = (MsgInfo *) folder_parse_msg(msgid_path);
	if (!msginfo || !(msginfo->header)) {
		procmsg_msginfo_free(msginfo);
		g_free(msgid_path);
		g_free(account_name);
		return NULL;
	}

	newmsgid = g_strdup(msgid);
	newmsgid[strlen(newmsgid) - 1] = option;

	newmsgid_path = get_new_path(OUTBOX_DIR, account_name, newmsgid);

	if (compose_move_msg(msgid_path, newmsgid_path) != SM_OK) {
		if (is_file_exist(msgid_path))
			SMTP_SET_MSG_CB(SMTP_COPYTO_OUT_FOLDER_ERROR, NULL, NULL);
		procmsg_msginfo_free(msginfo);
		g_free(msgid_path);
		g_free(newmsgid_path);
		g_free(account_name);
		g_free(newmsgid);
		return NULL;

	}
	g_free(msginfo->header->msgid);
	msginfo->header->msgid = g_strdup(msgid);

	/* Remove the message with old message ID from UI */
	SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);
	g_free(msginfo->header->msgid);
	msginfo->header->msgid = g_strdup(newmsgid);

	/*Add message with new message ID to UI */
	SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);

	/*add the new msgid to the unread list */
	if (is_message_unread(msgid)) {
		folder_message_del_list(msgid, UNREAD);
		folder_message_add_list(newmsgid, UNREAD);
	}

	procmsg_msginfo_free(msginfo);
	g_free(msgid_path);
	g_free(newmsgid_path);
	g_free(account_name);
	return newmsgid;
}

/**
    This function called by adaption layer to  resend mails. 

    @param msgid_list   list of message ids
    @return SM_OK on success (or) smtp errorcode on failure
*/
gint send_resend_mails(const GSList * msgid_list)
{
	MsgInfo *msginfo = NULL;
	gchar tmpfile[MAXPATHLEN + 1];
	GSList *cur = NULL;
	PrefsAccount *prefs_account = NULL;
	gchar *msgidpath = NULL;
	gchar *outpath = NULL;
	gchar *outbox_id = NULL;
	gchar *account_name = NULL;
	gchar *msgid = NULL;
	gint retval = SM_OK;
	gint msgidCount = 0;
	gint idx = 1;

	debug_print("resend mails starts:\n");
	smtpconnection_alive = FALSE;

	if (msgid_list == NULL) {
		return SMTP_SEND_ERROR;
	}

	msgidCount = g_slist_length((GSList *) msgid_list);
	send_wa.total_messages_size = send_get_msglist_file_size(msgid_list);
	send_wa.current_messages_size = 0;
	send_wa.cancel_status = 0;

	for (cur = (GSList *) msgid_list; cur != NULL; cur = cur->next, idx++) {
		msgid = (gchar *) cur->data;
		if (msgid == NULL) {
			continue;
		}
		/* Check for Valid message */
		if (send_is_folder_match(msgid, SENT_DIR) == FALSE) {
			continue;
		}

		/* Get prefs Account info */
		account_name = get_account_name(msgid);

		if ((prefs_account = send_get_prefs_account(account_name)) == NULL) {
			g_free(account_name);
			continue;
		}

		/* Get the MsgInfo */
		msgidpath = folder_get_message_file_path(msgid);
		if (msgidpath == NULL) {
			g_free(account_name);
			continue;
		}
		msginfo = (MsgInfo *) folder_parse_msg(msgidpath);
		if ((msginfo == NULL) || (msginfo->header == NULL)) {
			debug_print("msginfo not found from msg id = %s  \n", msgid);
			g_free(msgidpath);
			g_free(account_name);
			continue;
		}

		if (procmime_check_attachments(msgidpath, FALSE) == TRUE) {
			msginfo->header->flags.perm_flags |= MSG_ATTACH;
		}
		if (msginfo->header->msgid == NULL) {
			debug_print("MsgInfoHeader is NULL \n");
			procmsg_msginfo_free(msginfo);
			g_free(msgidpath);
			g_free(account_name);
			continue;
		}
		/* Set prefs Account options to resend mail */
		if (msginfo->settings == NULL) {
			msginfo->settings = g_new0(EmailSpecificSettings, 1);
		}

		if (msginfo->settings != NULL) {
			msginfo->settings->options = prefs_account->advanced_settings.email_sending;
		}

		/* Create a temporary file to write all the messages */
		g_snprintf(tmpfile, sizeof(tmpfile), "%s%ctmpmsg.%08x",
			   get_tmp_dir(), G_DIR_SEPARATOR, (gint) msginfo);

		if ((retval = compose_resend_mail(tmpfile, msgidpath,
						  prefs_account, msginfo)) != SM_OK) {
			debug_print(" Error in composing mail ....\n");
			g_free(msgidpath);
			procmsg_msginfo_free(msginfo);
			g_free(account_name);
			continue;
		}
		g_free(msgidpath);
		msginfo->settings->options = SEND_NEXT_CONNECTION;
		outbox_id = send_newmsgid_with_outbox_options(msginfo);
		msginfo->settings->options = SEND_IMMEDIATE;

		if ((outbox_id != NULL) && (prefs_account->user_settings.account_name != NULL)) {
			outpath = get_new_path(OUTBOX_DIR,
					       prefs_account->user_settings.account_name,
					       outbox_id);
		}

		if (compose_move_msg(tmpfile, outpath) != SM_OK) {
			retval = SMTP_COPYTO_OUT_FOLDER_ERROR;
			debug_print("Errorcode(%d):Mail moving to outfolder Fail\n", retval);
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		} else {
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(outbox_id);
			msginfo->header->msgtag.recvtype =
			    prefs_account->receive_settings.recv_type;
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);
#if 0
			if (send_wa.cancel_status != CANCEL) {
				if (msginfo->settings->options == SEND_IMMEDIATE) {
					cur_send_msgid = outbox_id;
					send_outbox_mail(prefs_account->user_settings.account_name,
							 outbox_id, 0, idx, msgidCount,
							 display_progress);
					display_progress = FALSE;
				}
			}
#endif
		}
		g_free(outbox_id);
		procmsg_msginfo_free(msginfo);
		g_free(outpath);
		g_free(account_name);
	}
	SMTP_SET_MSG_CB(SMTP_CHECK_FOR_OUT_MSGS, NULL, NULL);
	SMTP_SEND_COMPLETE_CB();
	return SM_OK;
}

/** 
   This function called by send_mail function. It opens the given file in 
   read mode and invokes the smtp specific send mail function.
  
   @param file  fileame which contains composed email 
   @param ac_prefs user preferences account information 
   @param to_list  the list of users the mail to be sent 
   @return SM_OK on success (or) smtp errorcode on failure  
*/
static gint
send_message(const gchar * file,
	     const PrefsAccount * ac_prefs, GSList * to_list, gboolean display_progress)
{
	FILE *fp = NULL;
	gint retval = SM_OK;
	gchar *msgid = NULL;
	Folder *folder = NULL;

	debug_print("[%s][%s]:starts here :\n", __FILE__, __FUNCTION__);

	IS_PARAM_NULL(file, SMTP_FILE_NAME_NULL, return SMTP_FILE_NAME_NULL);
	IS_PARAM_NULL(to_list, SMTP_RCPT_LIST_INVALID, return SMTP_FILE_NAME_NULL);
	IS_PARAM_NULL(ac_prefs, SMTP_ACCOUNT_NULL, return SMTP_FILE_NAME_NULL);

	sent_message = FALSE;
	if (display_progress == TRUE) {
		SMTP_SET_MSG_CB(SMTP_START, ac_prefs->user_settings.account_name, NULL);
	}
#if CL_USE
	retval = cl_iap_connect(ac_prefs->advanced_settings.iap_name, FALSE);
	if (retval != CL_OK) {
		debug_print("CL connection Error n smtp = %d \n", retval);
		engine_errno = ENGINE_ERROR_IAP_CONNECT;
		SMTP_SEND_COMPLETE_CB();
		return SMTP_SEND_ERROR;
	}
#endif
	msgid = folder_get_msgid_from_path(g_strdup(file));
	folder = folder_get_msg_folder(msgid);
	if (strncmp(OUTBOX_DIR, folder->name, strlen(OUTBOX_DIR)) == 0) {
		if (msgid[strlen(msgid) - 1] == (OUTBOX_SUSPENDED | 0x30)) {
			return SMTP_ERROR;
		}
	}
	if ((fp = fopen(file, "rb")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		retval = SMTP_FILE_OPEN_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}

	retval = send_message_smtp(ac_prefs, to_list, fp);
	cur_send_msgid = NULL;
	fclose(fp);
	return retval;
}


/** 
    This functionis to smtp specific send mail function. Using session 
    common API, it creates smtp session and opens asynchonous socket 
    connection to the SMTP server specified preferences account. It also 
    notifies the status of send mail functioality through call back function 

    @param ac_prefs  user preferences account information 
    @param to_list   the list of users the mail to be sent 
    @param file      file pointer supplied by send_mesage_smtp function  
    @return SM_OK on success (or) smtp errorcode on failure  
*/

static gint send_message_smtp(const PrefsAccount * ac_prefs, GSList * to_list, FILE * fp)
{
	Session *session = NULL;
	SMTPSession *smtp_session = NULL;
	gint retval = SM_OK;
	gint bytes = 0;
	gchar *outfile = NULL;
	gchar *servname = NULL;

	debug_print("[%s][%d]:starts here :\n", __FUNCTION__, __LINE__);

	IS_PARAM_NULL(fp, SMTP_FILE_POINTER_NULL, return SMTP_FILE_POINTER_NULL);
	IS_PARAM_NULL(to_list, SMTP_RCPT_LIST_INVALID, return SMTP_RCPT_LIST_INVALID);
	IS_PARAM_NULL(ac_prefs, SMTP_ACCOUNT_NULL, return SMTP_ACCOUNT_NULL);

	if ((retval = compose_check_for_valid_account(ac_prefs)) != SM_OK) {
		debug_print("Errorcode(%d):Invalid inputs to send mail...\n", retval);
		servname = send_message_get_smtp_config_servername(ac_prefs);
		if (!servname) {
			servname = ac_prefs->send_settings.smtp_server;
		}
		SMTP_SET_MSG_CB(retval, servname, ac_prefs->user_settings.account_name);
		return retval;
	}

	/* In some cases same smtp connection  is used for sending next mails  
	 * In such case check the connection is valid for that account. If not same
	 * destory the session and create new session
	 */

	if (get_device_mode_flag() == FALSE) {
		if (smtpsession != NULL) {
			session_destroy((Session *) smtpsession);
			smtpsession = NULL;
		}
		retval = SMTP_FLIGHT_MODE_ERROR;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		return retval;
	}
	if (smtpsession != NULL) {
		if ((smtpconnection_alive == FALSE) ||
		    (smtpsession->account == NULL) ||
		    (smtpsession->state != SMTP_KEEPCONNALIVE) ||
		    (strcmp(ac_prefs->user_settings.account_name, smtpsession->account)) != 0) {
			/* Free the session */
			session_destroy((Session *) smtpsession);
			smtpsession = NULL;
		}
	}

	if (smtpsession == NULL) {
		session = smtp_session_new();
		if (!session) {
			retval = SMTP_SESSION_INVALID;
			debug_print("[%d] Erorcode(%d) : Session creation error... \n",
				    __LINE__, retval);
			SMTP_SET_MSG_CB(retval, NULL, NULL);
			return retval;
		}
		/* Convert the session to SMTP session */
		smtp_session = SMTP_SESSION(session);
		smtp_session->account = g_strdup(ac_prefs->user_settings.account_name);
		send_message_get_smtp_config_data(ac_prefs, session);
	} else {
		session = (Session *) smtpsession;
		smtp_session = SMTP_SESSION(session);
	}

	/* Set the authorisation type */

	smtp_session->to_list = to_list;
	smtp_session->cur_to = to_list;

	/* Convert the data to send in Internet Message Format */
	smtp_session->send_data_fp = get_outgoing_rfc2822_file(fp, &bytes, &outfile);
	if (!smtp_session->send_data_fp) {
		session_destroy(session);
		retval = SMTP_SEND_ERROR;
		smtpsession = NULL;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
		unlink(outfile);
		g_free(outfile);
		return retval;
	} else {
		smtp_session->send_data_len = bytes;
	}
	/* Function to call when message is received */
	session_set_recv_message_notify(session, send_recv_message);

	/* Function to call when sending data is in progress */
	session_set_send_data_progressive_notify(session, send_send_data_progressive);
	/* Function to call when data sending is complete */
	session_set_send_data_notify(session, send_send_data_finished);

	if (SMTP_SESSION(session)->state != SMTP_KEEPCONNALIVE) {
		if (session_connect(session, smtp_session->hostname, smtp_session->port) < 0) {
			/* NOTE: session_destroy will fclose the smtp_session->send_data_fp that 
			 * we got above as well if it is not fclosed before, as it's aliased
			 * in session->write_buf_fp */
			session_destroy(session);
			smtpsession = NULL;
			retval = SMTP_SERVER_INVALID;
#if USE_SSL
			if (session->ssl_type != SSL_NONE) {
				if (ssl_error != SSL_SUCCESS) {
					retval = ssl_error;
				}
			}
#endif
			SMTP_SET_MSG_CB(retval, smtp_session->hostname, NULL);
			unlink(outfile);
			g_free(outfile);
			return retval;
		}
	} else {
		if (smtp_use_connectionalive(session) != SM_OK) {
			retval = SMTP_SEND_ERROR;
			SMTP_SET_MSG_CB(retval, NULL, NULL);
			if (session != NULL) {
				session_destroy(session);
				smtpsession = NULL;
			}
			unlink(outfile);
			g_free(outfile);
			return retval;
		}
	}

	/* Wait till the session state is SESSION_READY, SESSION_SEND
	 * or SESSION_RECV state
	 */

	while (session_is_connected(session) &&
	       ((SMTP_SESSION(session)->state != SMTP_ERROR) &&
		(SMTP_SESSION(session)->state != SMTP_CANCEL) &&
		(SMTP_SESSION(session)->state != SMTP_KEEPCONNALIVE))) {
		g_main_context_iteration(NULL, TRUE);
	}

	if (SMTP_SESSION(session)->error_val == SM_AUTHFAIL) {
		retval = SMTP_AUTHENTICATION_FAIL_ERROR;
		SMTP_SET_MSG_CB(retval, smtp_session->hostname, NULL);
	} else if (SMTP_SESSION(session)->error_val == SM_CANCEL) {
		retval = SMTP_CANCEL_MAIL;
		SMTP_SET_MSG_CB(retval, NULL, NULL);
	} else if (SMTP_SESSION(session)->state == SMTP_CANCEL) {
		/* though the SMTP has sent the message it reponds with 
		 * SMTP cancel. So if the SMTP message has been sent
		 * for the message sent, it will be SM_OK
		 */
		if (sent_message == TRUE) {
			SMTP_SESSION(session)->error_val = SM_OK;
		} else {
			retval = SMTP_CANCEL_MAIL;
			SMTP_SET_MSG_CB(retval, NULL, NULL);
		}
	} else if (session->state == SESSION_ERROR ||
		   (session->state == SESSION_EOF &&
		    (!session->remote_close_ok)) ||
		   SMTP_SESSION(session)->state == SMTP_ERROR ||
		   SMTP_SESSION(session)->error_val != SM_OK) {
		if (session->state == SESSION_ERROR) {
			retval = SMTP_SERVER_INVALID;
		} else {
			if (SMTP_SESSION(session)->error_val == SMTP_INVALID_EMAIL_ADDRESS) {
				retval = SMTP_INVALID_EMAIL_ADDRESS;
			} else {
				retval = SMTP_SEND_ERROR;
			}
		}
#if USE_SSL
		if (session->ssl_type != SSL_NONE) {
			if (ssl_error != SSL_SUCCESS) {
				retval = ssl_error;
			}
		}
#endif
		SMTP_SET_MSG_CB(retval, smtp_session->hostname, NULL);
	}
	if ((retval != SM_OK) || (smtpconnection_alive == FALSE)) {
		session_destroy(session);
		smtpsession = NULL;
	} else {
		g_free(smtp_session->send_data);
		g_free(smtp_session->error_msg);
		unlink(outfile);
		g_free(outfile);
		return SM_OK;
	}

	unlink(outfile);
	g_free(outfile);
	return retval;
}

/** 
    This function notifies the status of send mail to the UI process using
    call back mechanism. 
  
    @param session IN  smtp session created for sending mail   
    @return  SMTP_SEND_ERROR  when smtp status is SMTP_ERROR 
     otherwise returns SM_OK  
*/

static gint send_recv_message(Session * session)
{
	SMTPSession *smtp_session = SMTP_SESSION(session);

	debug_print("[%s][%s:%d] : starts here :\n", __FILE__, __FUNCTION__, __LINE__);

	IS_PARAM_NULL(session, SMTP_SESSION_INVALID, return SMTP_SESSION_INVALID);

	switch (smtp_session->state) {
	case SMTP_READY:
	case SMTP_CONNECTED:
		return SM_OK;
	case SMTP_HELO:
	case SMTP_EHLO:
	case SMTP_AUTH:
		/*SMTP_SET_MSG_CB ( SMTP_AUTHENTICATING, NULL , NULL);
		 * bug fix 8955
		 */
		/*SMTP_SET_MSG_CB ( SMTP_AUTHENTICATING, smtp_session->account , NULL);
		 */
		break;
	case SMTP_FROM:
	case SMTP_RCPT:
	case SMTP_DATA:
	case SMTP_EOM:
		SMTP_SET_MSG_CB(SMTP_SENDING, NULL, NULL);
		break;
	case SMTP_QUIT:
		SMTP_SET_MSG_CB(SMTP_QUITTING, NULL, NULL);
		break;
	case SMTP_ERROR:
		return SMTP_SEND_ERROR;
	default:
		return SM_OK;
	}
	return SM_OK;
}


/** 
    This function notifies the mail progress to the UI process using 
    call back mechanism. 
  
    @param session IN  smtp session created for sending mail   
    @param cur_len IN  portion of mail sent  
    @param total_len IN total length of mail to be sent to SMTP server 
    @return SM_OK on success or smtp error code   
*/

static gint send_send_data_progressive(Session * session, const guint cur_len, guint total_len)
{
	long bytes_sent = 0;
	guint percent = 0;
	debug_print("[%s][%s:%d]:starts here:\n", __FILE__, __FUNCTION__, __LINE__);

	IS_PARAM_NULL(session, SMTP_SESSION_INVALID, return SMTP_SESSION_INVALID);

	if (SMTP_SESSION(session)->state != SMTP_SEND_DATA &&
	    SMTP_SESSION(session)->state != SMTP_EOM) {
		return SM_OK;
	}
	if (total_len == 0) {
		return SM_OK;
	}

	if (send_wa.total_messages_size <= 0) {
		send_wa.total_messages_size = total_len;
	}
	bytes_sent = send_wa.current_messages_size + cur_len;
	percent = (bytes_sent * 100) / send_wa.total_messages_size;
	if (percent > 100) {
		percent = 100;
	}
	if ((smtp_cb) && (smtp_cb->smtp_set_percentage_cb)) {
		smtp_cb->smtp_set_percentage_cb(percent);
	}
	return SM_OK;
}


/** 
    This function notifies UI process when mail sent completly to server
    through call back mechanism. 

    @param session IN  smtp session created for sending mail   
    @param len IN  portion of mail sent  
    @returns SM_OK on success 	
*/

static gint send_send_data_finished(Session * session, guint len)
{
	debug_print("[%s][%s:%d] : starts here :\n", __FILE__, __FUNCTION__, __LINE__);
	sent_message = TRUE;
	send_send_data_progressive(session, len, len);
	return SM_OK;
}

/** 
   This function used for cancelling the list of mails. In case of list 
   is NULL, it will cancel the currently sending mail.

   @ param items_list  List of message ids.
   @returns SM_OK on changing the state otherwisr returns SMTP error code
*/

gint send_cancel_mails(const GSList * items_list)
{
	gchar *newmsgid = NULL;
	gchar *msgid = NULL;
	gchar *msgid_path = NULL;
	gchar *newmsgid_path = NULL;
	gchar *account_name = NULL;
	GSList *cur = NULL;
	Folder *msgfolder = NULL;
	MsgInfo *msginfo = NULL;
	gint retval = SM_OK;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (items_list == NULL) {
		if (smtpsession != NULL) {
			smtpsession->state = SMTP_CANCEL;
			send_wa.cancel_status = CANCEL;
		}
		return SM_OK;
	}
	for (cur = (GSList *) items_list; cur != NULL; cur = cur->next) {
		msgid = NULL;
		newmsgid = NULL;
		account_name = NULL;
		msgid_path = NULL;
		newmsgid_path = NULL;

		msgid = (gchar *) cur->data;
		if (msgid == NULL) {
			continue;
		}

		/* Mail under sending process state change to cancel */
		if (cur_send_msgid != NULL) {
			if (strncmp(msgid, cur_send_msgid, strlen(cur_send_msgid)) == 0) {
				if (smtpsession != NULL) {
					smtpsession->state = SMTP_CANCEL;
					//continue;
				}
			}
		}
		/* Check for Mail already suspended */
		if (msgid[strlen(msgid) - 1] == (OUTBOX_SUSPENDED | 0x30)) {
			continue;
		}
		msgfolder = folder_get_msg_folder(msgid);
		if (msgfolder == NULL) {
			continue;
		}
		if (strncmp(OUTBOX_DIR, msgfolder->name, strlen(OUTBOX_DIR)) != 0) {
			continue;
		}
		account_name = get_account_name(msgid);
		msgid_path = folder_get_message_file_path(msgid);
		msginfo = (MsgInfo *) folder_parse_msg(msgid_path);
		if (!msginfo || !(msginfo->header)) {
			g_free(msgid_path);
			g_free(account_name);
			procmsg_msginfo_free(msginfo);
			continue;
		}

		newmsgid = g_strdup(msgid);
		newmsgid[strlen(newmsgid) - 1] = (OUTBOX_SUSPENDED | 0x30);
		/*Bug Fix:8177 Attachment icon disappear for cancelled mail */
		if (procmime_check_attachments(msgid_path, FALSE) == TRUE) {
			msginfo->header->flags.perm_flags |= MSG_ATTACH;
		}
		newmsgid_path = get_new_path(OUTBOX_DIR, account_name, newmsgid);

		if (compose_move_msg(msgid_path, newmsgid_path) != SM_OK) {
			retval = SMTP_COPYTO_OUT_FOLDER_ERROR;
			if (is_file_exist(msgid_path))
				SMTP_SET_MSG_CB(retval, NULL, NULL);
		} else {
			/* Update the status to UI that mails 
			 *  status changed to suspended 
			 */
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(msgid);
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, FALSE, FALSE);

			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(newmsgid);
			SMTP_UPDATE_FOLDER_VIEW_CB(msginfo->header, TRUE, FALSE);

			/*add the new msgid to the unread list */
			if (is_message_unread(msgid)) {
				folder_message_del_list(msgid, UNREAD);
				folder_message_add_list(newmsgid, UNREAD);
			}
		}
		procmsg_msginfo_free(msginfo);
		g_free(newmsgid);
		g_free(msgid_path);
		g_free(newmsgid_path);
		g_free(account_name);
	}
	return SM_OK;
}

/**
    This function called by adaption layer to clean up global memory 
    allocated by the smtp module. 
    @param none
    @return none
*/

void send_message_cleanup(void)
{
	if (smtpsession != NULL) {
		smtpsession->state = SMTP_CANCEL;
		session_destroy((Session *) smtpsession);
		smtpsession = NULL;
	}
}

/**
   This function used to update the read & unread status of a mail. 

    @param mesgidpath  message file path
    @param flag TRUE on unread , FALSE on read
    @return none
*/

static void send_read_unread_update(const gchar * msgidpath, gboolean flag)
{
	gchar *tmpmsgid = NULL;

	if (msgidpath == NULL) {
		return;
	}
	tmpmsgid = strrchr(msgidpath, G_DIR_SEPARATOR);

	if ((tmpmsgid != NULL) && (tmpmsgid + 1 != NULL)) {
		tmpmsgid++;
		if (flag == TRUE) {
			folder_message_add_list(tmpmsgid, UNREAD);
		} else {
			folder_message_del_list(tmpmsgid, UNREAD);
		}
	}
}

/**
    This function used to get the prefs account information for the 
    given account 

    @param account  Name of the account 
    @return PrefsAccount information on success or NULl on failure
*/

static PrefsAccount *send_get_prefs_account(const gchar * account)
{
	PrefsAccount *prefs_account = NULL;
	if (account != NULL) {
		prefs_account = (PrefsAccount *) account_get_account(account);
		if (prefs_account == NULL) {
			debug_print(" Account(%s) not found. Try Default\n", account);
			prefs_account = account_get_default_account();
		}
	}
	if (prefs_account == NULL) {
		debug_print(" Error : Invalid Account .. \n");
		SMTP_SET_MSG_CB(SMTP_ACCOUNT_NULL, NULL, NULL);
		return NULL;
	}
	return prefs_account;
}


/**
    This function used to check the folder name in msg id matching 
    the required folder 

    @param msgid  msgid in the format (ACCOUNT_FOLDER_NAME_ID) 
    @param foldername  to check whether this name matches the 
     msgid folder name 
    @return TRUE on matching the folder name or FALSE on failure 
*/

static gboolean send_is_folder_match(const gchar * msgid, const gchar * foldername)
{
	gboolean result = FALSE;
	gchar *msgid_folder_name = NULL;
	/* Check for invalid msgid */
	if (check_msgid(msgid) == FALSE) {
		return result;
	}
	if ((msgid_folder_name = get_fold_name((gchar *) msgid)) != NULL) {
		if (strcmp(msgid_folder_name, foldername) == 0) {
			result = TRUE;
		}
		g_free(msgid_folder_name);
	}
	return result;
}

/**
    This function used to generate new msgid with msginfo options
    ( like OUTBOX_WAITING etc).

    @param  msginfo contains the message information 
    @return new msgid on success or NULL failure 
*/

static gchar *send_newmsgid_with_outbox_options(const MsgInfo * msginfo)
{
	gchar *newid = NULL;

	if ((msginfo == NULL) || (msginfo->header == NULL)
	    || (msginfo->settings == NULL)) {
		return NULL;
	}
	if (msginfo->header->msgid == NULL) {
		return NULL;
	}
	if (send_wa.cancel_status == CANCEL) {
		newid = g_strdup_printf("%s%1d", msginfo->header->msgid, OUTBOX_SUSPENDED);
	} else if (msginfo->settings->options == SEND_ON_REQUEST) {
		newid = g_strdup_printf("%s%1d", msginfo->header->msgid, OUTBOX_ON_REQUEST);
	} else if (msginfo->settings->options == SEND_NEXT_CONNECTION) {
		newid = g_strdup_printf("%s%1d", msginfo->header->msgid, OUTBOX_WAITING);
	} else {
		newid = g_strdup_printf("%s%1d", msginfo->header->msgid, OUTBOX_SENDING);
	}
	return newid;
}

/**
    This function used to get the outbox mails count of given type
    ( like OUTBOX_WAITING and outbox failedetc).

    @param  msglist  contains the message information list 
    @param  option   contains the type of mails to get count
    @return mails count on success
*/

static gint send_get_outbox_mails_count(const GSList * msglist, const gchar option)
{
	gint count = 0;
	GSList *cur = NULL;
	MsgInfo *msginfo = NULL;
	gint send_status = 0;
	gchar out_failed = OUTBOX_FAILED | 0x30;

	if (msglist == NULL) {
		return count;
	}

	for (cur = (GSList *) msglist; cur != NULL; cur = cur->next) {
		msginfo = (MsgInfo *) cur->data;
		if (msginfo == NULL) {
			continue;
		} else if (msginfo->header == NULL) {
			continue;
		} else if (msginfo->header->msgid == NULL) {
			continue;
		}

		send_status = send_get_sending_status(msginfo->header->msgid);
		if (msginfo->header->msgid[strlen(msginfo->header->msgid) - 1]
		    == option) {
			count++;
		}

		/*Bug ID: 16871. Increment the count even for failed
		 * outbox mails for sending
		 */
		else if ((msginfo->header->msgid[strlen(msginfo->header->msgid) - 1]
			  == out_failed) && (send_status != SEND_ON_REQUEST)) {
			count++;
		}

		/* fix for 18671. When UI closes engine closes in the middle
		   of sending and when UI is opened again it starts sending
		   the mail but the mail count is not taken hence sending 0
		   of 0 is displayed
		 */
		else if ((msginfo->header->msgid[strlen(msginfo->header->msgid) - 1] ==
			  (OUTBOX_SENDING | 0x30))) {

			count++;
		}
	}

	return count;
}

/**
    This function used to get the outbox mails count of given type
    ( like OUTBOX_WAITING etc).

    @param  msglist  contains the message information list 
    @param  option   contains the type of mails to get count
    @return mails count on success
*/
static gint
send_get_outbox_mails_list(const gchar * account,
			   const gchar * msgid, GSList ** msglist, gint * outboxcount)
{
	GSList *msginfo_list = NULL;
	GSList *msgheader_list = NULL;
	gchar *id_account = NULL;
	gchar *msgidpath = NULL;
	MsgInfo *msginfo = NULL;
	gchar *tmpmsgid = NULL;

	/* This is a special condition.  If the msgId is NULL engine received
	 * Send/Receive Trigger and it needs to send all the mails stored with
	 * option SEND_NEXT_CONNECTION. Initially gets the message info of all
	 * messages in the OUTBOX and sends out the mails except SEND_ON_REQUEST
	 * and mails opened for editing 
	 */

	if (msgid == NULL) {
		/* Get the all mails info from outbox */
		msgheader_list = folder_populate_folder(OUTBOX_DIR);
		if (msgheader_list == NULL) {
			debug_print("Outbox doesn't contain any messages ");
			return SMTP_MSG_HDR_NULL;
		}
		send_wa.current_messages_size = 0;
		send_wa.cancel_status = 0;
		/*  Construct msginfo list from above  msgheaders */
		while (msgheader_list) {
			tmpmsgid = ((MsgHeader *) msgheader_list->data)->msgid;

			if (send_is_folder_match(tmpmsgid, OUTBOX_DIR) == FALSE) {
				procmsg_msgheader_free(msgheader_list->data);
				msgheader_list =
				    g_slist_remove(msgheader_list, msgheader_list->data);
				continue;
			}
			id_account = get_account_name(tmpmsgid);

			if ((id_account != NULL) && (account != NULL) &&
			    (strcmp(account, id_account) == 0)) {
				msginfo = g_new0(MsgInfo, 1);
				if (msginfo != NULL) {
					msginfo->header = (MsgHeader *) msgheader_list->data;
				}
				send_wa.total_messages_size += send_get_msgid_file_size(tmpmsgid);
				msgheader_list = g_slist_remove(msgheader_list,
								msgheader_list->data);
				msginfo_list = g_slist_append(msginfo_list, msginfo);
			} else {
				procmsg_msgheader_free((MsgHeader *) msgheader_list->data);
				msgheader_list =
				    g_slist_remove(msgheader_list, msgheader_list->data);
			}
			g_free(id_account);
		}
		g_slist_free(msgheader_list);
		*outboxcount = send_get_outbox_mails_count(msginfo_list, OUTBOX_WAITING | 0x30);
	} else {
		/* msgid  is not NULL and msgid should exists in Outbox folder */
		if (send_is_folder_match(msgid, OUTBOX_DIR) == FALSE) {
			debug_print("[%s] : Msgid not belongs to outbox \n", __FUNCTION__);
			return SMTP_MSG_HDR_NULL;
		}
		msgidpath = folder_get_message_file_path(msgid);
		debug_print("[%s][%d]MsgId Path=%s\n", __FUNCTION__, __LINE__, msgidpath);
		msginfo = (MsgInfo *) folder_parse_msg(msgidpath);
		if ((msginfo == NULL) || (msginfo->header == NULL)) {
			debug_print("[%s] msginfo not found from msg id = %s  \n",
				    __FUNCTION__, msgid);
			g_free(msgidpath);
			procmsg_msginfo_free(msginfo);
			return SMTP_MSG_HDR_NULL;
		} else {
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(msgid);
			msginfo_list = g_slist_append(msginfo_list, msginfo);
		}
		g_free(msgidpath);
	}
	*msglist = msginfo_list;
	return SM_OK;
}


/**
    This function used to add msgid /delete msgid /Free list of
    editor open msgid list 

    @param  msgid  contains the message id 
    @param  option   contains the type of option (add/delete/free) 
    @return TRUE on success of FALSE on failure 
*/

gboolean send_editor_open_mail(const gchar * msgid, const gint option)
{
	GSList *cur = NULL;
	gchar *tmpid = NULL;

	/* Free the open editor msg list */
	if (option == EDITOR_OPEN_MSG_LIST_FREE) {
		slist_free_strings(open_msgid_list);
		g_slist_free(open_msgid_list);
		open_msgid_list = NULL;
		return TRUE;
	}

	if (send_is_folder_match(msgid, OUTBOX_DIR) == FALSE) {
		return FALSE;
	}

	/* Check whether the given msg is exist in the list */
	if (option == EDITOR_OPEN_MSG_EXISTS) {
		for (cur = open_msgid_list; cur != NULL; cur = cur->next) {
			if (cur->data == NULL) {
				continue;
			}
			if (strcmp(msgid, cur->data) == 0) {
				return TRUE;
			}
		}
		return FALSE;
	}
	/* Delete the message from  open editor msg list */
	if (option == EDITOR_OPEN_MSG_DEL) {
		for (cur = open_msgid_list; cur != NULL; cur = cur->next) {
			if (cur->data == NULL) {
				continue;
			}
			if (strcmp(msgid, cur->data) == 0) {
				open_msgid_list = g_slist_remove_link(open_msgid_list, cur);
				g_free(cur->data);
				return TRUE;
			}
		}
		return FALSE;
	}
	/* Add  the message to open editor msg list */
	if (option == EDITOR_OPEN_MSG_ADD) {
		tmpid = g_strdup(msgid);
		open_msgid_list = g_slist_append(open_msgid_list, tmpid);
		return TRUE;
	}
	return FALSE;
}

/**
    This function used to get the outbox mail status 

    @param  msgid  contains the message id 
    @return status of the outbox mail 
*/

static gint send_get_outboxmail_status(const gchar * msgid)
{
	gchar *mid = NULL;
	gchar status = 0;

	if (msgid != NULL) {
		mid = strrchr(msgid, MSGID_SEPARATOR);
		if (mid != NULL) {
			mid++;
			status = (mid[strlen(mid) - 1]) & 0x07;
		}
	}
	return (gint) status;
}

/**
  This function is used to get the sending option

  @param msgid - Message ID of the message whose status is required
  @return gint - Status SEND_IMMEDIATE =1, SEND_NEXT_CONNECTION = 2,
  		 SEND_ON_REQUEST = 3
*/
static gint send_get_sending_status(const gchar * msgid)
{
	gchar *mid = NULL;
	gchar status = 0;

	if (msgid != NULL) {
		mid = strrchr(msgid, MSGID_SEPARATOR);
		mid++;
		if (mid != NULL) {
			status = (mid[0]) & 0x07;
		}
	}
	return (gint) status;
}



/**
    This function used to calculate the given msgid file size
    @param  msgid  contains the message id 
    @return the size of given msgid file size 
*/

static gint send_get_msgid_file_size(gchar * msgid)
{
	gint fsize = 0;
	gchar *outpath = NULL;

	if (msgid == NULL) {
		return fsize;
	}
	outpath = folder_get_message_file_path(msgid);
	fsize = get_file_size(outpath);
	g_free(outpath);
	if (fsize < -1) {
		fsize = 0;
	}
	return fsize;
}

/**
    This function used to calculate the given msgid file size
    @param  msgid_list  contains the message id list 
    @return the size of given msgid_list files size 
*/

static gint send_get_msglist_file_size(const GSList * msgid_list)
{
	gint fsize = 0;
	GSList *cur = NULL;

	if (msgid_list == NULL) {
		return fsize;
	}
	for (cur = (GSList *) msgid_list; cur != NULL; cur = cur->next) {
		fsize += send_get_msgid_file_size(cur->data);
	}
	return fsize;
}

static gboolean send_validate_to_list(const GSList * to_list, MsgHeader * header)
{
	GSList *cur = NULL;
	gchar *at_the_rate = NULL;
	gboolean error = FALSE;
	gchar * newmsgid = NULL;

	if (to_list == NULL) {
		error = TRUE;
	}

	for (cur = (GSList *) to_list; cur != NULL; cur = cur->next) {

		if ((gchar *) cur->data == NULL) {
			error = TRUE;
			break;
		}

		at_the_rate = strchr((gchar *) cur->data, AT_THE_RATE);
		if (at_the_rate == NULL) {
			error = TRUE;
		        break;
		}
		if (strchr(at_the_rate + 1, SECTION_SEPARATOR) == NULL) {
			error = TRUE;
			break;
		}
		if (at_the_rate[strlen(at_the_rate) - 1] == SECTION_SEPARATOR) {
			error = TRUE;
			break;
		}

	}	
		if(error) {
			newmsgid = send_change_outbox_status(header->msgid,OUTBOX_FAILED | 0x30);			
			SMTP_SET_MSG_CB(SMTP_RCPT_LIST_INVALID, newmsgid, NULL);
			return FALSE;
		}
	return TRUE;
}

gboolean send_get_msg_id(MsgInfo * msginfo, const PrefsAccount * prefs_account, gboolean send_flag)
{
	gchar buf[BUFFSIZE];
	gchar *savepath = NULL;
	gchar *tmpid = NULL;
	gchar *old_id = NULL;
	Folder *folder = NULL;

	memset(buf, '\0', sizeof(buf));

	IS_PARAM_NULL(msginfo, SMTP_MESSAGE_INFO_NULL, return FALSE);
	IS_PARAM_NULL(prefs_account, SMTP_ACCOUNT_NULL, return FALSE);

	if (msginfo->header->msgid != NULL) {
		old_id = g_strdup(msginfo->header->msgid);
		folder = folder_get_msg_folder(old_id);
		/*Bug fix 4619 : Check added for NULL folder */
		if ((folder != NULL) && (strcmp(folder->name, ARCHIVE) == 0)) {
			folder = NULL;
		}
	}
	compose_generate_msgid(msginfo, prefs_account, buf, sizeof(buf), FALSE);

	/*generating Outbox msgid if we want to send the mail rather then saving */
	if (msginfo->header->msgid != NULL && send_flag == TRUE) {

		savepath = get_new_path(OUTBOX_DIR, prefs_account->user_settings.account_name,
					msginfo->header->msgid);
	} else {
		savepath = get_new_path(folder != NULL ? folder->name : DRAFT_DIR,
					prefs_account->user_settings.account_name,
					msginfo->header->msgid);
	}

	if (savepath != NULL) {
		tmpid = g_strdup(strrchr(savepath, G_DIR_SEPARATOR));
		if (tmpid != NULL) {
			g_free(msginfo->header->msgid);
			msginfo->header->msgid = g_strdup(tmpid + 1);
		}
	}
	g_free(savepath);
	g_free(old_id);
	g_free(tmpid);
	return TRUE;
}

/**
	This function used to get smtp specific configuration 
	@param account contains the account information used smtp
	@param session contains smtp specific session information	
	@return TRUE on getting smtp configuation, otherwise returns FALSE
*/

static gboolean send_message_get_smtp_config_data(const PrefsAccount * account, Session * session)
{
#if CL_USE
	const IapInfo *iapinfo = NULL;
#endif
	gboolean use_default = TRUE;
	SMTPSession *smtp_session = NULL;

	if ((!session) || (!account)) {
		return FALSE;
	}
	if (!account->user_settings.account_name) {
		return FALSE;
	}
	smtp_session = SMTP_SESSION(session);
	smtp_session->account = g_strdup(account->user_settings.account_name);
	smtpsession = smtp_session;

#if CL_USE
	if (account->advanced_settings.use_iap_smtp_serv == TRUE) {
		if ((smtp_cb) && (smtp_cb->smtp_get_iap_conf_info)) {
			iapinfo = smtp_cb->smtp_get_iap_conf_info(cl_active_iap());
		}
	}
	if (iapinfo) {
		smtp_session->forced_auth_type = iapinfo->smtp_auth_type;
		smtp_session->iap_name = iapinfo->iap_name;
		if (!iapinfo->smtp_servname) {
			use_default = TRUE;
		} else {
			if (iapinfo->smtp_username) {
				smtp_session->user = g_strdup(iapinfo->smtp_username);
			}
			smtp_session->hostname = g_strdup(iapinfo->smtp_servname);
			if (iapinfo->smtp_passwd) {
				smtp_session->pass = g_strdup(iapinfo->smtp_passwd);
			}
			use_default = FALSE;
		}
	}
#endif
	if (use_default == TRUE) {
		smtp_session->forced_auth_type = account->send_settings.smtp_auth_type;
		if (account->send_settings.username) {
			smtp_session->user = g_strdup(account->send_settings.username);
		}
		if (account->send_settings.password) {
			smtp_session->pass = g_strdup(account->send_settings.password);
		}
		if (account->send_settings.smtp_server) {
			smtp_session->hostname = g_strdup(account->send_settings.smtp_server);
		}
		smtp_session->iap_name = NULL;
	}
	session->ssl_type = SSL_NONE;
#if USE_SSL
	if (use_default == TRUE) {
		smtp_session->port = account->send_settings.smtp_port;
		if (account->advanced_settings.smtp_use_ssl) {
			if (account->advanced_settings.smtp_use_tls == TRUE) {
				session->ssl_type = SSL_STARTTLS;
			} else {
				session->ssl_type = SSL_TUNNEL;
			}
		}
#if CL_USE
	} else {
		smtp_session->port = iapinfo->smtp_port;
		if (iapinfo->smtp_use_ssl) {
			if (iapinfo->smtp_use_tls == TRUE) {
				session->ssl_type = SSL_STARTTLS;
			} else {
				session->ssl_type = SSL_TUNNEL;
			}
		}
#endif
	}
#else
	smtp_session->port = account->send_settings.smtp_port;
#endif
	if (smtp_session->port == 0) {
		smtp_session->port = SMTP_PORT;
	}
	if (account->user_settings.email != NULL) {
		smtp_session->from = g_strdup(account->user_settings.email);
	}
	return TRUE;
}

/**
        This function used to get smtp specific sever name
        @param account contains the account information used smtp
        @return servername incase server is configures or NULL
*/

static gchar *send_message_get_smtp_config_servername(const PrefsAccount * account)
{
#if CL_USE
	const IapInfo *iapinfo = NULL;
	if (account == NULL) {
		return NULL;
	}
	if (account->advanced_settings.use_iap_smtp_serv == TRUE) {
		if ((smtp_cb) && (smtp_cb->smtp_get_iap_conf_info)) {
			iapinfo = smtp_cb->smtp_get_iap_conf_info(cl_active_iap());
		}
	}
	if (iapinfo) {
		if (!iapinfo->smtp_servname) {
			return iapinfo->smtp_servname;
		}
	}
#endif
	return NULL;
}
