
/*
 * LibSylph -- E-Mail client library
 * Copyright (C) 1999-2006 Hiroyuki Yamamoto
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <glib.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "intl.h"
#include "callbacks.h"
#include "smtp.h"
#include "utils.h"
#include "base64.h"
#include "ssl.h"
#include "md5_hmac.h"
#include "smtp_errorcodes.h"
#include "send_message.h"
#include "account.h"

static void smtp_session_destroy(Session * session);
static gint smtp_from(SMTPSession * session);
static gint smtp_auth(SMTPSession * session);
static gint smtp_starttls(SMTPSession * session);
static gint smtp_auth_cram_md5(SMTPSession * session);
static gint smtp_auth_login(SMTPSession * session);
static gint smtp_ehlo(SMTPSession * session);
static gint smtp_ehlo_recv(SMTPSession * session, const gchar * msg);
static gint smtp_helo(SMTPSession * session);
static gint smtp_rcpt(SMTPSession * session);
static gint smtp_data(SMTPSession * session);
static gint smtp_send_data(SMTPSession * session);
static gint smtp_eom(SMTPSession * session);
static gint smtp_session_recv_msg(Session * session, const gchar * msg);
static gint smtp_session_send_data_finished(Session * session, guint len);

/**
  Function to connect with temporary passwd

  @param smtp_session Session pointer
  @param message
  @return SM_OK on success and error code on Failure
*/
static gint smtp_connect_tmp_passwd(SMTPSession * smtp_session, const gchar * msg);

SmtpCallback *smtp_cb = NULL;
extern gboolean smtpconnection_alive;	/* Defined in send_message.c */

/**
   This function does the initialises the smtp callback  structure.
   This function is called by Abstraction Layer at time of initialisation 

   @param SmtpCallback IN  call 
   @return TRUE on success otherwise FALSE 
*/

gboolean smtp_init(SmtpCallback * cb)
{
	if (cb) {
		smtp_cb = cb;
		return TRUE;
	} else {
		smtp_cb = NULL;
		return FALSE;
	}
}

/**
   This function is used to clean up the global memory allocated 
   to smtp module. This function is called by Abstraction Layer at 
   time of closing the application. 

   @return nothing 
*/

void smtp_cleanup(void)
{
	g_free(smtp_cb);
	smtp_cb = NULL;
	send_message_cleanup();
}

/**
   This function creates a new session initialises with default parameters
   For every new mail to be sent, a new session is to be created.

   @return smtp session with some default parameters (or) NULL 
*/
Session *smtp_session_new(void)
{
	SMTPSession *session = NULL;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	session = g_new0(SMTPSession, 1);
	if (session == NULL) {
		return NULL;
	}
	session_init(SESSION(session));

	SESSION(session)->type = SESSION_SMTP;

	/* Function to call to receive the message from SMTP server */
	SESSION(session)->recv_msg = smtp_session_recv_msg;

	/* Function to call when receiving data is finished */
	SESSION(session)->recv_data_finished = NULL;

	/* Functionc all when all the data is sent to SMTP server */
	SESSION(session)->send_data_finished = smtp_session_send_data_finished;

	SESSION(session)->destroy = smtp_session_destroy;

	session->state = SMTP_READY;

#if USE_SSL
	session->tls_init_done = FALSE;
#endif

	session->hostname = NULL;
	session->user = NULL;
	session->pass = NULL;

	session->from = NULL;
	session->to_list = NULL;
	session->cur_to = NULL;

	session->send_data = NULL;
	session->send_data_fp = NULL;
	session->send_data_len = 0;

	session->avail_auth_type = 0;
	session->forced_auth_type = 0;
	session->auth_type = 0;

	session->error_val = SM_OK;
	session->error_msg = NULL;

	session->account = NULL;

	return SESSION(session);
}


/**
   This function is to destroys the session
       
   @param  Session smtp session 
   @return nothing 
*/

static void smtp_session_destroy(Session * session)
{
	SMTPSession *smtp_session = SMTP_SESSION(session);

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		return;
	}
	g_free(smtp_session->hostname);
	g_free(smtp_session->user);
	g_free(smtp_session->pass);
	g_free(smtp_session->from);
	g_free(smtp_session->send_data);
	session->write_buf = session->write_buf_p = NULL;
	g_free(smtp_session->error_msg);
	g_free(smtp_session->account);
}


/**
   This function sends the email address<FROM> of the sender to the server
   which is part of SMTP protocol
 
   @param session  SMTP session
   @return SM_OK on success or error code on failure 
*/

static gint smtp_from(SMTPSession * session)
{
	gchar buf[MSGBUFSIZE];

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	if (session->from == NULL) {
		debug_print("[%s][%s] : session->from NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}

	memset(buf, '\0', MSGBUFSIZE);

	session->state = SMTP_FROM;

	if (strchr(session->from, '<')) {
		g_snprintf(buf, sizeof(buf), "MAIL FROM: %s", session->from);
	} else {
		g_snprintf(buf, sizeof(buf), "MAIL FROM: <%s>", session->from);
	}
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
	log_message("SMTP> %s \n", buf);

	return SM_OK;
}

/**
   This function does the authentication based on session auth type. 

   @param session  SMTP session
   @return SM_OK on success or error code on failure 
*/
static gint smtp_auth(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_AUTH;
	if ((session->forced_auth_type == SMTPAUTH_CRAM_MD5) &&
	    ((session->avail_auth_type & SMTPAUTH_CRAM_MD5) != 0)) {
		smtp_auth_cram_md5(session);
	} else if ((session->forced_auth_type == SMTPAUTH_LOGIN) &&
		   ((session->avail_auth_type & SMTPAUTH_LOGIN) != 0)) {
		smtp_auth_login(session);
	} else if (session->forced_auth_type == SMTPAUTH_NO_AUTH) {
		smtp_from(session);
	} else {
		log_message((const gchar *) ("SMTP AUTH not available\n"));
		return SM_AUTHFAIL;
	}

	return SM_OK;
}

/**
    This function  is called to process the response codes received from 
    server based on authentication. This is part of smtp protocol.

    @param session  SMTP session
    @param msg  response received from SMTP server 
    @return SM_OK on success or error code on failure 
*/

static gint smtp_auth_recv(SMTPSession * session, const gchar * msg)
{
	gchar buf[MSGBUFSIZE];
	gchar *response = NULL;
	gchar *response64 = NULL;
	gchar *challenge = NULL;
	gint challengelen = 0;
	guchar hexdigest[33];
	gchar *domain_name = NULL;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	if (session->hostname == NULL) {
		debug_print("[%s][%s]:Hostname NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}

	memset(buf, '\0', MSGBUFSIZE);

	switch (session->auth_type) {
	case SMTPAUTH_LOGIN:
		session->state = SMTP_AUTH_LOGIN_USER;

		if (strncmp(msg, "334 ", 4) == 0) {
			base64_encode(buf, (guchar *) session->user, strlen(session->user));

			session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
			log_message("ESMTP> [USERID] \n");
		} else {
			/* Server rejects AUTH */
			snprintf(buf, sizeof(buf), "*");
			session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
			log_message("ESMTP> *\n");
		}
		break;
	case SMTPAUTH_CRAM_MD5:
		session->state = SMTP_AUTH_CRAM_MD5;

		if (strncmp(msg, "334 ", 4) == 0) {
			if (inet_addr(session->hostname) == -1) {
				domain_name = g_strdup(session->hostname);
			} else {
				domain_name = (gchar *) hostaddr_to_hostname(session->hostname);
			}

			challenge = g_malloc(strlen(msg + 4) + 1);
			challengelen = base64_decode((guchar *) challenge, msg + 4, -1);
			challenge[challengelen] = '\0';
			log_message("ESMTP< [Decoded: %s]\n", challenge);

			g_snprintf(buf, sizeof(buf), "%s", session->pass);
			md5_hex_hmac((gchar *) hexdigest,
				     (guchar *) challenge,
				     challengelen, (guchar *) buf, strlen(session->pass));
			g_free(challenge);
			/* CRAM-MD5 username should be given as username@domain */
			response =
			    g_strdup_printf("%s@%s %s", session->user, domain_name, hexdigest);
			log_message("ESMTP> [Encoded: %s]\n", response);

			response64 = g_malloc((strlen(response) + 3) * 2 + 1);
			base64_encode(response64, (guchar *) response, strlen(response));
			g_free(response);
			g_free(domain_name);

			session_send_msg(SESSION(session), SESSION_MSG_NORMAL, response64);
			log_message("ESMTP> %s\n", response64);
			g_free(response64);
		} else {
			/* Server rejects AUTH */
			session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "*");
			log_message("ESMTP> * \n");
		}
		break;
	default:
		/* stop smtp_auth when no correct authtype */
		session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "*");
		log_message("ESMTP> *\n");
		break;
	}

	return SM_OK;
}


/**
    This function is called when USER ID has been sent and reply for that has
    been received. This function sends the password to server.

    @param session  SMTP session
    @param msg  response received from SMTP server 
    @return SM_OK on success or error code on failure 
*/

static gint smtp_auth_login_user_recv(SMTPSession * session, const gchar * msg)
{
	gchar buf[MSGBUFSIZE];
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}

	memset(buf, '\0', MSGBUFSIZE);

	session->state = SMTP_AUTH_LOGIN_PASS;

	if (!strncmp(msg, "334 ", 4)) {
		base64_encode(buf, (guchar *) session->pass, strlen(session->pass));
	} else {
		/* Server rejects AUTH */
		g_snprintf(buf, sizeof(buf), "*");
	}
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
	log_message("ESMTP> [PASSWORD]\n");

	return SM_OK;
}

/**
    This function sends the EHLO command to server

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_ehlo(SMTPSession * session)
{
	gchar buf[MSGBUFSIZE];
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	memset(buf, '\0', MSGBUFSIZE);
	session->state = SMTP_EHLO;

	session->avail_auth_type = 0;
#if 0
	g_snprintf(buf, sizeof(buf), "EHLO %s",
		   session->hostname ? session->hostname : get_domain_name());
#endif
	g_snprintf(buf, sizeof(buf), "EHLO localhost");

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
	log_message("ESMTP> %s\n", buf);

	return SM_OK;
}

/**
    This function  is called to process the response code received from 
    server for EHLO command. This is part of smtp protocol.

    @param session  SMTP session
    @param msg  IN  response received from SMTP server 
    @return SM_OK on success or error code on failure 
*/

static gint smtp_ehlo_recv(SMTPSession * session, const gchar * msg)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}

	if (strncmp(msg, "250", 3) == 0) {
		const gchar *p = msg;
		p += 3;
		if (*p == '-' || *p == ' ') {
			p++;
		}
		if (g_strncasecmp(p, "AUTH", 4) == 0) {
			p += 5;
			if (strcasestr(p, "LOGIN")) {
				session->avail_auth_type |= SMTPAUTH_LOGIN;
			}
			if (strcasestr(p, "CRAM-MD5")) {
				session->avail_auth_type |= SMTPAUTH_CRAM_MD5;
			}
		}
		return SM_OK;
	} else if ((msg[0] == '1' || msg[0] == '2' || msg[0] == '3') &&
		   (msg[3] == ' ' || msg[3] == '\0')) {
		return SM_OK;
	} else if (msg[0] == '5' && msg[1] == '0' &&
		   (msg[2] == '4' || msg[2] == '3' || msg[2] == '1')) {
		return SM_ERROR;
	}
	return SM_ERROR;
}

/**
    This function  is called to start tls session with the server. 

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_starttls(SMTPSession * session)
{
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_STARTTLS;

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "STARTTLS");
	log_print("ESMTP> STARTTLS\n");

	return SM_OK;
}


/**
    This function  is called to send authentication type (AUTH CRAM-MD5) 
    to server. This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_auth_cram_md5(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_AUTH;
	session->auth_type = SMTPAUTH_CRAM_MD5;

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "AUTH CRAM-MD5");
	log_message("ESMTP> AUTH CRAM-MD5\n");

	return SM_OK;
}

/**
    This function  is called to send authentication type (AUTH LOGIN) 
    to server. This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_auth_login(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_AUTH;
	session->auth_type = SMTPAUTH_LOGIN;

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "AUTH LOGIN");
	log_message("ESMTP> AUTH LOGIN\n");

	return SM_OK;
}

/**
    This function  is called to send HELO command to SMTP server.
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_helo(SMTPSession * session)
{
	gchar buf[MSGBUFSIZE];
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	memset(buf, '\0', MSGBUFSIZE);
	session->state = SMTP_HELO;
#if 0
	g_snprintf(buf, sizeof(buf), "HELO %s",
		   session->hostname ? session->hostname : get_domain_name());
#endif
	g_snprintf(buf, sizeof(buf), "HELO localhost");
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
	log_message("SMTP> %s\n", buf);

	return SM_OK;
}

/**
    This function is called to send the list of rcpt to server. 
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_rcpt(SMTPSession * session)
{
	gchar buf[MSGBUFSIZE];
	gchar *to = NULL;
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);

	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	if (session->cur_to == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	memset(buf, '\0', MSGBUFSIZE);

	session->state = SMTP_RCPT;

	to = (gchar *) session->cur_to->data;

	if (strchr(to, '<')) {
		g_snprintf(buf, sizeof(buf), "RCPT TO: %s", to);
	} else {
		g_snprintf(buf, sizeof(buf), "RCPT TO: <%s>", to);
	}
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, buf);
	log_message("SMTP> %s\n", buf);

	session->cur_to = session->cur_to->next;

	return SM_OK;
}


/**
    This function s called to send the DATA command 
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_data(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_DATA;

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "DATA");
	log_message("SMTP> DATA\n");

	return SM_OK;
}

/**
    This function actually sends the data to the server 
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_send_data(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_SEND_DATA;

	if (session_send_data_fp(SESSION(session), session->send_data_fp,
				 session->send_data_len) < 0) {
		SMTP_SEND_COMPLETE_CB();
	}

	return SM_OK;
}


/**
    This function is send to QUIT command to server. 
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

gint smtp_quit(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_QUIT;
	/* hack for gmail support: don't care if the remote side client closes 
	 * before sending the reply to the QUIT
	 */
	SESSION(session)->remote_close_ok = TRUE;
	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, "QUIT");
	log_message("SMTP> QUIT\n");

	return SM_OK;
}


/**
    This function writes a single .(dot) into the socket.
    This is part of smtp protocol.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_eom(SMTPSession * session)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	session->state = SMTP_EOM;

	session_send_msg(SESSION(session), SESSION_MSG_NORMAL, ".");
	log_message("SMTP> . (EOM)\n");

	return SM_OK;
}


/**
    This function is called once data has been read from the socket 
    server. It then process the received message and and returns the status
    or error code.

    @param Session  SMTP session
    @param message 
    @return SM_OK on success or error code on failure 
*/


static gint smtp_session_recv_msg(Session * session, const gchar * msg)
{
	SMTPSession *smtp_session = SMTP_SESSION(session);
	gboolean cont = FALSE;
	gint ok = 0;

	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}

	if (strlen(msg) < 4) {
		/* hack for gmail: it will close the connection after QUIT, 
		 * instead of returning the result code 
		 */
		if (smtp_session->state == SMTP_QUIT && session->remote_close_ok) {
			session_disconnect(session);
			return SM_OK;
		} else {
			return SM_ERROR;
		}
	}

	if ((strncmp(msg, "504", 3) == 0) ||
	    (strncmp(msg, "503", 3) == 0) ||
	    (strncmp(msg, "554", 3) == 0) || (strncmp(msg, "501", 3) == 0)) {
		smtp_session->state = SMTP_ERROR;
		smtp_session->error_val = SM_ERROR;
		g_free(smtp_session->error_msg);
		smtp_session->error_msg = g_strdup(msg);
		return SM_ERROR;

	}
	if ((strncmp(msg, "550", 3) == 0)) {
		smtp_session->state = SMTP_ERROR;
		smtp_session->error_val = SMTP_INVALID_EMAIL_ADDRESS;
		g_free(smtp_session->error_msg);
		smtp_session->error_msg = g_strdup(msg);
		return SMTP_INVALID_EMAIL_ADDRESS;
	}

	if (strncmp(msg, "535", 3) == 0) {
		if ((smtp_session->state == SMTP_AUTH_LOGIN_PASS) ||
		    (smtp_session->state == SMTP_AUTH_CRAM_MD5)) {
			ok = smtp_connect_tmp_passwd(smtp_session, msg);
		}
		return ok;
	}
	if (!((smtp_session->state == SMTP_RCPT))) {

		if (msg[0] != '1' && msg[0] != '2' && msg[0] != '3') {
			smtp_session->state = SMTP_ERROR;
			smtp_session->error_val = SM_ERROR;
			g_free(smtp_session->error_msg);
			smtp_session->error_msg = g_strdup(msg);
			return SM_ERROR;
		}

		if (msg[3] == '-') {
			cont = TRUE;
		} else if (msg[3] != ' ' && msg[3] != '\0') {
			smtp_session->state = SMTP_ERROR;
			smtp_session->error_val = SM_UNRECOVERABLE;
			return SM_UNRECOVERABLE;
		}
	}

	/* ignore all multiline responses except for EHLO */
	if (cont && smtp_session->state != SMTP_EHLO) {
		return session_recv_msg(session);
	}
	switch (smtp_session->state) {
	case SMTP_READY:
	case SMTP_CONNECTED:
#if USE_SSL
		if (smtp_session->user || session->ssl_type != SSL_NONE) {
#else
		if (smtp_session->user) {
#endif
			smtp_ehlo(smtp_session);
		} else {
			smtp_helo(smtp_session);
		}
		break;
	case SMTP_HELO:
		smtp_from(smtp_session);
		break;
	case SMTP_EHLO:
		smtp_ehlo_recv(smtp_session, msg);
		if (cont == TRUE) {
			break;
		}
#if USE_SSL
		if (session->ssl_type == SSL_STARTTLS && smtp_session->tls_init_done == FALSE) {
			smtp_starttls(smtp_session);
			break;
		}
#endif
		if ((smtp_session->user) && (smtp_session->forced_auth_type != SMTPAUTH_NO_AUTH)) {
			if (smtp_session->pass == NULL) {
				ok = smtp_connect_tmp_passwd(smtp_session, msg);
				return ok;
			}
			if (smtp_auth(smtp_session) != SM_OK) {
				smtp_session->error_val = SM_AUTHFAIL;
				smtp_session->state = SMTP_ERROR;
				return SM_ERROR;
				//smtp_from ( smtp_session );
			}
		} else {
			smtp_from(smtp_session);
		}
		break;
	case SMTP_STARTTLS:
#if USE_SSL
		if (session_start_tls(session) < 0) {
			log_warning("can't start TLS session\n");
			smtp_session->state = SMTP_ERROR;
			smtp_session->error_val = SM_ERROR;
			return SM_ERROR;
		}
		smtp_session->tls_init_done = TRUE;
		smtp_ehlo(smtp_session);
#endif
		break;
	case SMTP_AUTH:
		smtp_auth_recv(smtp_session, msg);
		break;
	case SMTP_AUTH_LOGIN_USER:
		smtp_auth_login_user_recv(smtp_session, msg);
		break;
	case SMTP_AUTH_LOGIN_PASS:
	case SMTP_AUTH_CRAM_MD5:
		smtp_from(smtp_session);
		break;
	case SMTP_FROM:
		if (smtp_session->cur_to) {
			smtp_rcpt(smtp_session);
		}
		break;
	case SMTP_RCPT:
		if (smtp_session->cur_to) {
			smtp_rcpt(smtp_session);
		} else {
			smtp_data(smtp_session);
		}
		break;
	case SMTP_DATA:
		smtp_send_data(smtp_session);
		break;
	case SMTP_EOM:
		if (smtpconnection_alive == TRUE) {
			debug_print("Keep Connetion alive set. No Quit here\n");
			smtp_session->state = SMTP_KEEPCONNALIVE;
			return SM_OK;
		} else {
			smtp_quit(smtp_session);
		}
		break;
	case SMTP_QUIT:
		session_disconnect(session);
		break;
	case SMTP_CANCEL:
		smtp_session->error_val = SM_CANCEL;
		break;
	case SMTP_ERROR:
	default:
		smtp_session->error_val = SM_ERROR;
		return SM_ERROR;
	}

	if (cont) {
		return session_recv_msg(session);
	}

	return SM_OK;
}

/**
    This function is called when data transfer is finished and it sends
    a  single .(dot) into the socket.

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

static gint smtp_session_send_data_finished(Session * session, guint len)
{
	debug_print("[%s][%s] : starts here :\n", __FILE__, __FUNCTION__);
	if (session == NULL) {
		debug_print("[%s][%s] : session NULL :\n", __FILE__, __FUNCTION__);
		return SMTP_SESSION_INVALID;
	}
	smtp_eom(SMTP_SESSION(session));
	return SM_OK;
}

/**
    This function is called when keep connection alive option selected
    and session is to use the previous connection

    @param session  SMTP session
    @return SM_OK on success or error code on failure 
*/

gint smtp_use_connectionalive(Session * session)
{
	SMTPSession *smtp_session = SMTP_SESSION(session);

	if (session == NULL) {
		debug_print(" session connection alive is NULL\n");
		return SMTP_SESSION_INVALID;
	}
	if (smtp_session->state == SMTP_KEEPCONNALIVE) {
		smtp_from(smtp_session);
	} else {
		return SM_ERROR;
	}
	return SM_OK;
}
static gint smtp_connect_tmp_passwd(SMTPSession * smtp_session, const gchar * msg)
{
	PrefsAccount *ac_prefs = NULL;

	ac_prefs = (PrefsAccount *) account_get_account_cb((gchar *)
							   smtp_session->account);
	if(ac_prefs == NULL) {
		smtp_session->state = SMTP_ERROR;
		smtp_session->error_val = SM_AUTHFAIL;
		g_free(smtp_session->error_msg);
		smtp_session->error_msg = g_strdup(msg);
		return SM_AUTHFAIL;
	}		

	if ((ac_prefs->user_settings.scheduled_op == TRUE) ||
	    (smtp_session->pass != NULL)) {
		smtp_session->state = SMTP_ERROR;
		smtp_session->error_val = SM_AUTHFAIL;
		g_free(smtp_session->error_msg);
		smtp_session->error_msg = g_strdup(msg);
		return SM_AUTHFAIL;
	}
	if (smtp_session->pass != NULL) {
		SMTP_SET_MSG_CB(SMTP_PASSWD_WRONG, ac_prefs->user_settings.account_name, smtp_session->iap_name);

	} else {
		SMTP_SET_MSG_CB(SMTP_PASSWD_FAIL, ac_prefs->user_settings.account_name, smtp_session->iap_name);
	}

	g_free(ac_prefs->user_settings.tmp_passwd);
	ac_prefs->user_settings.tmp_passwd = NULL;
	ac_prefs->user_settings.cancel = FALSE;
	while (1) {
		(void) g_main_context_iteration(NULL, FALSE);

		if (ac_prefs->user_settings.tmp_passwd != NULL) {
			g_free(smtp_session->pass);
			smtp_session->pass = ac_prefs->user_settings.tmp_passwd;
			ac_prefs->user_settings.tmp_passwd = NULL;
			break;
		}
		if (ac_prefs->user_settings.cancel == TRUE) {
			smtp_session->error_val = SM_CANCEL;
			smtp_quit(smtp_session);
			return SM_OK;
		}
	}
	if (smtp_auth(smtp_session) != SM_OK) {
		smtp_session->error_val = SM_AUTHFAIL;
		smtp_session->state = SMTP_ERROR;
		g_free(smtp_session->error_msg);
		smtp_session->error_msg = g_strdup(msg);
		return SM_AUTHFAIL;
	}
	return SM_OK;
}
